# checkin.tcl: Check-in support
#
# $Id: checkin.tcl,v 1.15 2000/05/02 00:01:28 johnr Exp $
#
# Copyright (c) 1998-2000 The Regents of the University of California.
# All rights reserved. See the file COPYRIGHT for details.
#

# Patterns for file/directories that are ignored by sync
set syncIgnore {
    JUNK
    TEMP
    Builds
    api
    bin
    doc
    junk
    scratch
    temp
    tmp*
    local.mk
    vars.include.mk
    vars.mk
    config.*
    *.core
    #*
    *~
}

# The confirm startup message
set confirmMessage {
    ### Confirming check-in validity.                 ###
    ###                                               ###
    ### This is a confidence check only: the snapshot ###
    ### build might still fail. To check completely:  ###
    ###     divadmin buildAndSmoke                    ###
}

# The sync startup message
set syncMessage {
    ### Syncing local Diva tree to the repository. ###
    ### This is intended for Java packages ONLY.   ###
    ### To check all types of files, run           ###
    ###     cvs -n -q update                       ###
}

#######################################################################
#### confirm
#
# Check that the repository has no major problems. This is
# a confidence check that should be run after every non-trivial
# check-in.
#
proc confirm {DIVA {mailto diva-cvs@gigascale.eecs.berkeley.edu}} {
    global debug silent confirmMessage
    
    # Initialize
    puts $confirmMessage

    # Update or build the check-in copy of the tree
    # It would be nice if we could ignore the doc directory!
    set dir [file join $DIVA Builds Checkin]
    set diva [file join $dir diva]
    if [file exists $diva] {
	updateDiva $DIVA $diva
    } else {
	file mkdir $dir
	checkoutDiva $DIVA $dir
    }
    cd $diva
    puts [pwd]
    puts [exec ls -alg]
    global env
    puts $env(PATH)
    lappend env(PATH) $diva
    puts $env(PATH)
    set foo [file join $diva configure]
    puts $foo
    set foo [file split $foo]
    set foo [lreplace $foo 0 0]
    set foo "/[join $foo /]"
puts [file size $foo]
    puts $foo
    exec /bin/sh $foo
    
    # Check package directories
    puts "Checking packages"
    set s $silent
    set silent 1
    set t [checkPackageDirs $diva $diva]
    set silent $s
    set badPackages ""
    if { $t != "" } {
	set badPackages "Packages that are not in vars.mk:\n"
	foreach p $t {
	    append badPackages "    [divaPath $diva $p]\n"
	}
	append badPackages \n
    }

    # Check makefiles
    puts "Checking makefiles"
    set s $silent
    set silent 1
    set t [checkMakeDirs $diva $diva]
    set silent $s
    set badMakedirs ""
    if { $t != "" } {
	set badMakedirs "Directories that are not in the makefile:\n"
	foreach d $t {
	    append badMakedirs "    [divaPath $diva $d]\n"
	}
	append badMakedirs \n
    }

    # Clean and rebuild
    cd $diva

    ###Don't need to clean, just use dependencies
    ###if !$silent {
    ###	puts "make clean"
    ###}
    ###catch {exec make clean}

    if !$silent {
	puts "Recompiling"
    }
    catch {exec make -k -w packages >& tmpfile}

    # Parse the result and generate error message
    if !$silent {
	puts "Checking compile"
    }
    set errors [parseJavacOutput $DIVA tmpfile]
    if { $errors != "" } {
	set errors "Compile errors:\n$errors"
    }
    file delete tmpfile
    if { $errors == "" && $badPackages == "" && $badMakedirs == "" } {
	set indicator PASSED
	set message "    ### Check-in confirm PASSED. Congratulations! ###"
    } else {
	set indicator FAILED
	set message "    ### Check-in confirm FAILED. Don't go home yet! ###\n"
	append message "    ### The errors are:\n\n"
	append message $badPackages
	append message $badMakedirs
	append message $errors
	append message "    ### Check-in confirm FAILED. Don't go home yet! ###"
    }
    # Dump to stdout
    puts $message

    # Mail results if possible
    if { $mailto != "" } {
	# puts "Mailing result to $mailto"
	set fd [open tmpfile "w"]
	puts $fd $message
	close $fd

	if [catch {
	    exec mailx -s "\[DIVA-BUILD\] Check-in confirm $indicator" \
		    $mailto < tmpfile
	}] {
	    ###puts "(Cannot mail message -- does Windows suck or what?)"
	}
    }
}

#######################################################################
#### sync
#
# Check that the repository is synchronized with the local tree.
# Should be run before/after checking in new stuff.
#
proc sync {DIVA} {
    global debug silent syncMessage syncIgnore

    # Initialize
    set ignore [join $syncIgnore]
    puts $syncMessage

    # Grab the status from CVS using update
    puts "You might need to type your CVS password in now."
    catch {exec cvs -n -q update} status

    # Split into lines and filter
    puts "Processing..."
    set unknown {}
    set unknownDirs {}
    set modified {}
    foreach line [split $status \n] {
	if [regexp {^\? (.*)$} $line _ path] {
	    if [lpatternDisjoint $ignore [file split $path]] {
		if [file isdirectory $path] {
		    lappend unknownDirs $path
		} else {
		    lappend unknown $path
		}
	    }
	} elseif [regexp {^M (.*)$} $line _ path] {
	    if [lpatternDisjoint $ignore [file split $path]] {
		lappend modified $path
	    }
	}
    }

    if { $unknownDirs != "" } {
	puts "The following directories need to be added to the repository:"
	foreach d $unknownDirs {
	    puts "    $d"
	}
    }
    if { $unknown != "" } {
	puts "The following files need to be added to the repository:"
	foreach f $unknown {
	    puts "    $f"
	}
    }
    if { $modified != "" } {
	puts "The following files need to be checked in:"
	foreach f $modified {
	    puts "    $f"
	}
    }
    puts ""
    if { $unknownDirs == "" && $unknown == "" && $modified == "" } {
	puts "    ### Sync PASSED. Please run divadmin confirm now. ###\n"
    } else {
	puts "    #### Sync FAILED. Please correct the above errors. ###\n"
    }

    set junk {	
	if [regexp {^\? (.*\.java)$} $line _ file] {
	    puts "File $file is unknown"
	} elseif [regexp {^\? (.*\makefile)$} $line _ file] {
	    puts "Makefile $file is unknown"
	} elseif [regexp {^M (.*\.java)$} $line _ file] {
	    puts "File $file needs check-in"
	} elseif [regexp {^M (.*\makefile)$} $line _ file] {
	    puts "Makefile $file needs check-in"
	} else {
	    if { [regexp {^\? (.*)$} $line _ dir] \
		    && [file isdirectory $dir] } {
		puts "Directory $dir needs to be added"
	    }
	}
    }
}







