/*
 * $Id: BasicFigure.java,v 1.3 2004/03/12 18:27:54 tims Exp $
 *
 * Copyright (c) 1998-2000 The Regents of the University of California.
 * All rights reserved. See the file COPYRIGHT for details.
 *
 */

package diva.canvas.toolbox;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Composite;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import diva.canvas.AbstractFigure;
import diva.canvas.connector.BoundsSite;
import diva.canvas.interactor.ShapedFigure;
import diva.util.java2d.AbstractPaintedGraphic;
import diva.util.java2d.PaintedShape;
import diva.util.java2d.ShapeUtilities;

/** A BasicFigure is one that contains a single instance of
 *  Shape. The figure can have a fill with optional compositing (for
 *  translucency), and a stroke with a different fill. With this
 *  class, simple objects can be created on-the-fly simply by passing
 *  an instance of java.awt.Shape to the constructor. This class
 *  is  mainly intended for use for closed shapes -- for open shapes,
 *  use the PathFigure class.
 *
 * @version	$Revision: 1.3 $
 * @author 	John Reekie
 * @author      Nick Zamora
 */
public class BasicFigure extends AbstractFigure implements ShapedFigure {

    /** The color compositing operator
     */
    private Composite _composite = AlphaComposite.SrcOver; // opaque
    
     /** The painted shape that we use to draw the connector.
     */
    private AbstractPaintedGraphic _paintedObject;
    
	/** The LabelFigure for this figure(node).
	 */
	private LabelFigure _labelFigure = null;
    
	/** The LabelFigure for the MachineID of this figure(node).
	 */
	private LabelFigure _labelFigureMachineID = null;
    
    /** The BoundsSite - used to attach label.
     */
	private BoundsSite _boundsSite;
    
    private BoundsSite _boundsSiteMachineID;

    /** Create a new figure with the given shape. The figure, by
     *  default, has a unit-width continuous black outline and no fill.
     */
    public BasicFigure (Shape shape) {
        super();
        _paintedObject = new PaintedShape(shape);
    }

    /** Create a new figure with the given shape and outline width.
     * It has no fill. The default outline paint is black.
     */
    public BasicFigure (Shape shape, float lineWidth) {
        super();
        _paintedObject = new PaintedShape(shape, null, lineWidth);
    }

    /** Create a new figure with the given painted shape.
     */
    public BasicFigure (AbstractPaintedGraphic painted_object) {
        super();
	_paintedObject = painted_object;
    }

    /** Create a new figure with the given paint pattern. The figure,
     *  by default, has no stroke.
     */
    public BasicFigure (Shape shape, Paint fill) {
        super();
        _paintedObject = new PaintedShape(shape, fill);
    }

    /** Create a new figure with the given paint pattern and outline width.
     * The default outline paint is black.
     */
    public BasicFigure (Shape shape, Paint fill, float lineWidth) {
        super();
        _paintedObject = new PaintedShape(shape, fill, lineWidth);
    }

    /** Get the bounding box of this figure. This method overrides
     * the inherited method to take account of the thickness of
     * the stroke and/or LabelFigure, if there is one.
     * @author Modified by Maylene Waltz - WPI
     */
    public Rectangle2D getBounds () {
    	//REMOVED by Maylene Waltz - WPI
		//return _paintedObject.getBounds();
		
		//ADDED by Maylene Waltz - WPI
		Rectangle2D bounds = (Rectangle2D)_paintedObject.getBounds().clone();
		if (_labelFigure != null) {
			Rectangle2D.union(bounds, _labelFigure.getBounds(), bounds);
		}
		if (_labelFigureMachineID != null) {
			Rectangle2D.union(bounds, _labelFigureMachineID.getBounds(), bounds);
		}
		return bounds;
    }

    /** Get the color composition operator of this figure.
     */
    public Composite getComposite () {
        return _composite;
    }

    /** Get the fill paint pattern of this figure if this figure
     *	represents a shape with a fill paint pattern, otherwise
     *	return null.
     */
    public Paint getFillPaint () {
        if (_paintedObject instanceof PaintedShape) {
	    return ((PaintedShape)_paintedObject).fillPaint;
	}
	else {
	    return null;
	}
    }

	/** Get the LabelFigure for this figure.
	 * @author Maylene Waltz - WPI
	 */
	public LabelFigure getLabelFigure(){
		return _labelFigure;
	}
	
	/** Get the LabelFigure for the MachineID of this figure.
	 * @author Maylene Waltz - WPI
	 */	
	public LabelFigure getLabelFigureMachineID(){
		return _labelFigureMachineID;
	}
	
	/** Get the AbstractPaintedGraphic that paints this figure
	 * @author Maylene Waltz - WPI
	 */
	public AbstractPaintedGraphic getPaintedObject(){
		return _paintedObject;
	}
	
    /** Get the line width of this figure.
     */
    public float getLineWidth () {
        return _paintedObject.getLineWidth();
    }

    /** Get the shape of this figure.
     */
    public Shape getShape () {
        return _paintedObject.shape;
    }

    /** Get the stroke of this figure.
     */
    public Stroke getStroke () {
        return _paintedObject.getStroke();
    }

    /** Get the stroke paint pattern of this figure.
     */
    public Paint getStrokePaint () {
        return _paintedObject.strokePaint;
    }

    /** Test if this figure intersects the given rectangle. If there
     * is a fill but no outline, then there is a hit if the shape
     * is intersected. If there is an outline but no fill, then the
     * area covered by the outline stroke is tested. If there
     * is both a fill and a stroke, the region bounded by the outside
     * edge of the stroke is tested. If there is neither a fill nor
     * a stroke, then return false. If the figure is not visible,
     * always return false.
     */
    public boolean hit (Rectangle2D r) {
        if (!isVisible()) {
             return false;
        }
        
        //ADDED by Maylene Waltz - WPI
		boolean hit = _paintedObject.intersects(r);
		if (_labelFigure != null) {
			hit = hit || _labelFigure.intersects(r);
		}
		if (_labelFigureMachineID != null) {
			hit = hit || _labelFigureMachineID.intersects(r);
		}		
		return hit;
		
        //return _paintedObject.hit(r);
    }

	/** Test if this figure intersects the given rectangle. This default
	  *  implementation checks to see if the rectangle intersects with the
	  *  figure or the label.
	  * @author Maylene Waltz - WPI
	  */
	public boolean intersects(Rectangle2D r){
		boolean hit = _paintedObject.intersects(r);
		if (_labelFigure != null) {
			hit = hit || _labelFigure.intersects(r);
		}
		if (_labelFigureMachineID != null) {
			hit = hit || _labelFigureMachineID.intersects(r);
		}
		return hit;
	}
	
    /** Paint the figure. The figure is redrawn with the current
     *  shape, fill, and outline.
     */
    public void paint (Graphics2D g) {
        if (!isVisible()) {
             return;
        }
        if (_composite != null) {
            g.setComposite(_composite);
        }
        _paintedObject.paint(g);
		if (_labelFigure != null) {
			_labelFigure.paint(g);
		}
		if (_labelFigureMachineID != null) {
			_labelFigureMachineID.paint(g);
		}
    }
    
	/** Tell the figure to reposition its label if it has one.
	 * The label is currently only positioned at the top of the figure.
	 */
	public void repositionLabel () {
		if (getLabelFigure() != null) {
			Point2D pt = _boundsSite.getPoint();
			getLabelFigure().translateTo(pt);
			getLabelFigure().autoAnchor(getPaintedObject().shape);
		}
		if (getLabelFigureMachineID() != null) {
			Point2D pt = _boundsSiteMachineID.getPoint();
			getLabelFigureMachineID().translateTo(pt);
			getLabelFigureMachineID().autoAnchor(getPaintedObject().shape);
		}
	}
	
    /** Set the color composition operator of this figure. If the
     * composite is set to null, then the composite will not be
     * changed when the figure is painted. By default, the composite
     * is set to opaque.
     */
    public void setComposite (Composite c) {
        _composite = c;
        repaint();
    }

    /** Set the fill paint pattern of this figure. The figure will be
     *  filled with this paint pattern. If no pattern is given, do not
     *  fill it.
     */
    public void setFillPaint (Paint p) {
        if (_paintedObject instanceof PaintedShape) {
	    ((PaintedShape)_paintedObject).fillPaint = p;
	}
	repaint();
    }
	
	/** Set the LabelFigure for the OperatorID of this figure.
	 * @author Maylene Waltz - WPI
	 */
	public void setLabelFigure(LabelFigure label, BoundsSite boundsSite){
		_labelFigure = label;
		_boundsSite = boundsSite;
	}
	
	/** Set the LabelFigure for the MachineID of this figure.
	 * @author Maylene Waltz - WPI
	 */
	public void setLabelFigureMachineID(LabelFigure label, BoundsSite boundsSite){
		_labelFigureMachineID = label;
		_boundsSiteMachineID = boundsSite;
	}
	
   /** Set the line width of this figure. If the width is zero,
    * then the stroke will be removed.
    */
    public void setLineWidth (float lineWidth) {
      repaint();
      _paintedObject.setLineWidth(lineWidth);
      repaint();
    }

    /** Set the shape of this figure.
     */
    public void setShape (Shape s) {
        repaint();
        _paintedObject.shape = s;
        repaint();
    }

    /** Set the stroke of this figure.
     */
    public void setStroke (BasicStroke s) {
        repaint();
        _paintedObject.stroke = s;
        repaint();
    }

    /** Set the stroke paint pattern of this figure.
     */
    public void setStrokePaint (Paint p) {
        _paintedObject.strokePaint = p;
	repaint();
    }

    /** Transform the figure with the supplied transform. This can be
     * used to perform arbitrary translation, scaling, shearing, and
     * rotation operations. As much as possible, this method attempts
     * to preserve the type of the shape: if the shape of this figure
     * is an of RectangularShape or Polyline, then the shape may be
     * modified directly. Otherwise, a general transformation is used
     * that loses the type of the shape, converting it into a
     * GeneralPath.
     */
    public void transform (AffineTransform at) {
        repaint();
        _paintedObject.shape = ShapeUtilities.transformModify(
                _paintedObject.shape, at);
        //ADDED by Maylene Waltz - WPI
      	if(getLabelFigure() != null){
      		ShapeUtilities.transformModify(getLabelFigure().getShape(), at);
      	}
		if(getLabelFigureMachineID() != null){
			ShapeUtilities.transformModify(getLabelFigureMachineID().getShape(), at);
		}
	repaint();
    }

    /** Translate the figure with by the given distance.
     * As much as possible, this method attempts
     * to preserve the type of the shape: if the shape of this figure
     * is an of RectangularShape or Polyline, then the shape may be
     * modified directly. Otherwise, a general transformation is used
     * that loses the type of the shape, converting it into a
     * GeneralPath.
     */
    public void translate (double x, double y) {
	repaint();
        _paintedObject.shape = ShapeUtilities.translateModify(_paintedObject.shape, x, y);
                
    	//ADDED by Maylene Waltz - WPI
    	if(getLabelFigure() != null){
    		//getLabelFigure().translate(x,y);
    		repositionLabel();
    	}
		if(getLabelFigureMachineID() != null){
			repositionLabel();
		}	
	repaint();
    }
}


