/*
 * $Id: BasicGraphModel.java,v 1.7 2004/04/27 03:54:08 mnwaltz Exp $
 *
 * Copyright (c) 2000 The Regents of the University of California.
 * All rights reserved.  See the file COPYRIGHT for details.
 */
package diva.graph.basic;
import java.util.ArrayList;
import java.util.Iterator;

import diva.canvas.toolbox.LabelFigure;
import diva.graph.modular.BasicModularGraphModel;
import diva.graph.modular.CompositeNode;
import diva.graph.modular.Edge;
import diva.graph.modular.Graph;
import diva.graph.modular.Node;
import diva.util.ArrayIterator;
import diva.util.BasicPropertyContainer;
import diva.util.SemanticObjectContainer;

/**
 * A basic implementation of a mutable graph model that stores its
 * graph structure as a collection of nodes and edges (as opposed to
 * an adjacency matrix).  This class is a good reference class for how
 * to use the diva.graph.modular classes and will suffice as a data
 * structure for simple, throw-away applications (a heavy-weight
 * application like a schematic editor will probalby want to implement
 * things differently.
 * 
 * @author Michael Shilman  (michaels@eecs.berkeley.edu)
 * @version $Revision: 1.7 $
 * @rating Red
 */
public class BasicGraphModel extends BasicModularGraphModel {
	/** The root nodes of the graph.
	 * Added by: Maylene Waltz - WPI
	 */
	private ArrayList _rootNodes = new ArrayList();
	
	/** All the nodes in the graph.
	 * Added by: Maylene Waltz - WPI
	 */
	private ArrayList _nodes = new ArrayList();
	
	/** All the edges in the graph.
	 * Added by: Maylene Waltz - WPI
	 */
	private ArrayList _edges = new ArrayList();
	 
    /**
     * Construct an empty graph model.
     */
    public BasicGraphModel() {
        super(new BasicCompositeNode(null));
    }

    /**
     * Make a new composite node object.
     */
    public CompositeNode createComposite(Object semanticObject) {
        return new BasicCompositeNode(semanticObject);
    }

    /**
     * Make a new edge object.
     */
    public Edge createEdge(Object semanticObject) {
    	Edge edge = new BasicEdge(semanticObject);
    	_edges.add(edge);
        return edge;
        //return new BasicEdge(semanticObject);
    }
    
    /**
     * Make a new node object.
     */
    public Node createNode(Object semanticObject) {
    	Node node = new BasicNode(semanticObject);
    	_nodes.add(node);
    	return node;
        //return new BasicNode(semanticObject);
    }
    
    /** Adds the given <code>root</code> to the <code>ArrayList</code> of root nodes. */
    public void addRootNode(Node root){
    	_rootNodes.add(root);
    }
    
    /** Returns an <code>Iterator</code> of all the graph's root nodes. */
 	public Iterator getRootNodes(){
		return _rootNodes.iterator();
 	}
 	
	/** Returns an <code>Iterator</code> of all the nodes in the graph. */
 	public Iterator getNodes(){
 		return _nodes.iterator();
 	}
 	
 	/** Returns an <code>Iterator</code> of all the edges in the graph. */
 	public Iterator getEdges(){
 		return _edges.iterator();
 	}
    
    /**
     * A class that represents an intermediary
     * between a semantic object and a visual
     * representation.  
     */
    private static abstract class Intermediate extends BasicPropertyContainer
        implements SemanticObjectContainer {
        /**
         * The semantic object of this intermediate.
         */
        private Object _semanticObject = null;

        /**
         * Set the semantic object of this object, i.e.  the
         * application object that is semantically equivalent to this
         * node.
         */
        public Object getSemanticObject() {
            return _semanticObject;
        }
    
        /**
         * Set the semantic object of this object, i.e.  the
         * application object that is semantically equivalent to this
         * node.
         */
        public void setSemanticObject(Object o) {
            _semanticObject = o;
        }
    }

    /**
     * A simple node implementation.
     */
    private static class BasicNode extends Intermediate implements Node {
        /**
         * The edges <b>into</b> this node.
         */
        private ArrayList _in = new ArrayList();

        /**
         * The edges <b>out of</b> this node.
         */
        private ArrayList _out = new ArrayList();
        
        /**
         * The parent of this node.
         */
        private CompositeNode _parent = null;
        
		/** The <code>LabelFigure</code> of this node.
		 * ADDED by Maylene Waltz - WPI
		 */
		private LabelFigure _labelFigure;
		
		//ADDED by Maylene Waltz - WPI
		private LabelFigure _labelFigureMachineID;
        
        /** The ToolTipText of this node.
         * ADDED by Maylene Waltz - WPI
         */
        private String _toolTipText;
        
        /**
         * Create a node containing some user data.
         */
        public BasicNode(Object userObject) {
            this.setSemanticObject(userObject);
        }

        public void addInEdge(Edge e) {
            _in.add(e);
        }

        public void addOutEdge(Edge e) {
            _out.add(e);
        }
        
        public Graph getParent() {
            return _parent;
        }
        
        /**
         * Return an iterator over the <i>in</i> edges of this
         * node. This iterator does not support removal operations.
         */
        public Iterator inEdges() {
            return new ArrayIterator(_in.toArray());
        }

        /**
         * Return an iterator over the source, or "in", nodes of this
         * node. This iterator does not support removal operations.
         * This method provides a convenient way of traversing a
         * graph.
         */
        public Iterator inNodes() {
            return new diva.util.IteratorAdapter () {
                Iterator edges = inEdges();
                public boolean hasNext() {
                    return edges.hasNext();
                }
                public Object next() {
                    return ((Edge)edges.next()).getTail();
                }
            };
        }

        /**
         * Return an iterator over the <i>out</i> edges of this
         * node.  This iterator does not support removal operations.
         */
        public Iterator outEdges() {
            return new ArrayIterator(_out.toArray());
        }
 
        /**
         * Return an iterator over the sink, or "out", nodes of this
         * node. This iterator does not support removal operations.
         * This method provides a convenient way of traversing a
         * graph.
         */
        public Iterator outNodes() {
            return new diva.util.IteratorAdapter () {
                Iterator edges = outEdges();
                public boolean hasNext() {
                    return edges.hasNext();
                }
                public Object next() {
                    return ((Edge)edges.next()).getHead();
                }
            };
        }

        public void removeInEdge(Edge e) {
            _in.remove(e);
        }

        public void removeOutEdge(Edge e) {
            _out.remove(e);
        }

        public void setParent(Graph parent) {
            if(_parent != null) {
                ((BasicCompositeNode)_parent).remove(this);
            }
            _parent = (BasicCompositeNode)parent;
            if(_parent != null) {
                ((BasicCompositeNode)_parent).add(this);
            }
        }

        public String toString() {
            Object o = this.getSemanticObject();
            return "BasicNode[" + o + "]";
        }
        
        //ADDED by Maylene Waltz - WPI
		public LabelFigure getLabelFigure(){
			return _labelFigure;
		}
		
		//ADDED by Maylene Waltz - WPI
		public LabelFigure getLabelFigureMachineID(){
			return _labelFigureMachineID;
		}
		
		//ADDED by Maylene Waltz - WPI
		public String getToolTipText(){
			return _toolTipText;
		}
		
		//ADDED by Maylene Waltz - WPI
		public void setLabelFigure(LabelFigure label){
			_labelFigure = label;
		}
		
		//ADDED by Maylene Waltz - WPI
		public void setLabelFigureMachineID(LabelFigure label){
			_labelFigureMachineID = label;
		}
		
		//ADDED by Maylene Waltz - WPI
		public void setToolTipText(String tip){
			_toolTipText = tip;
		}
    }

    /**
     * A simple composite node implementation.
     */
    private static class BasicCompositeNode extends BasicNode 
	implements CompositeNode {
        /**
         * The nodes that this composite node
         * contains.
         */
        private ArrayList _nodes = new ArrayList();

        /**
         * Construct an empty composite nodes with
         * the given semantic object.
         */
        public BasicCompositeNode(Object userObject) {
            super(userObject);
        }

        public void add(Node n) {
            _nodes.add(n);
        }

        public boolean contains(Node n) {
            return _nodes.contains(n);
        }


        public int getNodeCount() {
            return _nodes.size();
        }

        public Node getNode(int i) {
            return (Node)_nodes.get(i);
        }

        public int getIndex(Node n) {
            return _nodes.indexOf(n);
        }

        public Iterator nodes() {
            return _nodes.iterator();
        }

        public void remove(Node n) {
            _nodes.remove(n);
        }

        public String toString() {
            Object o = this.getSemanticObject();
            return "CompositeNode[" + o + "]";
        }
    }

    /**
     * A basic implementation of the Edge interface.
     */
    private static class BasicEdge extends Intermediate implements Edge {
        /**
         * Whether or not this edge is directed.
         */
        private boolean _directed = true;

        /**
         * The head of the edge.
         */
        private Node _head = null;

        /**
         * The tail of the edge.
         */
        private Node _tail = null;

        /**
         * The weight on this edge.
         */
        private double _weight = 1.0;
		
		/** The LabelFigure of this edge.
		 * ADDED by Maylene Waltz - WPI
		 */
		private LabelFigure _labelFigure;
		
		/** The ToolTipText of this node.
		 * ADDED by Maylene Waltz - WPI
		 */
		private String _toolTipText;

        /**
         * Create a new edge with the specified user object
         * but no tail or head.
         */
        public BasicEdge(Object userObject) {
            this(userObject, null, null);
        }

        /**
         * Create a new edge with the specified user object
         * and tail/head
         */
        public BasicEdge(Object userObject, Node tail, Node head) {
            this.setSemanticObject(userObject);
            setTail(tail);
            setHead(head);
        }
		
        public boolean acceptHead(Node head) {
            return true;
        }

        public boolean acceptTail(Node tail) {
            return true;
        }

        public Node getHead() {
            return _head;
        }
    
        public Node getTail() {
            return _tail;
        }
        
        //ADDED by Maylene Waltz - WPI
        public String getToolTipText(){
        	return _toolTipText;
        }
    	
    	//ADDED by Maylene Waltz - WPI
    	public LabelFigure getLabelFigure(){
    		return _labelFigure;
    	}
    	
        public double getWeight() {
            return _weight;
        }
    
        public boolean isDirected() {
            return _directed;
        }
    
        public void setDirected(boolean val) {
            _directed = val;
        }

        public void setHead(Node n) {
            if(_head != null) {
                ((BasicNode)_head).removeInEdge(this);
            }
            _head = n;
            if(_head != null) {
                ((BasicNode)_head).addInEdge(this);
            }
        }
        
        public void setTail(Node n) {
            if(_tail != null) {
                ((BasicNode)_tail).removeOutEdge(this);
            }
            _tail = n;
            if(_tail != null) {
                ((BasicNode)_tail).addOutEdge(this);
            }
        }
        
        //ADDED by Maylene Waltz - WPI
		public void setLabelFigure(LabelFigure label){
			_labelFigure = label;
		}
		
		//ADDED by Maylene Waltz - WPI
		public void setToolTipText(String text){
			_toolTipText = text;
		}
		
        public void setWeight(double weight ) {
            _weight = weight;
        }

        /** Print a readable description of this edge
         */
        public String toString() {
            Object o = this.getSemanticObject();
            return "Edge[" + o + "]";
        }
    }
}

