/*
 * $Id: BasicNodeRenderer.java,v 1.7 2006/01/05 04:41:15 rimma3 Exp $
 *
 * Copyright (c) 1998-2000 The Regents of the University of California.
 * All rights reserved. See the file COPYRIGHT for details.
 */
package diva.graph.basic;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;

import javax.swing.SwingConstants;

import diva.canvas.CompositeFigure;
import diva.canvas.Figure;
import diva.canvas.connector.BoundsSite;
import diva.canvas.toolbox.BasicFigure;
import diva.canvas.toolbox.LabelFigure;
import diva.graph.GraphController;
import diva.graph.GraphModel;
import diva.graph.NodeRenderer;
import diva.graph.modular.Node;
import diva.util.java2d.ShapeUtilities;

/**
 * A factory which creates and returns a NodeFigure given a node input
 * to render.
 *
 * @author  Michael Shilman  (michaels@eecs.berkeley.edu)
 * @version $Revision: 1.7 $
 * @rating  Red
 */
public class BasicNodeRenderer implements NodeRenderer {
    /**
     * The graph controller
     */
    private GraphController _controller;
      
    /**
     * The shape for nodes.
     */
    private Shape _nodeShape = null;

    /**
     * The shape for composite nodes.
     */
    private Shape _compositeShape = null;

    /**
     * The scaling factor for composite nodes.
     *
     * @see #setCompositeScale(double)
     */
    private double _compositeScale = 0;

    /**
     * The fill paint for nodes.
     */
    private Paint _nodeFill = null;

    /**
     * The fill paint for composite nodes.
     */
    private Paint _compositeFill = null;
    
    //ADDED by Maylene Waltz
    private BoundsSite _boundsSite = null;
    private BoundsSite _boundsSiteMachineID = null;
    
    /** The color of the label's text. Default color is BLACK.
     */
    private Paint _labelTextColor = Color.black;
    
    /** The text size of the label. Default is 12.
     */
    private int _labelTextSize = 12;

    /**
     * Create a renderer which renders nodes square and specified color.
     */
    public BasicNodeRenderer(GraphController controller) {
        this(controller, new Rectangle2D.Double(0.0,0.0,10.0,10.0),
                new Rectangle2D.Double(0.0,0.0,10.0,10.0),
                Color.red, Color.gray, 1);
    }

    /**
     * Create a renderer which renders nodes using the
     * given shape and fill paint.  The given shape must be
     * cloneable.
     */
    public BasicNodeRenderer(GraphController controller,
            Shape nodeShape, Shape compositeShape,
            Paint nodeFill, Paint compositeFill, double compositeScale) {
        _controller = controller;
        setNodeShape(nodeShape);
        setNodeFill(nodeFill);
        setCompositeShape(compositeShape);
        setCompositeFill(compositeFill);
        setCompositeScale(compositeScale);
    }

    /**
     * Return the fill that composites are painted with.
     */
    public Paint getCompositeFill() {
        return _compositeFill;
    }

    /**
     * Return the scaling factor for the composite nodes
     *
     * @see #setCompositeScale(double)
     */
    public double getCompositeScale() {
        return _compositeScale;
    }

    /**
     * Return the shape that composites are rendered in.
     */
    public Shape getCompositeShape() {
        return _compositeShape;
    }

    /**
     * Return the graph controller.
     */
    public GraphController getGraphController(){
        return _controller;
    }

    /**
     * Return the fill that nodes are painted with.
     */
    public Paint getNodeFill() {
        return _nodeFill;
    }

    /**
     * Return the shape that nodes are rendered in.
     */
    public Shape getNodeShape() {
        return _nodeShape;
    }

	//ADDED by Maylene Waltz
	public Paint getLabelTextColor(){
		return _labelTextColor;
	}
	
	public int getLabelTextSize(){
		return _labelTextSize;
	}
	
    /**
     * Return the rendered visual representation of this node.
     */
    public Figure render(Object node) 
    {
        GraphModel model = _controller.getGraphModel();
        
        //ADDED by Maylene Waltz - WPI
        Node nodePeer = (Node)node;
        
        Shape shape = (model.isComposite(node)) ? _compositeShape : _nodeShape;
        if(shape instanceof RectangularShape) {
            RectangularShape r = (RectangularShape)shape;
            shape = (Shape)(r.clone());
        }
        else {
            shape = new GeneralPath(shape);
        }
        Paint fill = model.isComposite(node) ? _compositeFill : _nodeFill;

        BasicFigure bf = new BasicFigure(shape);
        bf.setFillPaint(fill);
	
	if(model.isComposite(node)) {
            CompositeFigure rep = new CompositeFigure(bf);
	    double scale = getCompositeScale();
            rep.getTransformContext().getTransform().scale(scale, scale);
	    return rep;
        }

        bf.setToolTipText(nodePeer.getToolTipText());
        
		//ADDED by Maylene Waltz - WPI
		_boundsSite = new BoundsSite(bf, 1, 1, 50.0);
		_boundsSiteMachineID = new BoundsSite(bf, 2, 5, 50.0);
		
		if(nodePeer.getLabelFigure() != null){
        	bf.setLabelFigure(nodePeer.getLabelFigure(), _boundsSite);
        	if(bf.getLabelFigure() != null){
        		bf.getLabelFigure().setFillPaint(_labelTextColor);
        		bf.getLabelFigure().setSize(_labelTextSize);
        	}
		}
		if(nodePeer.getLabelFigureMachineID() != null){
			bf.setLabelFigureMachineID(nodePeer.getLabelFigureMachineID(), _boundsSiteMachineID);
			if(bf.getLabelFigureMachineID() != null){
				bf.getLabelFigureMachineID().setFillPaint(_labelTextColor);
				bf.getLabelFigureMachineID().setSize(_labelTextSize);
			}
		}
		
		//-----------------------------
		// SECURITY SHIELD
		//-----------------------------
		//Added by RIMMA:
		if ((String)model.getProperty(node, "isSS") == "true")
        {
			//bf.setShape(new Ellipse2D.Double(0.0,0.0,15.0,15.0));
			bf.setShape(ShapeUtilities.createShieldShape());
			bf.setFillPaint(Color.YELLOW);
        }

		//-----------------------------
		// SECURITY SHIELD
		//-----------------------------
		//Added by RIMMA:
		if ((String)model.getProperty(node, "isSSPlus") == "true")
        {
			//bf.setShape(new Ellipse2D.Double(0.0,0.0,15.0,15.0));
			bf.setShape(ShapeUtilities.createShieldShape());
			bf.setStroke(new BasicStroke (5.0F, BasicStroke.CAP_BUTT, 
                    BasicStroke.JOIN_BEVEL));
			bf.setStrokePaint(Color.YELLOW);
			bf.setFillPaint(Color.ORANGE);
        }
		
        return bf;
    }

    /**
     * Set the fill to paint the composites with.
     */
    public void setCompositeFill(Paint p) {
        _compositeFill = p;
    }

    /**
     * Set the scaling factor for the composite nodes.
     * Given factor must be greater than 0 and less than
     * or equal to 1.
     *
     * (XXX document this).
     */
    public void setCompositeScale(double scale) {
        if((scale <= 0) || (scale > 1)) {
            String err = "Scale must be between > 0 and <= 1.";
            throw new IllegalArgumentException(err);
        }
        _compositeScale = scale;
    }

    /**
     * Set the shape for composites to be rendered in.  The
     * shape must implement Cloneable.
     */
    public void setCompositeShape(Shape s) {
        _compositeShape = s;
    }

    /**
     * Set the fill to paint the nodes with.
     */
    public void setNodeFill(Paint p) {
        _nodeFill = p;
    }
        
    /**
     * Set the shape for nodes to be rendered in.  The
     * shape must implement Cloneable.
     */
    public void setNodeShape(Shape s) {
        _nodeShape = s;
    }

	//ADDED by Maylene Waltz
	public void setLabelTextColor(Paint p){
		_labelTextColor = p;
	}
	
	public void setLabelTextSize(int size){
		_labelTextSize = size;
	}
}


