package edu.wpi.cs.dsrg.xmldb.xat.common.operator;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

/**
 * This is the top level interface for statistics gathering.
 * This interface is generalized for all statistics gathering
 * and provides a standard means to access properties. These properties
 * will be set by either by the query objects themselves and accessed
 * by outside components (Cost Model, ExecutionEngine, etc) or vice versa.
 *
 * These properties will be used to retrieve information about a query object
 * such as its throughput -or- it will be used by an outside component
 * to customize how the object should run, such as how much work it should do.
 *
 * All property values are strings, although they can be converted into
 * ints or doubles, depending on the property.  Each implementation of this
 * interface should provide validation of a given property in the setProperty method.
 * Furthermore, each implementation is responsible for setting the default
 * properties that it will support.
 * 
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0    
 */
public interface XATProperties {

    /**
    * Returns an enumeration of property names supported by
    * this query object
    * @return an Enumeration
    * @since   1.0       
    */
    java.util.Enumeration getProperties();

    /**
    * Retrieve the value of the property with the given name.
    * @param name the name of the property to return
    * @return String the value of the property
    * @throws IllegalArgumentException if the given property cannot be found
    * @since   1.0       
    */
    String getProperty(String name);

    /**
     * Checks if the given property is valid for this QueryObject.
     * @param nameToCheck is the name to check   
     * @return boolean true if the property is supported by this object, false otherwise.
     * @since   1.0            
     */
    public boolean isValidPropertyName(String nameToCheck);

    /**
    * Sets the given property to the new value.  This method is responsible
    * for validating the given property.  An exception is thrown if
    * the value is not correct.
    * Validation should be done in the following manner: 
    * First the method should check if the object supports the given property by calling
    * <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
    * be thrown. <p>
    *
    * Once it is known that the object supports the property, the value must be validated.  Each implementor of
    * XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
    * For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
    * has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
    * to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
    * will reduce duplicate code. <p>
    *
    * When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
    * <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
    * >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
    * it is up to the subclass that created the property to validate it (as described above).
    * @param name the name of the property to set
    * @param value the new value of the property
    * @throws Exception if the given property cannot be found or
    * if the value is not valid for the given property.
    * @since   1.0           
    */
    void setProperty(String name, String value) throws Exception;
}