package edu.wpi.cs.dsrg.xmldb.xat.common.operator;
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Properties;
/**
 * Provides an abstract implementation of the XATProperties interface.  
 * Any type that wants to make use of Properties should derive from this type.  Note,
 * in order to make use of the StatisticsGatherer and Statistics the type should derive from XATQueryObjectImp
 * b/c this type does not provide an interface to the StatisticsGatherer.  <p>
 * The setProperty method is abstract so that implementing classes can provide some validation. This
 * class provides some basic validation- ensuring values are non zero.  If more is required,
 * it should be provided by the lower classes. <p> 
 * Each class that extends this class should set the initial properties in the 
 * setDefaultProperties method using addProperty. <p>
 *
 * The properties are used to configure the various objects in the system and this class
 * provides a standard means to access the properties.  The properties only support Strings for both name
 * and value.
 *
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0        
 */
public abstract class XATPropertiesImp implements XATProperties {
    /** the class variable that actually stores the properties */
    private java.util.Properties properties;

    /**
     * Create a new Properties, and calls setDefaultProperties().
     * @since   1.0            
     */
    public XATPropertiesImp() {
        super();
        properties = new Properties();
        setDefaultProperties();
    }

    /**
     * Adds a new property that will be supported with a given
     * default value.  
     * @param name java.lang.String the name of the property
     * @param defaultValue java.lang.String its default value.
     * @since   1.0            
     */
    protected void addProperty(String name, String defaultValue) {
        /** this method is not the same as this.setProperty */
        properties.setProperty(name, defaultValue);
    }

    /**
    * Returns an enumeration of property names supported by
    * this query object
    * @return an Enumeration
    * @since   1.0           
    */
    public java.util.Enumeration getProperties() {
        return properties.keys();
    }

    /**
    * retrieve the value of the property with the given name.
    * @param name the name of the property to return
    * @return String the value of the property
    * @throws IllegalArgumentException if the given property cannot be found
    * @since   1.0           
    */
    public String getProperty(String name) {
        String value = properties.getProperty(name);
        if (value == null) {
            throw new IllegalArgumentException(
                "The property named " + name + " is not supported by this object");
        }
        return value;
    }

    /**
    * Checks if the given property is valid for this QueryObject.  
    * @return boolean true if the property is supported by this object, false otherwise.
    * @since   1.0            
    */
    public boolean isValidPropertyName(String nameToCheck) {
        return properties.containsKey(nameToCheck);
    }

    /**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    * @since   1.0           
    */
    protected void setDefaultProperties() {
    }

    /**
    * This method actually sets the property value.  Because the properties variable
    * is private, no other class may actually access it.  Therefore, this method is
    * needed so that sub classes may update the internal values.  
    * @param name java.lang.String
    * @param value java.lang.String
    * @since   1.0            
    */
    protected void setNewPropertyValue(String name, String value) {
        properties.setProperty(name, value);
    }

    /**
    * sets the given property to the new value.  This method is responsible
    * for validating the given property.  An exception is thrown if
    * the value is not correct.
    * Validation should be done in the following manner: 
    * First the method should check if the object supports the given property by calling
    * <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
    * be thrown. <p>
    *
    * Once it is known that the object supports the property, the value must be validated.  Each implementor of
    * XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
    * For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
    * has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
    * to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
    * will reduce duplicate code. <p>
    *
    * When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
    * <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
    * >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
    * it is up to the subclass that created the property to validate it (as described above).
    * @param name the name of the property to set
    * @param value the new value of the property
    * @throws Exception if the given property cannot be found or
    * if the value is not valid for the given property.
    * @since   1.0           
    */
    public void setProperty(String name, String value) throws Exception {

        if (isValidPropertyName(name)) {
            /*
            try {
                //ok, then try a double			
                Double d1 = new Double(value);
                if (d1.compareTo(new Double(0)) >= 0) {
                    setNewPropertyValue(name, value);
                } else {
                    throw new IllegalArgumentException(
                        "The property named " + name + " requires a non negative number value ");
                }
            
            } catch (NumberFormatException e3) {
                //hrm, ok, must be a String
                setNewPropertyValue(name, value);
            }
            */
            setNewPropertyValue(name, value);
        } else {
            throw new Exception(
                "The object " + this +" does not support the property named " + name);
        }
    }
}