package edu.wpi.cs.dsrg.xmldb.xat.common.treestructure;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Vector;

import edu.wpi.cs.dsrg.utils.BindingTable;
import edu.wpi.cs.dsrg.utils.XATLeaf;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTreeNodeHasNoParentException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.Visitable;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/** 
 * This class defines an order un-sensitive XML algebra tree.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public interface XATTree extends Visitable, XATQueryObject {
    /**
     * Add a new parent node for a given DOMPatternNode node. 
     * The parent node is updated to have a link to the child node.
     * The child node is updates to have a link to the parent node.
     * @return void
     * @param parent The parent node it attached to.
     * @param child The child node attached.
     * @since 1.0
     */
    void addNode(XATNode child, XATNode newparent);

    /**
     * For all the leaves, add the subtree to it.
     * Note: We clone the subtree, and then added it to the tree.
     * Hence, further modifications will not by automatically updated.
     * Make sure create the subtree first, then added to the tree.
     * @return void
     * @param childTree The XATTree attached.
     * @since 1.0
     */
    public void addXATTree(XATTree childTree);

    /**
     * Append a XAT Node to the leaves of the XATTree.
     * @return void
     * @param XATNode The XATNode append as leaf of XATTree.
     * @since 1.0
     */
    void appendXATNode(XATNode leave);

    /**
     * Clone the XATTree.
     * @return Object The cloned tree.
     * @since 1.0
     */
    public Object clone();

    /**
     * Get all the leaves of this XATTree.
     * @return XATNode[] An array of XATNodes those are the leaf nodes.
     * @since 1.0     
     */
    XATNode[] convertToArray();

    /**
     * Delete node. It will affect number of nodes. All the children nodes of the
     * to-be-deleted node will be the children of the parent node of the to-be-delete node.
     * If the to-be-deleted node is the root node of the XATTree, the exception XATTreeNodeHasNoParentException
     * will the throwed. Instead, the root node should be changed by setRoot() method.
     * @return void
     * @param node the to-be-deleted child node.
     * @since 1.0     
     */
    void deleteNode(XATNode Node) throws XATTreeNodeHasNoParentException;

    /**
     * Get all the leaf node of given XATTree, which is rooted at given root node.
     * <pre>
     * This method is used to find all of the Source operators in an XATTree
     * Looks at all of the root's children and adds them to the vector
     * if that nodes' operator is an XML source .  The function is then
     * recurvsively called on each of root's children until there are no more nodes 
     * to look at.  
     * </pre>
     * @return Vector of all the leaf nodes including the subtree of FOR and GB.
     * @param root The root of a tree to iterate over
     * @see getLeaves();
     * @since 1.0     
     */    
    Vector getAllLeafNodes(XATNode root);
    
    /**
     * Get the binding table of this Tree.
     * @return BindingTable The binding table of XATTree.
     * @since 1.0
     */
    BindingTable getBindingTable();
    
    /**
     * Get all the leaves of this XATTree.
     * @return XATNode[] An array of XATNodes those are the leaf nodes.
     * @since 1.0
     */
    XATNode[] getLeaves();
    
    /**
     * Get all the leaves of this XATTree along with their depth.
     * @return XATLeaf[] An array of XATNodes those are the leaf nodes.
     * @since 1.0
     */
    XATLeaf[] getLeavesWithDepth();
    
    /** 
     * Get the root node of this XATTree.
     * @return XATNode The root of XATTree.
     * @since 1.0
     */
    XATNode getRoot();
    
    /**
     * Assuming the XATTree is a forest sometime, this method get all the roots of XAT.
     * @return java.util.Iterator All the roots of given XAT.
     * @since 1.0     
     */
    java.util.Iterator getRoots();
    
    /**
     * Insert a new node as the root of this tree.
     * @return void
     * @param root The root will be inserted in.
     * @since 1.0     
     */
    void insertRoot(XATNode root);
    
    /**
     * Move node. It will affect three nodes, the original parent node A, the child
     * node B and the new parent node C. The A's children is updated,
     * The B's parent is updated, and C's children is updated.
     * <p>
     * The child node will always move to the end of the children of the parent node.
     * @return void
     * @param child The to-be-moved child node.
     * @param new_parent The new parent node.
     * @since 1.0     
     */
    void moveNode(XATNode child, XATNode new_parent);
    
    /**
     * Move node. It will affect three nodes, the original parent node A, the child
     * node B and the new parent node C. The A's children is updated,
     * The B's parent is updated, and C's children is updated.
     * <p>
     * The child node will move to the specific position in the parent node.
     * @return void
     * @param child The to-be-moved child node.
     * @param new_parent The new parent node.
     * @param position The new position in the parent node.
     * @since 1.0          
     */
    void moveNode(XATNode child, XATNode new_parent, int position);
    
    /**
     * Replace oldNode in Tree with newNode.
     * @return void
     * @param oldNode The old XATNode will be replaced.
     * @param newNode The new XATNode used to replace the old one.
     * @since 1.0          
     */
    void replaceNode(XATNode oldNode, XATNode newNode);
    
    /**
     * Set the binding table of this Tree.
     * @return void
     * @param bind The new binding table for this tree.
     * @since 1.0          
     */
    void setBindingTable(BindingTable bind);
    
    /**
     * Set root node.
     * @return XATNode The original root node.
     * @param new_root The new root node.
     * @since 1.0               
     */
    XATNode setRoot(XATNode new_root);
    
    /**
     * Give number of nodes in this tree.
     * @return int The computed size of the tree.
     * @since 1.0               
     */
    int size();
    
    /**
     * This going to validate the parent-child
     * relationship.
     * It's a top down traversal.
     * And, if it find a children't parent is not set properly
     * in the tree, a warning message will be shown, and also
     * the parent link will be fixed.
     * @return boolean Tree: This is a valid XATTree. False: otherwise.
     * @since 1.0          
     */
    boolean validate();
}