package edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Vector;

/**
 * It used to visit the lowest level class to be visited first, then go to 
 * the parent class if cannot find the visit function.
 * It use the reflective functions. This will be the chosen visitor implementation 
 * for others.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */
  
public abstract class InheritableVisitor implements Visitor
{
    /** To visit a null object. */
    public abstract Object dispatch();
    /** To visit any object. */
    public abstract Object dispatch(Object obj);
/**
 * Get the ancestor of a specific class.
 * We get the interface before we get the superclass.
 * @return java.util.Vector
 * @param c java.lang.Class
 * @since 1.0
 */
protected Vector getAncestors(Class c) {
	Vector result = new Vector();
	if(c == null)
		return result;

	// add class itself.	
	result.addElement(c);

	// then add all the super class into the Vector.
	for(int i = 0; i < result.size(); i++){

		Class curClass = (Class) result.elementAt(i);
	
		// add class' interfaces.
		Class[] interfaces = curClass.getInterfaces();
		for(int j = 0; j < interfaces.length; j++){
			if(interfaces[j] != Visitable.class) 
				result.addElement(interfaces[j]);
		}

		// add class' super class.
		if(curClass.getSuperclass() != null && curClass.getSuperclass() != Object.class)
			result.addElement(curClass.getSuperclass());
	}

	return result;
}
/**
 * The getMethod function will trying to get all the classes and also the interfaces from the lower level to the
 * higher level to figure out which will be the proper method to invoke. Due to the reason the default JAVA will
 * always match the highest level of the abstract first.
 *
 * Algorithm:
 * 1) First, we get all the classes and interfaces into a vector from the bottom to the up.
 * 2) Second, we try to find the method from the beginning of that vector to the end.
 *
 * @return Method The found Method. 
 * @param c The class to be visited.
 * @since 1.0
 */
 
protected Method getMethod(Class c) {
    if (c == null)
        return null;
    Method m = null;

    // A vector for all the ancestors include itself.
    Vector ancestors = new Vector();

    // add class itself.	
    ancestors.addElement(c);

    // then add all the super class into the Vector.
    for (int i = 0; i < ancestors.size(); i++) {

        // get the next to be procesed Class.
        Class curClass = (Class) ancestors.elementAt(i);
        // check if we can find a method for it or not.
        try {
            m = getClass().getMethod("visit", new Class[] { curClass });
            // return the method if we found it.
            if (m != null) {
                return m;
            }
        } catch (NoSuchMethodException e) {
            // add class' interfaces.
            Class[] interfaces = curClass.getInterfaces();
            for (int j = 0; j < interfaces.length; j++) {
                if (interfaces[j] != Visitable.class)
                    ancestors.addElement(interfaces[j]);
            }

            // add class' super class.
            if (curClass.getSuperclass() != null
                && curClass.getSuperclass() != Object.class)
                ancestors.addElement(curClass.getSuperclass());
        }
    }

    // then, we can only call the dispatch method.
    try {
        m = getClass().getMethod("dispatch", new Class[] { Object.class });
    } catch (NoSuchMethodException e) {
        System.out.println("Cannot Happend!");
    }
    return m;
}
    /**
     * Well in some case, the visit() will call the visitable, which is
     * the highest interface in the visitor pattern. In that case, we just
     * lead it to the visit(object) method to dynamically find out where you should go. 
     * <p>
     * You should never ever visit(visitable) directly, but its subclasses.
     * @return Object
     * @param v 
     * @since 1.0
     */
    public Object visit(Visitable v)
    {
        return visit((Object)v);
    }
    /** 
     * If there is an inheretance between the visited objects, it will only match the 
     * highest class.
     * @return Object
     * @param object
     * @since 1.0
     */
    public Object visit(Object object)
    {
        if (object == null)
            return dispatch();
        Method m = getMethod(object.getClass());
        try
        {
            return m.invoke(this,
                new Object[] { object });
        }
        catch (InvocationTargetException e)
        {
            e.getTargetException().printStackTrace();
            return dispatch(object);
        }
        catch (IllegalAccessException e)
        {
            return dispatch(object);
        }
        catch (IllegalArgumentException e)
        {
            return dispatch(object);
        }
    }
}
