package edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration;

import edu.wpi.cs.dsrg.utils.BindingTable;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumnImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SourceImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.FunctionImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.CartesianProductImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.DistinctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupByImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.JoinImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderByImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.ProjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.SelectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Navigate;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.NavigationStep;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * SQLOperator Evaluator is used to help create SQL queries
 * @author: Brad Pielech
 */
public class SQLOperatorEvaluator extends OperatorEvaluator {
    /**
     * Call the super constructor
     * @param heuristic int
     * @param varTable BindingTable
     */
    public SQLOperatorEvaluator(int heuristic, BindingTable varTable) {
        super(heuristic, varTable);
    }
    /**
     * Call the super's constuctor
     * @param varTable java.util.Hashtable
     */
    public SQLOperatorEvaluator(BindingTable varTable) {
        super(varTable);
    }
/**
 * <pre>
 * Evaluates a Function Operator and determines the SQL equivalent 
 * The function name is the name parameter of the result
 * The function parameters go into the parameters field of the result
 * The function binding is the binding of the result
 * </pre>
 * @return java.lang.Object
 * @param param Function
 */
public Object visit(FunctionImp param) {
    QueryResult s1 = new QueryResult();
    if (isSourceVisited()) {
        String funName = param.getName();
        funName = funName.trim().toUpperCase();
        String[] sqlFuncs = { "MAX", "MIN", "AVG", "COUNT", "SUM", "MULTISTEPORDER", "SINGLESTEPORDER" };
        boolean sqlAble = false;
        for (int i = 0; i < sqlFuncs.length; i++) {   
            if (funName.equals(sqlFuncs[i])) {
                sqlAble = true;
            }
        }
        if (sqlAble) {
            //NavigationStep[] funParams = (NavigationStep[])param.getParams();
            String paramList = "";
            for (int i = 0; i < param.getParams().length; i++) {
                paramList = paramList + param.getParams()[i] /*.getStepValue() */
                +";";
            }
            if (!paramList.equals(""))
                paramList = paramList.substring(0, (paramList.length() - 1));
            s1.addFragment(
                new QueryFragment(
                    QueryTypes.FUNCTION,
                    funName,
                    param.getOutputColumnName(),
                    paramList));
            s1.setEvaluated(true);
        } else {
	        //if it isn't a SQL able function, stop creating SQL
            super.dispatch(param);
        }
    } else {
        s1.setEvaluated(false);
    }
    return s1;
}
    /**
     * <pre>
     * Evaluates a NameColumn Operator and determines the SQL equivalent of the destination
     * the old column name is return in the name parameter of the fragment, 
     * the new name is the binding
     * </pre>
     * @return java.lang.Object
     * @param param NameColumnImp
     * @deprecated NameColumn operators can now be pushed down so they will not occur in the tree to generate
     * a query from
     */
    public Object visit(NameColumnImp param) {
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        String newC = param.getNewColumnName();
        String oldC = param.getOriginalColumnName();
        s1.addFragment(new QueryFragment(1, oldC, newC));
        return s1;
    }
    /**
     * <pre>
     * Evaluates a Source Operator and determines the SQL equivalent of the destination
     * The Source doesnt have an SQL equivalent so it is just a blank fragment
     * </pre>
     * @return java.lang.Object
     * @param param Source
     */
    public Object visit(SourceImp param) {
        //set the sourceVisited to TRUE
        setSourceVisited(true);
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        //create an empty QueryFragment and then return the result object
        QueryFragment frag = new QueryFragment(-1, "");
        s1.addFragment(frag);
        return s1;
    }
    /**
     * Evaluates a CartesianProduct Operator and determines the SQL equivalent of the destination
     * ReWriting will take of this so it wont happen
     * @return java.lang.Object
     * @param param CartesianProductImp
     */
    public Object visit(CartesianProductImp param) {
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        //create an empty QueryFragment and then return the result object
        QueryFragment frag = new QueryFragment(-1, "");
        s1.addFragment(frag);
        return s1;
    }
    /**
     * <pre>
     * Evaluates a Distinct Operator and determines the SQL equivalent of the destination
     * The result contains a fragment with the Distinct's output name in it
     * </pre>
     * @return java.lang.Object
     * @param param Distinct
     */
    public Object visit(DistinctImp param) {
        QueryResult s1 = new QueryResult();
        //call the visit method for Distinct's child (Navigate)
        String name = param.getInputColumnName();
        QueryFragment frag = new QueryFragment(QueryTypes.DISTINCT, name);
        s1.addFragment(frag);
        s1.setEvaluated(true);
        return s1;

    }
    /**
     * </pre>
     * Evaluates an GroupBy Operator and determines the SQL equivalent of the destination
     * take all of the columns names in the orderBy and add them to the oderBy clause
     * the result will look like GroupBy col1, col2, etc
     * </pre> 
     *
     * GroupBy returns multiple SQL fragments, 1 for the groupBy and 1 for each of the operators in its subTree
     * Therefore, we first visit the subTree first
     * @return java.lang.Object
     * @param param GroupBy
     */
    public Object visit(GroupByImp param) {
        QueryResult s1 = new QueryResult();
		//first visit the subTree
		/***
		**  traverse the tree better
		*/
		XATTree t1 = param.getXATTree();
		XATNode x1 = t1.getRoot();
		QueryResult subTreeR = (QueryResult) visit(x1.getOperator());  		
		 
        if (isSourceVisited()) {
            s1.setEvaluated(true);
            QueryFragment frag = new QueryFragment(-1, "");
            s1.addFragment(frag);
        } else {
            s1.setEvaluated(false);
        }
        //before returning s1, add all fragments from the subTreeR into s1
        QueryFragment[] f1 = subTreeR.getFragment();
        for (int i = 0; i < f1.length; i++){
        	s1.addFragment(f1[i]);
        }
        return s1;
    }
    /**
     * <pre>
     * Evaluates a Join Operator and determines the SQL equivalent of the destination
     * This adds to the where clause, it will look like Where $a = '10'
     * </pre> 
     *
     * Same thing as select
     * @return java.lang.Object
     * @param param Join
     */
    public Object visit(JoinImp param) {
        QueryResult s1 = new QueryResult();
        //If the join only has 1 child, it can be deleted
        XATNode x1 = param.getNode();
        XATNode[] children = x1.getChildren();
        if(children.length < 2){
			QueryFragment frag = new QueryFragment(-1,"");
			s1.addFragment(frag);
			s1.setEvaluated(true);
			return s1;	
        }
        if (isSourceVisited()) {
            try {
                String exp = (String) param.getExpression().toSQL(this);
                s1.addFragment(new QueryFragment(QueryTypes.WHERE, exp));
                s1.setEvaluated(true);
	            return s1;
				//there will be a NullPointer when there is no exp associated with this Join
            } catch (NullPointerException e1) {
                //create an empty QueryFragment and then return the result object
                QueryFragment frag = new QueryFragment(-1, "");
                s1.addFragment(frag);
                return s1;
            }
        } else {
            s1.setEvaluated(false);
        }
        return s1;
    }
    /**
     * <pre>
     * Evaluates an OrderBy Operator and determines the SQL equivalent of the destination
     * take all of the columns names in the orderBy and add them to the oderBy clause
     * the result will look like OrderBy col1, col2, etc
     * </pre>
     * 
     * @return java.lang.Object
     * @param param OrderBy
     */
    public Object visit(OrderByImp param) {
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        //create an empty QueryFragment and then return the result object
        String[] columns = param.getColumnNames();
        String text = "";
        for (int i = 0; i < columns.length; i++) {
            text = text + columns[i] + ",";
        }
        //chop off the last ,
        text = text.substring(0, (text.length() - 1));
        QueryFragment frag = new QueryFragment(QueryTypes.ORDERBY, text);
        s1.addFragment(frag);
        return s1;
    }
    /**
     * Evaluates a Project Operator and determines the SQL equivalent of the destination
     * This wont happen either
     * @return java.lang.Object
     * @param param Project
     */
    public Object visit(ProjectImp param) {
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        //create an empty QueryFragment and then return the result object
        QueryFragment frag = new QueryFragment(-1, "");
        s1.addFragment(frag);
        return s1;
    }
    /**
     * <pre>
     * Evaluates a Select Operator and determines the SQL equivalent of the destination
     * This adds to the where clause, it will look like Where $a = '10'
     * </pre> 
     * 
     * @return java.lang.Object
     * @param param Select
     */
    public Object visit(SelectImp param) {
        QueryResult s1 = new QueryResult();
        if (isSourceVisited()) {
            String exp = (String) param.getExpression().toSQL(this);
            s1.addFragment(new QueryFragment(QueryTypes.WHERE, exp));
            s1.setEvaluated(true);
        } else {
            s1.setEvaluated(false);
        }
        return s1;

    }
/**
 * <pre>
 * Evaluates a Navigate Operator and determines the SQL equivalent of the destination
 * If the Navigate contains a row in the destination, it goes to the FROM, otherwise
 * it goes to the SELECT
 * </pre>
 * @return java.lang.Object
 * @param param Navigate
 */
public Object visit(Navigate param) {

    Navigate x1 = param;
    QueryResult s1 = new QueryResult();
    if (isSourceVisited()) {
        NavigationStep[] destination = x1.getDestination();

        String dest = "";
        if (destination != null) {
            dest = createString(destination);

            //if the destination contains row, then the Navigate will add to the FROM clause
            if (dest.indexOf("row") > 0 || dest.indexOf("ROW") > 0) {
                //find the step that is row, then go to step before this
                String table = "";
                for (int i = 0; i < destination.length; i++) {
                    String thisStep = destination[i].getStepValue();
                    if (thisStep.endsWith("row") || thisStep.endsWith("ROW")) {
                        table = thisStep.substring(0,thisStep.indexOf("."));
                    }
                }

                //then have to add in the binding so the table has a binding with it.
                String binding = x1.getBinding();
                s1.addFragment(new QueryFragment(QueryTypes.FROM, table, binding));
                s1.setEvaluated(true);
                return s1;

            } else {
                //have to examine the contents of the destination to figure out the select clause
                for (int i = 0; i < destination.length; i++) {
                    NavigationStep step = destination[i];
                    //make sure it isnt text()
                    if (!step.getStepValue().equals("text()") && !step.getStepValue().equals("#text")) {
                        String value = step.getStepValue();
                        String entryPoint = x1.getEntryPoint()[0].getStepValue();
                        //add in the entryPoint binding so you get entryPoint.destination
                        String select = "\"" + entryPoint + "\"." + value;
                        String binding = x1.getBinding();
                        //return the new fragment
                        s1.addFragment(new QueryFragment(QueryTypes.SELECT, select, binding));
                        s1.setEvaluated(true);
                        return s1;
                    }
                }

            }
        }
        s1.setEvaluated(true);
        //create an empty QueryFragment and then return the result object
        QueryFragment frag = new QueryFragment(-1, "");
        s1.addFragment(frag);
    } else {
        s1.setEvaluated(false);
    }
    return s1;
}
}
