package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATPropertiesImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
	/**
	* 
	* StatisticsGatherer is a large statistics engine that supports arbitrary statistics
	* and many different access methods.  See below for a description on using this Gatherer, 
	* including how to add new properties, retrieve values, retrive objects, and update values<p>
	*
	* The Gatherer works by organizing statistics into categories.  The statsTable will always have 
	* 2 categories called HISTORICAL and GENERIC(records are identified by a Object identifier right now).
	* HISTORICAL keeps track of the statistics for the operators
	* regardless of the scheduling algorithm or query plan.  HISTORICAL also
	* contains the aggregate stats for the whole system.	<p>
	*
	* GENERIC contains basic statistics that do not depend on the category.  These statistics,
	* such as NUMBER_OF_TUPLES_IN_INPUT_QUEUES, should have the same values regardless of the category,
	* otherwise, there are consistency issues.  <p>
	* 
	* Because the HISTORICAL name may (and has already) change, it is safest to
	* refer to it using <tt>StatisticsGatherer.OVERALL</tt> <p>
	*
	* Additional categories can be added to the statsTable that will correspond to smaller views
	* of the overall statistics.  For example, categories could be added that correspond to
	* a particular scheduling algorithm or a certain plan. <p>
	*
	* The individual categories are organized by property names.  The property names are
	* are gathered using <tt>XATQueryObject.getProperties()</tt>.  Each property name
	* has a ModifiedArrayList (extention of ArrayList with some different access methods)
	* that sorts the XATOperators based on the property.  Queries, such as what is the operator
	* with the lowest (insert propery name), can be made against the Gatherer <p>
	*	
	* The StatisticsGather should be used as follows:
	* <ol>
	* <li>  <tt> getStatisticsGatherer()</tt>  This will instantiate the statisticsGather.
	*	  Note that there can only be 1 statistics gatherer so subsequent calls will just return
	*	  a handle to the lone statsisticsGatherer
	* <li>  <tt>initStatsTable(XATTree) </tt> or <tt> initStatsTable(XATNode[]) </tt>
	*     This will initialize the statsTable with the given XATNodes
	*     This can only be called once.
	* <li>  The rest depends on what you want to do:
	*	<ol type = "a">
	*	  <li> To add a new XATQueryObject:
	*	  	 <tt>addNewQueryObject(newQueryObject)</tt>
	*	  <li> To add a new statistics category with a given description:
	*	     <tt>addNewCategory(Object)</tt> <p>
	*	  <li> To have the StatisticsGatherer work with a new property:
 	*			<ul>
 	*				<li> If the property requires a special calculation besides
 	*					 just replacing the old value with a new one (such as LAST_TIME_RUN), 
 	* 					 then:
 	*					 <ol>
 	*						<li> you <b>need</b> to create a new StatisticsObject type and implement
 	*							the <tt>calculateNewValue()</tt> or <tt>calculateNewValue(String)</tt>.  Use the 2nd method
 	*							if you require some outside information to calculate the value.  The first
 	*							is used if no outside information is used.  
 	*						<li> you need to update <tt>StatisticsReference.initReferenceTable()</tt>
 	*					 with the name of the property and the name of the StatisticsObject that calculates the property.
 	*					 You should also decide if it is a category specific or a generic statistics.  See StatisticsReference for more.
 	*					</ol>
 	*				<li> If the property does not require a special calculation, then you dont have to do anything extra
 	*			</ul>
	* 	  For the rest of the operations, it is important to note that the
	*	  category is not specified each time.  This makes it easier to
	*	  to access data without specifying the same parameter over and over.
	*	  To change the category, use <tt>setCurrentCategory(Object)</tt> <p>
	*
	*	  <li> To add a new value of a property to the Gatherer:
	*	     <tt>updateValue(String propertyName, XATOperator operator)</tt>
	*	     Note:  This will update the value for the current category as well as OVERALL
	*			(if the currentcategory != OVERALL)
	*	  <li> To get a value of a property of a QueryObject from the Gatherer:
	*	     <tt>getValue(String propertyName, XATQueryObject queryObject)</tt><p>
	*
	*	  <b>This last method is probably the most useful and powerful in the StatisticsGatherer</b><p>
	*	  <li> To get a QueryObject with a certain value for a given property:
	*		 <ul>
	*		  <li> <tt> getOperator(String propertyName, int identifier)</tt>
	*         <li> <tt> getQueue(String propertyName, int identifier)</tt>
	*		  <li> <tt> getQueryObject(String propertyName, int identifier</tt>
	*		 </ul>
	*		Note: The first two are merely convience methods and they avoid casting on the client end.
	*		They are identical to calling getQueryObject and then casting. <p> 	
	*		The identifier is any of the static constants found in StatisticsReference.  Some
	*		examples are FIRST, LAST, HIGHEST, LOWEST, and AVERAGE.
	*		Depending on the identifier, the method will return the XATQueryObject whose propertyValue
	*		matches the given identifier with respect to the other XATQueryObjects. <p>
	*	  </ol>
	* </ol> 
	*/	
 
public final class StatisticsGatherer extends XATPropertiesImp implements java.io.Serializable {
	private static StatisticsGatherer statsGatherer = null;
	/** This is where all of the statistics are kept */
	private Hashtable statsTable;
	private boolean init = false;
	private StatisticsReference ref; 
	public final static java.lang.String OVERALL = "HISTORICAL";
	public final static java.lang.String GENERIC = "GENERIC";
	/** 
	 * this is the current category of statistics we are looking at
	 * this is used as a convience so that anytime the stats engine is queried, you
	 * do not have to specifiy which category.  
	 */
	private java.lang.Object currentCategory;
	private boolean timer;
/**
 * StatisticsGatherer constructor comment.
 * In order to use the statisticsgatherer you need to set the property, like:
 * 	System.setProperty("StatisticsGatherer", "EXIST");
 */
private StatisticsGatherer() {
	super();
	statsTable = new Hashtable();
	currentCategory = this.OVERALL;
	ref = new StatisticsReference();
	timer = (new Boolean(getProperty("TIMER"))).booleanValue();
}
/**
 * Adds a new category of statistics into the StatisticsGatherer
 * A category is defined as a grouping of statistics. For example, some categories
 * would be Scheduling algorithms, XATTrees, or any other object.  The only category that
 * cannot be added is one with a String Identifier named "OVERALL".  This category
 * is used in the StatisticsGather by default so adding another would create problems  <p>
 *
 * Before categories can be added, the statsgather must have been initialized.  <p>
 *
 * One problem with this is that the new catgory will have its starting values set to thato
 * of the OVERALL category.  I am not sure how to get around this, nor do I know if it is a major problem.
 * If all of the categories are initialized at the beginning, then it will not matter much, if at all. <p>
 * @param identifier java.lang.Object
 * @throws XATStatisticsException if the category named identifier already exists in the statistic gatherer
 */
public void addNewCategory(Object identifier) throws Exception{
	
	if(!init){
		throw new XATStatisticsException("The statsTable must be initialized before any new categories may be added");	
	}
	if(statsTable.containsKey(identifier)){
		throw new XATStatisticsException("Error adding new category.  The category already exists");
	}
	/**
	 * clone the hashtable associated with OVERALL and 
	 * place it into the statsTable with the given identifier	  
	 */	
	try {
		Hashtable row = (Hashtable)statsTable.get(this.OVERALL);
		//Hashtable newCategory = (Hashtable)edu.wpi.cs.dsrg.utils.DeepCopyMaker.makeDeepCopy(row);
		/** now put the new category into the statsTable */
		statsTable.put(identifier, copyHashTable(row, identifier));
	} catch (Exception e) {
		throw e;
	}
	
	
	}
/**
 * Adds an array of QueryObject to the statsTable.  This seems like a relatively
 * expensive operation.
 * @param newQueryObject 
 */
public void addNewQueryObject(XATQueryObject[] newQueryObjects) {
    for (int i = 0; i < newQueryObjects.length; i++) {
        /** 
         * iterate through the different categories and add this operator to each 
         * row in the category         
         */
        for (Enumeration keys = statsTable.keys(); keys.hasMoreElements();) {
            Object o = keys.nextElement();
            try {
                if (o instanceof String) {
                    String s = (String) o;
                    if (s == this.GENERIC) {
                        insertIntoStatsTable(newQueryObjects[i], (Hashtable) statsTable.get(o), true);
                    } else {
                        insertIntoStatsTable(newQueryObjects[i], (Hashtable) statsTable.get(o), false);
                    }
                } else {
                    insertIntoStatsTable(newQueryObjects[i], (Hashtable) statsTable.get(o), false);
                }

            } catch (Exception e) {
                System.out.println(e.toString());
            }
        }
    }
}
/**
 * Add an QueryObject to the statsTable.  This seems like a relatively
 * expensive operation.
 * @param newQueryObject 
 */
public void addNewQueryObject(XATQueryObject newQueryObject) {
	/** create an array with 1 element in it and then call addNewQueryObject(Array)*/
    XATQueryObject[] array = new XATQueryObject[]{newQueryObject};
    addNewQueryObject(array);
}
/**
 * Iterates through this hashtable and copies all of the keys
 * into the new table.  For the values, the associated ArrayList will
 * be copied, but not cloned.  A new StatisticsObject will be created, but its
 * corresponding operator will be copied to the new StatisticsObject
 * Creation date: (1/14/2003 10:36:49 AM)
 * @return java.util.Hashtable
 * @param oldTable java.util.Hashtable
 * @param newIdentifier the name of the new category.  This is needed to create new StatisticsObjects with
 * the appropriate category
 */
private Hashtable copyHashTable(Hashtable oldTable, Object newIdentifier) throws Exception {
    try {

        Hashtable newTable = new Hashtable();
        /** iterate through the keys */
        for (Enumeration e = oldTable.keys(); e.hasMoreElements();) {
            String key = (String) e.nextElement();

            /** get the ArrayList associated with the key */
            ArrayList keyList = (ArrayList) oldTable.get(key);
            ArrayList newList = new ModifiedArrayList();
            for (int i = 0; i < keyList.size(); i++) {
                /** get the next statisticsObject */
                StatisticsObject statsObject = (StatisticsObject) keyList.get(i);
                /** instantiate a new statsobject of the same type as the old */
                StatisticsObject newStatsObject =
                    (StatisticsObject) Class
                        .forName(statsObject.getClass().getName())
                        .newInstance();
                /** set the new object's operator to be the same as the old */
                newStatsObject.setCategory(newIdentifier);
                newStatsObject.setAssociatedQueryObject(statsObject.getAssociatedQueryObject());
                newStatsObject.setPropertyName(statsObject.getPropertyName());
                newStatsObject.setPropertyValue(statsObject.getPropertyValue());

                newList.add(newStatsObject);
            }
            /** put the key and a new ArrayList into the hashtable */
            newTable.put(key, newList);
        }
        return newTable;
    } catch (Exception e) {
        //hrm, big problems
        throw e;
    }

}
/**
 * Returns an Iterator of all of the query objects that
 * are known in this stats gatherer.  This is useful for
 * making queries against the statsGather when the query objects are not known. <p>
 *
 * This does not traverse the GENERIC query objects b/c those generally are not useful except
 * to calculate other statistics. <p>
 * Note, this is an expensive operation.  It runs on the order of n^2 where n is the number of
 * query objects in the system. 
 * @return java.util.Iterator
 */
public synchronized Iterator getAllKnownQueryObjects() {
	/** wait while available = false.  This means someone else is accessing the stats gatherer*/
	/*
	while(available == false){
		try {
			wait();	
		} catch (InterruptedException e) {
			
		}		
	}
	/** lock the stats gatherer*/
	//available = false;
	
	ArrayList list = new ArrayList();
	Hashtable table = (Hashtable)statsTable.get(currentCategory);
	for(Enumeration e = table.keys();e.hasMoreElements();){
		ArrayList nextList = (ArrayList)table.get(e.nextElement());		
		for (int i = 0; i < nextList.size(); i++){
			StatisticsObject stats = (StatisticsObject)nextList.get(i);
			if(!list.contains(stats.getAssociatedQueryObject())){
				list.add(stats.getAssociatedQueryObject());
			}
		}		
	}
	/** make this available to others and notify them as such */
	/*
    available = true;
    notifyAll();
    */
    /** this may create a concurrent modification exception */
    //Collections.sort(list);
	return list.iterator();
}
/**
 * Returns the currentCategory
 * @return java.lang.Object
 */
public Enumeration getCategories() {
	return statsTable.keys();
}
/**
 * Returns the currentCategory
 * @return java.lang.Object
 */
public java.lang.Object getCurrentCategory() {
	return currentCategory;
}
/**
 * 
 * Gets the XATQueryObject whose propertyValue corresponds to the propertyName and valueIdentifier
 * in the currentCategory.<p>
 * For example, getQueryObject("SELECTIVITY", StatisticsReference.LOWEST)
 * will return the XATQueryObject with the lowest selectivity within the currentCategory.
 * getOperator("LAST_TIME_RUN", StatisticsReference.HIGHEST) will return the XATQueryObject
 * ran most recently (lastTimeRun is highest).    <p>
 *
 * See StatisticsReference for a list of supported identifiers
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 * @param propertyName java.lang.String
 * @param identifier int
 */
public synchronized XATQueryObject getQueryObject(String propertyName, int valueIdentifier) {
	return getQueryObject(propertyName, valueIdentifier, false);
}
/**
 * 
 * Gets the XATQueryObject whose propertyValue corresponds to the propertyName and valueIdentifier
 * in the currentCategory.<p>
 * For example, getQueryObject("SELECTIVITY", StatisticsReference.LOWEST)
 * will return the XATQueryObject with the lowest selectivity within the currentCategory.
 * getOperator("LAST_TIME_RUN", StatisticsReference.HIGHEST) will return the XATQueryObject
 * ran most recently (lastTimeRun is highest).    <p>
 *
 * See StatisticsReference for a list of supported identifiers <p>
 *
 * Added 5/26: You can now ask for any position in the list, not just highest, lowest, average, etc.  Just
 * pass in the position you want.  To retrieve highest, lowest, etc, still make use of the constants in StatisticsReference
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 * @param propertyName java.lang.String
 * @param identifier int
 * @param updateValue boolean true if the values should be updated first before the first, last, etc value is retrieved
 */
public synchronized XATQueryObject getQueryObject(String propertyName, int valueIdentifier, boolean updateValue) {
	
    try{
	
	/** again, if it is a generic property, get the list from generic category, o/w get
	    it from the current category
	*/
	long time = 0; 
    if(timer){
	    time = System.currentTimeMillis();
    } 
	
	ModifiedArrayList list = null;
	if(ref.isGenericProperty(propertyName)){
		list = (ModifiedArrayList)((Hashtable)statsTable.get(this.GENERIC)).get(propertyName);
	} else{
		list = (ModifiedArrayList)((Hashtable)statsTable.get(currentCategory)).get(propertyName);
	}
	// get the query object according to the given identifier
    XATQueryObject o1 = ((StatisticsObject)list.getByIdentifier(valueIdentifier, updateValue)).getAssociatedQueryObject();
	          
    
    if(timer){	    
		time = System.currentTimeMillis() - time;		
		long oldTime = Long.parseLong(getProperty("TOTAL_RETRIEVE_OBJECT_TIME"));
		time = time + oldTime;
		long timesRun = (Long.parseLong(getProperty("TOTAL_RETRIEVE_OBJECT_REQUESTS"))) + 1;
		try{
			setProperty("TOTAL_RETRIEVE_OBJECT_TIME", "" + time);
			setProperty("TOTAL_RETRIEVE_OBJECT_REQUESTS", "" + timesRun);
		} catch(Exception e){
			System.out.println(e.toString());
		}		
    }
    
    return o1;
    
    } catch(NullPointerException e){	    
		throw new IllegalArgumentException("The statistic " + propertyName + " was not found in the StatisticsGatherer");
    }
}
/**
* we want there to be only 1 statistics gatherer so
* a singleton design pattern is used.
* @return StatisticsGatherer a handle to the statistics gatherer 
*/
public static StatisticsGatherer getStatisticsGather(){
	if(statsGatherer == null){
		statsGatherer = new StatisticsGatherer();
	}
	return statsGatherer;
}
/**
 * Returns the value for given property of the queryObject in the current category
 * Creation date: (1/15/2003 2:05:37 PM)
 * @return java.lang.String
 * @param propertyName java.lang.String
 * @param queryObject XATQueryObject
 */
public synchronized String getValue(String propertyName, XATQueryObject queryObject) {
	return retrieveValue(propertyName, queryObject, ref.VALUE);
}
/**
 * Returns the value for given property of the queryObject in the current category
 * If this statistics object does not support this operation, an error will be thrown
 * @return java.lang.String
 * @param propertyName java.lang.String
 * @param queryObject XATQueryObject
 * @param description the type of value to retrieve.  Available constants are found in StatisticsReference
 */
public synchronized String getValue(String propertyName, XATQueryObject queryObject, int description) {	
	return retrieveValue(propertyName, queryObject, description);
}
/**
 * Allows you to specify an initial array of XATNodes. The ordering (structure) of the 
 * nodes is irrelevant because the query plan may change later. If the operators themselves 
 * change (new operators added) then this statistics gather will have to be updated. <p>
 *
 * The properties of each operator will be stored inside of a hashtable.  The
 * key of the hashtable will be the property name (String).  The value will be 
 * an ArrayList where each entry in the List is a StatisticsObject (which points to an operator).  
 * Each StatisticsObject is a Comparator, this will allow the StatisticsObjects to 
 * be sorted and organized later.    <p>
 * Creation date: (1/10/2003 3:47:27 PM)
 * @param initialTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 * @throws XATStatisticsException if the Statistics Gatherer has already been initialized.
 * This prevents the statsTable from being initialized multiple times.  
 */
private void initStatsTable(XATNode[] initNodes) throws XATStatisticsException {

    if (init) {
        //stats table has already been initialized
        throw new XATStatisticsException("The statsTable has already been initialized. \nMake changes to the query objects through the appropriate interfaces in the StatisticsGatherer");
    }
    //only init once 
    init = true;
    /** 
     * create the hashtables that will store all of the property names and 
     * their associated ArrayLists.
     * generic stores generic properties and specific is specific properties
     */
    Hashtable specific = new Hashtable();
    Hashtable generic  = new Hashtable();
    /** 
     * For each node that we see in the initNodes array,
     * 1. iterate through the supported properties
     * 2. IF the next property is already in the hashtable THEN
     *     a. get the ArrayList "attached" to the entry in the hashtable
     *     b. create a new StatisticsObject with the given property name and operator
     *	   c. insert the StatsObject into the array (position doesnt matter)
     *    ELSE (if the property is not in the hashtable
     *     d. insert property into table
     *     e. create new array list
     *     f. execute steps 2b and 2c
     * 3. then add the properties for the attached output datasource. 
     *	  this is a little tricky because XATOperators have DataSources while the StreamOperators have
     *	  queues.  So, have to try to cast the operator to a StreamOperator, then get the queues.
     *	  otherwise, just get the output datasource
     * 4. This is repeated twice, once for the generic properties, once for the category-specific properties
     */
    for (int i = 0; i < initNodes.length; i++) {
        XATNode nextNode = initNodes[i];
        XATOperator nextOperator = nextNode.getOperator();
        try {
            insertIntoStatsTable(nextOperator, generic, true);
            insertIntoStatsTable(nextOperator, specific, false);
        } catch (Exception ex) {
            throw new XATStatisticsException(ex.toString());
        }

        /** now try to iterate through the Output datasources of the XAT operator */
        //XATStreamOperator streamOp = (XATStreamOperator) nextOperator;
        for (Iterator dataSources = nextOperator.getOutputDataSources(); dataSources.hasNext();) {            
            try {
	            XATQueryObject qo = (XATQueryObject)dataSources.next();                
	            insertIntoStatsTable(qo, generic, true);
            	insertIntoStatsTable(qo, specific, false);
            } catch (Exception ex) {
                throw new XATStatisticsException(ex.toString());
            }
        }
    }
    /**
     * now we will sort all of the initial ArrayLists
     */
    /* 
    for (Iterator i = specific.values().iterator(); i.hasNext();) {
        Collections.sort((ArrayList) i.next());
    }
    for (Iterator i = generic.values().iterator(); i.hasNext();) {
        Collections.sort((ArrayList) i.next());
    }
    */
    /**
     * now the basic table is set up, we will put the generic into the 
     * statsTable in with GENERIC as the key.
     * the specific will go into the statsTable with SPECIFIC as the key 
     */
    statsTable.put(this.OVERALL, specific);
    statsTable.put(this.GENERIC, generic);
    return;
}
/**
 * Allows you to specify an initial XATTree that
 * contains all of the operators.  The ordering (structure) of the operators is irrelevant
 * because the query plan may change later. If the operators themselves change (new operators added)
 * then this statistics gather will have to be updated. <p>
 *
 * The properties of each operator will be stored inside of a hashtable along
 * with a Comparator so they can be sorted later and organized later. <p>   
 * @param initialTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 * @throws XATStatisticsException if the Statistics Gatherer has already been initialized.
 * This prevents the statsTable from being initialized multiple times. 
 */
public void initStatsTable(XATTree initTree) throws XATStatisticsException {
	initStatsTable(initTree.convertToArray());
	this.addNewQueryObject(initTree);
	timer = (new Boolean(getProperty("TIMER"))).booleanValue();
}
/**
 * Inserts all of the properties of this queryObject into the StatsTable
 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 * @param row java.util.Hashtable
 * @boolean generic If the row we are inserting objects into corresponds to the GENERIC row
 */
private synchronized void insertIntoStatsTable(
    XATQueryObject queryObject,
    Hashtable row,
    boolean generic)
    throws Exception {    
    for (Enumeration e = queryObject.getStatistics(); e.hasMoreElements();) {
        String nextProperty = (String) e.nextElement();
        /** 
         * if the property is generic and we are looking for generic properties 
         * 			-or-
         * if the property is not generic and we are not looking for generic properties
         * then add the property
        */ 
        if (ref.isGenericProperty(nextProperty) == generic) {
            if (!row.containsKey(nextProperty)) {
                row.put(nextProperty, new ModifiedArrayList());
            }
            try {
                /** get the array list*/
                ModifiedArrayList list = (ModifiedArrayList) row.get(nextProperty);
                /** look at the reference table to see which statistics object to instantiate*/
                StatisticsObject statsObject =
                    (StatisticsObject) Class
                        .forName(ref.getStatisticsObjectName(nextProperty))
                        .newInstance();
                statsObject.setPropertyName(nextProperty);
                /** set the initial value of this statsobject*/
                statsObject.setPropertyValue("0");
                statsObject.setAssociatedQueryObject(queryObject);
                /** statsObject starts out with OVERALL category */
                statsObject.setCategory(this.OVERALL);                
                list.add(statsObject);
                /** put the updated list back into the hashtable, this will overwrite the original, which is ok*/
                row.put(nextProperty, list);
            } catch (Exception e1) {
                throw e1;
            }
        }
    }    

}
/**
 * Checks if the propertyName is in the StatisticsGatherer
 * @return boolean
 * @param propertyName java.lang.String
 */
public boolean isStatisticSupported(String propertyName) {
	if(!((Hashtable)statsTable.get(this.OVERALL)).containsKey(propertyName)){
		return ((Hashtable)statsTable.get(this.GENERIC)).containsKey(propertyName);
	} else{
		return true;
	}
}
/**
 * Removes a query object from the statistics gatherer.  This will
 * help to reduce the memory footprint of the statistics gatherer by removing
 * unused and unnessary query objects.  Of course this could be abused by removing
 * necessary query objects.  
 * @param object edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 */
public synchronized void removeQueryObject(XATQueryObject object) {

    /**
     * iterate through each list in each category and
     * remove the statistics object that matches this object. 
     */
    for (Enumeration e = statsTable.keys(); e.hasMoreElements();) {
        Hashtable categoryTable = (Hashtable) statsTable.get(e.nextElement());
        //iterate through each array list for the category 
        for (Iterator lists = categoryTable.values().iterator(); lists.hasNext();) {
            ModifiedArrayList list = (ModifiedArrayList) lists.next();
            StatisticsObject statsObject = list.getStatisticsObject(object);
            //if there is a statsobject for this query object, remove it now 
            if (statsObject != null) {
                list.remove(statsObject);
            }
        }
    }

}
/**
 * This will reset the statistics gatherer to its pre-init state.  This
 * will remove all categories and all statistics.  Care should be given with this method, otherwise
 * it could really mess up execution 
 */
public void reset() {	
	statsTable.clear();
	init = false;	
}
/**
 * Returns the value for given property of the queryObject in the current category
 * If mean is true, then this will return the mean
 * @return java.lang.String
 * @param propertyName java.lang.String
 * @param queryObject XATQueryObject
 * @param descr String describing the type of value to retrieve 
 */
private synchronized String retrieveValue(
    String propertyName,
    XATQueryObject queryObject,
    int desc) {

    if (!queryObject.isValidStatistic(propertyName)) {
        throw new IllegalArgumentException(
            "The statistic " + propertyName + " is not supported by this QueryObject " + queryObject);
    }
    long time = 0;
    if (timer) {
        time = System.currentTimeMillis();
    }
    ModifiedArrayList list = null;
    /** 
     * if it is a generic property, just get the value from the generic category 
     * if it isnt generic, then get the value from the current category
     */
    if (ref.isGenericProperty(propertyName)) {
        list =
            (ModifiedArrayList) ((Hashtable) statsTable.get(this.GENERIC)).get(
                propertyName);
    } else {
        list =
            (ModifiedArrayList) ((Hashtable) statsTable.get(currentCategory)).get(
                propertyName);
    }
    StatisticsObject statsObject = list.getStatisticsObject(queryObject);
    if (timer) {
        time = System.currentTimeMillis() - time;
        long oldTime = Long.parseLong(getProperty("TOTAL_RETRIEVE_VALUE_TIME"));
        time = oldTime + time;
        int timesRun =
            (Integer.parseInt(getProperty("TOTAL_RETRIEVE_VALUE_REQUESTS"))) + 1;
        try {
            setProperty("TOTAL_RETRIEVE_VALUE_TIME", "" + time);
            setProperty("TOTAL_RETRIEVE_VALUE_REQUESTS", "" + timesRun);
        } catch (Exception e) {
            System.out.println(e.toString());
        }

    }
    
    if(desc == ref.VALUE) return statsObject.getPropertyValue();
    else if(desc == ref.HIGHEST) return "" + ((AggregateStatisticsObject) statsObject).getMaxValue();
    else if(desc == ref.LAST) return "" + ((AggregateStatisticsObject) statsObject).getMostRecent();
    else if(desc == ref.AVERAGE) return "" + ((AggregateStatisticsObject) statsObject).getPropertyValueMean();
    else if(desc == ref.LOWEST) return "" + ((AggregateStatisticsObject) statsObject).getMinValue();
    else throw new IllegalArgumentException("Error trying to retrieve the mean, max, or min of a statistic that does not support it");
    
}
/**
 * Set the current category.  An XATStatisticsException will
 * be thrown if the category does not belong to the stats table
 * @param newCurrentCategory java.lang.Object
 * @throws XATStatisticsException
 */
public void setCurrentCategory(java.lang.Object newCurrentCategory) {
	if(!statsTable.containsKey(newCurrentCategory)){
		throw new XATStatisticsException("The new current category does not belong to the stats table");
	}
	currentCategory = newCurrentCategory;
}
	/**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    */   
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		addProperty("TIMER", "false");
		addProperty("TOTAL_UPDATE_VALUE_TIME","0");
		addProperty("TOTAL_UPDATE_VALUE_REQUESTS","0");
		addProperty("TOTAL_RETRIEVE_VALUE_TIME","0");
		addProperty("TOTAL_RETRIEVE_VALUE_REQUESTS","0");
		addProperty("TOTAL_RETRIEVE_OBJECT_TIME","0");		
		addProperty("TOTAL_RETRIEVE_OBJECT_REQUESTS","0");
	}
/**
 * Insert the method's description here.
 * Creation date: (1/14/2003 1:17:54 PM)
 * @return java.lang.String
 */
public String toString() {
	String rs = new String();
	Hashtable row = (Hashtable)statsTable.get(this.OVERALL);
	rs = "=========================\n";
	rs = rs + "StatsTable statistics: \nNumber of categories: " + statsTable.size();
	rs = rs + "\nNumber of properties: " + row.size();
	rs = rs + "\nProperties:\n";
	rs = rs + "------------------------\n";
	for(Enumeration e = row.keys();e.hasMoreElements();){
		String next = (String)e.nextElement();		
		ArrayList list = (ArrayList)row.get(next);
		rs = rs + "\n\tName: " + next + " # of operators: " + list.size();
		for (Iterator i = list.iterator(); i.hasNext(); ){
			StatisticsObject so = (StatisticsObject)i.next();
			rs = rs + "\n\t\t" + so.toString();
		}  
	}
	rs = rs + "\n=========================\n";
	return rs;
}
/**
 * This will update the given propertyName for the currentCategory.  If
 * the currentCategory is something besides "OVERALL", then the OVERALL
 * will be updated as well.  The OVERALL will be updated <emph>after</emph>
 * the other is updated.
 * The value is updated by calling the calculateNewValue method in the corresponding StatisticsObject
 * @param propertyName java.lang.String
 * @param queryObject the XATQueryObject whose property is being updated
 */
public void updateValue(String propertyName, XATQueryObject queryObject) {

	updateValue(propertyName, queryObject, null);
}
/**
 * This will update the given propertyName for the currentCategory.  If
 * the currentCategory is something besides "OVERALL", then the OVERALL
 * will be updated as well.  The OVERALL will be updated <emph>after</emph>
 * the other is updated.
 * The value is updated by calling the calculateNewValue method in the corresponding StatisticsObject
 * @param propertyName java.lang.String
 * @param queryObject the XATQueryObject whose property is being updated
 * @param newPropertyValue a string representing the new propertyvalue that will
 * assist the StatisticsObject in calculating its new value 
 */
public synchronized void updateValue(
    String propertyName,
    XATQueryObject queryObject,
    String newPropertyValue) {    
    if (!queryObject.isValidStatistic(propertyName)) {        
        throw new IllegalArgumentException(
            "The statistic " + propertyName + " is not supported by this QueryObject " + queryObject);
    }

    /** 
     * if it is a generic property, just get the value from the generic category 
     * if it isnt generic, then get the value from the current category
     */
    long time = 0; 
    if(timer){
	    time = System.currentTimeMillis();
    } 
    ModifiedArrayList list = null;
    if (ref.isGenericProperty(propertyName)) {
        list =
            (ModifiedArrayList) ((Hashtable) statsTable.get(this.GENERIC)).get(
                propertyName);
    } else {
        list =
            (ModifiedArrayList) ((Hashtable) statsTable.get(currentCategory)).get(
                propertyName);
    }
    StatisticsObject statsObject = list.getStatisticsObject(queryObject);

    /** 
     * we have to make this available to others
     * because the calculate new value methods will need to access the stats gatherer
     * however, we wont notifyAll because that could cause other threads to access this.  
     */
    //available = true;

    if (statsObject == null) {
        throw new XATStatisticsException("Error looking up the query object in the statistics gatherer. \n This is probably due to the queryobject not being initialized in the stats gatherer");
    }
    /** actually calculate the new value */
    if (newPropertyValue != null) {
        statsObject.calculateNewValue(newPropertyValue);
    } else {
        statsObject.calculateNewValue();
    }
    /** sort the list */
    //Collections.sort(list);
    /** now check if we have to update the OVERALL category */
    if (!ref.isGenericProperty(propertyName)) {
        if (!currentCategory.equals(StatisticsGatherer.OVERALL)) {
            list =
                (ModifiedArrayList)
                    ((Hashtable) statsTable.get(StatisticsGatherer.OVERALL)).get(
                    propertyName);
            statsObject = list.getStatisticsObject(queryObject);
            if (newPropertyValue != null) {
                statsObject.calculateNewValue(newPropertyValue);
            } else {
                statsObject.calculateNewValue();
            }
            /** sort the list */
            //Collections.sort(list);
        }
    }
    /** update some properties about the stats gatherer */
    if(timer){
		time = System.currentTimeMillis() - time;		
		long oldTime = Long.parseLong(getProperty("TOTAL_UPDATE_VALUE_TIME"));
		time = oldTime + time;
		long timesRun = (Long.parseLong(getProperty("TOTAL_UPDATE_VALUE_REQUESTS"))) + 1;
		try{
			setProperty("TOTAL_UPDATE_VALUE_TIME", "" + time);
			setProperty("TOTAL_UPDATE_VALUE_REQUESTS", "" + timesRun);
		} catch(Exception e){
			System.out.println(e.toString());
		}
		
    }
}
}
