package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;

/**
 * <b> Subclasses of this should override the calculateNewValue() and calculateNewValue(String)
 * method. </b> The compareTo(Object) method provides a default comparision (on propertyValue),
 * but should be overwritten if more is desired.  The calculateNewValue just copies the value from the operator <p>
 * 
 * Statistics Object that stores a propertyName, propertyValue,
 * and an associated operator.  The StatisticsObject stores an operator's value
 * of a property at a certain time.  The propertyValue here does not necessarily
 * correspond to the value in the operator because the operator's value is updated
 * continuosly (after each time the operator is run) while this value is updated
 * only when the execution controller decides to do so. <p>
 *
 * This object implements Comparable so that one StatisticsObject may be compared
 * to another.  This allows the StatisticsGatherer to sort these objects based
 * on the given property. <p>
 *
 * The calculateNewValue is called by the statisticsGatherer.  
 * </pre>
 * 
 * Creation date: (1/10/2003 4:45:18 PM)
 * @author: Brad Pielech
 */
public class StatisticsObject extends XATQueryObjectImp implements Comparable {

	private String propertyName;
	private XATQueryObject associatedQueryObject;
	private String propertyValue;
	/** 
	 * Each StatisticsObject is associated with a certain
	 * category.  The category corresponds to a category found in the statsTable
	 */
	private java.lang.Object category;
/**
 * Default Constructor
 */
public StatisticsObject() {
	super();
	init();
}
/**
 * StatisticsObject constructor comment.
 */
public StatisticsObject(String name, XATQueryObject queryObject) {
	super();
	propertyName = name;
	associatedQueryObject = queryObject;	
	init();
}
/**
 * Calculates the new Value of the given property of this operator.  This
 * should be overwritten by subclasses who can tailor this method to calculating 1
 * specific property. <p>
 * This method should update at least the local propertyValue and possibly the
 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
 *
 * The procedure for calculating new values should be as follows:
 *
 * <ol>
 * <li> Consult the statsTable for any propertyValues that are needed.
 * <li> Perform the required calculations using the propertyValues from step 1
 * <li> Update the local propertyValue.
 * 
 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
 * </ol> 
 */
public void calculateNewValue() {
		
}
/**
 * This is semantically the same as <tt>calculateNewValue()</tt>, but this
 * provides a parameter which assists in the calculation.
 * Calculates the new Value of the given property of this operator.  This
 * should be overwritten by subclasses who can tailor this method to calculating 1
 * specific property. <p>
 * This method should update at least the local propertyValue and possibly the
 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
 *
 * The procedure for calculating new values should be as follows:
 *
 * <ol>
 * <li> Consult the statsTable for any propertyValues that are needed.
 * <li> Perform the required calculations using the propertyValues from step 1
 * <li> Update the local propertyValue.
 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
 * </ol> 
 */
public void calculateNewValue(String newValue) {
	
	/**
	* dummy implementation that just sets this value to whatever is in the operator
	*/
	this.setPropertyValue(newValue);	
	
}
/**
 * It is important that care is taken when implementing this method.  Implementing
 * it incorrectly will lead to incorrect behavior later when the array is sorted
 * and the ArrayList tries to get values by identifiers. <p>
 *
 * This particular implementation tries to determine the data type of
 * the propertyValue and then do a standard comparison.		 
 * 
 * Compares this object with the specified object for order.  Returns a
 * negative integer, zero, or a positive integer as this object is less
 * than, equal to, or greater than the specified object.<p>
 *
 * The implementor must ensure <tt>sgn(x.compareTo(y)) ==
 * -sgn(y.compareTo(x))</tt> for all <tt>x</tt> and <tt>y</tt>.  (This
 * implies that <tt>x.compareTo(y)</tt> must throw an exception iff
 * <tt>y.compareTo(x)</tt> throws an exception.)<p>
 *
 * The implementor must also ensure that the relation is transitive:
 * <tt>(x.compareTo(y)&gt;0 &amp;&amp; y.compareTo(z)&gt;0)</tt> implies
 * <tt>x.compareTo(z)&gt;0</tt>.<p>
 *
 * Finally, the implementer must ensure that <tt>x.compareTo(y)==0</tt>
 * implies that <tt>sgn(x.compareTo(z)) == sgn(y.compareTo(z))</tt>, for
 * all <tt>z</tt>.<p>
 *
 * It is strongly recommended, but <i>not</i> strictly required that
 * <tt>(x.compareTo(y)==0) == (x.equals(y))</tt>.  Generally speaking, any
 * class that implements the <tt>Comparable</tt> interface and violates
 * this condition should clearly indicate this fact.  The recommended
 * language is "Note: this class has a natural ordering that is
 * inconsistent with equals."
 * 
 * @param   o the Object to be compared.
 * @return  a negative integer, zero, or a positive integer as this object
 *		is less than, equal to, or greater than the specified object.
 * 
 * @throws ClassCastException if the specified object's type prevents it
 *         from being compared to this Object.
 */
public int compareTo(java.lang.Object o) {
    StatisticsObject statsO = (StatisticsObject) o;
    String compareValue = statsO.getPropertyValue();
    /** 
     * first compare the names of the properties, if they are the same,
     * compare the values, otherwise, return the result of that comparison
     */
    int compareName = propertyName.compareTo(statsO.getPropertyName());
    if (compareName != 0) {
        return compareName;
    } else {
	    try{
        //ok, try a double			
        Double d1 = new Double(propertyValue);
        Double d2 = new Double(compareValue);
        return d1.compareTo(d2);
    } catch (NumberFormatException e3) {
    //hrm, ok, must be a String
    return propertyValue.compareTo(compareValue);
}

}
}
/**
 * Compare this to another StatsObject. They are the same
 * if the operator and propertyName are the same. 
 * @return int
 * @param otherObject java.lang.Object
 */
public boolean equals(Object otherObject) {
	try {
		StatisticsObject stats = (StatisticsObject) otherObject;
    return (
        (this.associatedQueryObject == stats.getAssociatedQueryObject())
            && (this.propertyName == stats.getPropertyName()));
	} catch (ClassCastException e) {
		return false;
	}
    
}
/**
 * Insert the method's description here.
 * Creation date: (1/14/2003 10:45:41 AM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator
 */
public XATQueryObject getAssociatedQueryObject() {
	return associatedQueryObject;
}
/**
 * Insert the method's description here.
 * Creation date: (1/17/2003 3:10:07 PM)
 * @return java.lang.Object
 */
public java.lang.Object getCategory() {
	return category;
}
/**
 * Insert the method's description here.
 * Creation date: (1/14/2003 10:47:52 AM)
 * @return java.lang.String
 */
public java.lang.String getPropertyName() {
	return propertyName;
}
/**
 * Get the value of the property specfied by this.getPropertyName().  This does not
 * necessarily equal this.getAssocitatedOperator().getProperty(this.getPropertyName)
 * because the operator's value is updated continuosly (after each time the operator is run) 
 * while this value is updated
 * only when the execution controller decides to do so.  Therefore, this value may be a historical value.
 * The last time it was update may be accessed through this.getLastTimeUpdated();
 * Creation date: (1/10/2003 4:57:46 PM)
 * @return java.lang.Object
 */
public java.lang.String getPropertyValue() {
	return propertyValue;
}
/**
 * Insert the method's description here.
 * Creation date: (2/23/2003 6:22:45 PM)
 */
private void init() {
	}
/**
 * Insert the method's description here.
 * Creation date: (1/14/2003 10:45:41 AM)
 * @param newAssociatedOperator edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator
 */
public void setAssociatedQueryObject(XATQueryObject newAssociatedQueryObject) {
	associatedQueryObject = newAssociatedQueryObject;
}
/**
 * Insert the method's description here.
 * Creation date: (1/17/2003 3:10:07 PM)
 * @param newCategory java.lang.Object
 */
public void setCategory(java.lang.Object newCategory) {
	category = newCategory;
}
/**
 * Insert the method's description here.
 * Creation date: (1/14/2003 10:47:52 AM)
 * @param newPropertyName java.lang.String
 */
public void setPropertyName(java.lang.String newPropertyName) {
	propertyName = newPropertyName;
}
/**
 * Sets the property value.  It is better to use the calculateNewValue() method
 * to update the property because this method will just blindly take the parameter and
 * set it to the internal value.  Using calculateNewValue() will allow the subclasses to perform 
 * required calculations
 * @param newPropertyValue java.lang.String
 */
public void setPropertyValue(String newPropertyValue) {
	propertyValue = newPropertyValue;
}
/**
 * Insert the method's description here.
 * Creation date: (1/20/03 8:52:33 PM)
 * @return java.lang.String
 */
public String toString() {
	String rs = new String();
	rs = "\n\t\t\tStatisticsObject" + hashCode();
	rs = rs + "\n\t\t\t QueryObject: " + getAssociatedQueryObject();
	rs = rs + "\n\t\t\t propertyName: " + getPropertyName();
	rs = rs + "\n\t\t\t propertyValue: " + getPropertyValue();
	return rs;
}
}
