package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.apache.html.dom.HTMLDOMImplementationImpl;
import org.apache.xml.serialize.LineSeparator;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.NameValuePair;
/**
 * StatisticsPrinter is used to display the statistics gatherer. It contains
 * a list of property names to display such that only the relevant or desired
 * properies can be printed.  <p>
 * The output format should be set to either HTML, XML (doesnt work), or CSV.  
 * The output file name and columnHeaders should be set as well.  Then,
 * you can either set it up to periodically print out statistics by calling
 * <tt> run() </tt>.  Make sure to correctly configure the numberOfPrintOuts and frequency.
 * To print out on demand, call 
 * <tt>getHTMLStatistics()</tt> (or the appropriate getStatistics method) and then calling
 * <tt> print() </tt>.    <p> 
 */
public class StatisticsPrinter extends XATQueryObjectImp implements Runnable{
	
	/** constant to represent HTML output */
	public static final int HTML = 0;
	/** constant to represent XML output */
	public static final int XML = 1;
	/** constant to represent comma separated output */
	public static final int CSV = 2;

	public static final long ONE_SECOND = 1000;
	/** the array of column header (column names) that will be printed */
	private java.util.ArrayList columnHeaders;
	/** this is the category in the statsTable that will be printed */
	private java.lang.Object category;
	/** this is the old category in the statsTable */
	private java.lang.Object oldCategory;
	/** the document that is built up each time this is run */
	private org.w3c.dom.Document document;
	/** the time that the thread begins to run*/
	private long startTime;
	/** the comma separated values string that will be printed out */
	private StringBuffer csvString;
	/** true if we should print the overall statistics every time */
	private boolean printOverall;
	/** 
	 the comma separated values string that will contain the OVERALL category
	 if the user wishes to print that out to a separate file
	 */
	private StringBuffer overallCSVString;
	/** the sequence number that will be appended to the name of the output file*/	
	private long sequenceNum;

	private boolean printBlankRows;
	/** true if the headers should always be printed */
	private boolean alwaysPrintHeader;
	private boolean headersPrinted;
/**
 * Sets default values for output file name, output format, and frequency
 */
public StatisticsPrinter() {
	super();	
	init();
}
/**
 * Add a column header to the printer at the given position.  Throws IllegalArgumentException if the propertyName is not valid
 * @param position int
 * @param propertyName String
 * @throws IllegalArgumentException if the propertyName is not valid
 */
public void addColumnHeader(int position, String propertyName) {
	
	if(stats.isStatisticSupported(propertyName)){
		columnHeaders.add(position, propertyName);
	} else{
		throw new IllegalArgumentException("The propertyName is not supported by the StatisticsGatherer");
	}
	
}
/**
 * Add a column header to the printer.  Throws IllegalArgumentException if the propertyName is not valid
 * The column name will be inserted at the end
 * @param propertyName String
 * @throws IllegalArgumentException if the propertyName is not valid
 */
public void addColumnHeader(String propertyName) {
	addColumnHeader(columnHeaders.size(), propertyName);	
}
/**
 * calls the correct getStatistics method (either getHTML, getCSV, or getXML)
 * @param format either this.XML, HTML, or CSV
 */
public void gatherStatistics() {
	int format = getFormat();	
	if(format == StatisticsPrinter.HTML){
		getHTMLStatistics();		
	}
	if(format == StatisticsPrinter.CSV){
		csvString.append(getCSVStatistics(getCategory()));
		if(printOverall){
			overallCSVString.append(getCSVStatistics(stats.OVERALL));
		}
	}
}
/**
 * Returns the category that this printer will display 
 * @return java.lang.Object
 */
public java.lang.Object getCategory() {
	return category;
}
/**
 * Encodes the statistics table as a comma separated document
 * in a tabular format with only the specified columns. 
 * This doesnt return the csv statistics, but rather it gatherers them for printout
 * @param Object category, the category of statistics to gather
 */
private StringBuffer getCSVStatistics(Object category) {
	StringBuffer s = new StringBuffer();
    stats.setCurrentCategory(category);
    s.append(category).append("\ntimestamp: ").append(System.currentTimeMillis() - startTime).append("\n");    

    if((alwaysPrintHeader) || ((!alwaysPrintHeader) && !headersPrinted)) {
	    s.append("QueryObject,");
    	for (Iterator i = columnHeaders.iterator(); i.hasNext();) {
        	s.append(((NameValuePair) i.next()).getName()).append(",");
    	}
    	headersPrinted = true;
    }
    
    s.delete(s.length() - 1, s.length());
    s.append("\n");
    for (Iterator i = stats.getAllKnownQueryObjects(); i.hasNext();) {
        XATQueryObject qo = (XATQueryObject) i.next();
        StringBuffer row = new StringBuffer();
        /** dont want the whole query plan printed out here, it is just messy*/
        if (qo instanceof XATTree) {
            row.append("Tree,");
        } else {
            row.append(qo).append(",");
        }
        /** count how many times that the query object did not support the property */
        int noValueCount = 0;
        /** iterate through the columnHeaders */        
        for (int i2 = 0; i2 < columnHeaders.size(); i2++) {
            NameValuePair pair = (NameValuePair) columnHeaders.get(i2);
            String nextColumn = pair.getName();
            String quantifier = pair.getValue();
            String value = null;
            try {
                if (quantifier.equalsIgnoreCase("average")) {
                    value = stats.getValue(nextColumn, qo, StatisticsReference.AVERAGE);
                } else {
                    value = stats.getValue(nextColumn, qo);

                }
            } catch (IllegalArgumentException e) {
                /* 
                 * if the property is not supported by this query object, just
                 * print dashes in its place 
                 */
                value = "------";
                noValueCount++;
            }
            row.append(value).append(",");
        }
        /* 
          if we are printing empty rows or the previous query object had 1 value that
          was not empty (-----), then we will append the row

          else, if we do not want to print empty rows and all of the rows were empty,
          we will not append the previous row to the csvString 
        */
        if(printBlankRows || (noValueCount != columnHeaders.size())){
			s.append(row.toString());
        } else{
			//dont append 
        }
			        
        s.delete(s.length() - 1, s.length());
        s.append("\n");
    }
    //add another blank line so there is a spacing between categories
	s.append("\n");
    stats.setCurrentCategory(oldCategory);
    return s;
}
/**
 * Returns the fileName to print to, including the appended sequence number.
 * The parameter is used to determine if we are printing to the overall file or
 * the regular.  
 */
private String getFileName(boolean overall) {
	String fileName = null;
	if(!overall){
		fileName = this.getProperty("FILE_NAME");
	} else{
		fileName = this.getProperty("OVERALL_FILENAME");
	}
    /** insert the sequence num into the fileName */
    int length = 4;
    int outputFormat = getFormat();
    if (outputFormat == this.HTML) {
        length = 5;
    }
    String ext = fileName.substring(fileName.length() - length);
    fileName = fileName.substring(0, fileName.length() - length);
    fileName = fileName + "" + sequenceNum + ext;
    return fileName;
}
/**
 * Parses the format property by converting the HTML, CSV, or XML string
 * into the corresponding class variable constant
 * @return int
 */
public int getFormat() {
	String format = this.getProperty("FORMAT");
	if(format.equalsIgnoreCase("HTML")){
		return this.HTML;
	} 
	if(format.equalsIgnoreCase("XML")){
		return this.XML;
	}
	if(format.equalsIgnoreCase("CSV")){
		return this.CSV;
	}
	return this.HTML;	
}
/**
 * Encodes the statistics table as an HTML document 
 * in a tabular format with only the specified columns.
 * DOM is used here to create the HTML document.  This allows for the overall document to be updated
 * easily by inserting new children.  This is a little more cumbersome than creating a String, but
 * this is the correct way to do it.   
 */
private void getHTMLStatistics() {
    stats.setCurrentCategory(category);

    if (document == null) {
        document =
            HTMLDOMImplementationImpl.getHTMLDOMImplementation().createHTMLDocument(
                "Stats table printout");
    }
    Element body = ((org.w3c.dom.html.HTMLDocument) document).getBody();
    Element heading = document.createElement("h3");
    heading.appendChild(
        document.createTextNode(
            "Category: "
                + getCategory()
                + " timestamp: "
                + (System.currentTimeMillis() - startTime)));
    body.appendChild(heading);
    body.appendChild(document.createElement("br"));
    Element table = document.createElement("table");
    body.appendChild(table);
    table.setAttribute("rules", "all");
    table.setAttribute("cellspacing", "0%");
    table.setAttribute("cellpadding", "5%");
    Element row = document.createElement("tr");
    table.appendChild(row);
    Element tableHeader = document.createElement("th");
    row.appendChild(tableHeader);

    if ((alwaysPrintHeader) || ((!alwaysPrintHeader) && !headersPrinted)) {
        headersPrinted = true;
        tableHeader.appendChild(document.createTextNode("QueryObject:"));
        for (Iterator i = columnHeaders.iterator(); i.hasNext();) {
            String nextColumn = ((NameValuePair) i.next()).getName();
            tableHeader = document.createElement("th");
            row.appendChild(tableHeader);
            tableHeader.appendChild(document.createTextNode(nextColumn));
        }
    }

    for (Iterator i = stats.getAllKnownQueryObjects(); i.hasNext();) {
        XATQueryObject qo = (XATQueryObject) i.next();
        //make a new row in the table for this next query object
        row = document.createElement("tr");
        table.appendChild(row);
        Element cell = document.createElement("td");
        row.appendChild(cell);
        if (qo instanceof XATTree) {
            cell.appendChild(document.createTextNode("Tree"));
        } else {
            cell.appendChild(document.createTextNode(qo.toString()));
        }
        for (Iterator i2 = columnHeaders.iterator(); i2.hasNext();) {
            NameValuePair pair = (NameValuePair) i2.next();
            try {
                String nextColumn = pair.getName();
                String quantifier = pair.getValue();
                String value = null;
                if (quantifier.equalsIgnoreCase("average")) {
                    value = stats.getValue(nextColumn, qo, StatisticsReference.AVERAGE);
                } else {
                    value = stats.getValue(nextColumn, qo);
                }
                cell = document.createElement("td");
                row.appendChild(cell);
                cell.appendChild(document.createTextNode(value));
            } catch (IllegalArgumentException e) {
                /* 
                 * if the property is not supported by this query object, just
                 * print dashes in its place 
                 */
                cell = document.createElement("td");
                row.appendChild(cell);
                cell.appendChild(document.createTextNode("------"));
            }
        }
    }

    stats.setCurrentCategory(oldCategory);
}
/**
 * Initializes more things about the statistics printer
 */
private void init() {
    columnHeaders = new ArrayList();    
    setCategory(StatisticsGatherer.OVERALL);
    document = null;
    csvString = new StringBuffer();    
    /** set up the sequence num */
    String name = "printer_sequence_num";
    try {
        File f = new File(name);
        long number = 0;
        // if file exist read from it.
        if (f.exists()) {

            DataInputStream in = new DataInputStream(new FileInputStream(name));
            number = in.readLong();
            in.close();
        }
        // then we can write it to the file again.
        DataOutputStream out = new DataOutputStream(new FileOutputStream(name));
        out.writeLong(number + 10);
        out.close();
        sequenceNum = number;       
    } catch (Exception e) {
        sequenceNum = 0;
    }
    headersPrinted = false;
	startTime = System.currentTimeMillis();
}
/**
 * This will interuppt the printer
 */
public void interrupt() {
	Thread.currentThread().interrupt();
	
	}
/**
 * Actually prints the document to the file system
 */
public void print() {
	String fileName = getFileName(false);
	if(getFormat() == this.HTML){
		printHTML(fileName);
	} else if(getFormat() == this.CSV){
		printCSV(fileName, csvString);
		if(printOverall){
			fileName = getFileName(true);
			printCSV(fileName, overallCSVString);
		}
	}

}
/**
 * Actually prints the document to the file system.  This method
 * allows for a heading to be placed at the top. This heading is just plain text.
 * @param fileName the filename to print too
 * @param data the string buffer that contains the data to print
 */
private void printCSV(String fileName, StringBuffer data) {

    try {
        FileWriter out = new FileWriter(fileName, true);
        /** print the csvString*/
        out.write(data.toString());
        out.close();
        /** now erase the string so we dont print it out again later*/
        data = new StringBuffer();
    } catch (IOException e) {
        System.out.println(e.toString());
    }

}
/**
 * Actually prints the document to the file system.  This method
 * allows for a heading to be placed at the top. This heading is just plain text.
 */
private void printHTML(String fileName) {

    OutputFormat format = new OutputFormat(document);
    format.setLineSeparator(LineSeparator.Windows);
    format.setIndenting(true);
    format.setLineWidth(0);
    format.setPreserveSpace(true);
    try {
        XMLSerializer serializer =
            new XMLSerializer(new FileWriter(fileName, true), format);
        serializer.asDOMSerializer();
        serializer.serialize(document);
    } catch (IOException e) {
        System.out.println(e.toString());
    }

}
/**
 * Runs the StatisticsPrinter thread.  First this will sleep for
 * a specified amount of time, then it will print out to a file the contents
 * of the statistics gatherer.
 * This runs either for numberOfPrintouts or until it is interuppted <p>
 * This method should not be called from the outside, but rather <tt>this.start()</tt> should
 * be invoked.  This method is only public b/c this class implements Runnable, which calls for
 * a public method named run().
 */
public void run() {
	
	int numberOfPrintouts = Integer.parseInt(this.getProperty("NUMBER_OF_PRINTOUTS"));
	long frequency = Long.parseLong(this.getProperty("FREQUENCY"));
	printBlankRows = new Boolean(this.getProperty("PRINT_EMPTY_ROW")).booleanValue();
	alwaysPrintHeader = new Boolean(this.getProperty("ALWAYS_PRINT_HEADERS")).booleanValue();
	printOverall = new Boolean(this.getProperty("ALWAYS_PRINT_OVERALL")).booleanValue();
	if(printOverall){
		overallCSVString = new StringBuffer();
	}
	if(frequency == 0){
		return ; 
	}
	int outputFormat = this.getFormat();
	
    while (numberOfPrintouts > 1) {
	    
        try {	        
//			System.out.println("******** Printing out stats table *******");
            if (outputFormat == this.HTML) {	            
                getHTMLStatistics();
            }
            if (outputFormat == this.XML) {
                //		printXML();
            }
            if (outputFormat == this.CSV) {
                csvString.append(getCSVStatistics(getCategory()));
                if(printOverall){
					overallCSVString.append(getCSVStatistics(stats.OVERALL));
                }
            }
            long timeBefore = System.currentTimeMillis();
            Thread.currentThread().sleep(frequency);
            long timeAfter = System.currentTimeMillis();
            //System.out.println("******** Slept for " + (timeAfter - timeBefore) + " *********");
        } catch (InterruptedException e) {
	        //print();
            return;
        }
        numberOfPrintouts--;
    }
//    print();
	return ;
}
/**
 * Set the category in the stats table that will be printed
 * This will also set the oldCategory to be equal to what was the category before
 * @param newCategory java.lang.Object
 * @throws XATStatisticsException if the category is not valid
 */
public void setCategory(java.lang.Object newCategory) {
	oldCategory = stats.getCurrentCategory();
	/** 
	 * try to set the statsTable's category to the new category 
	 * if it is not valid, then an exception will be thrown 
	 */
	stats.setCurrentCategory(newCategory);
	category = newCategory;
	/** now reset it so this doesnt mess up other operations */
	stats.setCurrentCategory(oldCategory);
}
/**
 * Adds all of the column headers to the printer. The existing column headers
 * are replaced.  The Collection should contain strings representing valid property Names.
 * The properties will be added in the order specified by the Collection's iterator. <p>
 * Throws IllegalArgumentException if any of the propertyName are not valid
 * If any of the names are not valid, then none of them will be added and the column
 * headers will remain unchanges.  
 * @param propertyNames Collection
 * @throws IllegalArgumentException if any of the propertyNames are not valid
 */
public void setColumnHeaders(Collection propertyNames) {

	for (Iterator i = propertyNames.iterator();i.hasNext();){
		String nextName = ((NameValuePair)i.next()).getName();
		if(!stats.isStatisticSupported(nextName)){
			throw new IllegalArgumentException("The propertyName [" + nextName + "] is not supported by the StatisticsGatherer");
		}
	}
	columnHeaders.clear();
	columnHeaders.addAll(propertyNames);	
}
	/**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    */   
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		addProperty("FILE_NAME", "output");
		/** if frequency is set to 0, then this will not print out periodically */
		addProperty("FREQUENCY", "0");
		addProperty("FORMAT", "HTML");
		addProperty("NUMBER_OF_PRINTOUTS", "10000");
		/** 
		 if this is set to true, this printer will print out every row, even if no
		 statistics supported by the query object are being printed.  
		*/
		addProperty("PRINT_EMPTY_ROW", "true");
		/** 
		 if this is set to true, this printer will print out the headers
		 every time (for the duration of 1 experiment).  Setting this to false
		 will cause the printer to only output the headers once per experiment.  This
		 will make the output file 
		*/
		addProperty("ALWAYS_PRINT_HEADERS", "true");

		/** 
		 if this is set to true, this printer will print out the statistics for
		 the OVERALL category every time to a separate file.
		 
		 This will only work for CSV printing because the HTML is really just a toy 
		*/
		addProperty("ALWAYS_PRINT_OVERALL", "false");

		/** 
		 if this PRINT_OVERALL is set to true, then this property will contain the
		 file name that the overall is printed to
		*/
		addProperty("OVERALL_FILENAME", "overall.csv");
		
		
		
	}
	
/**
 * Set the meta information about the experiment.  This will
 * appear at the top of the output
 * @param meta java.lang.String
 */
public void setMetaInformation(String meta) {
	
	int format = getFormat();	
	if(format == this.HTML){		
		if(document == null){
			document = HTMLDOMImplementationImpl.getHTMLDOMImplementation().createHTMLDocument("Stats table printout"); 
		}
		Element body = ((org.w3c.dom.html.HTMLDocument)document).getBody();	
		Node first = body.getFirstChild();
		Element metaE = document.createElement("p");
		StringTokenizer st = new StringTokenizer(meta,"|");
	    while (st.hasMoreTokens()) {
		    metaE.appendChild(document.createTextNode(st.nextToken()));
      		metaE.appendChild(document.createElement("br"));  	
     	}			
		body.insertBefore(metaE, first);				
	}
	if(format == this.CSV){
		csvString.insert(0,meta + "\n");
	}
	
	
	}
/**
* sets the given property to the new value.  This method is responsible
* for validating the given property.  An exception is thrown if
* the value is not correct.
* Validation should be done in the following manner: 
* First the method should check if the object supports the given property by calling
* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
* be thrown. <p>
*
* Once it is known that the object supports the property, the value must be validated.  Each implementor of
* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
* will reduce duplicate code. <p>
*
* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
* it is up to the subclass that created the property to validate it (as described above).
* @param name the name of the property to set
* @param value the new value of the property
* @throws Exception if the given property cannot be found or
* if the value is not valid for the given property.
*/

public void setProperty(String name, String value) throws Exception {

    if (isValidPropertyName(name)) {
        if(name == "FORMAT"){
	        String[] s = new String[]{"HTML", "XML","CSV"};
	        for (int i = 0; i < s.length; i++){
	        	if(value.equalsIgnoreCase(s[i])){
					this.setNewPropertyValue(name, value);
	        	}
	        }
	        throw new IllegalArgumentException(
                        "The property named " + name + " must be either HTML, XML, or CSV"); 	
        } else{
			super.setProperty(name, value);
        }

    } else {
        throw new Exception("The object does not support the property named " + name);
    }
}
}
