package edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import edu.wpi.cs.dsrg.utils.BindingTable;
import edu.wpi.cs.dsrg.utils.BindingTableComputer;
import edu.wpi.cs.dsrg.utils.RewriteUtilities;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumnImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Navigate;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * This Class is used to cancel out Navigate and Tagger by using the binding tables.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class NavTaggerCancelout {
/**
 * Default NavTaggerCancelout constructor.
 * @since 1.0
 */
public NavTaggerCancelout() {
    super();
}
/**
 * Cancel out Navigate and Tagger if possible.
 * This method first compute the Simplified Full Path for each variable in the Binding Table.
 * Then it will compare the values for each keys, and find alias.
 * Third, search through the trees to change the Navigation with alias into rename operators.
 * Forth, do the schema clean up to get rid of the unnecessary rename and tagger operators.
 * @return void
 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 * @since 1.0
 * @see XATTree
 */
public static void cancelOut(XATTree tree) {
    // computs its binding table.
    BindingTable btable = new BindingTableComputer().computeBindingTable(tree);

    // compute the full path for each keys in the binding table and sort those keys by their values in String.
    // There are two hashtables used.
    // AliasHash  is (String value, vector alias) to hash between values of keys and the keys.
    // KeyHash is (String key, vector alias|String key) to hash between key and their alias, or their rename Key.
    // ChainHash is (String key, vector alias) to hash between key and their replacement on their chains.
    Hashtable AliasHash = new Hashtable();
    Hashtable KeyHash = new Hashtable();
    Hashtable ChainHash = new Hashtable();

    // now, populate the three hashtables.
    Enumeration keys = btable.keys();
    while (keys.hasMoreElements()) {
        String key = (String) keys.nextElement();
        String value = btable.navStepsToString(btable.computeSimplifiedFullPath(key));
        // check if the hashtable has this value already or not.
        Object alias = AliasHash.get(value);
        if (alias == null) {
            alias = new Vector();
        }
        ((Vector) alias).add(key);
        // update the alias hashtable.
        AliasHash.put(value, alias);
        // update the key hash table.
        KeyHash.put(key, alias);
        // update the chain hash table.
        ChainHash.put(key, computeAlias(key, btable));
    }

    if (isDebug()) {

        // print out the binding table.
        System.out.println("***NavTagger Cancel Out: Binding Table.");
        System.out.println(btable.toString());
        System.out.println("***NavTagger Cancel Out: Binding Table (FULL PATH).");
        System.out.println(btable.toString(true));
        // print out the AliasHash table.
        System.out.println("***NavTagger Cancel Out: AliasHash Table.");
        System.out.println(printHashtable(AliasHash));
        // print out the KeyHash table.
        System.out.println("***NavTagger Cancel Out: KeyHash Table.");
        System.out.println(printHashtable(KeyHash));
        // print out the KeyHash table.
        System.out.println("***NavTagger Cancel Out: ChainHash Table.");
        System.out.println(printHashtable(ChainHash));
    }

    // the ChainHash used later to help the rewriting of the tree.

    // now, traverse the tree to find all the navigate operators has the key as the output column.
    replaceNavToName(tree.getRoot(), ChainHash);

    // update Tree's binding table.
    tree.setBindingTable(new BindingTableComputer().computeBindingTable(tree));

    //** clean up the useless rename and taggers.
    //** This is done by calling the rewrite modules.
    RewriteUtilities.cleanXATAfterGeneration(tree);
}
/**
 * Compute the alias of given column.
 * This method will return a Vector of columns on this Key's chain columns,
 * And then if the columns'full path same as this Key's simplifiedpath,
 * Then, it is its alias.
 * @return Vector All the alias of this Key. The lowest one is at the end of this Vector.
 * @param key The column name searching alias for.
 * @param bind The BindingTable used to search alias for the given column name 'key'.
 * @since 1.0
 * @see BindingTable
 */
public static Vector computeAlias(String key, BindingTable btable) {

    /**
     * Note: This alias computation method will only limited on this Key's Chain.
     * It's good for Nav-tagger cancel out, but
     * It's not good for find-common sub-expressions on columns.
     */
    Vector result = new Vector();

    if (key == null || btable == null)
        return result;

    // first, get this Key's simplified full path.
    String simpPath =
        btable.navStepsToString(btable.computeSimplifiedFullPath(key));

    // second, get this Key's chain.
    Vector chain = btable.getChain(key);

    if (isDebug()) {
        System.err.println("Chain for key [" + key + "] is: " + chain);
    }

    // third, for each key on the chain, check if their fullPath same as  this key's simplificed Path.
    for (int i = 0; i < chain.size(); i++) {
        String step = (String) chain.elementAt(i);
        String fullPath = btable.navStepsToString(btable.getVariableValue(step, true));
        if (simpPath.equals(fullPath)) {
            result.add(step);
        }
    }

    return result;
}
/**
 * Get the value of the propertity 'DEBUG_NavTaggerCancelout'.
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
public static boolean isDebug() {
	return System.getProperty("DEBUG_NavTaggerCancelout")!=null;
}
/**
 * Print out a Hashtable in a good format.
 * @return java.lang.String
 * @param hash java.util.Hashtable
 * @since 1.0
 */
static String printHashtable(Hashtable hash) {
	StringWriter strWrt = new StringWriter();
	PrintWriter pout = new PrintWriter(strWrt);

	if(hash == null)
		return "NULL hashtable.";

	Enumeration keys = hash.keys();
	while(keys.hasMoreElements()){
		Object key = keys.nextElement();
		Object value = hash.get(key);
		pout.println(key + "\t|\t" + value);
	}
	
	return strWrt.toString();
}
/**
 * A recursive function to replace Nav to Name.
 * The ChainHash has (String key, Vector alias).
 * @return void
 * @param root The root XATNode of the given tree.
 * @param ChainHash java.util.Hashtable
 * @since 1.0
 * @see Navigate
 * @see NameColumn
 */
static void replaceNavToName(XATNode root, Hashtable ChainHash) {
    XATOperator op = root.getOperator();

    if (op instanceof Navigate) {
        // get the output column of this navigate.
        String outCol = op.getOutputColumnName();

        // get the value from the ChainHash.
        Vector value = (Vector) ChainHash.get(outCol);

        if (value == null) {
	        // CANNOT BE.
	        if (isDebug())
	        	System.err.println("NavTaggerCancelOut.replaceNavToName CANNOT BE HERE. Check ChainHash generation!");
        }else if(value.size() == 0){
            // Then there is no alias for this Key in its lower chain.
            // do nothing.
            if (isDebug())
                System.out.println(op + " cannot be reduced further.");
        } else {

	        // get the lowest possible alias.
	        String rename = (String)value.elementAt(value.size() -1);

            // make the operator as a rename.
            NameColumn nameOp = new NameColumnImp(root, rename, outCol);
            // update this node.
            root.setOperator(nameOp);
            if (isDebug()) {
                System.out.println("Rename: " + rename + " --> " + outCol);
            }
        }
    } // else, its not an Nav, do nothing.

    // recursively process all of its children.
    XATNode[] children = root.getChildren();
    if (children == null)
        return;
    else {
        for (int iChild = 0; iChild < children.length; iChild++) {
            replaceNavToName(children[iChild], ChainHash);
        }
    }
}
/**
 * Set the propertity 'DEBUG_NavTaggerCancelout'.
 * @return void
 * @param debug The flag indicate: True - with debug information output. 
 * False - without debug information output.
 * @since 1.0
 */
public void setDebug(boolean debug) {
	System.setProperty("DEBUG_NavTaggerCancelout","OK");
}
}
