package edu.wpi.cs.dsrg.utils;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.*;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.*;

/**
 * This class compute the bindingTable for a given XAT.
 * Usage eg : BindingTableComputer btc = new BindingTableComputer();
 *            BindingTable bt = btc.computeBindingTable(xat);
 *            xat.setBindingTable( bt );
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class BindingTableComputer extends InheritableVisitor {

	/**
	 * The XATTree for computing binding table.
	 * @since 1.0
	 */
    private XATTree xat;

	/**
	 * The binding table computed.
	 * @since 1.0
	 */
    private BindingTable bindingTable;

    /**
     * The default BindingTableComputer constructor.
     * @since 1.0
     */
    public BindingTableComputer() {
    }
    /**
     * Compute the binding table for a given XAT.
     * @return BindingTable The result BindingTable computed.
     * @param xatTree The XATTree computing binding table for.
     * @since 1.0
     */
    public BindingTable computeBindingTable(XATTree xatTree) {

	    xat = xatTree;
	    
        bindingTable = new BindingTable();
        visit( xat.getRoot().getOperator() );

        return bindingTable;
    }
    public Object dispatch() {
        return null;

    }
public Object dispatch(Object o1) {
    if (System.getProperty("DEBUG_BindingTableComputer") != null) {
        System.err.println(
            "Dispatch!! BindingTable not computed for " + o1.getClass().toString());
    }
    XATOperator op = (XATOperator) o1;
    XATNode[] children = op.getNode().getChildren();
    for (int i = 0; i < children.length; i++)
        visit(children[i].getOperator());
    return null;
}
/**
 * Recursively compute all variable bindings of a subtree rooted at a "FOR" node.
 * @return The computing result.
 * @param fr The FOR operator.
 * @since 1.0
 */
public Object visit(FOR fr) {

	// to visit operators in for tree
	XATTree xat = fr.getXtreFunctions();
	visit(xat.getRoot().getOperator() );

	// to visit child operators of the for operator
	XATNode[] children = fr.getNode().getChildren();
	for( int i=0; i < children.length; i++ )
	visit( children[i].getOperator() );
	
	return null;
}
/**
 * Recursively compute all variable bindings of a subtree rooted at a "Function" node.
 * @return The computing result.
 * @param func The Function operator.
 * @since 1.0
 */
public Object visit(Function func) {

    // build the navigation step array with all the parameters
    // of this function.
    Object[] params = func.getParams();

    NavigationStep[] steps = new NavigationStep[params.length];

    for (int i = 0; i < params.length; i++) {
        if (params[i] instanceof ColumnName) {
            steps[i] = new VarStepImp(((ColumnName) params[i]).getColumnName());
        } else
            steps[i] = new StringStepImp(params[i].toString());
    }

    bindingTable.updateVariableTable(func.getOutputColumnName(), steps);

    XATNode[] children = func.getNode().getChildren();
    for (int i = 0; i < children.length; i++)
        visit(children[i].getOperator());

    return null;
}
/**
 * Recursively compute all variable bindings of a subtree rooted at a "NameColumn" node.
 * @return The computing result.
 * @param name The NameColumn operator.
 * @since 1.0
 */
 
public Object visit(NameColumn name) {

    String oldName = name.getOriginalColumnName();
    String newName = name.getNewColumnName();

    // If this is not redundant rename node then put it in the binding table
    // and if the new name and old name are different,
    if (oldName != null && !newName.equals(oldName))
        bindingTable.updateVariableTable(
            newName,
            new NavigationStep[] { new VarStepImp(oldName)});

    XATNode[] children = name.getNode().getChildren();
    for (int i = 0; i < children.length; i++)
        visit(children[i].getOperator());

    return null;
}
/**
 * Recursively compute all variable bindings of a subtree rooted at a "Source" node.
 * @return The computing result.
 * @param source The Source operator.
 * @since 1.0
 */
public Object visit(Source source) {
    NavigationStep[] value = null;
    BindingTable btable = xat.getBindingTable();
    if (btable != null)
        value = btable.getVariableValue(source.getOutputColumnName());
        
    if(value == null) 
        value = new NavigationStepImp[] { new NodeStepImp("SOURCE")};
    bindingTable.updateVariableTable(source.getOutputColumnName(), value);

    return null;
}
/**
 * Recursively compute all variable bindings of a subtree rooted at a "GroupBy" node.
 * @return The computing result.
 * @param groupby The GroupBy operator.
 * @since 1.0
 */
 
public Object visit(edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy groupby) {
    // to visit operators in for tree
    XATTree xat = groupby.getXATTree();
    if ((xat.getRoot()!=null)&&(xat.getRoot().getOperator()!=null))
    visit(xat.getRoot().getOperator());

    // to visit child operators of the for operator
    XATNode[] children = groupby.getNode().getChildren();
    for (int i = 0; i < children.length; i++)
        visit(children[i].getOperator());
    return null;
}
/**
 * Recursively compute all variable bindings of a subtree rooted at a "Navigate" node.
 * @return The computing result.
 * @param nav The Navigate operator.
 * @since 1.0
 */

public Object visit(Navigate nav) {

    NavigationStep[] entrypoint = nav.getEntryPoint();
    NavigationStep[] destination = nav.getDestination();
    Vector value = new Vector();
    for (int i = 0; i < entrypoint.length; i++)
        value.addElement(entrypoint[i]);

    if (destination != null) {
        for (int j = 0; j < destination.length; j++)
            value.addElement(destination[j]);
    } else if (nav instanceof NavUnnest){
	    // if that's an navigation unnest, a UNNEST step will be added to the binding table.
	    value.addElement(new NodeStepImp("UNNEST"));
    }else{
	    System.err.println("WARNING: Empty Destination. It might be used in NavUnnest.");
    }

    bindingTable
        .updateVariableTable(
            nav.getOutputColumnName(),
            (NavigationStep[]) value.toArray(new NavigationStep[] {
    }));

    XATNode[] children = nav.getNode().getChildren();
    for (int i = 0; i < children.length; i++)
        visit(children[i].getOperator());

    return null;
}
/**
 * Recursively compute all variable bindings of a subtree rooted at a "Tagger" node.
 * @return The computing result.
 * @param tag The Tagger operator.
 * @since 1.0
 */
public Object visit(Tagger tag) {
    ListPattern lp = null;
    if ((tag.getPattern() instanceof DOMPattern)) {
        lp =
            ((DOMPattern) tag.getPattern()).toListPattern(
                ((DOMPattern) tag.getPattern()).getRoot());
    } else {
        lp = (ListPattern) tag.getPattern();
    }

    bindingTable.updateVariableTable(tag.getOutputColumnName(), lp);

    XATNode[] children = tag.getNode().getChildren();
    for (int i = 0; i < children.length; i++)
        visit(children[i].getOperator());
    return null;
}
}
