package edu.wpi.cs.dsrg.xmldb.xat.common.xatnode;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;

/**
 * The class defines the data model of the XAT, which is a table of XML fragments.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public interface XATTable extends XATDataSource, ColumnAccessor, RowAccessor {
    /**
    * Append this table to given table.
    * @return void  
    * @param sour The source table will be appended.
    * @since 1.0
    */
    public void appendTable(XATTable sour);
    /**
     * Append a new tuple to the end of the table.
     * @return int The position of the new appended tuple.
     * @param new_tuple The new tuple going to be appended.
     * @since 1.0
     */
    int appendTuple(Vector new_tuple) throws XATTableSchemaNoMatchException;
    /**
     * Construct a new XATTable which has the same column names as this XATTable.
     * @return XATTable The new constructed XATTable.
     * @since 1.0
     */
    XATTable cloneSchema();
    /**
     * Delete all the tuples in this XATTable.
     * @return void
     * @since 1.0
     */
    void deleteAllTuples();
    /**
     * Delete a tuple of the specified position.
     * @return Vector The content of the deleted tuple.
     * @param position The position of the tuple going to be deleted.
     * @since 1.0
     */
    Vector deleteTuple(int position)
        throws XATTableOutOfRangeException, XATTableEmptyException;
    /**
     * This method takes two XATTables, finds all matching columns. 
     * @return Vector The vector containning all common columns of two XATTable.
     * @param tab2 The another source XATTable.
     * @since 1.0
     */
    public Vector getCommonColumnNames(XATTable tab2);
    /**
     * Get the Name of the XATTable.
     * @return String The name of the XATTable.
     * @since 1.0
     */
    String getName();
    /** 
     * Get total number of tuples in this table.
     * @return int The total number of tuples with this XATTable.
     * @since 1.0
     */

    int getNumberOfTuples();
    /**
     * Get the vector containing the columns contents.
     * @return Vector The Vector containing the columns contents.
     * @since 1.0
     */
    Vector getRowList();
    /**
     * Get a tuple in a specified position.
     * @return Vector The Vector of cells in one tuple.
     * @param position The position of the tuple start from 1.
     * @since 1.0
     */
    Vector getTuple(int position)
        throws XATTableEmptyException, XATTableOutOfRangeException;
    /**
     * This method takes two XATTables, and finds all un-matching columns. 
     * @return Vector Containing all the un-matching columns.
     * @param tab2 The another source XATTable.
     * @since 1.0
     */
    public Vector getUnCommonColumnNames(XATTable tab2);
    /**
     * Insert a new tuple to the table before specific position.
     * @return int The number of tuples after the insertion.
     * @param positoin The inserted tuple position.
     * @param new_tuple The new tuple content.
     * @since 1.0
     */
    int insertTuple(int position, Vector new_tuple)
        throws XATTableOutOfRangeException, XATTableSchemaNoMatchException;
    /**
     * Move a tuple to a specific position.
     * @return void
     * @param from_position The position of tuple to be moved.
     * @param to_position Identify the dest position moved to.
     * @since 1.0
     */
    void moveTuple(int from_position, int to_position)
        throws XATTableOutOfRangeException, XATTableEmptyException;
    /**
    * Prints the XATTable
    * @return void 
    * @since 1.0
    */
    void print();

    /** 
     * Set Column Name in the specified position.
     * @return void
     * @param newName The new name for column.
     * @param position The position for this new column name.
     * @since 1.0
     */

    void setColumnName(String newName, int position)
        throws XATTableOutOfRangeException;
    /**
     * Set the Name of the Table.
     * @return The old name of the XATTable.
     * @param newName The new name of the XATTable.
     * @since 1.0
     */
    String setName(String newName);
    /**
     * Convert the XATTable into a String for output.
     * @return String The result string.
     * @since 1.0
     */

    String toString();
    /**
     * Update a tuple.
     * @return Vector The old tuple updated. 
     * @param position The position of tuple to be updated.
     * @param new_tuple The new tuple.
     * @since 1.0
     */
    Vector updateTuple(int position, Vector new_tuple)
        throws
            XATTableOutOfRangeException,
            XATTableEmptyException,
            XATTableSchemaNoMatchException;
}