/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * Implements XATQueue.
 * All methods are synchronized for multi-thread accessing.
 * 
 * Creation date: (4/27/2003 1:23:37 AM)
 * @author: Jinhui Jian
 */
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.queuemanager.PlaceHolder;
import edu.wpi.cs.dsrg.xmldb.xat.component.queuemanager.XATQueueManager;

public class XATPersistentQueueImp extends XATQueueImp {
	
	/** The index of the place-holder in the queue. If phIndex < 0, then
	 *  there are no tuples stored on disk. This is only for performance
	 *  purposes to avoid iterating over the queue every time we need to
	 *  read from or write to disk. */
	private int phIndex = -1;
	
	/** A handle for the queue manager. */
	private XATQueueManager queueManager = XATQueueManager.getQueueManager();	

	/** The size of this queue, including tuples stored on disk. This is 
	 *  for performance purposes only, to avoid iterating over the queue
	 *  every time we need to know its size. */
	private int queueSize = 0;	
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:37:04 PM)
	 */
	public XATPersistentQueueImp() {
		super();
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:59 PM)
	 * @param owner Object
	 */
	public XATPersistentQueueImp(Object owner) {
		super(owner);
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:53 PM)
	 * @param owner Object
	 * @param queueName String
	 */
	public XATPersistentQueueImp(Object owner, String queueName) {
		super(owner, queueName);
	}

	/**
	 * Clean up tuples that have been dequeued by all customers.
	 *
	 * Creation date: (4/27/2003 3:10:56 AM)
	 */
	public void cleanTuples() {
		// if no more customers, clear the queue
		if (this.cachedCustomer == null)
			this.contents.clear();
	
		// otherwise, clean up tuples when necessary.
		// the number "5" is a tradeoff between speed and storage:
		// larger number requires less cleanup,
		// smaller number reducees the storage overhead for old tuples
		if (this.cachedCursor > CLEANTHRESHOLD && this.minCursor > CLEANTHRESHOLD) {
	
			int min = Math.min(this.minCursor, this.cachedCursor);
	
			// update the cached customer
			this.cachedCursor -= min;
	
			// update the rest of the customers
			if (this.customerMap != null) {
				Set keySet = this.customerMap.keySet();
				for (Iterator it = keySet.iterator(); it.hasNext();) {
					Object curCustomer = it.next();
					XATQueueCursor oldCursor = (XATQueueCursor) this.customerMap.get(curCustomer);
					oldCursor.setCursor(oldCursor.getCursor() - min);
				}
	
				// also update the minimal cursor
				this.minCursor -= min;
			}

			// clean up tuples by deleting all the tuples up to the min cursor.
			// don't need to worry about PlaceHolder here because min can only
			// point as far as the first tuples in the PlaceHolder.
			for (int i = 0; i < min; i++)
				this.contents.remove(0);
		}
		// re-calculate the size of the queue
		this.computeQueueSize();
	}
	
	/**
	 * Iterates through the contents of the queue and calculates the total
	 * number of tuples in this queue, including ones on disk.
	 * 
	 * Creation Date: (Jun 11, 2004 3:39:37 PM)
	 */
	private synchronized void computeQueueSize() {
		boolean noPH = true;
		this.queueSize = 0;
		for (int i = 0; i < this.contents.size(); i++) {
			if (this.contents.get(i) instanceof PlaceHolder) {
				this.queueSize += ((PlaceHolder)this.contents.get(i)).getNumTuples();
				this.phIndex = i;
				noPH = false;
			}
			else {
				this.queueSize++;
			}
		}
		if (noPH) {
			this.phIndex = -1;
		}
		
		/*
		try {
			FileWriter out = new FileWriter("queue_size.txt", true);
			out.write(System.currentTimeMillis() + "   " + this.queueSize + "\n");
			out.close();
		} catch (Exception ie) {
			System.out.println(ie);
			ie.printStackTrace();
		}
		*/
	}

	/**
	 * Dequeues a number of tuples into an array.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:22 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	public synchronized XATTuple[] dequeue(int number, Object customer) {
		//System.out.println("Start number: " + number);
		// make sure the customer can dequeue
		if (!this.isRegistered(customer)) {
			this.register(customer);
		}

		// also make sure there are enough tuples to dequeue
		// if not, set to maximum possible to dequeue for this customer
		int cursor = this.getCursor(customer);
		if (cursor + number > this.getQueueSize()) {
			number = this.getQueueSize() - cursor;
		}
		//System.out.println("Final number: " + number);
		// get the result from the queueManager
		XATTuple[] result = this.queueManager.dequeue(this, number, customer);

		// clean up the queue
		this.cleanTuples();
		
		// update queue attributes
		this.setCursor(customer, this.getCursor(customer) + number);
		this.computeQueueSize();
		this.totalDequeued += result.length;
		
		return result;
	}

	/**
	 * Dequeues a tuple.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:31 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple dequeue(Object customer) throws XATQueueException {
		
		int number = 1;
		
		// make sure the customer can dequeue
		if (!this.isRegistered(customer))
			this.register(customer);
			
		// also make sure there is a tuple to dequeue
		// if not, then cannot read anything from file 
		// (number will be set to 0)
		int cursor = this.getCursor(customer);
		if (cursor + number > this.getQueueSize()) {
			throw new XATQueueException("Not enough tuples to dequeue");
		}
		else {		
			// get the result from the queueManager and convert it into an XATTuple
			XATTuple result = this.queueManager.dequeue(this, customer);
			
			// clean up the queue
			this.cleanTuples();
			
			// update queue attributes
			this.setCursor(customer, this.getCursor(customer) + 1);
			this.computeQueueSize();
			this.totalDequeued++;
		
			return result;
		}
	}
	
	/**
	 * Dequeues a number of tuples into a List.
	 * 
	 * Creation Date: (Jun 16, 2004 2:34:18 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	public synchronized List dequeueN(int number, Object customer) {		
		// make sure the customer can dequeue
		if (!this.isRegistered(customer)) {
			this.register(customer);
		}

		// also make sure there are enough tuples to dequeue
		// if not, set to maximum possible to dequeue for this customer
		int cursor = this.getCursor(customer);
		if (cursor + number > this.getQueueSize()) {
			number = this.getQueueSize() - cursor;
		}
		
		// get the result from the queueManager
		List result = this.queueManager.dequeueN(this, number, customer);

		// clean up the queue
		this.cleanTuples();

		// update queue attributes
		this.setCursor(customer, this.getCursor(customer) + number);
		this.computeQueueSize();
		this.totalDequeued += result.size();
		
		return result;
	}

	/**
	 * Enqueues a list of tuples.
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (Jun 15, 2004 11:13:36 AM)
	 * @param newTuples List
	 * @param provider Object
	 */
	public synchronized void enqueue(List newTuples, Object provider) {
		queueManager.enqueue(this, newTuples);
		
		// update queue attributes
		this.computeQueueSize();
		this.totalEnqueued += newTuples.size();
	}

	/**
	 * Enqueues a tuple. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation Date: (Jun 16, 2004 2:36:21 PM)
	 * @param newTuple XATTuple
	 * @param provider Object
	 */
	public synchronized void enqueue(XATTuple newTuple, Object provider) {
		//try {
		queueManager.enqueue(this, newTuple);
		//} catch (Exception e) {			
		//	System.out.println("Exception enqueuing tuple: " + newTuple.toString());
		//	try {
		//		e.printStackTrace(new PrintStream(new FileOutputStream("exceptions.txt")));
		//	} catch (Exception ex) {
		//		ex.printStackTrace();
		//	}
		//}
		
		// update queue attributes
		this.computeQueueSize();
		this.totalEnqueued++;
	}

	/**
	 * Enqueues an array of tuples. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[] 
	 * @param provider java.lang.Object
	 */
	public synchronized void enqueue(XATTuple[] newTuples, Object provider) {
		queueManager.enqueue(this, newTuples);
		
		// update queue attributes
		this.computeQueueSize();
		this.totalEnqueued += newTuples.length;
	}
	
	/**
	 * Returns a light weight query object of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:39:22 PM)
	 * @return LightWeightQueryObject
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("XATQueueImp", this.getClass().getName(), true);
	}

	/**
	 * Returns the index of the place-holder of this queue. If it is 
	 * negative, then there are no tuples stored on disk.
	 * 
	 * Creation Date: (Jun 15, 2004 3:33:31 PM)
	 * @return int
	 */
	public synchronized int getPHIndex() {
		return this.phIndex;
	}

	/**
	 * Gets the string representation of the queue's layout.
	 * 
	 * Creation Date: (Jun 29, 2004 4:14:05 PM)
	 * @return String
	 */
	public String getQueueLayout() {
		String str = "";
		str += "|";
		int tupCount = 0;
		for (int i = 0; i < this.getContents().size(); i++) {
			Object curr = this.getContents().get(i);
			if (this.getContents().get(i) instanceof PlaceHolder) {
				str += "T->" + tupCount + "|";
				str += "PH->" + ((PlaceHolder)curr).getNumTuples() + "|";
				tupCount = 0;
			}
			else {
				tupCount++;
			}
		}
		str += "T->" + tupCount + "|";
		return str;
	}

	/**
	 * Gets a handle for the queue manager through the queue.
	 * 
	 * Creation Date: (Jun 8, 2004 11:29:28 AM)
	 * @return XATQueueManager
	 */
	public XATQueueManager getQueueManager() {
		return this.queueManager;
	}

	/**
	 * Returns the number of tuples that are in the queue, including the
	 * tuples on disk.
	 * 
	 * Creation Date: (Jun 16, 2004 4:40:12 PM)
	 * @return int
	 */
	public synchronized int getQueueSize() {
		return this.queueSize;
	}
	
	/**
	 * position starts from 0
	 * @param position
	 * @param tuple
	 * @param provider
	 * used for inserting migration sync flag in stream queues.
	 */
	public synchronized void insertTuple(int position, XATTuple tuple, Object provider) {
		System.out.println("The method is not supported in this class!!!");
		System.exit(-1);
	}


	/**
	 * Retrives a number of tuples from the queue. This is analagous to 
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:59 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	public XATTuple[] peek(int number, Object customer) {	
		// make sure the customer can peek
		if (!this.isRegistered(customer)) {
			this.register(customer);
		}

		// also make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		int cursor = this.getCursor(customer);
		if (cursor + number > this.getQueueSize()) {
			number = this.getQueueSize() - cursor;
		}
		
		// get the result from the queueManager
		return this.queueManager.peek(this, number, customer);
	}
	
	/**
	 * Retrieves a tuple from the queue. This is analagous to dequeue, but
	 * this will not increment or cleanup cursors.
	 *
	 * Creation Date: (Jun 16, 2004 4:45:49 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple peek(Object customer) throws XATQueueException {
		// make sure the customer can peek
		if (!this.isRegistered(customer)) {
			this.register(customer);
		}
	
		// also make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		int cursor = this.getCursor(customer);
		if (cursor + 1 > this.getQueueSize()) {
			throw new XATQueueException("Not enough tuples to dequeue");
		}
			
		// get the result from the queueManager
		return this.queueManager.peek(this, customer);
	}

	/**
	 * Retrieves the last tuple in the queue. This is for XATSubQueueImp.
	 * XATSubQueueImp needs to merge a "subtuple" into the last tuple 
	 * (which must also be a "subtuple") to form a "full tuple". This is 
	 * analagous to dequeue, but this will not increment or cleanup 
	 * cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:47:31 PM)
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple peekLast() throws XATQueueException {	
		// make sure there is a tuple to peek
		// if not, throw an exception
		if (1 > this.getQueueSize()) {
			throw new XATQueueException("Not enough tuples to dequeue");
		}
			
		// get the result from the queueManager
		return this.queueManager.peekLast(this);
	}
	
	/**
	 * Retrieves a number of tuples from the queue. This is analagous to
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:49:09 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	public synchronized List peekN(int number, Object customer) {
		// make sure the customer can peek
		if (!this.isRegistered(customer)) {
			this.register(customer);
		}
	
		// also make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		int cursor = this.getCursor(customer);
		if (cursor + number > this.getQueueSize()) {
			number = this.getQueueSize() - cursor;
		}
			
		// get the result from the queueManager
		return this.queueManager.peekN(this, number, customer);
	}
	
	public synchronized XATTuple peekNthTuple(int number, Object customer) {
		System.out.println("The method is not supported in this class!!!");
		System.exit(-1);
		return null;
	}

	/**
	 * Returns the contents of the queue by calling toString() on each 
	 * tuple. Caller is the consumer of the queue you want to print out.
	 * Caller can be null if this is the top output queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:50:17 PM)
	 * @param caller Object
	 * @return String
	 */
	public String printContent(Object caller) {
	    StringBuffer rs = new StringBuffer();
	
	    rs.append(this.queueSize);
	    rs.append(" tuples: \n");
	
	    if (caller != null) {
	    	int n = this.queueSize - this.getCursor(caller);
	    	XATTuple[] tempResult = this.peek(n, caller);
	        for (int i = 0; i < n; i++) {
	            XATTuple tuple = tempResult[i];
	            rs.append(tuple.toString());
	            rs.append("\n");
	        }
	    }
	    else {
	    	caller = "new-caller";
	    	this.register(caller);
	    	this.setCursor(caller, 0);
	    	XATTuple[] tempResult = this.peek(this.queueSize, caller);
			for (int i = 0; i < this.queueSize; i++) {
				XATTuple tuple = tempResult[i];
				rs.append(tuple.toString());
				rs.append("\n");
			}
	    }
	    
	    return rs.toString();
	}
	
	/**
	 * Sets the place-holder index to the given index number.
	 * 
	 * Creation Date: (Jun 15, 2004 3:31:03 PM)
	 * @param newIndex int
	 */
	public synchronized void setPHIndex(int newIndex) {
		this.phIndex = newIndex;
	}

	/**
	 * Sets the tuples of this queue to the new tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:20 PM)
	 * @param newTuples List
	 */
	public synchronized void setTuples (List newTuples) {
		 this.contents = newTuples;
		 this.computeQueueSize();
	}
	
	/**
	 * Skips the next tuple.
	 *
	 * Creation Date: (Jun 16, 2004 5:14:34 PM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	public synchronized void skip(Object customer) throws XATQueueException {
		//get cursor
		int cursor = this.getCursor(customer);
	
		//check validity
		if(cursor >= this.queueSize)
			throw new XATQueueException("Not enough tuples to skip!");
			
		//increase cursor
		this.setCursor(customer, this.getCursor(customer) + 1);		
	}
	
	/**
	 * Skips a number of tuples.
	 *
	 * Creation Date: (Jun 16, 2004 5:16:24 PM)
	 * @param number int
	 * @param customer Object
	 */
	public synchronized void skipN(int number, Object customer) throws XATQueueException {
		// get cursor
		int cursor = this.getCursor(customer);
	
		// check validity
		if(cursor + number > this.queueSize)
			throw new XATQueueException("Not enough tuples to dequeue!");
			
		//increase cursor
		this.setCursor(customer, this.getCursor(customer) + number);		
	}

	@Override
	public void enqueue(XATTupleNegativeImp newTuple, Object provider) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void enqueue(XATTupleNegativeImp[] newTuples, Object provider) {
		// TODO Auto-generated method stub
		
	}
}
