/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.util.ArrayList;
import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataSource;

/**
 * <pre>
 * Simulates a FIFO Queue and provides default access methods to enqueue 
 * and dequeue data. Each XATQueue is associated with a list of owners. 
 * Only the owner may enqueue tuples. In order to dequeue tuples, the 
 * customer must first register with the queue. This is done through 
 * <code>queue.register(Object customer)</code>.
 *
 * The owner may not dequeue tuples. This prevents what could be an 
 * infinite loop.  
 *
 * Each queue has a schema associated with it. There is <emph>no</emph> 
 * validation done when a tuple is enqueued, meaning that the queue does 
 * not validate the tuples' schema to the queue's. Doing so would reduce
 * scalability significantly.
 *
 * The queue also has one property associated with it: TUPLE_COUNT. This 
 * specifies how many tuples are in the queue at any given time. This is 
 * different than <code>queue.getQueueSize(Object customer)</code>, which 
 * specifies how many tuples are available to a given customer.  
 * </pre> 
 */
public interface XATQueue extends XATDataSource{
	public int getBoxInputQID();
	public void setBoxInputQID(int newID);
	/**
	 * Adds a new owner for this queue. The newOwner object will now be 
	 * able to enqueue tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:47 PM)
	 * @param newOwner Object
	 */
	void addNewOwner(Object newOwner);
	
	/**
	 * Adds a new object to the end of the queue. The object can be either
	 * a PlaceHolder or an XATTuple.
	 * 
	 * Creation Date: (Jun 16, 2004 2:42:03 PM)
	 * @param newTuple Object 
	 */
	void addTuple (Object newTuple);

	/**
	 * Adds the new tuples to the end of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:42:13 PM)
	 * @param newTuples List
	 */
	void addTuples (List newTuples);

	/**
	 * Clean up tuples that have been dequeued by all customers.
	 *
	 * Creation date: (4/27/2003 3:10:56 AM)
	 */
	void cleanTuples();

	/**
	 * Dequeues a number of tuples into an array.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:22 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	XATTuple[] dequeue(int number, Object customer);

	/**
	 * Dequeues a tuple.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:31 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	XATTuple dequeue(Object customer);
	
	/**
	 * Dequeues a number of tuples into a List.
	 * 
	 * Creation Date: (Jun 16, 2004 2:34:18 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	List dequeueN(int number, Object customer);

	/**
	 * Enqueues a list of tuples.
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (Jun 15, 2004 11:13:36 AM)
	 * @param newTuples List
	 * @param provider Object
	 */
	void enqueue(List newTuples, Object provider);

	/**
	 * Enqueues a tuple. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation Date: (Jun 16, 2004 2:36:21 PM)
	 * @param newTuple XATTuple
	 * @param provider Object
	 */
	void enqueue(XATTuple newTuple, Object provider);

	/**
	 * Enqueues an array of tuples. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[] 
	 * @param provider java.lang.Object
	 */
	void enqueue(XATTuple[] newTuples, Object provider);
	
	/**
	 * Returns the contents of the queue.
	 * 
	 * Creation Date: (May 25, 2004 11:38:36 AM)
	 * @return List
	 */
	List getContents();
	
	/**
	 * Returns the cursor of the customer. There is a side effect: the 
	 * customer is cached. This method needs to be protected such that any
	 * subclass can extend enqueue or dequeue.
	 * 
	 * Creation date: (4/27/2003 1:40:07 AM)
	 * @param customer Object
	 * @return int
	 */
	int getCursor(Object customer);
	
	/**
	 * Gets the cursor of the given customer.
	 * 
	 * Creation Date: (Jun 16, 2004 2:38:44 PM)
	 * @param customer Object
	 * @return int 
	 */
	int getCustCursor(Object customer);

	/**
	 * Returns a light weight query object of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:39:22 PM)
	 * @return LightWeightQueryObject
	 */
	LightWeightQueryObject getLightWeightQueryObject();

	/**
	 * Gets the customers of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:38:05 PM)
	 * @return List
	 */
	List getQueueCustomer();

	/**
	 * Gets the string representation of the queue's layout.
	 * 
	 * Creation Date: (Jun 29, 2004 4:14:05 PM)
	 * @return String
	 */
	String getQueueLayout();

	public void setQueueID(int num);
	
	public int getQueueID();
	
	/**
	 * Sets the name of the queue.
	 * 
	 * Creation Date: (Jun 16, 200
	/**
	 * Gets the queue name.
	 * 
	 * Creation Date: (Jun 16, 2004 4:39:30 PM)
	 * @return String
	 */
	String getQueueName();

	/**
	 * Returns the queue number.
	 * 
	 * Creation Date: (Jun 16, 2004 4:39:40 PM)
	 * @return int
	 */
	int getQueueNumber();

	/**
	 * Gets the owner(s) of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:39:51 PM)
	 * @return List
	 */
	List getQueueOwner();
	
	/**
	 * Returns the number of tuples that are in the queue, including the
	 * tuples on disk.
	 * 
	 * Creation Date: (Jun 16, 2004 4:40:12 PM)
	 * @return int
	 */
	int getQueueSize();

	/**
	 * Returns the number of tuples in the queue for the given customer.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @return int
	 * @param customer Object
	 */
	int getQueueSize(Object customer);

	/**
	 * Returns the schema of the queue.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @return XATQueueSchema
	 */
	XATQueueSchema getSchema();
	
	/**
	 * Returns the contents of the queue.
	 * Note: This is left over from the old queue implementation in case
	 * other components use this method.
	 * 
	 * Creation Date: (May 25, 2004 11:38:36 AM)
	 * @author Natasha Bogdanova
	 * @return List
	 */
	List getTuples();

	/**
	 * Increase the cursor of the cached customer by the given value.
	 *
	 * Creation date: (4/27/2003 1:53:53 AM)
	 * @param number int
	 */
	void increaseCursor(int number);
	
	void insertTuple(int position, XATTuple tuple, Object provider);
	/**
	 * Returns the isOldBoxInput attribute of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:42:33 PM)
	 * @return boolean
	 */
	boolean isOldBoxInput();

	/**
	 * Checks to see if the caller is the owner of the queue
	 * 
	 * Creation Date: (Jun 16, 2004 4:42:25 PM)
	 * @param caller Object
	 * @return boolean
	 */
	boolean isOwner(Object caller);

	/**
	 * Checks whether the customer is registered with this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:07 PM)
	 * @param customer Object
	 * @return boolean
	 */
	boolean isRegistered(Object customer);

	/**
	 * Retrives a number of tuples from the queue. This is analagous to 
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:59 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	XATTuple[] peek(int number, Object customer);
	
	/**
	 * Retrieves a tuple from the queue. This is analagous to dequeue, but
	 * this will not increment or cleanup cursors.
	 *
	 * Creation Date: (Jun 16, 2004 4:45:49 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	XATTuple peek(Object customer) throws XATQueueException;

	/**
	 * Retrieves a number of tuples from the queue. This is analagous to
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:49:09 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	List peekN(int number, Object customer);

	XATTuple peekNthTuple(int number, Object customer);


	/**
	 * Returns the contents of the queue by calling toString() on each 
	 * tuple. Caller is the consumer of the queue you want to print out.
	 * Caller can be null if this is the top output queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:50:17 PM)
	 * @param caller Object
	 * @return String
	 */
	String printContent(Object caller);
	
	/**
	 * Registers a customer so that the customer can dequeue tuples from 
	 * this queue.
	 *
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param customer Object
	 */
	void register(Object customer);
	
	/**
	 * This will forcefully remove all tuples from the queue, even if the
	 * queue may still have some customers. Note: this method obviously 
	 * can create havoc if it is used indiscriminately, but it is a huge 
	 * speedup for the XATApplication.
	 *  
	 * Creation date: (4/13/2003 1:38:36 AM)
	 */
	void removeAllTuples();
	
	/**
	 * Removes the owner from the queue. This will prevent the owner from 
	 * enqueuing tuples in the future.
	 * 
	 * Creation Date: (Jun 16, 2004 5:03:42 PM)
	 * @param owner Object
	 */
	void removeOwner(Object owner);
	
	/**
	 * Sets the cursor of the given customer to a new value. Also, caches
	 * the customer and the new cursor.
	 * 
	 * Creation Date: (Jun 16, 2004 5:03:47 PM)
	 * @param customer Object
	 * @param newCursor int
	 */
	void setCursor(Object customer, int newCursor);
	
	/**
	 * Sets the isOldBoxInput attribute of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 5:06:20 PM)
	 * @param flag boolean
	 */
	void setOldBoxInput(boolean flag);
	
	/**
	* Sets the given property to the new value. This method is responsible
	* for validating the given property. An exception is thrown if the 
	* value is not correct. 
	* 
	* Validation should be done in the following manner: 
	* First the method should check if the object supports the given 
	* property by calling <tt>isValidPropertyName(String)<tt>. If it is 
	* not a valid property name, then an exception should be thrown. <p>
	*
	* Once it is known that the object supports the property, the value 
	* must be validated. Each implementor of XATStatistics has a set of 
	* properties that are common to all of its types and some that are 
	* specific to types. For example, all XATOperators (which implement 
	* XATStatistics) have certain properties while an individual 
	* XATOperator has specific properties. The individual types should 
	* only validate their own specific types. If the caller is trying to 
	* set a common type, it should be validated by the super class by 
	* calling <tt>super.setProperty(name, value)<tt>. This will reduce 
	* duplicate code. <p>
	*
	* When it is determined the name/value pair is correct, the pair is 
	* updated using <tt>setNewPropertyValue()<tt> <p>
	* 
	* <tt>XATStatisticsImp<tt> provides default validation by checking to 
	* see if the property's value is >=0. Property values that are 
	* Strings will just be set. If the property requires more strict 
	* constraints than this, it is up to the subclass that created the 
	* property to validate it (as described above).
	* 
	* Creation Date: (Jun 16, 2004 5:11:15 PM)
	* @param name String
	* @param value String
	* @throws Exception
	*/
	void setProperty(String name, String value) throws Exception;
	
	/**
	 * Sets the name of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 5:11:56 PM)
	 * @param newName String
	 */
	void setQueueName(String newName);
	
	/**
	 * Sets the number of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 5:12:34 PM)
	 * @param num int
	 */
	void setQueueNumber(int num);
	
	void setQueueOwner(Object owner);
		
	/**
	 * Sets the schema of the queue.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param schema XATQueueSchema
	 */
	void setSchema(XATQueueSchema schema);
		
	/**
	 * Sets the tuples of this queue to the new tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:20 PM)
	 * @param newTuples List
	 */
	void setTuples (List newTuples);
	
	/**
	 * Returns the size of this queue. This is mainly for the the GUI.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:52 PM)
	 * @return int
	 */
	int size();
	
	/**
	 * Skips the next tuple.
	 *
	 * Creation Date: (Jun 16, 2004 5:14:34 PM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	void skip(Object customer) throws XATQueueException;
	
	/**
	 * Skips a number of tuples.
	 *
	 * Creation Date: (Jun 16, 2004 5:16:24 PM)
	 * @param number int
	 * @param customer Object
	 */
	void skipN(int number, Object customer) throws XATQueueException;
	
	/**
	 * Converts the queue to a string.
	 * 
	 * Creation Date: (Jun 16, 2004 5:17:41 PM)
	 * @return String
	 */
	String toString();
	
	/**
	 * Unregisters a customer. The customer can not dequeue tuples after 
	 * it has been unregistred.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	void unregister(Object customer) throws XATQueueException;

	String getCurrentResult();
	
	void clearCurrentResult();
	
	boolean isFinalQueue();
	
	void setAsFinalQueue(boolean finalQueue);
	
	
	/**Following is added for support multiple schemas in a single queue 
	 * The schemas of this queue. 
	 * Added by Song Wang for supporting multiple schemas in one queue
	*/
	public void addSchema(XATQueueSchema schema);
	public void setSchemas(ArrayList schemas);
	public XATQueueSchema getSchema(int i);
	public ArrayList getSchemas();
	public boolean initialized();
	public void setInitialized(boolean b);
}
