/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataSourceImp;

/**
 * The abstract class for all types of queues. This class implements all the
 * methods that are shared by all queues. However, the implementation of queue
 * access methods, such as enqueue and dequeue, is left to each individual
 * queue implementation.
 * 
 * Currently, an XATQueue can be an instance of
 * - XATMemoryQueue, or
 * - XATPersistentQueue
 * 
 * Creation Date: (Jul 27, 2004 9:51:34 AM)
 * @author Natasha Bogdanova
 */
public abstract class XATQueueImp extends XATDataSourceImp implements XATQueue {
	
	/** useful during migration to connect new and old boxes' input queues */
	protected int boxInputQID = 0;

	/** This threshold is used to determine when the queue should be cleaned. */
	public static int CLEANTHRESHOLD = 0;

	/** The cachedCustomer records the most recent customer. If the same
	 *  customer access a queue again, the cachedCursor is used avoiding 
	 *  checking the customerMap. This is solely for performance purposes. */
	protected Object cachedCustomer = null;
	
	/** The cursor of the cached customer. */
	protected int cachedCursor = 0;
	
	/** The list containing all contents of the queue. */ 
	protected List contents = new ArrayList();

	/** Customer list that maps a customer to its cursor:
	 *  Object => XATQueueCursor.
	 *  The map contains all customers except for the cached one. */
	protected Map customerMap = null;
	
	/** isOldBoxInput attribute. */
	protected boolean isOldBoxInput = false;

	/** The minimal cursor of all customers except for the cached one. 
	 *  This variable is solely for performance purposes - it avoids 
	 *  iteration over all customers to find the minimal cursor. */
	protected int minCursor = Integer.MAX_VALUE;

	/** The name of the queue. It is set as the IP address of the machine its feeding tuple to. */
	protected String queueName = new String();
	
	/** The number of the queue. It indicates the position of this queue as to other input queues to an operator. */
	protected int queueNumber = -1;
	
	/** Each queue in a query plan has a unique queue id
	  * This id is initialized at the query generation time, along with operator id and state id.
	  * this id can be used during distributed migration to mark the subplan that is changed by optimizer
	  * queue id starts from 0 and is incremented each time a new queue is added to the system */
	protected int queueID = -1;

	/** The owners of this queue. */
	protected List queueOwners = new ArrayList();

	/** The schema of this queue. */
	protected XATQueueSchema schema = null;
	
	/** The schemas of this queue. 
	 * Added by Song Wang for supporting multiple schemas in one queue
	*/
	protected ArrayList schemas = new ArrayList();
	
	protected boolean initialized= false;
	
	/** The total number of dequeued tuples. This is only for the GUI. */
	protected int totalDequeued = 0;
	
	/** The total number of enqueued tuples. This is only for the GUI. */
	protected int totalEnqueued = 0;

	protected boolean isFinalQueue = false;
	protected StringBuffer currentResultBuffer = new StringBuffer();

	/*------------- BEGIN ABSTRACT CLASSES ----------------------------------*/	
	/**
	 * Clean up tuples that have been dequeued by all customers.
	 *
	 * Creation date: (4/27/2003 3:10:56 AM)
	 */
	abstract public void cleanTuples();
	
	/**
	 * Dequeues a number of tuples into an array.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:22 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	
	abstract public XATTuple[] dequeue(int number, Object customer);

	/**
	 * Dequeues a tuple.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:31 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	abstract public XATTuple dequeue(Object customer) throws XATQueueException;
	
	/**
	 * Dequeues a number of tuples into a List.
	 * 
	 * Creation Date: (Jun 16, 2004 2:34:18 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	abstract public List dequeueN(int number, Object customer);

	/**
	 * Enqueues a list of tuples.
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (Jun 15, 2004 11:13:36 AM)
	 * @param newTuples List
	 * @param provider Object
	 */
	abstract public void enqueue(List newTuples, Object provider);

	/**
	 * Enqueues a Negative tuple. 
	 * Note: the provider is reserved for future usage.
	 * @author Mingrui and Abhishek
	 * Creation Date: (Feb, 2007 2:36:21 PM)
	 * @param newTuple XATTuple
	 * @param provider Object
	 */
	abstract public void enqueue(XATTupleNegativeImp newTuple,Object provider);
	
	
	/**
	 * Enqueues a tuple. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation Date: (Jun 16, 2004 2:36:21 PM)
	 * @param newTuple XATTuple
	 * @param provider Object
	 */
	
	
	abstract public void enqueue(XATTuple newTuple, Object provider);
	
	/**
	 * Enqueues an array of negative tuples. 
	 * Note: the provider is reserved for future usage.
	 *@author Mingrui and Abhishek
	 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[] 
	 * @param provider java.lang.Object
	 */
	abstract public void enqueue(XATTupleNegativeImp[] newTuples, Object provider);
	
	/**
	 * Enqueues an array of tuples. 
	 * Note: the provider is reserved for future usage.
	 * 
	 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[] 
	 * @param provider java.lang.Object
	 */
	abstract public void enqueue(XATTuple[] newTuples, Object provider);
	
	/**
	 * Returns a light weight query object of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:39:22 PM)
	 * @return LightWeightQueryObject
	 */
	abstract public LightWeightQueryObject getLightWeightQueryObject();

	/**
	 * Gets the string representation of the queue's layout.
	 * 
	 * Creation Date: (Jun 29, 2004 4:14:05 PM)
	 * @return String
	 */
	abstract public String getQueueLayout();
	
	/**
	 * Returns the total number of tuples that are in the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:40:12 PM)
	 * @return int
	 */
	abstract public int getQueueSize();

	/**
	 * Retrives a number of tuples from the queue. This is analagous to 
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:59 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	abstract public XATTuple[] peek(int number, Object customer);
	
	/**
	 * Retrieves a tuple from the queue. This is analagous to dequeue, but
	 * this will not increment or cleanup cursors.
	 *
	 * Creation Date: (Jun 16, 2004 4:45:49 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	abstract public XATTuple peek(Object customer) throws XATQueueException;

	/**
	 * Retrieves the last tuple in the queue. This is for XATSubQueueImp.
	 * XATSubQueueImp needs to merge a "subtuple" into the last tuple 
	 * (which must also be a "subtuple") to form a "full tuple". This is 
	 * analagous to dequeue, but this will not increment or cleanup 
	 * cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:47:31 PM)
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	abstract public XATTuple peekLast() throws XATQueueException;
	
	/**
	 * Retrieves a number of tuples from the queue. This is analagous to
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:49:09 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	abstract public List peekN(int number, Object customer);

	abstract public XATTuple peekNthTuple(int number, Object customer);
	/**
	 * Returns the contents of the queue by calling toString() on each 
	 * tuple. Caller is the consumer of the queue you want to print out.
	 * Caller can be null if this is the top output queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:50:17 PM)
	 * @param caller Object
	 * @return String
	 */
	abstract public String printContent(Object caller);
		
	/**
	 * Sets the tuples of this queue to the new tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:20 PM)
	 * @param newTuples List
	 */
	abstract public void setTuples (List newTuples);
	
	/**
	 * Skips the next tuple.
	 *
	 * Creation Date: (Jun 16, 2004 5:14:34 PM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	abstract public void skip(Object customer) throws XATQueueException;
	
	/**
	 * Skips a number of tuples.
	 *
	 * Creation Date: (Jun 16, 2004 5:16:24 PM)
	 * @param number int
	 * @param customer Object
	 */
	abstract public void skipN(int number, Object customer) throws XATQueueException;

	/*------------- END ABSTRACT CLASSES ------------------------------------*/

	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:37:04 PM)
	 */
	public XATQueueImp() {
		super();
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:59 PM)
	 * @param owner Object
	 */
	public XATQueueImp(Object owner) {
		super();
		this.queueOwners.add(owner);
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:53 PM)
	 * @param owner Object
	 * @param queueName String
	 */
	public XATQueueImp(Object owner, String queueName) {
		super();
		this.queueName = queueName;
		this.queueOwners.add(owner);
	}

	/**
	 * Adds a new owner for this queue. The newOwner object will now be 
	 * able to enqueue tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:47 PM)
	 * @param newOwner Object
	 */
	public synchronized void addNewOwner(Object newOwner) {	
		if(!queueOwners.contains(newOwner))
			queueOwners.add(newOwner);
	}
	
	/**
	 * Adds a new object to the end of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:42:03 PM)
	 * @param newTuple Object 
	 */
	public synchronized void addTuple (Object newTuple) {
		 this.contents.add(newTuple);
	}

	/**
	 * Adds the new tuples to the end of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:42:13 PM)
	 * @param newTuples List
	 */
	public synchronized void addTuples (List newTuples) {
		 this.contents.addAll(newTuples);
	}
	
	/**
	 * Compute the minimal cursor in the customer map.
	 * 
	 * Creation date: (4/27/2003 3:21:38 AM)
	 */
	private void computeMinCursor() {
		this.minCursor = Integer.MAX_VALUE;
	
		Collection cursors = this.customerMap.values();
		for(Iterator it = cursors.iterator(); it.hasNext();){
			XATQueueCursor cursor = (XATQueueCursor)it.next();
			if(cursor.getCursor() < this.minCursor) {
				this.minCursor = cursor.getCursor();
			}
		}
	}
	
	/**
	 * Returns the contents of the queue.
	 * 
	 * Creation Date: (May 25, 2004 11:38:36 AM)
	 * @return List
	 */
	public synchronized List getContents() {
		return this.contents;
	}
	
	/**
	 * Returns the cursor of the customer. There is a side effect: the 
	 * customer is cached. This method needs to be protected such that any
	 * subclass can extend enqueue or dequeue.
	 * 
	 * Creation date: (4/27/2003 1:40:07 AM)
	 * @param customer Object
	 * @return int
	 */
	public int getCursor(Object customer) {	
		//if cached, return directly
		if (this.cachedCustomer == customer) {
			return this.cachedCursor;
		}
	
		int result = 0;
		
		//if not cached, check the customer map
		if(this.customerMap == null || !this.isRegistered(customer)) {
			this.register(customer);
		}
	
		XATQueueCursor cursor = (XATQueueCursor)this.customerMap.get(customer);
		if(cursor == null) {
			this.register(customer);
			cursor = new XATQueueCursor();
		}
		
		result = cursor.getCursor();
		
		//before return, first save the old cache in customer map
		this.customerMap.put(this.cachedCustomer, new XATQueueCursor(this.cachedCursor));
		this.customerMap.remove(customer);
		
		//also recompute the minCursor because the 
		//old minCursor may be now moved out to be cached
		if(this.cachedCursor < this.minCursor){
			this.minCursor = this.cachedCursor;
		}
		else{
			this.computeMinCursor();
		}
		
		//cache the customer
		this.cachedCustomer = customer;
		this.cachedCursor = result;
		
		return result;
	}

	/**
	 * Gets the cursor of the given customer.
	 * 
	 * Creation Date: (Jun 16, 2004 2:38:44 PM)
	 * @param customer Object
	 * @return int 
	 */
	public int getCustCursor(Object customer) {
		return this.getCursor(customer);
	}

	/**
	 * Gets the customers of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:38:05 PM)
	 * @return List
	 */
	public List getQueueCustomer() {
		List customers = new ArrayList();
		
		if(cachedCustomer != null) {
			customers.add(cachedCustomer);
		}
		if(customerMap!= null && !customerMap.isEmpty()) {
			customers.addAll(customerMap.keySet());
		}
		return customers;
	}

	/**
	 * Gets the queue name.
	 * 
	 * Creation Date: (Jun 16, 2004 4:39:30 PM)
	 * @return String
	 */
	public String getQueueName() {
		return this.queueName;
	}

	/**
	 * Returns the queue number.
	 * 
	 * Creation Date: (Jun 16, 2004 4:39:40 PM)
	 * @return int
	 */
	public int getQueueNumber() {
		return this.queueNumber;
	}

	/**
	 * Gets the owner(s) of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:39:51 PM)
	 * @return List
	 */
	public List getQueueOwner() {
		return this.queueOwners;
	}
	
	
	/**
	 * Returns the number of tuples in the queue for the given customer.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @return int
	 * @param customer Object
	 */
	public synchronized int getQueueSize(Object customer) {		
		int i = this.getQueueSize() - this.getCursor(customer);
		try{
			if(i<0) 
				throw new Exception("Queue size for " + customer + "is negative! Wrong cursor!");
		} catch (Exception e) {
			System.out.println(e);
			e.printStackTrace();
			
		}
		return i;
	}

	/**
	 * Returns the schema of the queue.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @return XATQueueSchema
	 */
	public synchronized XATQueueSchema getSchema() {
		return this.schema;
	}
	
	/** The schemas of this queue. 
	 * Added by Song Wang for supporting multiple schemas in one queue
	*/
	public synchronized XATQueueSchema getSchema(int i) {
		return (XATQueueSchema)this.schemas.get(i);
	}
	/** The schemas of this queue. 
	 * Added by Song Wang for supporting multiple schemas in one queue
	*/
	public synchronized ArrayList getSchemas() {
		return this.schemas;
	}
	public boolean initialized(){
		return this.initialized;
	}
	public void setInitialized(boolean b){
		this.initialized=b;
	}
	

	/**
	 * Returns the contents of the queue.
	 * Note: This is left over from the old queue implementation in case
	 * other components use this method.
	 * 
	 * Creation Date: (May 25, 2004 11:38:36 AM)
	 * @author Natasha Bogdanova
	 * @return List
	 */
	public synchronized List getTuples() {
		return this.contents;
	}

	/**
	 * Increase the cursor of the cached customer by the given value.
	 *
	 * Creation date: (4/27/2003 1:53:53 AM)
	 * @param number int
	 */
	public void increaseCursor(int number) {	
		this.cachedCursor += number;
	
		//reset the cursor
		if (cachedCursor == this.getQueueSize()) {
			cachedCursor = 0;
		}	
	}
	
	abstract public void insertTuple(int position, XATTuple tuple, Object provider);
    
	/**
	 * Returns the isOldBoxInput attribute of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:42:33 PM)
	 * @return boolean
	 */
	public boolean isOldBoxInput() {
		return this.isOldBoxInput;
	}

	/**
	 * Checks to see if the caller is the owner of the queue
	 * 
	 * Creation Date: (Jun 16, 2004 4:42:25 PM)
	 * @param caller Object
	 * @return boolean
	 */
	public boolean isOwner(Object caller) {		
		//This if statement is needed for GUI, see Tim if you need it changed.
		if(caller instanceof XATStreamOperator)
			if(queueOwners.size() > 0) {
			if (queueOwners.get(0) instanceof XATStreamOperator) 
				if(((XATStreamOperator) queueOwners.get(0)) == (XATStreamOperator)caller)
					return true;
			}
		return queueOwners.contains(caller);
	}

	/**
	 * Checks whether the customer is registered with this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:07 PM)
	 * @param customer Object
	 * @return boolean
	 */
	public synchronized boolean isRegistered(Object customer){			
		if(this.cachedCustomer == null){
			return false;
		}
	
		// first check if it has been registered
		if(this.cachedCustomer == customer)
			return true;
			
		if(this.customerMap.containsKey(customer)) 
			return true;
		
		// if it isn't in the cache or isn't in the map, it's not registered.
		
		return false;
	}
	
	/**
	 * Registers a customer so that the customer can dequeue tuples from 
	 * this queue.
	 *
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param customer Object
	 */
	public synchronized void register(Object customer) {
		//System.out.println("Register "+customer+ " to queue "+this);
		if(this.customerMap == null) {
			this.customerMap = new HashMap();
		}
	
		if(this.cachedCustomer == null){
			this.cachedCustomer = customer;
			//System.out.println(" Add customer to cache***");
			this.cachedCursor = 0;
		}
		else{
			//first check if it has been registered
//			if((this.cachedCustomer instanceof XATStreamOperator) && (customer instanceof XATStreamOperator)) {
//				if(((XATStreamOperator)this.cachedCustomer).getObjectID() == ((XATStreamOperator)customer).getObjectID()) {
//					System.out.println("queue "+this.hashCode()+ " customer already in cache***" + customer.hashCode()+customer+ "***");
//					return;
//				}
//			} else 
			if(this.cachedCustomer == customer) {
				//System.out.println(" customer already in cache***");
				return;
			}
			if(this.customerMap != null && this.customerMap.containsKey(customer)) {
				//System.out.println(" customer already in map***");
				return;
			}
	
			//add to customer map
			this.customerMap.put(customer, new XATQueueCursor(0));
			//System.out.println(" Add customer to map***");
	
			//the new customer must have the minimal cursor
			this.minCursor = 0;
		}				
	}
	
	/**
	 * This will forcefully remove all tuples from the queue, even if the
	 * queue may still have some customers. Note: this method obviously 
	 * can create havoc if it is used indiscriminately, but it is a huge 
	 * speedup for the XATApplication.
	 *  
	 * Creation date: (4/13/2003 1:38:36 AM)
	 */
	public synchronized void removeAllTuples() {	
		this.contents.clear();
	
		// reset all cursors
		// update the cached customer
		this.cachedCursor = 0;
	
		//update the rest customers
		if (this.customerMap != null) {
			Set keySet = this.customerMap.keySet();
			for (Iterator it = keySet.iterator(); it.hasNext();) {
				Object curCustomer = it.next();
				XATQueueCursor oldCursor = (XATQueueCursor) this.customerMap.get(curCustomer);
				oldCursor.setCursor(0);
			}
	
			//also update the minimal cursor
			this.minCursor = Integer.MAX_VALUE;
		}	
	}
	
	/**
	 * Removes the owner from the queue. This will prevent the owner from 
	 * enqueuing tuples in the future.
	 * 
	 * Creation Date: (Jun 16, 2004 5:03:42 PM)
	 * @param owner Object
	 */
	public void removeOwner(Object owner) {		
		queueOwners.remove(owner);		
	}
	
	/**
	 * Sets the cursor of the given customer to a new value. Also, caches
	 * the customer and the new cursor.
	 * 
	 * Creation Date: (Jun 16, 2004 5:03:47 PM)
	 * @param customer Object
	 * @param newCursor int
	 */
	public void setCursor(Object customer, int newCursor) {
		// make sure the customer is registered
		if (!this.isRegistered(customer)) {
			this.register(customer);
			//throw new XATQueueException("Customer is not registered!");
		}

		if (this.cachedCustomer == customer) {
			// if cached, update cached cursor directly
			this.cachedCursor = newCursor;
		}
		else if (this.customerMap == null) {
			this.register(customer);
			// if not cached, check the customer map
			//throw new XATQueueException("Customer is not registered!");
		}
	
		// before return, first save the old cache in customer map
		this.customerMap.put(this.cachedCustomer, new XATQueueCursor(this.cachedCursor));
		this.customerMap.remove(customer);
	
		// also recompute the minCursor because the 
		// old minCursor may be now moved out to be cached
		if (this.cachedCursor < this.minCursor)
			this.minCursor = this.cachedCursor;
		else
			this.computeMinCursor();
	
		//cache the customer
		this.cachedCustomer = customer;
		this.cachedCursor = newCursor;
	}
	
	/**
	* Called by the constructor. This will set the default properties that 
	* ALL query objects will support. Each subclass should override this 
	* method and add their own unique properties, but should also call
	* super.setDefaultProperties() so that the common properties may be 
	* set.
	* 
	* Creation Date: (Jun 16, 2004 5:03:53 PM)
	*/   
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		// this property should be set in each operator's run() method
		// when certain number of tuples are to be dequeued from this queue.
		addProperty("DEQUEUED_TUPLE_COUNT", "0");		
	}
	
	/**
	* Called by the constructor. This will set the default Statistics that
	* ALL query objects will support. Each subclass should override this 
	* method and add their own unique properties, but should also call
	* super.setDefaultStatistics() so that the common statistics may be 
	* set.
	* 
	* Creation Date: (Jun 16, 2004 5:04:47 PM)
	*/   
	protected void setDefaultStatistics(){
		/**
		* This statistic will just reflect queue.getQueueSize(), but this
		* will allow the statistics gatherer to store how many tuples are
		* in the queue at certain intervals. This will allow for the SG to
		* infer how many tuples were dequeued from the queue the previous
		* time.
		*/
		super.setDefaultStatistics();
		addStatistic("TUPLE_COUNT");		
	}
	
	/**
	 * Sets the isOldBoxInput attribute of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 5:06:20 PM)
	 * @param flag boolean
	 */
	public void setOldBoxInput(boolean flag) {
		isOldBoxInput = flag;
	}
	
	/**
	* Sets the given property to the new value. This method is responsible
	* for validating the given property. An exception is thrown if the 
	* value is not correct. 
	* 
	* Validation should be done in the following manner: 
	* First the method should check if the object supports the given 
	* property by calling <tt>isValidPropertyName(String)<tt>. If it is 
	* not a valid property name, then an exception should be thrown. <p>
	*
	* Once it is known that the object supports the property, the value 
	* must be validated. Each implementor of XATStatistics has a set of 
	* properties that are common to all of its types and some that are 
	* specific to types. For example, all XATOperators (which implement 
	* XATStatistics) have certain properties while an individual 
	* XATOperator has specific properties. The individual types should 
	* only validate their own specific types. If the caller is trying to 
	* set a common type, it should be validated by the super class by 
	* calling <tt>super.setProperty(name, value)<tt>. This will reduce 
	* duplicate code. <p>
	*
	* When it is determined the name/value pair is correct, the pair is 
	* updated using <tt>setNewPropertyValue()<tt> <p>
	* 
	* <tt>XATStatisticsImp<tt> provides default validation by checking to 
	* see if the property's value is >=0. Property values that are 
	* Strings will just be set. If the property requires more strict 
	* constraints than this, it is up to the subclass that created the 
	* property to validate it (as described above).
	* 
	* Creation Date: (Jun 16, 2004 5:11:15 PM)
	* @param name String
	* @param value String
	* @throws Exception
	*/
	public void setProperty(String name, String value) throws Exception{
		super.setProperty(name, value);
	}
	
	/**
	 * sets the unique id of the queue.
	 * 
	 * @param num
	 */
	public void setQueueID(int num) {
		this.queueID = num;
	}
	
	public int getQueueID() {
		return this.queueID;
	}
	
	/**
	 * Sets the name of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 5:11:56 PM)
	 * @param newName String
	 */
	public void setQueueName(String newName) {
		this.queueName = newName;
	}
	
	/**
	 * Sets the number of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 5:12:34 PM)
	 * @param num int
	 */
	public void setQueueNumber(int num) {
		this.queueNumber = num;
	}
	
	public void setQueueOwner(Object owner) {
		this.queueOwners.clear();
		this.queueOwners.add(owner);
	}
		
	/**
	 * Sets the schema of the queue.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param schema XATQueueSchema
	 */
	public synchronized void setSchema(XATQueueSchema schema) {
		this.schema = schema;	
	}
	
	/** The schemas of this queue. 
	 * Added by Song Wang for supporting multiple schemas in one queue
	*/
	public synchronized void addSchema(XATQueueSchema schema) {
		this.schemas.add(schema);	
	}
	/** Set the schemas of this queue. 
	 * Added by Song Wang for supporting multiple schemas in one queue
	*/
	public void setSchemas(ArrayList inputSchemas){
		this.schemas = inputSchemas;
	}
	
	/**
	 * Returns the size of this queue. This is mainly for the the GUI.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:52 PM)
	 * @return int
	 */
	public int size(){
		return this.getQueueSize();
	}
	
	/**
	 * Converts the queue to a string.
	 * 
	 * Creation Date: (Jun 16, 2004 5:17:41 PM)
	 * @return String
	 */
	public String toString(){
		String ret = "Queue: ";
		ret += this.getQueueName() + " ";
		ret += this.getQueueNumber() + " ";
		ret += this.getQueueSize() + " ";
		ret += "Owner: " + this.getQueueOwner().toString(); 
		ret += "\nContents:\n" + this.contents.toString();
		
		return ret;
	}
	
	/**
	 * Unregisters a customer. The customer can not dequeue tuples after 
	 * it has been unregistred.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	public synchronized void unregister(Object customer) throws XATQueueException {	
		// if the cached one is to be moved
		//System.out.println("Catched:" + this.cachedCustomer + "To be unregistered:" + customer.toString());
		
		if (this.cachedCustomer == customer) {
			this.cachedCustomer = null;
			
			// no other customers, easy
			if ((this.customerMap == null) || (this.customerMap.size() == 0)) {
				this.cachedCursor = 0;
				this.cachedCustomer = null;
	
				// also flush queues
				this.contents.clear();
			}
			// otherwise, move some lucky guy from the customer map to the
			// cache
			else {
				Set keySet = this.customerMap.keySet();
				Iterator it = keySet.iterator();
				while (it.hasNext()) {
					this.cachedCustomer = it.next();
					if (this.cachedCustomer == customer) {
						this.customerMap.remove(this.cachedCustomer);
						this.cachedCustomer = null;
					}
					else break;
				}
				this.cachedCursor =
					((XATQueueCursor) this.customerMap.get(this.cachedCustomer)).getCursor();
				this.customerMap.remove(this.cachedCustomer);
	
				//if no other customers, delete the map
				if (this.customerMap.isEmpty()) {
					this.customerMap = null;
					this.minCursor = Integer.MAX_VALUE;
	
					//also flush queues
					for(int i = 0; i < this.cachedCursor; i++)
						this.contents.remove(0);
					this.cachedCursor = 0;
				}
				//otherwise it may need to recompute the minimal cursor 
				//because the old one may have been removed
				else {
					if (this.cachedCursor == this.minCursor) {
						this.computeMinCursor();
					}
				}	
			}
		}
		//otherwise remove some one in the customer map
		else {
			//first check whether customer has registerd
			if (this.customerMap == null || !this.customerMap.containsKey(customer)) {
				return;
			}
			//throw new XATQueueException("The customer to be deleted does not exist!");			
	
			XATQueueCursor cursor = (XATQueueCursor) this.customerMap.get(customer);
			if (cursor == null)
				throw new XATQueueException("The customer to be deleted does not exist!");
	
			//remove customer from map
			this.customerMap.remove(customer);
	
			//if no other customers, delete the map
			if (this.customerMap.isEmpty()) {
				this.customerMap = null;
				this.minCursor = Integer.MAX_VALUE;
			}
			//otherwise it may need to recompute the minimal cursor 
			//because the old one may have been removed
			else {
				if (cursor.getCursor() == this.minCursor) {
					this.computeMinCursor();
				}
				else {
					this.minCursor = Integer.MAX_VALUE;
				}
			}
	
		}
	}
	
	public int getBoxInputQID() {
		return this.boxInputQID;
	}
	
	public void setBoxInputQID(int newID) {
		this.boxInputQID = newID;
	}
	protected void appendNewResult(String newResult) {
		currentResultBuffer.append(newResult);
	}

	public String getCurrentResult() {
		return currentResultBuffer.toString();
	}

	public void clearCurrentResult() {
		currentResultBuffer = new StringBuffer();
	}

	public boolean isFinalQueue() {
		return this.isFinalQueue;
	}

	public void setAsFinalQueue(boolean finalQueue) {
		this.isFinalQueue = finalQueue;
	}

}
