/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * Implements XATQueue.
 * All methods are synchronized for multi-thread accessing.
 * 
 * Creation date: (4/27/2003 1:23:37 AM)
 * @author: Jinhui Jian
 */
import java.util.ArrayList;
//import java.util.Iterator;
import java.util.List;
//import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;

/**
 * This class implements a SIMPLE queue where all tuples are always stored in main
 * memory. It is assumed that there will always be enough memory to process the
 * tuples, so no backup persistent storage is required.
 * 
 * It is "simple" because it just serves as a temp buffer to hold tuples. 
 * It assumes there is only one queue customer and one queue owner.
 * The enqueue, dequeue and cleartuple methods do not check the customer and owner information.
 * Any operator can enqueue from or dequeue to this simple queue, and tuples can be cleared from the queue 
 * right away after being dequeued.  
 * 
 * Creation Date: (Jul 27, 2004 9:53:48 AM)
 */
public class XATSimpleQueueImp extends XATQueueImp {
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:37:04 PM)
	 */
	public XATSimpleQueueImp() {
		super();
		CLEANTHRESHOLD = 0;
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:59 PM)
	 * @param owner Object
	 */
	public XATSimpleQueueImp(Object owner) {
		super(owner);
		CLEANTHRESHOLD = 0;
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:53 PM)
	 * @param owner Object
	 * @param queueName String
	 */
	public XATSimpleQueueImp(Object owner, String queueName) {
		super(owner, queueName);
		CLEANTHRESHOLD = 0;
	}

	/**
	 * Clean up tuples that have been dequeued by all customers.
	 *
	 * Creation date: (4/27/2003 3:10:56 AM)
	 */
	public synchronized void cleanTuples() {

		// the number CLEANTHRESHOLD is a tradeoff between speed and storage:
		// larger number requires less cleanup,
		// smaller number reduces the storage overhead for old tuples
		if (this.cachedCursor > CLEANTHRESHOLD) {
			// clean up tuples
			this.contents.subList(0, cachedCursor).clear();
			cachedCursor = 0;
		}
		//for (int i = 0; i < min; i++) {
		//	this.contents.remove(0);
		//}
	}
	
	/**
	 * Dequeues a number of tuples into an array.
	 * 
	 * The parameter customer has no effects on this method.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:22 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	public synchronized XATTuple[] dequeue(int number, Object customer) {

		// make sure there are enough tuples to dequeue
		// if not, set to maximum possible to dequeue for this customer
		if (number > this.contents.size() - cachedCursor) {
			number = this.contents.size();
		}
		
		// get the tuples
		//int cursor = this.getCursor(customer);
		XATTuple[] result = new XATTuple[number];
		for (int i = 0; i < number; i++) {
			result[i] = (XATTuple)this.contents.get(cachedCursor + i);	
		}
		
		cachedCursor += number;
		
		// clean up the queue
		this.cleanTuples();
		
		// update queue attributes
		//this.setCursor(customer, cursor + number);
		this.totalDequeued += result.length;
		
		return result;
	}

	/**
	 * Dequeues a tuple.
	 * 
	 * The parameter customer has no effects on this method.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:31 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple dequeue(Object customer) throws XATQueueException {
			
		// also make sure there is a tuple to dequeue
		// if not, then cannot dequeue
		if ((this.contents.size() - cachedCursor) == 0) {
			throw new XATQueueException("Not enough tuples to dequeue");
		}
		else {		
			// get the tuple
			//int cursor = this.getCursor(customer);
			XATTuple result = (XATTuple)this.contents.get(cachedCursor);
			
			cachedCursor++;
			
			// clean up the queue
			this.cleanTuples();
			
			// update queue attributes
			//this.setCursor(customer, this.getCursor(customer) + 1);
			this.totalDequeued++;
		
			return result;
		}
	}
	
	/**
	 * Dequeues a number of tuples into a List.
	 * The parameter customer has no effects on this method.
	 * Creation Date: (Jun 16, 2004 2:34:18 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	public synchronized List dequeueN(int number, Object customer) {		

		// make sure there are enough tuples to dequeue
		// if not, set to maximum possible to dequeue for this customer
		if (number > this.contents.size() - cachedCursor) {
			number = this.contents.size() - cachedCursor;
		}
		
		// get the tuples
		//int cursor = this.getCursor(customer);
		List result = new ArrayList(number);
		//result = this.contents.
		for(int i = 0; i < number; i++) {
			result.add(this.contents.get(cachedCursor + i));
		}

		// clean up the queue
		cachedCursor += number;
		
		this.cleanTuples();

		// update queue attributes
		//this.setCursor(customer, cursor + number);
		this.totalDequeued += result.size();
		
		return result;
	}

	/**
	 * Enqueues a list of tuples.
	 * The parameter provider has no effects on this method.
	 *
	 * Creation date: (Jun 15, 2004 11:13:36 AM)
	 * @param newTuples List
	 * @param provider Object
	 */
	public synchronized void enqueue(List newTuples, Object provider) {
		// add the tuples
		this.contents.addAll(newTuples);
		
		// update queue attributes
		this.totalEnqueued += newTuples.size();
		/*
		if (this.isFinalQueue) {
			for (int i=0; i<newTuples.size(); i++) {
				XATTuple tuple = (XATTuple)newTuples.get(i);
				if (!tuple.isPunctuation()) {
					this.appendNewResult(tuple.toString()+"\n");
				}
			}
			if (this.totalEnqueued%100 == 0) {
				this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
			}
		} else {
			this.contents.addAll(newTuples);
		}*/
	}

	/**
	 * Enqueues a tuple. 
	 * The parameter provider has no effects on this method.
	 *
	 * Creation Date: (Jun 16, 2004 2:36:21 PM)
	 * @param newTuple XATTuple
	 * @param provider Object
	 */
	public synchronized void enqueue(XATTuple newTuple, Object provider) {
		// add the tuple
		this.contents.add(newTuple);
		
		// update queue attributes
		this.totalEnqueued++;
		
		/*
		if (this.isFinalQueue) {
			if (!newTuple.isPunctuation()) {
				this.appendNewResult(newTuple.toString()+"\n");
				if (this.totalEnqueued%100 == 0) {
					this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
				}
			}
		} else {
			this.contents.add(newTuple);
		}*/
	}

	/**
	 * Enqueues an array of tuples. 
	 * The parameter provider has no effects on this method.
	 *
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[] 
	 * @param provider java.lang.Object
	 */
	public synchronized void enqueue(XATTuple[] newTuples, Object provider) {
		// add the tuples
		int numNewTuples = newTuples.length;
		for(int i = 0; i < numNewTuples; i++) {
			this.contents.add(newTuples[i]);
		}
		
		// update queue attributes
		this.totalEnqueued += numNewTuples;
		/*
		if (this.isFinalQueue) {
			for(int i=0; i<newTuples.length; i++) {
				if (!newTuples[i].isPunctuation()) {
					this.appendNewResult(newTuples[i].toString()+"\n");
				}
			}
			if (this.totalEnqueued%100 == 0) {
				this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
			}
		} else {
			for(int i = 0; i < newTuples.length; i++) {
				this.contents.add(newTuples[i]);
			}
		}*/
	}

	/**
	 * Returns a light weight query object of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:39:22 PM)
	 * @return LightWeightQueryObject
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("XATSimpleQueue", this.getClass().getName(), true);
	}

	/**
	 * Gets the string representation of the queue's layout.
	 * 
	 * Creation Date: (Jun 29, 2004 4:14:05 PM)
	 * @return String
	 */
	public String getQueueLayout() {
		return "|T->" + this.getQueueSize() + "|";
	}

	/**
	 * Returns the number of tuples that are in the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:40:12 PM)
	 * @return int
	 */
	public synchronized int getQueueSize() {
		return this.contents.size() - cachedCursor;
	}
	
	/**
	 * Returns the number of tuples that are in the queue.
	 * The parameter customer has no effects on this method.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @return int
	 * @param customer Object
	 */
	public synchronized int getQueueSize(Object customer) {		
		return this.contents.size() - cachedCursor;
	}

	/**
	 * Retrives a number of tuples from the queue. This is analagous to 
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * The parameter customer has no effects on this method.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:59 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	public XATTuple[] peek(int number, Object customer) {	
		// make sure the customer can peek
		//if (!this.isRegistered(customer)) {
		//	this.register(customer);
		//}

		// also make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		if (number > this.contents.size() - cachedCursor) {
			number = this.contents.size() - cachedCursor;
		}
		
		// get the tuples
		//int cursor = this.getCursor(customer);
		XATTuple[] result = new XATTuple[number];
		for(int i = 0; i < number; i++) {
			result[i] = (XATTuple)this.contents.get(cachedCursor + i);
		}
		
		return result;
	}
	
	/**
	 * Retrieves a tuple from the queue. This is analagous to dequeue, but
	 * this will not increment or cleanup cursors.
	 *
	 * The parameter customer has no effects on this method.
	 * 
	 * Creation Date: (Jun 16, 2004 4:45:49 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple peek(Object customer) throws XATQueueException {
	
		// also make sure there is a tuple to peek
		// if not, then cannot peek
		if (this.contents.size() - cachedCursor == 0) {
			throw new XATQueueException("Not enough tuples to peek.");
		}
			
		// get the tuple
		XATTuple result = (XATTuple)this.contents.get(cachedCursor);
		
		return result;
	}

	/**
	 * Retrieves the last tuple in the queue. This is for XATSubQueueImp.
	 * XATSubQueueImp needs to merge a "subtuple" into the last tuple 
	 * (which must also be a "subtuple") to form a "full tuple". This is 
	 * analagous to dequeue, but this will not increment or cleanup 
	 * cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:47:31 PM)
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple peekLast() throws XATQueueException {	
		// make sure there is a tuple to peek
		// if not, throw an exception
		if (this.contents.size() - cachedCursor == 0) {
			throw new XATQueueException("Not enough tuples to peek.");
		}
		
		return (XATTuple)contents.get(this.contents.size() - 1);
	}
	
	/**
	 * Retrieves a number of tuples from the queue. This is analagous to
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:49:09 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	public synchronized List peekN(int number, Object customer) {
	
		// make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		if (number > this.contents.size() - cachedCursor) {
			number = this.contents.size() - cachedCursor;
		}
		
		// get the tuples
		List result = new ArrayList();
		for(int i = 0; i < number; i++) {
			result.add(this.contents.get(cachedCursor + i));
		}
		
		return result;
	}
	
	/**
	 * Parameter "number" starts from 0.
	 */
	public synchronized XATTuple peekNthTuple(int number, Object customer) {
		// make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		if (number > this.contents.size() - cachedCursor) {
			number = this.contents.size() - cachedCursor;
		}
		
		// get the tuples
		XATTuple result = (XATTuple)this.contents.get(cachedCursor + number);
		
		return result;
	}
	
	/**
	 * Sets the tuples of this queue to the new tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:20 PM)
	 * @param newTuples List
	 */
	public synchronized void setTuples (List newTuples) {
		this.contents = newTuples;
	}
	
	/**
	 * Returns the contents of the queue by calling toString() on each 
	 * tuple. Caller is the consumer of the queue you want to print out.
	 * Caller can be null if this is the top output queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:50:17 PM)
	 * @param caller Object
	 * @return String
	 */
	public String printContent(Object customer) {
	    StringBuffer rs = new StringBuffer();
	
	    rs.append(this.getQueueSize());
	    rs.append(" tuples: \n");
	
	    int number = this.contents.size() - cachedCursor;
	    XATTuple[] result = this.peek(number, customer);
	    for (int i = 0; i < number; i++) {
	         rs.append(result[i].toString() + "\n");
	    }    
	    return rs.toString();
	}
	
	/**
	 * Skips the next tuple.
	 *
	 * Creation Date: (Jun 16, 2004 5:14:34 PM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	public synchronized void skip(Object customer) throws XATQueueException {	
		// check validity
		if(this.contents.size() - cachedCursor == 0)
			throw new XATQueueException("Not enough tuples to skip!");
			
		// increase cursor
		cachedCursor++;
	}
	
	/**
	 * Skips a number of tuples.
	 *
	 * Creation Date: (Jun 16, 2004 5:16:24 PM)
	 * @param number int
	 * @param customer Object
	 */
	public synchronized void skipN(int number, Object customer) throws XATQueueException {
		// check validity
		if(number > this.contents.size() - cachedCursor)
			throw new XATQueueException("Not enough tuples to dequeue!");
			
		// increase cursor
		cachedCursor += number;	
	}

	/**
	 * Returns the cursor of the customer. There is a side effect: the 
	 * customer is cached. This method needs to be protected such that any
	 * subclass can extend enqueue or dequeue.
	 * 
	 * Creation date: (4/27/2003 1:40:07 AM)
	 * @param customer Object
	 * @return int
	 */
	public int getCursor(Object customer) {	
		return cachedCursor;
	}

	/**
	 * Gets the cursor of the given customer.
	 * 
	 * Creation Date: (Jun 16, 2004 2:38:44 PM)
	 * @param customer Object
	 * @return int 
	 */
	public int getCustCursor(Object customer) {
		return this.cachedCursor;
	}

	/**
	 * Gets the customers of the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:38:05 PM)
	 * @return List
	 */
	public List getQueueCustomer() {
		List customers = new ArrayList();
		
		if(cachedCustomer != null) {
			customers.add(cachedCustomer);
		}
		
		return customers;
	}

	/**
	 * Checks whether the customer is registered with this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:07 PM)
	 * @param customer Object
	 * @return boolean
	 */
	public synchronized boolean isRegistered(Object customer){			
		if(this.cachedCustomer == null){
			return false;
		}
	
		// first check if it has been registered
		if(this.cachedCustomer == customer)
			return true;
		
		return false;
	}
	
	/**
	 * position starts from 0
	 * @param position
	 * @param tuple
	 * @param provider
	 * used for inserting migration sync flag in stream queues.
	 */
	public synchronized void insertTuple(int position, XATTuple tuple, Object provider) {
		this.contents.add(cachedCursor+position, tuple);
	}
	
public int size(){
	return this.contents.size();
}
	/**
	 * Registers a customer so that the customer can dequeue tuples from this queue.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param customer Object
	 */
	public synchronized void register(Object customer) {
		//System.out.println("Register "+customer+ " to queue "+this);

		if (this.cachedCustomer == customer) {
			//System.out.println(" customer already in cache***");
			return;
		}
		
		this.cachedCustomer = customer;
		//System.out.println(" Add customer to cache***");
		this.cachedCursor = 0;
		return;
	}

	/**
	 * This will forcefully remove all tuples from the queue, even if the queue
	 * may still have some customers. Note: this method obviously can create
	 * havoc if it is used indiscriminately, but it is a huge speedup for the
	 * XATApplication.
	 * 
	 * Creation date: (4/13/2003 1:38:36 AM)
	 */
	public synchronized void removeAllTuples() {	
		this.contents.clear();
	
		// reset all cursors
		// update the cached customer
		this.cachedCursor = 0;
	}

	/**
	 * Converts the queue to a string.
	 * 
	 * Creation Date: (Jun 16, 2004 5:17:41 PM)
	 * @return String
	 */
	public String toString(){
		String ret = "Queue: ";
		ret += this.getQueueName() + " ";
		ret += this.getQueueNumber() + " ";
		ret += this.getQueueSize() + " ";
		ret += "Owner: " + this.getQueueOwner().toString(); 
		ret += "Customer: " + this.getQueueCustomer().toString(); 
		
		return ret;
	}

	/**
	 * Unregisters a customer. The customer can not dequeue tuples after 
	 * it has been unregistred.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	public synchronized void unregister(Object customer) throws XATQueueException {	
		// if the cached one is to be moved
		System.out.println("Cached:" + this.cachedCustomer + "To be unregistered:" + customer.toString());
		
		if (this.cachedCustomer == customer) {
			this.cachedCustomer = null;
			this.cleanTuples();
		}
	}

	@Override
	public synchronized void enqueue(XATTupleNegativeImp newTuple, Object provider) {
		// TODO Auto-generated method stub
		
//		 add the tuple
		this.contents.add(newTuple);
		
		// update queue attributes
		this.totalEnqueued++;
		
		/*
		if (this.isFinalQueue) {
			if (!newTuple.isPunctuation()) {
				this.appendNewResult(newTuple.toString()+"\n");
				if (this.totalEnqueued%100 == 0) {
					this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
				}
			}
		} else {
			this.contents.add(newTuple);
		}*/
		
	}

	@Override
	public synchronized void enqueue(XATTupleNegativeImp[] newTuples, Object provider) {
		// add the tuples
		int numNewTuples = newTuples.length;
		for(int i = 0; i < numNewTuples; i++) {
			this.contents.add(newTuples[i]);
		}
		
		// update queue attributes
		this.totalEnqueued += numNewTuples;
		/*
		if (this.isFinalQueue) {
			for(int i=0; i<newTuples.length; i++) {
				if (!newTuples[i].isPunctuation()) {
					this.appendNewResult(newTuples[i].toString()+"\n");
				}
			}
			if (this.totalEnqueued%100 == 0) {
				this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
			}
		} else {
			for(int i = 0; i < newTuples.length; i++) {
				this.contents.add(newTuples[i]);
			}
		}*/
		
	}

	/**
	 * Increase the cursor of the cached customer by the given value.
	 *
	 * Creation date: (4/27/2003 1:53:53 AM)
	 * @param number int
	 */
	//public void increaseCursor(int number) {
	//	if(this.cachedCursor + number <= this.contents.size())
	//		this.cachedCursor += number;
	//	else
	//		this.cachedCursor = this.contents.size();
		
		//reset the cursor
		//if (cachedCursor == this.getQueueSize()) {
		//	cachedCursor = 0;
		//}	
	//}
}
