/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct 21, 2004
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.state;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;

import java.io.Serializable;
import java.util.ArrayList;
/**
 * @author Bin Liu
 *	This class holds all WEquiJoinStateImp from input queues of the join operator having the same
 *	partition ID. It is used by class PMJHashStates. It is the basic unit when moving states
 *	across machines or pushing into the disks
 *	Here is the relationship of states in a Partitioned-Mway-Symmetric-Hash-Join:
 *	(1) All states (from multiple input queues) are partitioned into many partitions
 *	(2) Partitions with the same ID from input queues are put into this HashStatePartition
 *	(3) Multiple tuples may exist for each input queue, it is captured in WEquiJoinStateImp 
 */
public class HashStatePartitionGroup implements Serializable {
	private int partitionID = -1;
	private WEquiJoinStateImp[] hashStatePartitions;	// hold all states correspond to multiple input queues
	private int inputQueueSize = -1;	// number of input queues in the join operator
	private int[] joinAttrPositions;	// indicates the join attribute positions of each input queue
	private int[] numOfCells;			// indicates number of cells in each input Queue
	
	// some book keeping variables that describe basic information about the partition group
	private long totalTuples = 0; // record the total number of tuples have been inserted, sum of all tuples in each WEquiJoinStateImp
	private int[] totalTuplesArray; // to record the total number of tuples have been insterted, in terms of each WEquiJoinStateImp
									// use this to provide a precise cost of states composed of different join columns
	private double partitionGroupSize = 0.0; // to record the size of the partition group, by multiply the number of tuples in each WEquiJoinState and
											// the corresponding the unit_size in terms of number of attributes
	
	private long totalTuplesOutputed = 0; // record the total number of tuples have been outputed in this partition
	
	/*
	 * This is used in a query tree case, if the tuple does not make through the whole query tree that generates
	 * final output, then it is treated as intermediate result. The pure intermediate result of a single operator
	 * is captured by the totalTuples. While here, it will trace back to all previous joins, the effects recorded 
	 * each partition is the global (in terms of query tree)
	 */
	private long totalProjectedTupleNumber = 0;
	
	public void setStatePartition(int position, WEquiJoinStateImp state) {
	    hashStatePartitions[position] = state;
	}
	
	public double getPartitionGroupSize(){
		return partitionGroupSize;
	}
	public void addTotalProjectedTupleNumber(int pt){
		totalProjectedTupleNumber += pt;
	}
	/*
	 * This is used to estimate the main memory usage when tracing the intermediate results, multiple the number
	 * of intermediate results by the unit_rate of the tuple, now, the returned totalProjectedTupleNumber is the 
	 * estimation of the main memory size instead of the number of tuples -- Bin, May-18-2005
	 */	
	public void addTotalProjectedTupleNumber(int pt,double inter_rate){
		totalProjectedTupleNumber += pt*(int)inter_rate;
	}

	public int getTotalProjectedTupleNumber(){
		return (int)totalProjectedTupleNumber;
	}
	// some put and get functions of totalTuplesOutputed
	public void addTotalTuplesOutputed(int tt){
		totalTuplesOutputed += tt;
	}
	public int getTotalTuplesOutputed(){
		return (int)totalTuplesOutputed;
	}
	
	public void setPartitionID (int id){
		partitionID = id;
	}
	public int getPartitionID(){
		return partitionID;
	}
	
	public int getTotalNumOfTuples(){
		return (int)totalTuples;
	}
	
	public void setTotalNumOfTuples(int total) {
	    this.totalTuples = total;
	}
	
	public int[] getTotalNumofTuplesArray(){
		return totalTuplesArray;
	}
	
	/*
	 * return all states of this partition, it is used in merging disk residented partitions 
	 */
	public WEquiJoinStateImp[] getHashStates(){
		return hashStatePartitions;
	}
	
	/**
	 * Constructor: jap = join attribute positions, ncl = number of cell per queue
	 * @param jap
	 */
	public HashStatePartitionGroup(int[] jap, int[] ncl){
	    //this.partitionID = partitionid;
		numOfCells = new int[jap.length];
		joinAttrPositions = new int[jap.length]; // make sure it is not a reference, not sure whether i need to do this?
		for (int i=0; i<jap.length; i++){
			joinAttrPositions[i] = jap[i];
			numOfCells[i] = ncl[i]; //initialization
		}
		inputQueueSize = jap.length;
		hashStatePartitions = new WEquiJoinStateImp[inputQueueSize];
		for (int i=0; i<inputQueueSize; i++){
			hashStatePartitions[i] = new WEquiJoinStateImp();
		}
		totalTuplesArray = new int[inputQueueSize];
	}

	/**
	 * Constructor: jap = join attribute positions, ncl = number of cell per queue
	 * sts = the actual WEquiJoinState of the HashStatePartition
	 * This construction is used to compose the hashstatepartition that contains states that
	 * have been invalidated.
	 * @param jap
	 */
	public HashStatePartitionGroup(int[] jap, int[] ncl, WEquiJoinStateImp[] sts){
	    //this.partitionID = partitionid;
		numOfCells = new int[jap.length];
		joinAttrPositions = new int[jap.length]; // make sure it is not a reference, not sure whether i need to do this?
		for (int i=0; i<jap.length; i++){
			joinAttrPositions[i] = jap[i];
			numOfCells[i] = ncl[i]; //initialization
		}
		inputQueueSize = jap.length;
		if (sts.length != inputQueueSize){
			System.out.println("ERROR. The number of QEuilJoinState is not the same with the number of queues");
		}
		hashStatePartitions = sts;
		totalTuplesArray = new int[inputQueueSize]; 
	}
	
	/**
	 * Add one tuple into the HashBucket that indicates by queueIndex
	 * @param tp
	 * @param queueIndex
	 */
	public boolean insertTuple (Object key, XATTuple tp, int queueIndex){
		if (queueIndex >= inputQueueSize) {
			System.out.println("Tuple can not be inserted. queueIndex > inputQueueSize?");
			return false;
		}
		
		hashStatePartitions[queueIndex].insertTuple(key, tp);
		totalTuples ++;
		totalTuplesArray[queueIndex]++;    
		partitionGroupSize += numOfCells[queueIndex];
		return true;
	} //end addTuple
	
	public void decreaseTotalTuplesByOne(int queueIndex) {
	    totalTuples--;
		totalTuplesArray[queueIndex]--;
		totalTuplesArray[queueIndex]--;    
		partitionGroupSize -= numOfCells[queueIndex];
	}
	
	/**
	 * Encapuslate the major logic of m-way hash join, the probe of multiple queues are based on a SWEEP
	 * algorithm. It first probe the queues left to the queueIndex, then to the right of the queueIndex
	 * the result tuples are combined without any projections 
	 * @param tp
	 * @param queueIndex
	 * @return
	 */
	public XATTuple[] probe(XATTuple tp, int queueIndex){
		int currentJoinPos = -1;
		Object joinValue = null;
		ArrayList sourceTuples = new ArrayList();
		ArrayList matchedTuples = new ArrayList();
		
		// using a sweep algorithm to probe all other queue states 
		// queueIndex from 0~(inputQueueSize-1)
		currentJoinPos = joinAttrPositions[queueIndex];
		sourceTuples.add(tp);
		
		for (int i=queueIndex-1; i>=0; i--){
			
			for (int j=0; j<sourceTuples.size(); j++){
				XATTuple currTuple = (XATTuple)sourceTuples.get(j);
				joinValue = currTuple.getCell(currentJoinPos).getValue();
				XATTuple[] currMatchedTuples = hashStatePartitions[i].probe(joinValue);				
				// merge currTuple with currMatchedTuples with currTuple on the RIGHT
				// put the combined tuple in the matchedTuples
				if (currMatchedTuples != null){
					for (int k=0; k<currMatchedTuples.length; k++){
						XATTuple cmbTuple = XATTupleImp.combineTuples(currMatchedTuples[k],currTuple);
						matchedTuples.add(cmbTuple);
					}
				}
			}			
			if (matchedTuples.isEmpty()){
				// if no match tuples were found, just return
				return null; // or empty XATTuple?
			}
			// if matchedTuples are found, then put the matchedTuples into sourceTuples, and loop
			sourceTuples.clear();
			for (int k=0; k<matchedTuples.size();k++){
				sourceTuples.add(matchedTuples.get(k));
			}
			matchedTuples.clear();
			// change the join posistion index, in a left <-- scan, the global join index will exactally 
			// the same with the one joinAttrPositions
			currentJoinPos = joinAttrPositions[i];
		} //
		
		// prepare the current base position
		int currBasePosition = 0; // indicates number of cells in one XATTuple so far without the newly added one
		for (int i=0; i<queueIndex; i++){
			currBasePosition += numOfCells[i];
		}
		
		for (int i=queueIndex+1; i<inputQueueSize; i++){
			
			for (int j=0; j<sourceTuples.size(); j++){
				XATTuple currTuple = (XATTuple)sourceTuples.get(j);
				currentJoinPos = currBasePosition + joinAttrPositions[i-1];				
				joinValue = currTuple.getCell(currentJoinPos).getValue();
				XATTuple[] currMatchedTuples = hashStatePartitions[i].probe(joinValue);				
				// merge currTuple with currMatchedTuples with currTuple on the LEFT
				// put the combined tuple in the matchedTuples
				if (currMatchedTuples != null){
					for (int k=0; k<currMatchedTuples.length; k++){
						XATTuple cmbTuple = XATTupleImp.combineTuples(currTuple,currMatchedTuples[k]);
						matchedTuples.add(cmbTuple);
					}
				}
			}			
			if (matchedTuples.isEmpty()){
				// if no match tuples were found, just return
				return null; // or empty XATTuple?
			}
			// if matchedTuples are found, then put the matchedTuples into sourceTuples, and loop
			sourceTuples.clear();
			for (int k=0; k<matchedTuples.size();k++){
				sourceTuples.add(matchedTuples.get(k));
			}
			matchedTuples.clear();
			// change the base join posistion index, in a --> right scan, the global join index appends the
			// existing currBasePosition
			currBasePosition += numOfCells[i];			
		} //
		if (sourceTuples.isEmpty()) {
			return null;
		} else {
			return (XATTuple[])sourceTuples.toArray(new XATTuple[]{});
		}
		
	} // end probe

	/**
	 * A tuple (usually a singleton tuple) from a different partition group is used to purge 
	 * to-be-expired group during parallel track partition redistribution. This tuple only 
	 * purged group with different partition ID from itself.
	 * This can expire tuples in to-be-expired partitions faster. 
	 * @param tp
	 * @param windowSize
	 * @param isTopJoinDuringPT
	 * @return
	 */
	public void fastPurge(XATTuple tp, long windowSize, boolean isSenderJoinDuringPT, int excludingPosition) {
	    for(int i=0; i<hashStatePartitions.length; i++) {
	        if(i != excludingPosition) {
	            int num_purged = hashStatePartitions[i].purgeByWindow(windowSize, tp, isSenderJoinDuringPT);
	            //System.out.println("tuples fast purged: " + num_purged);
	            totalTuples = totalTuples - num_purged;
	        }
	    }
	}
	
	/**
	 * Encapuslate the major logic of window-based m-way hash join, the probe of multiple queues are based on a SWEEP
	 * algorithm. It first purges all out-of-dated tuples in the state, it then 
	 * probes the queues left to the queueIndex, then to the right of the queueIndex
	 * the result tuples are combined without any projections 
	 * @param tp
	 * @param queueIndex
	 * @return
	 */
	public XATTuple[] purgeProbe(XATTuple tp, int queueIndex, long windowSize, boolean isSenderJoinDuringPT, boolean isTopJoinDuringPT){
		int currentJoinPos = -1;
		Object joinValue = null;
		ArrayList sourceTuples = new ArrayList();
		ArrayList matchedTuples = new ArrayList();
		
		// using a sweep algorithm to probe all other queue states 
		// queueIndex from 0~(inputQueueSize-1)
		currentJoinPos = joinAttrPositions[queueIndex];
		sourceTuples.add(tp);
		
		for (int i=queueIndex-1; i>=0; i--){
			
			for (int j=0; j<sourceTuples.size(); j++){			
				XATTuple currTuple = (XATTuple)sourceTuples.get(j);

				/// purge out-of-date states, DOWN_SWEEP ///
				//long latestProcessedTime = currTuple.getNthTimeStamp(currentJoinPos);
				//long latestProcessedTime = currTuple.getMinTimeStamp();
				int dest_joinPos = joinAttrPositions[i];
                long latesttimestamp = currTuple.getMinTimeStamp();
                int num_purged = hashStatePartitions[i].purgeByWindow(windowSize, latesttimestamp);
				//int num_purged = hashStatePartitions[i].purgeByWindow(windowSize, currTuple, isSenderJoinDuringPT);
				//System.out.println("num_purged:" + num_purged);
				totalTuples = totalTuples - num_purged;
				
				joinValue = currTuple.getCell(currentJoinPos).getValue();
				XATTuple[] currMatchedTuples = hashStatePartitions[i].probe(joinValue);				
				// merge currTuple with currMatchedTuples with currTuple on the RIGHT
				// put the combined tuple in the matchedTuples
				if (currMatchedTuples != null){
					for (int k=0; k<currMatchedTuples.length; k++){
					    if(isJoinable(windowSize, currMatchedTuples[k], currTuple, isTopJoinDuringPT)) {
					        XATTuple cmbTuple = XATTupleImp.combineTuples(currMatchedTuples[k],currTuple);
					        cmbTuple.setPartitionID(currTuple.getPartitionID()); //assume transitive join predicates
					        matchedTuples.add(cmbTuple);
					    }
					}
				}
			}			
			if (matchedTuples.isEmpty()){
				// if no match tuples were found, just return
				return null; // or empty XATTuple?
			}
			// if matchedTuples are found, then put the matchedTuples into sourceTuples, and loop
			sourceTuples.clear();
			for (int k=0; k<matchedTuples.size();k++){
				sourceTuples.add(matchedTuples.get(k));
			}
			matchedTuples.clear();
			// change the join posistion index, in a left <-- scan, the global join index will exactally 
			// the same with the one joinAttrPositions
			currentJoinPos = joinAttrPositions[i];
		} 
		
		// prepare the current base position
		int currBasePosition = 0; // indicates number of cells in one XATTuple so far without the newly added one
		for (int i=0; i<queueIndex; i++){
			currBasePosition += numOfCells[i];
		}
		
		for (int i=queueIndex+1; i<inputQueueSize; i++){
			
			for (int j=0; j<sourceTuples.size(); j++){
				XATTuple currTuple = (XATTuple)sourceTuples.get(j);
				currentJoinPos = currBasePosition + joinAttrPositions[i-1];
				
				/// purge out-of-date states, UP-SWEEP ///
				long lastestProcessedTime = currTuple.getNthTimeStamp(currentJoinPos);
				int dest_joinPos = joinAttrPositions[i];
                long latesttimestamp = currTuple.getMinTimeStamp();
                int num_purged = hashStatePartitions[i].purgeByWindow(windowSize, latesttimestamp);
				//int num_purged = hashStatePartitions[i].purgeByWindow(windowSize, currTuple, isSenderJoinDuringPT);
				totalTuples = totalTuples - num_purged;
				
				joinValue = currTuple.getCell(currentJoinPos).getValue();
				XATTuple[] currMatchedTuples = hashStatePartitions[i].probe(joinValue);				
				// merge currTuple with currMatchedTuples with currTuple on the LEFT
				// put the combined tuple in the matchedTuples
				if (currMatchedTuples != null){
					for (int k=0; k<currMatchedTuples.length; k++){
					    if(isJoinable(windowSize, currMatchedTuples[k], currTuple, isTopJoinDuringPT)) {
					        XATTuple cmbTuple = XATTupleImp.combineTuples(currTuple,currMatchedTuples[k]);
					        cmbTuple.setPartitionID(currTuple.getPartitionID()); //assume transitive join predicates
					        matchedTuples.add(cmbTuple);
					    }
					}
				}
			}			
			if (matchedTuples.isEmpty()){
				// if no match tuples were found, just return
				return null; // or empty XATTuple?
			}
			// if matchedTuples are found, then put the matchedTuples into sourceTuples, and loop
			sourceTuples.clear();
			for (int k=0; k<matchedTuples.size();k++){
				sourceTuples.add(matchedTuples.get(k));
			}
			matchedTuples.clear();
			// change the base join posistion index, in a --> right scan, the global join index appends the
			// existing currBasePosition
			currBasePosition += numOfCells[i];			
		} 
		if (sourceTuples.isEmpty()) {
			return null;
		} else {
			return (XATTuple[])sourceTuples.toArray(new XATTuple[]{});
		}
		
	} // end purgeProbe
	
	/*
	 * In case the hashstatepartition has not been pushed into the disk, then we can directly purge the states
	 * that accumulated in each corresponding WEquiJoinStateImp
	 */
	public void purgeState(){
		int total_purged = 0;
		for (int i=0; i<hashStatePartitions.length; i++){
			total_purged += hashStatePartitions[i].purgeInvalidatedStates();
		}
		System.out.println("total purged = " + total_purged);
		totalTuples = totalTuples - total_purged;
	}
	
	/*
	 * compose a HashStatePartition from the invalidate_states, 
	 * maybe we need to check whether the states are not null?
	 */
	public HashStatePartitionGroup getInvalidatedStates(int[] joinAP, int[] ncl){
		WEquiJoinStateImp[] sts = new WEquiJoinStateImp[hashStatePartitions.length];
		for (int i=0; i<hashStatePartitions.length; i++){
			sts[i] = hashStatePartitions[i].getInvalidatedStates();
		}		
		HashStatePartitionGroup ret = new HashStatePartitionGroup(joinAP, ncl, sts);
		
		return ret;
	}
	/**
	 * Check if two tuples to be joined are within window frame from each other.
	 * @param windowsize
	 * @param tuple1
	 * @param tuples2
	 * @return
	 */
	public boolean isJoinable(long windowsize, XATTuple tuple1, XATTuple tuple2, boolean isTopJoinDuringPT) {
	    boolean joinable = false;
	    if(isTopJoinDuringPT) {
	        if(tuple1.isNew() && tuple2.isNew())
	            return false;
	    }
	    if ((Math.abs(tuple1.getMaxTimeStamp() - tuple2.getMinTimeStamp()) < windowsize) 
	            && (Math.abs(tuple1.getMinTimeStamp() - tuple2.getMaxTimeStamp()) < windowsize))
	        return joinable = true;
	    
	    return joinable;
	}
	
	// add more functions such as getPartition, removePartition, getLargestPartitionID ...
	// for chooseing partitions to move or to push into the disk.
	
	public String toString() {
	    StringBuffer s = new StringBuffer();
	    s.append("\nTuples in PartitionGroup: " + this.getTotalNumOfTuples());
	    for(int i=0; i<this.hashStatePartitions.length; i++) {
	        s.append("\npartition: " + i);
	        s.append(hashStatePartitions[i]);
	    }
	    return s.toString();
	}
	
	
} //end of HashStatePartition
