/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import java.util.Iterator;
//import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATWBinaryJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
	 * DummyJoinOperator that simulates a real operator.  The DummyJoinOperator
	 * does not have any semantic meaning, but rather, it is used to simulate a
	 * real operator so that operator scheduling will have more meaning <p>
	 *
	 * This implementation is basically the same as DummyOperator, except that it dequeues
	 * tuples from 2 queues and can possibly have selectivity > 1. 
	 * The operator will dequeue the required number of tuples (from the NUMBER_OF_TUPLES_TO_DEQUEUE property),
	 * figure out how many to keep, and then sleep for a certain amount of time. <p>
	 *
	 * selectivity is used as follows: the number of tuples that will be outputted
	 * is equal to the number in the input buffer, N, times REAL_SELECTIVITY, r.
	 * so if 10 tuples come in and the selectivity is 20%, only 2 tuples will
	 * be placed in the output buffer <p>
	 *
	 * Because the tuples dont really matter, we first will figure out how
	 * many tuples to output N', then sleep for RESULT_DELAY * N' milleseconds.
	 * This will simulate actual processing. <p>
	 *
	 * Once it is done sleeping, then the tuples will be placed in the output buffer.  
	 */
public class ProbabilisticMultiStreamOperatorImp
	extends XATMultiQueueStreamOperatorImp
	implements ProbabilisticMultiStreamOperator {

	public ProbabilisticMultiStreamOperatorImp(ProbabilisticMultiStreamOperatorImp oldop, XATNode newnode) {
		super(oldop, newnode);
	}
	/**
	 * DummyJoinOperatorImp constructor comment.
	 * @param init edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
	 */
	public ProbabilisticMultiStreamOperatorImp(XATNode init) {
		super(init);
	}
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		Iterator i = this.getInputQueues();
		while (i.hasNext()) {
			XATQueue q = (XATQueue) i.next();
			if (q.getSchema() == null){
				return false;
			}
		}
		if (this.getOutputQueue().getSchema() == null) {
			init();
		}

		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		double selectivity = Double.parseDouble(this.getProperty("REAL_SELECTIVITY"));
		double delay = Double.parseDouble(this.getProperty("RESULT_DELAY"));

		int tuplesInLeft = this.getLeftInputQueue().getQueueSize(this);
		int tuplesInRight = this.getRightInputQueue().getQueueSize(this);

		if (tuplesInLeft == 0 || tuplesInRight == 0) {
			return false;
		}
		
		/** represent how many we will actually dequeue from each side */
		try {
		
			XATTuple[] left = this.getLeftInputQueue().dequeue(tuplesInLeft, this);
			XATTuple[] right = this.getRightInputQueue().dequeue(tuplesInRight, this);
		
			int numberToOutput = (int) (selectivity * (left.length + right.length));
		
			int counter = 0;
			int counter2 = 0;
			XATTuple[] output = new XATTuple[numberToOutput];
		
			for (int j = 0; j < numberToOutput; j++) {
		
				if (counter2 >= left.length-1)
					counter2 = 0;
				
				output[counter] = new XATTupleImp(left[counter2].getArrayData());
		
				XATDataValue[] rightData = right[0].getArrayData();
		
				for (int m = 0; m < rightData.length; m++) {
		
					output[counter].addCell(rightData[m]);
					//XATTimeStamp ts = new XATTimeStampImp();
					
					output[counter].combineTimeStamps(left[counter2], right[0]);
				}
		
				counter++;
				counter2++;
		
			}
		
			/** now enqueue the tuples */
			int time = new Double(delay * 1000).intValue();
			/*
			Thread.sleep(time);
			*/
		
			this.getOutputQueue().enqueue(output, this);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return true;
	}
	/**
	* Runs the actual operator.
	*
	* Unlike in Rainbow, the Stream operators have the implementation tied to the operator
	* because the StreamOperators represent physical operators instead of Logic.
	* Creation date: (1/6/2003 5:15:27 PM)
	*/
	public boolean run() {

		Iterator i = this.getInputQueues();

		while (i.hasNext()) {

			XATQueue q = (XATQueue) i.next();
			if (q.getSchema() == null){
				return false;
			}
		}

		if (this.getOutputQueue().getSchema() == null) {
			init();
		}

		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		int howMany =
			Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		double selectivity =
			Double.parseDouble(this.getProperty("REAL_SELECTIVITY"));
		double delay = Double.parseDouble(this.getProperty("RESULT_DELAY"));

		/** 
		 * get the tuples 
		 * first figure out how many are in the left queue and then the right queue
		 * and try to take evenly from each
		 */
		int tuplesInLeft = this.getLeftInputQueue().getQueueSize(this);
		int tuplesInRight = this.getRightInputQueue().getQueueSize(this);

		if (tuplesInLeft == 0 || tuplesInRight == 0) {

			return false;
		}
		/** represent how many we will actually dequeue from each side */
		try {
		
			XATTuple[] left = this.getLeftInputQueue().dequeue(tuplesInLeft, this);
			XATTuple[] right = this.getRightInputQueue().dequeue(tuplesInRight, this);
		
			int numberToOutput = (int) (selectivity * (left.length + right.length));
		
			int counter = 0;
			int counter2 = 0;
			XATTuple[] output = new XATTuple[numberToOutput];
		
			for (int j = 0; j < numberToOutput; j++) {
		
				if (counter2 >= left.length-1)
					counter2 = 0;
				
				output[counter] = new XATTupleImp(left[counter2].getArrayData());
		
				XATDataValue[] rightData = right[0].getArrayData();
		
				for (int m = 0; m < rightData.length; m++) {
		
					output[counter].addCell(rightData[m]);
					//XATTimeStamp ts = new XATTimeStampImp();
					
					output[counter].combineTimeStamps(left[counter2], right[0]);
//					newtimes.addAll(left[counter2].getTimeStamp());
//					
//					ts.combine(left[counter2].getTimeStamp(), right[0].getTimeStamp());
//					output[counter].setTimeStamp(ts);
				}
		
				counter++;
				counter2++;
		
			}
		
			/** now enqueue the tuples */
		
			int time = new Double(delay * 1000).intValue();
			/*
			Thread.sleep(time);
			*/
		
			this.getOutputQueue().enqueue(output, this);
		
//		} catch (XATTimeStampException e) {
//			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return true;

	}
	/**
	 * The Dummy operator has 2 parameters.  The SELECTIVITY
	 * is a value between 0 and 1 and RESULT_DELAY is the
	 * time it takes (in milleseconds) for this operator to place a tuple in its output queue
	 * 
	 */
	protected void setDefaultProperties() {
		super.setDefaultProperties();
		addProperty("RESULT_DELAY", "100");
		addProperty("REAL_SELECTIVITY", ".5");
	}
	/**
	* sets the given property to the new value.  This method is responsible
	* for validating the given property.  An exception is thrown if
	* the value is not correct.
	* Validation should be done in the following manner: 
	* First the method should check if the object supports the given property by calling
	* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
	* be thrown. <p>
	*
	* Once it is known that the object supports the property, the value must be validated.  Each implementor of
	* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
	* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
	* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
	* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
	* will reduce duplicate code. <p>
	*
	* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
	* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
	* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
	* it is up to the subclass that created the property to validate it (as described above).
	* @param name the name of the property to set
	* @param value the new value of the property
	* @throws Exception if the given property cannot be found or
	* if the value is not valid for the given property.
	*/
	public void setProperty(String name, String value)
		throws java.lang.Exception {
		//first see if the DummyOperator supports the property name	
		if (isValidPropertyName(name)) {
			if (name.equals("REAL_SELECTIVITY")) {
				try {
					double selectivity = Double.parseDouble(value);
					if ((selectivity >= 0) && (selectivity <= 2)) {
						setNewPropertyValue(name, value);
					} else {
						throw new NumberFormatException(
							"The property named "
								+ name
								+ " requires a double value between 0 and 2 (inclusive)");
					}
				} catch (NumberFormatException e) {
					throw new NumberFormatException(
						"The property named "
							+ name
							+ " requires a double value between 0 and 2 (inclusive)");
				}
			} else {
				//the property name is valid, but its value is validated in the super class
				super.setProperty(name, value);
			}
		} else
			throw new Exception(
				"The DummyOperator does not support the property named "
					+ name);

	}

	public String toToolTipString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for (index = sName.length() - 1; index >= 0; index--)
			if (sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if (id == "-1") {
			rs = rs + hashCode();
		} else {
			rs = rs + id;
		}
		if (this.isActive())
			rs += " ACTIVE ";
		else
			rs += " INACTIVE ";

		/** now add the join predicate and window size info */
		rs += "\n";
		rs += "Properties: \n";
		rs += "RESULT_DELAY = " + this.getProperty("RESULT_DELAY") + "\n";
		rs += "REAL_SELECTIVITY = "
			+ this.getProperty("REAL_SELECTIVITY")
			+ "\n";

		return rs;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator#setLeftInputQueue(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue)
	 */
	public void setLeftInputQueue(XATQueue newQueue) throws XATQueueException {
		super.setLeftInputQueue(newQueue);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator#setLeftOutputQueue(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue)
	 */
	public void setLeftOutputQueue(XATQueue newQueue)
		throws XATQueueException {
		super.setLeftOutputQueue(newQueue);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
	 */
/*	public void init() {
		Iterator i = this.getInputQueues();
		String schemaName = new String();
		XATQueueSchema s;
		XATQueueSchema outSchema = new XATQueueSchema();

		while (i.hasNext()) {

			s = ((XATQueue) i.next()).getSchema();
			schemaName += "#" + s.getName();
			
			Iterator j = s.getColNames().iterator();
			while(j.hasNext()){
				outSchema.addColumn((String) j.next());
			}
		}

		outSchema.setName(schemaName);
		this.getOutputQueue().setSchema(outSchema);

	} */

	public boolean hasState(){
		return false;
	}
	
	public int getStateSize() {
		return 0;
	}
	
	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		Iterator i = this.getInputQueues();
		while (i.hasNext()) {
			XATQueue q = (XATQueue) i.next();
			if (q.getSchema() == null){
				return false;
			}
		}
		if (this.getOutputQueue().getSchema() == null) {
			init();
		}

		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		double selectivity = Double.parseDouble(this.getProperty("REAL_SELECTIVITY"));
		double delay = Double.parseDouble(this.getProperty("RESULT_DELAY"));

		int tuplesInLeft = this.getLeftInputQueue().getQueueSize(this);
		int tuplesInRight = this.getRightInputQueue().getQueueSize(this);

		if (tuplesInLeft == 0 || tuplesInRight == 0) {
			return false;
		}
		
		/** represent how many we will actually dequeue from each side */
		try {
		
			XATTuple[] left = this.getLeftInputQueue().dequeue(tuplesInLeft, this);
			XATTuple[] right = this.getRightInputQueue().dequeue(tuplesInRight, this);
		
			int numberToOutput = (int) (selectivity * (left.length + right.length));
		
			int counter = 0;
			int counter2 = 0;
			XATTuple[] output = new XATTuple[numberToOutput];
		
			for (int j = 0; j < numberToOutput; j++) {
		
				if (counter2 >= left.length-1)
					counter2 = 0;
				
				output[counter] = new XATTupleImp(left[counter2].getArrayData());
		
				XATDataValue[] rightData = right[0].getArrayData();
		
				for (int m = 0; m < rightData.length; m++) {
		
					output[counter].addCell(rightData[m]);
					//XATTimeStamp ts = new XATTimeStampImp();
					
					output[counter].combineTimeStamps(left[counter2], right[0]);
				}
		
				counter++;
				counter2++;
		
			}
		
			/** now enqueue the tuples */
			int time = new Double(delay * 1000).intValue();
			/*
			Thread.sleep(time);
			*/
		
			this.getOutputQueue().enqueue(output, this);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return true;
	}
	
	/*
	 * compare the states of this operaotr and the passed-in operator,
	 * the current operator is served as the old operator in old box, while the passed-in operator 
	 * is served as the new operator in new box.
	 * 
	 * when a pair of matching states is found, set the matched flag of matching states, copy the contents of the 
	 * old matching state to the matching new state by calling the copyStateContent() method in the corresponding state class.
	 * return:
	 * TRUE -- is this operator and the given operator have matched states.
	 * FALSE -- otherwise.
	 */
	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}
	/*
	 * recompute the unmatched states inside this operator.
	 * Usually the unmatched states are computed based on the matched states inside one of the 
	 * children operator of this operator.
	 * 
	 * return: 
	 * TRUE -- if the states (inside child operator) that the recomputation of the unmatched states are based on are 
	 *      all matched, and the recomputation is successful.
	 * FALSE -- if any of the states (inside child operator) that are used to recompute the unmatched states
	 *          is unmatched, or the recomputation is unsucessful.  
	 */
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	//public int setStateID(int nextid) {
	//	return nextid;
	//}

	public boolean finalizeOperation(){
		return true;
	}
	
}

