/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import java.util.Iterator;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * DummyOperator that simulates a real operator.  The DummyOperator
 * does not have any semantic meaning, but rather, it is used to simulate a
 * real operator so that operator scheduling will have more meaning <p>
 *
 * The operator will dequeue the required number of tuples (from the NUMBER_OF_TUPLES_TO_DEQUEUE property),
 * figure out how many to keep, and then sleep for a certain amount of time. <p>
 *
 * selectivity is used as follows: the number of tuples that will be outputted
 * is equal to the number in the input buffer, N, times REAL_SELECTIVITY, r.
 * so if 10 tuples come in and the selectivity is 20%, only 2 tuples will
 * be placed in the output buffer <p>
 *
 * Because the tuples dont really matter, we first will figure out how
 * many tuples to output N', then sleep for RESULT_DELAY * N' milleseconds.
 * This will simulate actual processing. <p>
 *
 * Once it is done sleeping, then the tuples will be placed in the output buffer.  
 */
public class SchemaNameChangeOperatorImp
	extends XATSingleQueueStreamOperatorImp
	implements SchemaNameChangeOperator {

	boolean leaf;
	String newStreamName;
	String[] newColNames;

	/**
	 * DummyOperatorImp constructor comment.
	 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
	 */
	public SchemaNameChangeOperatorImp(XATNode node) {
		super(node);
		leaf = false;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/11/2003 4:50:24 PM)
	 * @return java.lang.String
	 */
	public String getStreamName() {
		return newStreamName;
	}
	
	public String[] getColNames() {
		return newColNames;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/11/2003 4:45:25 PM)
	 * @return boolean
	 */
	public boolean isLeaf() {
		return leaf;
	}
	/**
	 * Runs the actual operator.
	 *
	 * Unlike in Rainbow, the Stream operators have the implementation tied to the operator
	 * because the StreamOperators represent physical operators instead of Logic.
	 * Creation date: (1/6/2003 5:15:23 PM)
	 */
	public boolean run() {
	
		if (!hasInit()) {
			init();
			if (!hasInit()) {
				return false;
			}
		}
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		double selectivity =
			Double.parseDouble(this.getProperty("REAL_SELECTIVITY"));
		double delay = Double.parseDouble(this.getProperty("RESULT_DELAY"));
		
		/** get the tuples */

		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);
		/** figure out how many tuples to keep*/
		double dHowManyToKeep = selectivity * tuples.length;
		int howManyToKeep = (new Double(dHowManyToKeep)).intValue();


		XATTuple[] kept = new XATTuple[howManyToKeep];
		
		for(int i = 0; i < howManyToKeep ; i++){
			kept[i] = tuples[i];

		}
      
		/** now enqueue the tuples */
		try {
		  int time = new Double(delay * 1000).intValue();
		  Thread.sleep(time);
		} catch (InterruptedException e) {
		//sleeping was interrupted, so just continue
		 }   
		
		this.getOutputQueue().enqueue(kept, this);
		return true;
	}
	/**
	 * The Dummy operator has 2 parameters.  The SELECTIVITY
	 * is a value between 0 and 1 and RESULT_DELAY is the
	 * time it takes (in milleseconds) for this operator to place a tuple in its output queue
	 * 
	 */
	protected void setDefaultProperties() {
		super.setDefaultProperties();
		addProperty("RESULT_DELAY", "0");
		addProperty("REAL_SELECTIVITY", ".5");

	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/11/2003 4:45:25 PM)
	 * @param newLeaf boolean
	 */
	public void setLeaf(boolean newLeaf) {
		leaf = newLeaf;
	}
	/**
	* sets the given property to the new value.  This method is responsible
	* for validating the given property.  An exception is thrown if
	* the value is not correct.
	* Validation should be done in the following manner: 
	* First the method should check if the object supports the given property by calling
	* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
	* be thrown. <p>
	*
	* Once it is known that the object supports the property, the value must be validated.  Each implementor of
	* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
	* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
	* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
	* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
	* will reduce duplicate code. <p>
	*
	* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
	* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
	* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
	* it is up to the subclass that created the property to validate it (as described above).
	* @param name the name of the property to set
	* @param value the new value of the property
	* @throws Exception if the given property cannot be found or
	* if the value is not valid for the given property.
	*/
	public void setProperty(String name, String value)
		throws java.lang.Exception {
		//first see if the DummyOperator supports the property name	
		if (isValidPropertyName(name)) {
			if (name.equals("REAL_SELECTIVITY")) {
				try {
					double selectivity = Double.parseDouble(value);
					if ((selectivity >= 0) && (selectivity <= 1)) {
						setNewPropertyValue(name, value);
					} else {
						//System.out.println("Trying to set REAL_SELECTIVITY to: " + selectivity);
						throw new NumberFormatException(
							"The property named "
								+ name
								+ " requires a double value between 0 and 1 (inclusive)");
					}
				} catch (NumberFormatException e) {
					throw new NumberFormatException(
						"The property named "
							+ name
							+ " requires a double value between 0 and 1 (inclusive)");
				}
			} else {
				//the property name is valid, but its value is validated in the super class
				super.setProperty(name, value);
			}
		} else
			throw new Exception(
				"The DummyOperator does not support the property named "
					+ name);

	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/11/2003 4:50:24 PM)
	 * @param newStreamName java.lang.String
	 */
	public void setStreamName(String newStreamName) {
		this.newStreamName = newStreamName;
	}
	
	public void setColNames(String[] newColNames) {
		this.newColNames = newColNames;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/11/2003 4:45:54 PM)
	 * @return java.lang.String
	 */
	public String toString() {
		String s = super.toString();
		if (isLeaf()) {
			s = s + " LEAF" + " Stream: " + getStreamName();
		}
		return s;
	}

	public String toToolTipString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for (index = sName.length() - 1; index >= 0; index--)
			if (sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if (id == "-1") {
			rs = rs + hashCode();
		} else {
			rs = rs + id;
		}
		if (this.isActive())
			rs += " ACTIVE ";
		else
			rs += " INACTIVE ";

		/** now add the join predicate and window size info */
		rs += "\n";

		return rs;

	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
	 */
	public void init() {
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength()==0) {
			return;
		}
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength()==0) {
			outSchema = new XATQueueSchema();
			outSchema.setName(newStreamName);
			for (int i=0; i<newColNames.length; i++) {
				outSchema.addColumn(newColNames[i]);
			}
			this.getOutputQueue().setSchema(outSchema);
		}
		this.computeOutputPositions();
		this.setHasInit(true);
	}
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}


	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}



	public boolean matchStates(XATStreamOperator op) {
		return false;
	}


	public boolean recomputeUnmatchedState() {
		return true;
	}
	


	public int setStateID(int nextid) {
		return nextid;
	}

	public boolean finalizeOperation(){
		return true;
	}

		
		
	}

