/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Apr 12, 2004
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Random;
import java.util.TreeMap;
import java.util.Vector;
import java.util.zip.CRC32;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;

/* TreeMap - containsKey, get, put - O(log(n))
 * LinkedList - removeFirst, removeLast, getFirst, getLast - O(1) 
 */

/**
 * Wrapper for storing a list of XATTuples.  Internally this class stores all
 * the tuples in a LinkedList, and builds an index on the tuples.  The index is
 * just a TreeMap whose key is the crc32 of the concatenated fields of the tuple
 * and value whose is an ArrayList that holds the indexes of all tuples whose
 * crc32 values where the same although this most likely will only occur with 
 * duplicate tuples.
 * @author Steven Torance, Chris Vitone
 * @version 1.0
 */
public class TuplesList implements Serializable{
	private TreeMap tuplesHashIndex = new TreeMap();
	private LinkedList tuples = new LinkedList();
	private XATQueueSchema schema;
	private String ignorefield;
	
	/**
	 * Constructor that takes in the Queue's schema, and field that is to be ignored.
	 * @param schema The queue's schema.
	 * @param Field to ignore.
	 * @author Steve
	 */
	public TuplesList(XATQueueSchema schema, String ignorefield) {
		this.schema = schema;
		this.ignorefield = ignorefield;
	}
	
	/**
	 * Compares the fields of two tables exluding the ignorefield.
	 * @param t1 First tuple to compare.
	 * @param t2 Second tuple to compare.
	 * @return True if the tuples' fields are the same, false otherwise.
	 * @author Steve
	 */
	private boolean compareFields(XATTuple t1, XATTuple t2) {
		boolean found = true;
		XATDataValue[] data1 = t1.getArrayData();
		XATDataValue[] data2 = t2.getArrayData();
		
		// iterate through each field and compare them unless its one of
		// the ignored fields.
		for (int j = 0; j < data1.length && found; j++)
		{
			// check if the field is an ignored field, and if it is
			// continue iterating through the rest of the fields
			if (schema.getColNames().get(j).equals(ignorefield))
				continue;
			// compare the fields
			if (!data1[j].getValue().equals(data2[j].getValue()))
				found = false;
		}	
		
		return found;
	}
	
	/**
	 * Concatenates all the fields of the given tuple.
	 * The crc32 of the returned string which is used
	 * as a hash value for inserting into the TreeMap.
	 * @param t XATTuple whose fields to concatenate.
	 * @return The String of all concatenates fields.
	 * @author Chris
	 */
	private String concatTupleFields(XATTuple t) {
		StringBuffer concatString = new StringBuffer();
		
		// concatenate all fields except the ignored one
		XATDataValue[] data = t.getArrayData();
		for (int j = 0; j < data.length-1; j++)
		{
			// if its the ignoredfield, move along!
			if (schema.getColNames().get(j).equals(ignorefield))
				continue;
			concatString.append(data[j]);
		}
		
		return concatString.toString();
	}
	
	/**
	 * Adds the given tuple to the end of the list.
	 * This function at worst case is O(logn)
	 * @param t XATTuple to add.
	 * @author Steve
	 */
	public void addToEnd(XATTuple t) {
		addToEnd(t, false);
	}
	
	/**
	 * Adds the given tuple to the end of the list.
	 * This function at worst case is O(logn)
	 * @param t XATTuple to add.
	 * @param unique Whether the tuple should be inserted only if it's unique.
	 * @author Steve
	 */
	public void addToEnd(XATTuple t, boolean unique) {
		
		// Concatenate the fields and calculate the CRC32 of the
		// the concatenated fields.
		CRC32 crc32 = new CRC32();
		String hashString = concatTupleFields(t);
		
		crc32.update(hashString.getBytes());
		Long crc32val = new Long(crc32.getValue());
		
		// Add it to the TreeMap index.
		// Look for the crc32 value in the TreeMap, and get the list of tuples.
		LinkedList hashedTuples = (LinkedList) tuplesHashIndex.get(crc32val);
		// The CRC32 value already exists
		if (hashedTuples != null) {
			// if we only want unique values, check all tuples in the list
			// to make sure its just a CRC32 clash and not a duplicate entry.
			if (unique) {
				ListIterator li = hashedTuples.listIterator();
				while (li.hasNext()) {
					XATTuple indexTuple = (XATTuple)li.next();
					if (compareFields(t, indexTuple))
						return;
				}
			}
			hashedTuples.add(t);
		} else {
			// The CRC32 value does not exist, so just add it.
			hashedTuples = new LinkedList();
			hashedTuples.add(t);
			tuplesHashIndex.put(crc32val, hashedTuples);
		}
		// Now add it to the linked list of tuples.
		tuples.addLast(t);
	}
		
	/**
	 * Deletes count tuples from the beginning of the list.
	 * At worst case this functions is O(nlogn)
	 * @param count Number of tuples to delete.
	 * @author Steve
	 */
	public void deleteStartTo(int count) {
		// If there is nothing to delete, just return
		if (tuples.size() <= 0)
			return;
		
		// Makes sure we're not trying to delete more tuples than we have.
		count = Math.min(count, tuples.size());
		// Iterate through count times.
		for (int i = 0; i < count; i++) {
			// Get the tuple at the beginning (the one we're deleting),
			// concatenate its fields, then calculate the CRC32 of the 
			// concatenated fields.
			XATTuple t = (XATTuple) tuples.getFirst();
			
			// Calculate the CRC32 Hash
			CRC32 crc32 = new CRC32();
			String hashString = concatTupleFields(t);
			
			crc32.update(hashString.getBytes());
			Long crc32val = new Long(crc32.getValue());
			
			// Now find the tuples by using its crc32 value in the TreeMap index.
			LinkedList hashedTuples = (LinkedList)tuplesHashIndex.get(crc32val);
			// If it does not exist in the TreeMap then something is very wrong.
			if (hashedTuples == null) {
				//something is very wrong
				System.err.print("Tuples being deleted with no entry in Hash Index");
			}
			// If the TreeMap has only one entry in the list of tuples,
			// just delete the entry from the TreeMap.
			else if (hashedTuples.size() <= 1) {
				tuplesHashIndex.remove(crc32val);
			// Otherwise remove the first entry in the list of tuples.
			} else {
				hashedTuples.removeFirst();
			}
			
			// Now finally remove it from the LinkedList
			tuples.removeFirst();
		}
	}
	
	/**
	 * Deletes all tuples from the start of the list whose timestamp is less than
	 * the given timestamp.
	 * @param timestamp Maximum timstamp of tuples to delete from list.
	 * @throws XATTimeStampException
	 * @author Steve
	 */
	public void deleteStartToTimestamp(long timestamp) 
		throws XATTimeStampException {
		// If there is nothing to delete, just return
		if (tuples.size() <= 0)
			return;
		
		for (int i = 0; i < tuples.size() ; i++) {
			
			// Get the tuple at the beginning (the one we're deleting),
			// concatenate its fields, then calculate the CRC32 of the 
			// concatenated fields.
			XATTuple t = (XATTuple) tuples.getFirst();
			
			// Before we waste anytime calculating the CRC32 of the concatenated
			// fields, determine based on the timestamp if the tuple should be
			// deleted.  If the tuple should not be deleted then break out of the 
			// loop since subsequent tuples will also not need to be deleted.
			if (t.getMinTimeStamp() >= timestamp)
				break;
			
			// Calculate the CRC32 Hash
			CRC32 crc32 = new CRC32();
			String hashString = concatTupleFields(t);
			
			crc32.update(hashString.getBytes());
			Long crc32val = new Long(crc32.getValue());
			
			// Now find the tuples by using its crc32 value in the TreeMap index.
			LinkedList hashedTuples = (LinkedList)tuplesHashIndex.get(crc32val);
			// If it does not exist in the TreeMap then something is very wrong.
			if (hashedTuples == null) {
				//something is very wrong
				System.err.print("Tuples being deleted with no entry in Hash Index");
			}
			// If the TreeMap has only one entry in the list of tuples,
			// just delete the entry from the TreeMap.
			else if (hashedTuples.size() <= 1) {
				tuplesHashIndex.remove(crc32val);
			// Otherwise remove the first entry in the list of tuples.
			} else {
				hashedTuples.removeFirst();
			}
			
			// Now finally remove it from the LinkedList
			tuples.removeFirst();
		}
	}
	
	/**
	 * Determines if the given tuple exist in this list.
	 * At worst case this function is O(nlogn) but this is
	 * only if every tuple has the same crc32 which would
	 * be extremely rare.  Therefore, this function usually
	 * is O(logn).
	 * @param t The XATTuple whose existance to determine.
	 * @return True if the tuples exists, false otherwise.
	 * @author Chris
	 */
	public boolean hasTuple(XATTuple t) {
		// Concatenate the fields and calculate the CRC32 of the
		// the concatenated fields.
		CRC32 crc32 = new CRC32();
		String hashString = concatTupleFields(t);
		
		crc32.update(hashString.getBytes());
		Long crc32val = new Long(crc32.getValue());
		
		// Search for the tuple, in the TreeMap.
		LinkedList hashedTuples = (LinkedList)tuplesHashIndex.get(crc32val);
		// The tuple does not exist.
		if (hashedTuples == null) {
			// tuple does not exist
			System.out.println("nothing in the tuplesList");
			return false;
		} else {
			if (hashedTuples.size() == 0) {
				System.err.print("Tuple has no index entry in Hash Index");
			} else {
				// Iterate through each tuple in the list of tuples for the TreeMap
				// entry, and compare the passed in tuple with each stored tuple.
				// If any of the tuples match, return true.
				ListIterator li = hashedTuples.listIterator();
				while(li.hasNext()) {
					XATTuple indexTuple = (XATTuple)li.next();
					if (compareFields(t, indexTuple))
						return true;
				}
				
				return false;
			}
		}
		
		return false;
	}
	
	
	
	/**
	 * Determines if the given tuple exist in this list only if
	 * the tuple's timestamp is greater than or equal to the given
	 *  timestamp.
	 * @param t The XATTuple whose existance to determine.
	 * @param timestamp Minimum timestamp of tuple's to search.
	 * @return True if the tuples exists, false otherwise.
	 * @author Chris
	 */
	public boolean hasTupleGTETimestamp(XATTuple t, long timestamp) 
		throws XATTimeStampException {
		// Concatenate the fields and calculate the CRC32 of the
		// the concatenated fields.
		CRC32 crc32 = new CRC32();
		String hashString = concatTupleFields(t);
		
		crc32.update(hashString.getBytes());
		Long crc32val = new Long(crc32.getValue());
		
		// Search for the tuple, in the TreeMap.
		LinkedList hashedTuples = (LinkedList)tuplesHashIndex.get(crc32val);
		// The tuple does not exist.
		if (hashedTuples == null) {
			// tuple does not exist
			return false;
		} else {
			if (hashedTuples.size() == 0) {
				System.err.print("Tuple has no entry in Hash Index");
			} else {
				// Iterate through each tuple in the list of tuples for the TreeMap
				// entry, and compare the passed in tuple with each stored tuple.
				// If any of the tuples match, return true.
				ListIterator li = hashedTuples.listIterator();
				while(li.hasNext()) {
					XATTuple indexTuple = (XATTuple)li.next();
					// Check to make sure the timestamp if greater than or equal to
					// the passed in timestamp.  If it's not, keep iterating.
					if (indexTuple.getMinTimeStamp() < timestamp)
						continue;
					// otherwise do the tuple comparison
					if (compareFields(t, indexTuple))
						return true;
				}
				
				return false;
			}
		}
		
		return false;
	}
	
	/**
	 * Determines if the given tuple exist in this list only if
	 * the tuple's timestamp is less than or equal to the given
	 *  timestamp.
	 * @param t The XATTuple whose existance to determine.
	 * @param timestamp Maximum timestamp of tuple's to search.
	 * @return True if the tuples exists, false otherwise.
	 * @author Chris
	 */
	public boolean hasTupleLTETimestamp(XATTuple t, long timestamp)
		throws XATTimeStampException {
		// Concatenate the fields and calculate the CRC32 of the
		// the concatenated fields.
		CRC32 crc32 = new CRC32();
		String hashString = concatTupleFields(t);
		
		crc32.update(hashString.getBytes());
		Long crc32val = new Long(crc32.getValue());
		
		// Search for the tuple, in the TreeMap.
		LinkedList hashedTuples = (LinkedList)tuplesHashIndex.get(crc32val);
		// The tuple does not exist.
		if (hashedTuples == null) {
			// tuple does not exist
			return false;
		} else {
			if (hashedTuples.size() == 0) {
				System.err.print("Tuple has no entry in Hash Index");
			} else {
				// Iterate through each tuple in the list of tuples for the TreeMap
				// entry, and compare the passed in tuple with each stored tuple.
				// If any of the tuples match, return true.
				ListIterator li = hashedTuples.listIterator();
				while(li.hasNext()) {
					XATTuple indexTuple = (XATTuple)li.next();
					// Check to make sure the timestamp if less than or equal to
					// the passed in timestamp.  If it's not, stop iterating
					// since all subsequent tuples should also have timestamps
					// greater than the passed in timstamp.
					if (indexTuple.getMinTimeStamp() > timestamp)
						break;
					// otherwise do the tuple comparison
					if (compareFields(t, indexTuple))
						return true;
				}
				
				return false;
			}
		}
		
		return false;
	}
	
	/**
	 * Removes the first occurance of the given tuple.
	 * At worst this functions is O(logn + n).
	 * @param t XATTuple to remove.
	 * @author Chris
	 */
	public void remove(XATTuple t) {
		// Concatenate the fields and calculate the CRC32 of the
		// the concatenated fields.
		CRC32 crc32 = new CRC32();
		String hashString = concatTupleFields(t);
		
		crc32.update(hashString.getBytes());
		Long crc32val = new Long(crc32.getValue());
		
		// Now find the tuples by using its crc32 value in the TreeMap index.
		LinkedList hashedTuples = (LinkedList)tuplesHashIndex.get(crc32val);
		// The tuple doesn't exist, and thefore can not be deleted so just return.
		if (hashedTuples == null) {
			return;
		}
		// If there is only one tuple in the list of tuples, delete the entire entry.
		else if (hashedTuples.size() <= 1) {
			tuplesHashIndex.remove(crc32val);
		} else {
			ListIterator li = hashedTuples.listIterator();
			// Find the first tuple that matches and delete it.
			while(li.hasNext()) {
				XATTuple indexTuple = (XATTuple)li.next();
				if (compareFields(t, indexTuple)) {
					li.remove();
					break;
				}
			}
		}
		
		// Now find the tuple in the LinkedList and delete it.
		ListIterator li = tuples.listIterator();
		while(li.hasNext()) {
			
			XATTuple listTuple = (XATTuple)li.next();
			if (compareFields(t, listTuple)) {
				li.remove();
				return;
			}
		}
	}
	
	/**
	 * Retrieves the total number of stored tuples.
	 * @return The number of stored tuples.
	 * @author Chris
	 */
	public int size() {
		return tuples.size();
	}
	
	/**
	 * Removes all the stored tuples.
	 * @author Chris
	 */
	public void clear() {
		tuples.clear();
		tuplesHashIndex.clear();
	}
	
	/**
	 * Gets a tuple at the given index.
	 * This should not be used unless absolutely necessary due 
	 * to the ineffiency of a LinkedList at random access.
	 * @param index Index of tuple to retrieve.
	 * @return XATTuple at the given index.
	 * @author Steve
	 */
	public XATTuple get(int index) {
		return (XATTuple) tuples.get(index);
	}
	
	/**
	 * Gets a ListIterator starting at the beginning of the the LinkedList.
	 * @return A ListIterator.
	 * @author Steve
	 */
	public ListIterator listIterator() {
		return tuples.listIterator();
	}
	
	/**
	 * Gets a ListIterator starting at start of the the LinkedList.
	 * @param start Index to start the ListIterator at.
	 * @return A ListIterator.
	 * @author Steve
	 */
	public ListIterator listIterator(int start) {
		return tuples.listIterator(start);
	}
	
	/**
	 * Gets a ListIterator starting at the first tuple with a timestamp greater than
	 * or equal to the given timestamp;
	 * @param timestamp Timestamp to start ListIterator at.
 	 * @return A ListIterator.
 	 * @author Chris
	 */
	public ListIterator listIteratorTimestamp(long timestamp) 
		throws XATTimeStampException {
		ListIterator li = tuples.listIterator();
		
		while(li.hasNext()) {
			XATTuple t = (XATTuple) li.next();
			
			if (t.getMinTimeStamp() < timestamp)
				continue;
			else {
				li.previous();
				break;
			}
		}
		
		return li;
	}
	
	/**
	 * Just used for testing.
	 * @author Steve
	 */
	public static void main(String args[]) throws XATTimeStampException {
		
		final int TOTAL_TUPLES = 20000;
		final int SEARCHES = 10000;
		final int RANDOM_STRLEN = 100;
		final int FIELDS = 4;
		final boolean RANDOM_FIELDS = false;
		Random ran = new Random();
		XATQueueSchema schema = new XATQueueSchema();
		Vector cols = new Vector();
		for (int k = 0; k < FIELDS; k++)
			cols.add("field" + k);
		schema.setColNames(cols);
		
		TuplesList tl = new TuplesList(schema, "");
		Calendar start, end;
		long totaltime = 0;

		ArrayList al = new ArrayList();
		// Generate a bunch of tuples
		for (int i = 0; i < TOTAL_TUPLES; i++) {		
			XATTupleImp t = new XATTupleImp();
			if (RANDOM_FIELDS) {
				/* Random fields*/
				for (int j = 0; j < FIELDS; j++) {
					byte[] randstr = new byte[100];
					ran.nextBytes(randstr);
					t.addCell(new XATDataValueImp(new String(randstr) + i));
				}
			} else {
				/* Non Random fields*/
				for (int j = 0; j < FIELDS; j++) {
					t.addCell(new XATDataValueImp("field" + j + ":" + i));
				}
			}
			long[] newtimes = new long[t.getLength()];
			for(int x=0; x<t.getLength(); x++) {
				newtimes[x] = 100000 + (1000 * (i / 10)) + (i % 10);
			}
			t.setTimeStamp(newtimes);
			
//			XATTimeStampImp ts = new XATTimeStampImp();
//			try {
//				ts.setTimeStamp(100000 + (1000 * (i / 10)) + (i % 10), t);
//			} catch (XATTimeStampException e) {
//				e.printStackTrace();
//				return;
//			}
//			t.setTimeStamp(ts);
			
			al.add(t);
		}
		// Add those tuples to the TuplesList
		start = Calendar.getInstance();
		ListIterator li = al.listIterator();
		while (li.hasNext()) {
			tl.addToEnd((XATTuple)li.next());
		}
		end = Calendar.getInstance();
		totaltime = end.getTimeInMillis() - start.getTimeInMillis();
		
		System.out.println("Added " + tl.size() + " tuples in " + totaltime + " ms." );
		
		// Now do a bunch of random searches if the tuples weren't randomly generated.
		if (!RANDOM_FIELDS && tl.size() > 0) {
			start = Calendar.getInstance();
			for (int j = 0; j < SEARCHES; j++)
			{
				int index = ran.nextInt(TOTAL_TUPLES);
				
				XATTupleImp t = new XATTupleImp();
				for (int k = 0; k < FIELDS; k++) {
					t.addCell(new XATDataValueImp("field" + k + ":" + index));
				}
				
				if (!tl.hasTuple(t)) {
					System.err.println("Did not find tuple that should exist.");
				}
			}
			end = Calendar.getInstance();
			totaltime = end.getTimeInMillis() - start.getTimeInMillis();
			System.out.println("Finished " + SEARCHES + " searches in " + totaltime + " ms.");
			double avgsearchtime = (double)totaltime / (double)SEARCHES;
			System.out.println("Average Search Time is " + avgsearchtime + " ms.");
		}
				
		// Delete tuples with a timsestamp <= 100000.
		System.out.println("Tuple #3: " + tl.get(2));
		System.out.println("Deleting tuples with timestamp < 101000.");
		start = Calendar.getInstance();
		try {
			tl.deleteStartToTimestamp(101000);
		} catch (XATTimeStampException e) {
			e.printStackTrace();
			return;
		}
		end = Calendar.getInstance();
		totaltime = end.getTimeInMillis() - start.getTimeInMillis();
		System.out.println("New size is " + tl.size());
		System.out.println("Deletes took " + totaltime + " ms.");
		
		// Delete 100 Tuples
		System.out.println("Deleting next 100 elements.");
		start = Calendar.getInstance();
		tl.deleteStartTo(100);
		System.out.println("New size is " + tl.size());
		end = Calendar.getInstance();
		totaltime = end.getTimeInMillis() - start.getTimeInMillis();
		System.out.println("Deletes took " + totaltime + " ms.");
		
		int totaldeleted = 110;
		
		// Now do more searches to test the correctness of the deletes.
		// Once again only do this if the tuples were not randomly generated.
		long timestamp;
		if (!RANDOM_FIELDS && tl.size() > 0) {
			start = Calendar.getInstance();
			for (int j = 0; j < SEARCHES; j++)
			{
				int index = ran.nextInt(tl.size()-totaldeleted) + totaldeleted;
				
				XATTupleImp t = new XATTupleImp();
				for (int k = 0; k < FIELDS; k++) {
					t.addCell(new XATDataValueImp("field" + k + ":" + index));
				}
				
				switch (ran.nextInt(3)) {
				case 0:
				    if (!tl.hasTuple(t)) {
						System.err.println("Did not find tuple that should exist.");
					}
				    break;
				case 1:
				    timestamp = 100000 + (1000 * (index / 10));
					if (!tl.hasTupleGTETimestamp(t, timestamp)) {
						System.err.println("Did not find tuple that should exist with hasTupleGTETimstamp.");
					}
				    break;
				case 2:
				    timestamp = 100000 + (1000 * ((index / 10) + 1));
					if (!tl.hasTupleLTETimestamp(t, timestamp)) {
						System.err.println("Did not find tuple that should exist with hasTupleLTETimstamp.");
					}
				    break;
				}
			}
			end = Calendar.getInstance();
			totaltime = end.getTimeInMillis() - start.getTimeInMillis();
			System.out.println("Finished " + SEARCHES + " searches in " + totaltime + " ms.");
			double avgsearchtime = (double)totaltime / (double)SEARCHES;
			System.out.println("Average Search Time is " + avgsearchtime + " ms.");
		}	
		
		// Now remove just the last tuple.
		System.out.println("Deleting the last tuple.");
		start = Calendar.getInstance();
		XATTupleImp t = new XATTupleImp();
		int index = (tl.size()-1);
		for (int k = 0; k < FIELDS; k++) {
			t.addCell(new XATDataValueImp("field" + k + ":" + index));
		}
		
		tl.remove(t);
		end = Calendar.getInstance();
		totaltime = end.getTimeInMillis() - start.getTimeInMillis();
		System.out.println("New size is " + tl.size());
		System.out.println("Delete took " + totaltime + " ms.");
		
		// Finally delete all the tuples
		System.out.println("Deleting the rest.");
		start = Calendar.getInstance();
		tl.deleteStartTo(tl.size());
		end = Calendar.getInstance();
		totaltime = end.getTimeInMillis() - start.getTimeInMillis();
		System.out.println("New size is " + tl.size());
		System.out.println("Deletes took " + totaltime + " ms.");
		
		System.out.print("Done!");
		
	}
}
