//--------------------------------------------------
// DecisionTreeID3.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh;

import java.io.*;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorInputSpecification;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.translation.DecisionTreeNodeSignature;

/* A simple implementation of the ID3 algorithm */
@SuppressWarnings({ "serial", "rawtypes" })
public class DecisionTreeID3 implements java.io.Serializable {
	// -----------------------------------------------------------------------------

	static int maxNestedLevel = 1;
	static int nestedLevel = 1;
	static String xmlDT = "";
	int numAttributes; // The number of attributes including the output attribute
	// The names of all attributes. It is an array of dimension numAttributes. The last attribute
	// is the output attribute
	String[] attributeNames; 
	private SimpleOperatorInputSpecification input;

	private static String currDTSignature; // contains the signature of the current decision tree

	// -----------------------------------------------------------------------------

	public DecisionTreeID3(SimpleOperatorInputSpecification inputVal) {
		input = inputVal;
	}

	// -----------------------------------------------------------------------------

	/**
	 * Possible values for each attribute is stored in a vector. domains is an
	 * array of dimension numAttributes. Each element of this array is a vector
	 * that contains values for the corresponding attribute domains[0] is a
	 * vector containing the values of the 0-th attribute, etc.. The last
	 * attribute is the output attribute
	 */
	Vector[] domains;

	// -----------------------------------------------------------------------------

	/**
	 * The class to represent a data point consisting of numAttributes values of
	 * attributes
	 */
	class DataPoint {

		/**
		 * The values of all attributes stored in this array. i-th element in
		 * this array is the index to the element in the vector domains
		 * representing the symbolic value of the attribute. For example, if
		 * attributes[2] is 1, then the actual value of the 2-nd attribute is
		 * obtained by domains[2].elementAt(1). This representation makes
		 * comparing values of attributes easier - it involves only integer
		 * comparison and no string comparison. The last attribute is the output
		 * attribute
		 */
		public int[] attributes;

		public DataPoint(int numattributes) {
			attributes = new int[numattributes];
		}
	};

	// -----------------------------------------------------------------------------

	/* The class to represent a node in the decomposition tree */
	class TreeNode {
		public double entropy; // The entropy of data points if this node is a leaf node
		public Vector data; // The set of data points if this is a leaf node
		// If this is not a leaf node, the attribute that is used to divide
		// the set of data points
		public int decompositionAttribute; 
		public int decompositionValue; // the attribute-value that is used to divide the parent node
		public TreeNode[] children; // If this is not a leaf node, references to the children nodes
		public TreeNode parent; // The parent to this node. The root has parent == null

		public TreeNode() {
			data = new Vector();
		}

	};

	// -----------------------------------------------------------------------------

	/* The root of the decomposition tree */
	TreeNode root = new TreeNode();

	// -----------------------------------------------------------------------------

	/*
	 * This function returns an integer corresponding to the symbolic value of
	 * the attribute. If the symbol does not exist in the domain, the symbol is
	 * added to the domain of the attribute
	 */
	@SuppressWarnings("unchecked")
	public int getSymbolValue(int attribute, String symbol) {
		int index = domains[attribute].indexOf(symbol);
		if (index < 0) {
			domains[attribute].addElement(symbol);
			return domains[attribute].size() - 1;
		}
		return index;
	}

	// -----------------------------------------------------------------------------

	/* Returns all the values of the specified attribute in the data set */
	public int[] getAllValues(Vector data, int attribute) {
		Vector<String> values = new Vector<String>();
		int num = data.size();
		for (int i = 0; i < num; i++) {
			DataPoint point = (DataPoint) data.elementAt(i);
			String symbol = (String) domains[attribute]
					.elementAt(point.attributes[attribute]);
			int index = values.indexOf(symbol);
			if (index < 0) {
				values.addElement(symbol);
			}
		}

		int[] array = new int[values.size()];
		for (int i = 0; i < array.length; i++) {
			String symbol = (String) values.elementAt(i);
			array[i] = domains[attribute].indexOf(symbol);
		}
		values = null;
		return array;
	}

	// -----------------------------------------------------------------------------

	/**
	 * Returns a subset of data, in which the value of the specified attribute of
	 * all data points is the specified value
	 */
	public Vector getSubset(Vector data, int attribute, int value) {
		Vector<DataPoint> subset = new Vector<DataPoint>();

		int num = data.size();
		for (int i = 0; i < num; i++) {
			DataPoint point = (DataPoint) data.elementAt(i);
			if (point.attributes[attribute] == value)
				subset.addElement(point);
		}
		return subset;
	}

	// -----------------------------------------------------------------------------

	/**
	 * Calculates the entropy of the set of data points. The entropy is
	 * calculated using the values of the output attribute which is the last
	 * element in the array attributes
	 */
	public double calculateEntropy(Vector data) {
		int numdata = data.size(); // number of tuples in the training tuple set
		if (numdata == 0)
			return 0;

		int attribute = numAttributes - 1;
		int numvalues = domains[attribute].size();

		double sum = 0;
		for (int i = 0; i < numvalues; i++) {
			int count = 0;
			for (int j = 0; j < numdata; j++) {
				DataPoint point = (DataPoint) data.elementAt(j);
				if (point.attributes[attribute] == i)
					count++;
			}
			double probability = 1. * count / numdata;
			if (count > 0)
				sum += -probability * Math.log(probability);
		}
		return sum;
	}

	// -----------------------------------------------------------------------------

	/**
	 * This function checks if the specified attribute is used to decompose the
	 * data set in any of the parents of the specified node in the decomposition
	 * tree. Recursively checks the specified node as well as all parents
	 */
	public boolean alreadyUsedToDecompose(TreeNode node, int attribute) {
		if (node.children != null) {
			if (node.decompositionAttribute == attribute)
				return true;
		}
		if (node.parent == null)
			return false;
		return alreadyUsedToDecompose(node.parent, attribute);
	}

	// -----------------------------------------------------------------------------

	/*
	 * This function decomposes the specified node according to the ID3
	 * algorithm. Recursively divides all children nodes until it is not
	 * possible to divide any further I have changed this code from my earlier
	 * version. I believe that the code in my earlier version prevents useless
	 * decomposition and results in a better decision tree! This is a more
	 * faithful implementation of the standard ID3 algorithm
	 */
	public void decomposeNode(TreeNode node) {
		double bestEntropy = 0;
		boolean selected = false;
		int selectedAttribute = 0;

		int numdata = node.data.size();
		int numinputattributes = numAttributes - 1;

		node.entropy = calculateEntropy(node.data);
		if (node.entropy == 0)
			return;

		/*
		 * In the following two loops, the best attribute is located which
		 * causes maximum decrease in entropy
		 */
		for (int i = 0; i < numinputattributes; i++) {
			int numvalues = domains[i].size();

			if (alreadyUsedToDecompose(node, i))
				continue;
			// Use the following variable to store the entropy for the test node
			// created with the attribute i

			double averageentropy = 0;
			for (int j = 0; j < numvalues; j++) {
				Vector subset = getSubset(node.data, i, j);
				if (subset.size() == 0)
					continue;
				double subentropy = calculateEntropy(subset);
				averageentropy += subentropy * subset.size(); // Weighted sum
			}

			averageentropy = averageentropy / numdata; // Taking the weighted average
			if (selected == false) {
				selected = true;
				bestEntropy = averageentropy;
				selectedAttribute = i;
			} else {
				if (averageentropy < bestEntropy) {
					selected = true;
					bestEntropy = averageentropy;
					selectedAttribute = i;
				}
			}
		}

		if (selected == false)
			return;

		currDTSignature += " " + selectedAttribute + ":" + bestEntropy;

		// Now divide the dataset using the selected attribute
		int numvalues = domains[selectedAttribute].size();
		node.decompositionAttribute = selectedAttribute;
		node.children = new TreeNode[numvalues];
		for (int j = 0; j < numvalues; j++) {
			node.children[j] = new TreeNode();
			node.children[j].parent = node;
			node.children[j].data = getSubset(node.data, selectedAttribute, j);
			node.children[j].decompositionValue = j;
		}

		// Recursively divides children nodes
		for (int j = 0; j < numvalues; j++) {
			decomposeNode(node.children[j]);
		}

		// There is no more any need to keep the original vector.
		// Release this memory
		node.data = null; // Let the garbage collector recover this memory
	}

	// -----------------------------------------------------------------------------

	/**
	 * Function to read the data file. The first line of the data file should
	 * contain the names of all attributes. The number of attributes is inferred
	 * from the number of words in this line. The last word is taken as the name
	 * of the output attribute. Each subsequent line contains the values of
	 * attributes for a data point. If any line starts with // it is taken as a
	 * comment and ignored. Blank lines are also ignored.
	 */
	@SuppressWarnings("unchecked")
	public int readData(String filename) throws Exception {
		FileInputStream in = null;

		try {
			File inputFile = new File(filename);
			in = new FileInputStream(inputFile);
		} catch (Exception e) {
			System.err.println("Unable to open data file: " + filename + "\n"
					+ e);
			return 0;
		}

		BufferedReader bin = new BufferedReader(new InputStreamReader(in));
		String input;
		input = bin.readLine();
		if (input == null) {
			System.err.println("No data found in the data file: " + filename
					+ "\n");
			return 0;
		}

		String[] tokens = input.split("  ");
		numAttributes = tokens.length;
		if (numAttributes <= 1) {
			System.err.println("Read line: " + input);
			System.err
					.println("Could not obtain the names of attributes in the line");
			System.err
					.println("Expecting at least one input attribute and one output attribute");
			return 0;
		}

		// domains initialization
		domains = new Vector[numAttributes];
		for (int i = 0; i < numAttributes; i++) {
			domains[i] = new Vector();
		}

		attributeNames = new String[numAttributes];

		for (int i = 0; i < numAttributes; i++) {
			attributeNames[i] = tokens[i];
		}

		while (true) {
			input = bin.readLine();
			if (input == null)
				break;
			if (input.startsWith("//"))
				continue;
			if (input.equals(""))
				continue;

			// tokenizer = new StringTokenizer(input);
			String[] dataTokens = input.split("  ");
			// int numtokens = tokenizer.countTokens();
			int numtokens = dataTokens.length;

			if (numtokens != numAttributes) {
				System.err.println("Read " + root.data.size() + " data");
				System.err.println("Last line read: " + input);
				System.err
						.println("Expecting " + numAttributes + " attributes");
				return 0;
			}

			DataPoint point = new DataPoint(numAttributes);
			for (int i = 0; i < numAttributes; i++) {
				point.attributes[i] = getSymbolValue(i, dataTokens[i]);
			}
			root.data.addElement(point);
		}

		bin.close();
		return 1;
	} // End of function readData

	// -----------------------------------------------------------------------

	/*
	 * This function prints the decision tree in the form of rules. The action
	 * part of the rule is of the form outputAttribute = "symbolicValue" or
	 * outputAttribute = { "Value1", "Value2", .. } The second form is printed
	 * if the node cannot be decomposed any further into an homogenous set
	 */
	public void printTree(TreeNode node, String tab) {
		int outputattr = numAttributes - 1;
		if (node.children == null) {
			int[] values = getAllValues(node.data, outputattr);
			if (values.length == 1) {
				System.out.println(tab + "\t" + attributeNames[outputattr]
						+ " = \"" + domains[outputattr].elementAt(values[0])
						+ "\";");
				return;
			}
			System.out.print(tab + "\t" + attributeNames[outputattr] + " = {");
			for (int i = 0; i < values.length; i++) {
				System.out.print("\""
						+ domains[outputattr].elementAt(values[i]) + "\" ");
				if (i != values.length - 1)
					System.out.print(" , ");
			}
			System.out.println(" };");
			return;
		}

		int numvalues = node.children.length;
		for (int i = 0; i < numvalues; i++) {
			System.out.println(tab + "if( "
					+ attributeNames[node.decompositionAttribute] + " == \""
					+ domains[node.decompositionAttribute].elementAt(i)
					+ "\") {");
			printTree(node.children[i], tab + "\t");

			if (i != numvalues - 1)
				System.out.print(tab + "} else ");
			else
				System.out.println(tab + "}");
		}
	}

	// -----------------------------------------------------------------------------

	public String getTreeStr(TreeNode node, String tab) {
		String rc = "";
		int outputattr = numAttributes - 1;
		if (node.children == null) {
			int[] values = getAllValues(node.data, outputattr);
			if (values.length == 1) {
				rc += tab + "\t" + attributeNames[outputattr] + " = \""
						+ domains[outputattr].elementAt(values[0]) + "\";";
				return rc;
			}
			rc += tab + "\t" + attributeNames[outputattr] + " = {";
			for (int i = 0; i < values.length; i++) {
				rc += "\"" + domains[outputattr].elementAt(values[i]) + "\" ";
				if (i != values.length - 1)
					rc += " , ";
			}
			rc += " };";
			return rc;
		}

		int numvalues = node.children.length;
		for (int i = 0; i < numvalues; i++) {
			rc += tab + "if( " + attributeNames[node.decompositionAttribute]
					+ " == \""
					+ domains[node.decompositionAttribute].elementAt(i)
					+ "\") {";
			
			// recursive call
			rc += getTreeStr(node.children[i], tab + "\t");

			if (i != numvalues - 1)
				rc += tab + "} else ";
			else
				rc += tab + "}";
		}
		return rc;
	}

	// -----------------------------------------------------------------------------

	public static int NodeID = 0;

	// -----------------------------------------------------------------------------

	@SuppressWarnings("unchecked")
	public ArrayList prepareNodeSignatures(String parentNodeId, TreeNode node,
			ArrayList allNodes) {
		ArrayList childrenIDs = new ArrayList();
		String str = "";
		int outputattr = numAttributes - 1;
		if (node.children == null) {
			int[] values = getAllValues(node.data, outputattr);
			if (values.length == 1) {
				// ------------------------------------------------------------------------------
				// SINGLE LEAF NODE
				// ------------------------------------------------------------------------------
				String nodeID = "" + NodeID;
				String nodeType = "leaf";
				// ruster name
				// ruster name will be used for probing for its route when
				// generating the query plan
				String nodeValue = ""
						+ domains[outputattr].elementAt(values[0]);
				
				String streamId = input.getId();

				DecisionTreeNodeSignature nodeLeaf = new DecisionTreeNodeSignature(
						parentNodeId, nodeID, nodeType, streamId, "",// attrIdx,
						"",// attrName,
						"",// isBucketized,
						"",// nodeOperation,
						nodeValue,// nodeOperationVal);
						null);
				NodeID++; // increment node id for upcoming nodes
				// str += "<ValueNode value=\"" + attributeNames[outputattr] +
				// "='" + domains[outputattr].elementAt(values[0]) +
				// "'\"</ValueNode>" + NEW_LINE;
				str = "<ValueNode value=\""
						+ domains[outputattr].elementAt(values[0])
						+ "\"></ValueNode>";

				allNodes.add(nodeLeaf); // add node to the collection
				str = nodeID;
				childrenIDs.add(str);
				return childrenIDs;
			}

			assert false; // WANT TO CATCH THIS CASE

			// ------------------------------------------------------------------------------
			// MULTIPLE LEAF NODES
			// ------------------------------------------------------------------------------

			str = "<ValueNodeList>\n<ValueNode value=\""
					+ attributeNames[outputattr] + "={";
			for (int i = 0; i < values.length; i++) {
				str = "'" + domains[outputattr].elementAt(values[i]) + "' ";

				String nodeID = "" + NodeID;
				String nodeType = "leaf";
				String nodeValue = ""
						+ domains[outputattr].elementAt(values[i]);// ruster
																	// name
				// ruster name will be used for probing for its route when
				// generating the query plan

				String streamId = input.getId();

				DecisionTreeNodeSignature nodeLeaf = new DecisionTreeNodeSignature(
						parentNodeId, nodeID, nodeType, streamId, "",// attrIdx,
						"",// attrName,
						"",// isBucketized,
						"",// nodeOperation,
						nodeValue,// nodeOperationVal);
						null);
				NodeID++; // increment node id for upcoming nodes
				// str += "<ValueNode value=\"" + attributeNames[outputattr] +
				// "='" + domains[outputattr].elementAt(values[0]) +
				// "'\"</ValueNode>" + NEW_LINE;

				allNodes.add(nodeLeaf); // add node to the collection
				str = nodeID;
				childrenIDs.add(str);
			}// done iterating through multiple children

			return childrenIDs;
		}

		ArrayList allChildren = new ArrayList();
		int numvalues = node.children.length;
		for (int i = 0; i < numvalues; i++) {
			// ------------------------------------------------------------------------------
			// INTERNAL NODE
			// ------------------------------------------------------------------------------
			// retrieve attribute id from the schema mapping
			String attr_idx = (String) input.getSchemaMapping().get(
					attributeNames[node.decompositionAttribute]);

			String streamId = input.getId();
			String attrIdx = "" + attr_idx;
			String attrName = "" + attributeNames[node.decompositionAttribute];
			String isBucketized = "false";
			String nodeOperation = "EQ";
			String nodeOperationVal = ""
					+ domains[node.decompositionAttribute].elementAt(i);
			String nodeID = "" + NodeID;
			String nodeType = "internal";

//			String parenID = "";

			childrenIDs.add(nodeID);
			NodeID++; // increment node id for upcoming nodes

			// ------------------------------------------------------------------------------
			// RECURSIVELY CALL TO CHILDREN
			// ------------------------------------------------------------------------------
			allChildren = prepareNodeSignatures(nodeID, node.children[i],
					allNodes);

			DecisionTreeNodeSignature nodeInternal = new DecisionTreeNodeSignature(
					parentNodeId, nodeID, nodeType, streamId, attrIdx,
					attrName, isBucketized, nodeOperation, nodeOperationVal,
					allChildren);
			allNodes.add(nodeInternal); // add node to the collection
		}
		return childrenIDs;
	}

	// -----------------------------------------------------------------------------

	public String printXmlTree(TreeNode node, String tab) {
		String str = "";
		String NEW_LINE = System.getProperty("line.separator");

		int outputattr = numAttributes - 1;
		if (node.children == null) {
			int[] values = getAllValues(node.data, outputattr);
			if (values.length == 1) {
				// str += "<ValueNode value=\"" + attributeNames[outputattr] +
				// "='" + domains[outputattr].elementAt(values[0]) +
				// "'\"</ValueNode>" + NEW_LINE;
				str += "<ValueNode value=\""
						+ domains[outputattr].elementAt(values[0])
						+ "\"></ValueNode>" + NEW_LINE;
				return str;
			}
			str += "<ValueNodeList>\n<ValueNode value=\""
					+ attributeNames[outputattr] + "={";
			for (int i = 0; i < values.length; i++) {
				str += "'" + domains[outputattr].elementAt(values[i]) + "' ";
				if (i != values.length - 1) {
					str += " , ";
				}
			}
			str += "}\" ></ValueNodeList>" + NEW_LINE;
			return str;
		} else {
			maxNestedLevel++;
		}

		int numvalues = node.children.length;
		for (int i = 0; i < numvalues; i++) {
			// retrieve attribute id from the schema mapping
			String attr_idx = (String) input.getSchemaMapping().get(
					attributeNames[node.decompositionAttribute]);
			str += "<IfNode> " + NEW_LINE;
			str += "<LHS type=\"attr_idx\" value=\"" + attr_idx
					+ "\" attr_name_value=\""
					+ attributeNames[node.decompositionAttribute] + "\"></LHS>"
					+ NEW_LINE;
			str += "<RHS type=\"data\" value=\""
					+ domains[node.decompositionAttribute].elementAt(i)
					+ "\"></RHS>" + NEW_LINE;
			str += "<Operation value=\"" + "EQ" + "\"></Operation>" + NEW_LINE;
			nestedLevel++;
			str += printXmlTree(node.children[i], "" + nestedLevel);

			if (i != numvalues - 1) {
				str += "</IfNode>" + NEW_LINE + "<ElseNode/>" + NEW_LINE;
			} else {
				str += "</IfNode>" + NEW_LINE;
			}
		}
		return str;
	}

	// -----------------------------------------------------------------------------

	public double getDTCost(String statisticsFileName) {
		String tabs = "\t";

		double cost = 0.0; // + maxNestedLevel;
		int count = 0; // Line number of count

		// Stream to read file
		try {
			FileReader input = new FileReader(statisticsFileName);
			/*
			 * Filter FileReader through a Buffered read to read a line at a time
			 */
			BufferedReader bufRead = new BufferedReader(input);
			String line; // String that holds current file line

			// Read first line
			line = bufRead.readLine();
			count++;

			// Read through file one line at time. Traverse the DT
			while (line != null) {
				// System.out.println(count+": "+line);
				line = bufRead.readLine();

				if (line == null)
					continue;
				if (line.startsWith("//"))
					continue;

				String[] tupleTokens = line.split("  ");
				int traveralLevel = 0;
				double currCost = getCostByTraversal(tupleTokens, root,
						traveralLevel);

				cost += currCost;
				count++;
			}
			// Close our input stream
			bufRead.close();
		}
		// Catches any error conditions
		catch (IOException e) {
			System.err.println("!!! Unable to read from file ["
					+ statisticsFileName + "]!!!");
		}

		System.out.println(tabs + "Decision tree absolute cost: " + cost);
		// normalize the cost
		cost = cost / count;
		System.out.println(tabs + "Decision tree normalized cost: " + cost);
		return cost;
	}

	// -----------------------------------------------------------------------------

	public double getCostByTraversal(String[] tupleTokens, TreeNode node,
			int level) {
		double traversalCost = 1.0; // default cost

		int outputattr = numAttributes - 1;
		if (node.children == null) {
			int[] values = getAllValues(node.data, outputattr);
			if (values.length == 1) {
				// check that the value is correct
				assert tupleTokens[outputattr]
						.equalsIgnoreCase(domains[outputattr].elementAt(
								values[0]).toString());

				traversalCost++;
				// System.out.println("\t" + attributeNames[outputattr] +
				// " = \"" + domains[outputattr].elementAt(values[0]) + "\";");
				return traversalCost;
			}
			// System.out.print("\t" + attributeNames[outputattr] + " = {");
			for (int i = 0; i < values.length; i++) {
				// System.out.print("\"" +
				// domains[outputattr].elementAt(values[i]) + "\" ");
				if (i != values.length - 1) {
					// System.out.print(" , ");
				}
			}
			// System.out.println(" };");
			return traversalCost;
		}

		int numvalues = node.children.length;
		for (int i = 0; i < numvalues; i++) {
			try {
				if (tupleTokens.length == 0) {
					traversalCost += 0;
				} else {
					// check if the tuple matches the decomposing attribute
					// value
					if (tupleTokens[node.decompositionAttribute]
							.equalsIgnoreCase(domains[node.decompositionAttribute]
									.elementAt(i).toString())) {
						// increment the cost
						traversalCost++;

						// traverse deeper
						traversalCost += getCostByTraversal(tupleTokens,
								node.children[i], level++);
					}
				}
			} catch (Exception ex) {
				String message = ex.getMessage();
				System.err.println(message);
			}
		}
		return traversalCost;
	}

	// -----------------------------------------------------------------------------
	// This function creates the decision tree and prints it in the form of
	// rules on the console

	public void createDecisionTree() {
//		String tabs = "\t";
		decomposeNode(root);
//		System.out.println("");
//		System.out.println(tabs +
//		"--------------------DECISION TREE------------------------");
//		printTree(root, "");
//		System.out.println(tabs +
//		"--------------------END DECISION TREE--------------------");
	}

	// -----------------------------------------------------------------------------

	public void createDecisionTreeSignature() {
//		String tabs = "\t";
		decomposeNode(root);
//		 System.out.println("");
//		 System.out.println(tabs +
//		 "--------------------DECISION TREE------------------------");
//		 printTree(root, "");
//		 System.out.println(tabs +
//		 "--------------------END DECISION TREE--------------------");
	}

	// -----------------------------------------------------------------------------

	public TreeNode GetRoot() {
		return root;
	}
	

	// -----------------------------------------------------------------------------

	/* Here is the definition of the main function */
	public static void main(String[] args) throws Exception {
		int num = args.length;
		if (num != 2) {
			System.out.println("USAGE:");
			System.out
					.println("\tYou need to specify the name of the training tuples set datafile at the command line ");
			System.out
					.println("\tYou also need to specify the output file name where decision tree will be output");
			return;
		}

		// rimma: below is a hack for manual testing - need to fix it later on
		DecisionTreeID3 me = new DecisionTreeID3(null);
//		long startTime = System.currentTimeMillis(); // To print the time taken to process the data
		// int status =
		// me.readData("C:\\Rimma\\Research\\QueryMesh\\Code\\test-data1.txt");
		// int status =
		// me.readData("C:\\Rimma\\Research\\QueryMesh\\sample_dumps\\SAMPLE_DUMP.txt");

		String inputfileName = args[0];
		String outputfileName = args[1];

		int status = me.readData(inputfileName);

		if (status <= 0)
			return;

		me.createDecisionTree();

//		long endTime = System.currentTimeMillis();
//		long totalTime = (endTime - startTime) / 1000;
//		System.out.println(Decision tree build time + " Seconds");

		WriteToDTDumpFile(outputfileName, "<DecisionTree>\n" + xmlDT
				+ "</DecisionTree>\n");
		System.in.read();
	}

	/* End of the main function */

	// -----------------------------------------------------------------------------

	/* Here is the definition of the main function */
	public static DecisionTreeID3 optimizerCreateDT(
			SimpleOperatorInputSpecification input, String inputFileName,
			String outputFileName) {
		DecisionTreeID3 me = new DecisionTreeID3(input);
//		long startTime = System.currentTimeMillis(); // To print the time taken to process the data
		// int status =
		// me.readData("C:\\Rimma\\Research\\QueryMesh\\Code\\test-data1.txt");
		// int status =
		// me.readData("C:\\Rimma\\Research\\QueryMesh\\sample_dumps\\SAMPLE_DUMP.txt");

		int status;
		try {
			status = me.readData(inputFileName);
			if (status <= 0)
				return null;
			me.createDecisionTree();
//			long endTime = System.currentTimeMillis();
//			long totalTime = (endTime - startTime) / 1000;
			// System.out.println("Decision tree construction time: " +
			// totalTime + " Seconds");

			WriteToDTDumpFile(outputFileName, "<DecisionTree>\n" + xmlDT
					+ "\n</DecisionTree>\n");
		} catch (Exception e) {
			e.printStackTrace();
		}
		return me;
	}
	

	/* End of the main function */

	// -----------------------------------------------------------------------------

	public static String createDTSplitCriteriaSignature(String inputFileName) {
		currDTSignature = "";

		DecisionTreeID3 me = new DecisionTreeID3(null);

		int status;
		try {
			status = me.readData(inputFileName);

			if (status <= 0)
				return null;

			me.createDecisionTreeSignature();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return currDTSignature;
	}
	

	// -----------------------------------------------------------------------------

	public static void WriteToDTDumpFile(String fileLocation,
			String stringToWrite) {
		try {
			// Create file
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			out.newLine();
			// Close the output stream
			out.close();
		} catch (Exception e) {
			// Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}

	// -----------------------------------------------------------------------------
}
