//--------------------------------------------------
// STQM_Analyzer.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2008
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.multiquery;

import java.io.File;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.xml.sax.SAXException;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Coin;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.QueryMeshOptimizer.QM_SEARCH_START_SOLUTION;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.QueryMeshOptimizer.QM_SEARCH_STRATEGY;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.translation.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.*;


public class MQM_Analyzer 
{
    //-----------------------------------------------------------------------------	
    //                         PHASE I: CONCEPT DRIFT DETECTION
    //-----------------------------------------------------------------------------	
   
	public enum STQM_VCD_METHOD 
    {
    	MISCLASSIFICATION_RATE,
    	SIGNATURE_BASED
    }
    
    //-----------------------------------------------------------------------------
    
    public enum STQM_RCD_METHOD 
    {
    	NUMBER_OF_AFFECTED_ROUTES,
    	ROUTE_EDIT_DISTANCE,
    	DISSIMILAR_SUBSEQUENCE
    }
    
    //-----------------------------------------------------------------------------
	
    //constructor
	public MQM_Analyzer()
	{
		System.out.println("Constructor call to STQM Analyzer");
	}
	
    //-----------------------------------------------------------------------------
	
	public void DoQMAnalysis(Hashtable hashSamplesPerStream, 
							 Hashtable hashSamplesPerOperator,
							 Hashtable hashSTQMRouteSignaturesGlobalStorage,
							 OnlineClassifierOperatorImp OnlineClassifierRef,
							 STQM_VCD_METHOD vcdMethod,
		 					 STQM_RCD_METHOD rcdMethod,
		 					 Hashtable hashVirtualConceptDriftResults,
		 					 Hashtable hashRealConceptDriftResults)
	{		 

		//------------------------------------------------------------------
  	    // 1. LOOP THROUGH EACH STREAM AND CHECK IF VIRTUAL CONCEPT DRIFT HAS OCCURRED
		//------------------------------------------------------------------	
		Helper.WriteToFile(Helper.LOG_FILE, "\t"+Helper.GetCurrentTimeString() + ": --- Checking each stream for virtual concept drift --- ");
		Enumeration e = hashSamplesPerStream.keys();
		while (e.hasMoreElements()) 
		{		
			int key = Integer.parseInt(e.nextElement().toString()); //stream id
			int streamId = key;
			StreamOldAndNewSamples ss = (StreamOldAndNewSamples)hashSamplesPerStream.get(key);	
			boolean rcVirtual = false;
						
			if (ss.GetOldSamples().size() > 0)
			{
				rcVirtual = IsVirtualConceptDrift(streamId,
												  ss.GetOldSamples(), 
						 					      ss.GetNewSamples(),
						 					      OnlineClassifierRef,
						 					      hashSTQMRouteSignaturesGlobalStorage,
						 						  vcdMethod);
			}
			else
			{
				//if old sample is empty, this means this is the very first execution, hence by default we want to enforce a virtual concept drift
				rcVirtual = true;
			}
			//store the result in a hash table where for each stream, a result is stored (indicating if virtula concept drift has happened) is stored
			hashVirtualConceptDriftResults.put(key, rcVirtual);
		}
		
		//------------------------------------------------------------------
  	    // 2. LOOP THROUGH ALL RUSTERS, COMPUTE NEW ROUTES BASED ON UPDATED STATISTICS
		//    AFTER THE ROUTE IS COMPUTED, CHECK IF REAL CONCEPT DRIFT HAS OCCURRED FOR THE RUSTER
		//------------------------------------------------------------------
		Helper.WriteToFile(Helper.LOG_FILE, "\t" + Helper.GetCurrentTimeString() + ": --- Checking each ruster groups for real concept drift -- ");
		
		STQM.NEW_ROUTES = new Hashtable();
		
		//iterate through all streams
		e = hashSTQMRouteSignaturesGlobalStorage.keys();
		while (e.hasMoreElements()) 
		{	
			//select the hashtable storing all the routes for the current stream
			String key = (String) e.nextElement();	//stream id
			Hashtable hashStreamRoutesSignatures = (Hashtable)hashSTQMRouteSignaturesGlobalStorage.get(key);
			
			//iterate through all stream's routes
			Enumeration eRoutes = hashStreamRoutesSignatures.keys();
			while (eRoutes.hasMoreElements()) 
			{		
				String keyRoute = (String) eRoutes.nextElement(); //key = route id
				
				RouteSignatureStorage rss = (RouteSignatureStorage)hashStreamRoutesSignatures.get(keyRoute);

				//ordered list of operators
				String [] operators = rss.GetRouteSignature().split(",");

				//compute the new route for the current "ruster"				
				String globalStatsKey = key + "->" + rss.GetRouteSignature();
				
				String [] operatorsNew = ComputeRusterRouteBasedOnStatistics(globalStatsKey, 
																			 hashSamplesPerOperator, 
																			 hashStreamRoutesSignatures,
																			 operators.clone());
				
				//determine if real concept drift has occurred (based on operator order)
				boolean rcReal = IsRealConceptDrift(operators, 
												    operatorsNew,
												    rcdMethod);
				
				//store the result in a hash table where for each stream, a result is stored (indicating if virtula concept drift has happened) is stored				
				hashRealConceptDriftResults.put(globalStatsKey, rcReal);
				
				//store "new" routes in the STQM "cache"
				STQM.NEW_ROUTES.put(rss.GetStreamId() + ":" + rss.routeId, operatorsNew);
			}			
		}		
		return;
	}
	
    //-----------------------------------------------------------------------------
    
	public String[] ComputeRusterRouteBasedOnStatistics(String globalStatsKey, 
			 											Hashtable hashSamplesPerOperator, 
			 											Hashtable hashStreamRoutesSignatures,
			 										    String[] operators)
	{
    	String [] oldOperatorOrder = operators.clone();
    	ArrayList arrOperatorsAndSelectivities = new ArrayList();
    	
    	//operators returned will have the order based on the latest statistics
    	String [] newOperatorOrder = new String[operators.length];
    	
    	//iterate through the list of operators in the current route
    	for(int p = 0; p < operators.length; p++)
    	{
    		int key = Integer.parseInt(operators[p].toString()); 
    		//get the current statistics for the operator for the current ruster
    		Hashtable hashNewSTQMRusterExploitationStatisticsLocalStorage = (Hashtable)hashSamplesPerOperator.get(key);    		
    		
    		//get the op selectivity for the ruster
    		OperatorIndividualSample opStatsSampleForRuster = (OperatorIndividualSample)hashNewSTQMRusterExploitationStatisticsLocalStorage.get(globalStatsKey);
    		if (opStatsSampleForRuster == null)
    		{
    			System.out.println("Op: "+key+". No exe stats for : " + globalStatsKey);
    			OperatorAndSelectivity opAndSelectivity = new OperatorAndSelectivity(Integer.toString(p), 0.0);
        		arrOperatorsAndSelectivities.add(opAndSelectivity);
    		}
    		else
    		{
	    		double selectivity = 0;
	    		if (opStatsSampleForRuster.GetTuplesIn() > 0)
	    		{
	    			selectivity = opStatsSampleForRuster.GetTuplesOut()/opStatsSampleForRuster.GetTuplesIn();
	    		}
	    		OperatorAndSelectivity opAndSelectivity = new OperatorAndSelectivity(Integer.toString(p), selectivity);
	    		arrOperatorsAndSelectivities.add(opAndSelectivity);
    		}
    	}
    	
    	//sort the operators based on selectivities
		Object [] opAndSelsArray = arrOperatorsAndSelectivities.toArray();
		Arrays.sort(opAndSelsArray, new OperatorAndSelectivityComparer());
		
		for(int p = 0; p < operators.length; p++)
		{
			newOperatorOrder[p] = operators[((OperatorAndSelectivity)opAndSelsArray[p]).GetOperatorIndex()];
		}
		
		System.out.println("");
		
    	return newOperatorOrder;
	}
	
    //-----------------------------------------------------------------------------

    public boolean IsVirtualConceptDrift(int streamId,
    									 ArrayList oldSample, 
			   							 ArrayList newSample,
			   							 OnlineClassifierOperatorImp OnlineClassifierRef,
			   							 Hashtable hashSTQMRouteSignaturesGlobalStorage,
			   							 STQM_VCD_METHOD vcdMethod)
	{
    	boolean rc = false;
    	double maxThreshold = 0.2; //20% - max threhold. If the difference is greater than this value, then virtual concept drift is assumed
    	//--------------------------------------------------------
    	// DETECT VIRTUAL CONCEPT DRIFT BASED MISCLASSIFICATION RATE
    	//--------------------------------------------------------
    	if (vcdMethod == STQM_VCD_METHOD.MISCLASSIFICATION_RATE)
    	{
    		
    	}
    	//--------------------------------------------------------
    	// DETECT VIRTUAL CONCEPT DRIFT BASED MISCLASSIFICATION RATE
    	//--------------------------------------------------------
    	else if (vcdMethod == STQM_VCD_METHOD.SIGNATURE_BASED)
    	{    		
    		//form signatures
    		Hashtable<Integer, ArrayList> hashCurrentSampleClassificationTuples = new Hashtable<Integer, ArrayList>();
    		String oldSampleSignature = OnlineClassifierRef.GetCountBasedClassificationSignature(streamId, oldSample, hashSTQMRouteSignaturesGlobalStorage, hashCurrentSampleClassificationTuples);
    		Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + "Virtual Concept Drift Analysis [Signature-based]: Stream ["+streamId+"] OLD sample signature: " + oldSampleSignature); 
    		Hashtable<Integer, ArrayList> hashNewSampleClassificationTuples = new Hashtable<Integer, ArrayList>();
    		String newSampleSignature = OnlineClassifierRef.GetCountBasedClassificationSignature(streamId, newSample, hashSTQMRouteSignaturesGlobalStorage, hashNewSampleClassificationTuples);
    		Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + "Virtual Concept Drift Analysis [Signature-based]: Stream ["+streamId+"] NEW sample signature: " + newSampleSignature);
    		int absDiff = countAbsoluteSignatureDifferences(oldSampleSignature, newSampleSignature);
    		
    		double relDiff = absDiff/getSignatureLength(oldSampleSignature); 
    		if (relDiff > maxThreshold)
    		{
    			rc = true; //virtual concept drift is present
    		}
    		else
    		{
    			rc = false; //no virtual concept drift is present
    		}    			
    	}
    	return rc;
	}

    //-----------------------------------------------------------------------------
    
    private int getSignatureLength(String currSignature)
    {    	        	
    	int signatureLength = 0; //count of operators in the signature
    	String [] arrCurrSignature = currSignature.split(" ");    	
    	signatureLength = arrCurrSignature.length;    	
    	return signatureLength;
    }
    
    //-----------------------------------------------------------------------------
    
    private int countAbsoluteSignatureDifferences(String oldSignature, String newSignature)
    {
    	int absCount = 0;
    	String [] arrOldSignature = oldSignature.split(" ");
    	String [] arrNewSignature = newSignature.split(" ");

    	for (int i = 0; i < arrOldSignature.length; i++)
    	{    		
    		String oldSignatureOpCounter = arrOldSignature[i].toString().substring(arrOldSignature[i].toString().indexOf(":") + 1, arrOldSignature[i].toString().length());
    		String newSignatureOpCounter = arrNewSignature[i].toString().substring(arrNewSignature[i].toString().indexOf(":") + 1, arrNewSignature[i].toString().length());
    		
    		if (Integer.parseInt(oldSignatureOpCounter) != Integer.parseInt(newSignatureOpCounter))
    		{
    			absCount++;
    		}
    	}
    	
    	return absCount;
    }
    
    //-----------------------------------------------------------------------------
    
    public boolean IsRealConceptDrift(String[] operatorsOldOrder, 
				 					  String[] operatorsNewOrder,
				 					  STQM_RCD_METHOD rcdMethod)
	{
    	boolean rc = false;
    	int absDiffCount = 0;
    	//--------------------------------------------------------
    	// DETECT REAL CONCEPT DRIFT BASED ON NUMBER OF AFFECTED ROUTES
    	//--------------------------------------------------------
    	if (rcdMethod == STQM_RCD_METHOD.NUMBER_OF_AFFECTED_ROUTES)
    	{
    		for (int i = 0; i < operatorsOldOrder.length; i++)
    		{
    			if (operatorsOldOrder[i] != operatorsNewOrder[i])
    			{   
    				absDiffCount++;
    			}
        	}    	
    		
    		if (absDiffCount > 0)
    		{
    			rc = true;
    		}
    	}
    	//--------------------------------------------------------
    	// DETECT REAL CONCEPT DRIFT BASED ON ROUTE EDIT DISTANCE
    	//--------------------------------------------------------
    	else if (rcdMethod == STQM_RCD_METHOD.ROUTE_EDIT_DISTANCE)
    	{
    		
    	}
    	//--------------------------------------------------------
    	// DETECT REAL CONCEPT DRIFT BASED ON DISSIMILAR SUBSEQUENCE
    	//--------------------------------------------------------
    	else if (rcdMethod == STQM_RCD_METHOD.DISSIMILAR_SUBSEQUENCE)
    	{
    		
    	}
		return rc;
	}    

    //-----------------------------------------------------------------------------	
    //                      PHASE II: TUNING RECOMMENDATIONS
    //-----------------------------------------------------------------------------	

    public enum STQM_TUNING_RECOMMENDATIONS 
    {
    	NEW_CLASSIFIER_OLD_ROUTES,
    	OLD_CLASSIFIER_NEW_ROUTES,
    	NEW_CLASSIFIER_NEW_ROUTES,
    	NONE
    }
    
    //-----------------------------------------------------------------------------
	
	public STQM_TUNING_RECOMMENDATIONS MakeQMRecommendations(Hashtable hashSTQMDataSamplesGlobalStorage,
															 Hashtable hashVirtualConceptDriftResults,
		 					 		  					     Hashtable hashRealConceptDriftResults)
	{	
		int countVirtualDrifts = 0;
		int countRealDrifts = 0;
		
		STQM_TUNING_RECOMMENDATIONS currentRecommendations = STQM_TUNING_RECOMMENDATIONS.NONE; //default
		
		//-----------------------------------------------------------------------
		// VIRTUTAL CONCEPT DRIFTS - do the count
		//-----------------------------------------------------------------------		
		Enumeration eConceptDrifts = hashVirtualConceptDriftResults.elements(); 		
		while (eConceptDrifts.hasMoreElements()) 
		{
			if (eConceptDrifts.nextElement().toString().equalsIgnoreCase("true"))
			{
				countVirtualDrifts++;
			}
		}
		
		//-----------------------------------------------------------------------
		// REAL CONCEPT DRIFTS - do the count
		//-----------------------------------------------------------------------		
		eConceptDrifts = hashRealConceptDriftResults.elements(); 		
		while (eConceptDrifts.hasMoreElements()) 
		{
			if (eConceptDrifts.nextElement().toString().equalsIgnoreCase("true"))
			{
				countRealDrifts++;
			}
		}
		
		//analyze the cases		
		//-----------------------------------------------------------------------
		// HYBRID CONCEPT DRIFT
		//-----------------------------------------------------------------------		
		if ((countVirtualDrifts > 0) && (countRealDrifts > 0))
		{
			//------------------------------------------------
			// BUILD NEW QM
			//------------------------------------------------
			BuildNewQueryMesh(hashSTQMDataSamplesGlobalStorage);
			
			currentRecommendations = STQM_TUNING_RECOMMENDATIONS.NEW_CLASSIFIER_NEW_ROUTES;
		}
		//-----------------------------------------------------------------------
		// VIRTUAL CONCEPT DRIFT
		//-----------------------------------------------------------------------
		else if ((countVirtualDrifts > 0) && (countRealDrifts == 0))
		{		
			
			//------------------------------------------------
			// BUILD NEW QM
			//------------------------------------------------
			BuildNewQueryMesh(hashSTQMDataSamplesGlobalStorage);
			
			//compute new classifier for the QM

			/*
			System.out.println("\nCreating local query meshes...");
			int qmCount = 0;
			while (e.hasMoreElements()) 
			{
				qmCount++;
				System.out.println("\n\n*************************************************************************************************");
				System.out.println("*************************************************************************************************");
				System.out.println("\n--------------------QUERY MESH ["+qmCount+"]------------------------");
				SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  // do whatever you need with x 

				//--------------------------------------------------------------------
				// CALL LOCAL QUERY MESH SEARCH FOR EACH STREAM
				//--------------------------------------------------------------------	
				streamQM = produceLocalQueryMesh(currInput, heuristic, m, startSolution);	 
				bestQM.addQM(streamQM);
			}//end while (iterating all stream inputs)
			*/
			
			currentRecommendations = STQM_TUNING_RECOMMENDATIONS.NEW_CLASSIFIER_OLD_ROUTES;
		}
		//----------------------------------------------------------------------
		// REAL CONCEPT DRIFT
		//----------------------------------------------------------------------
		else if ((countVirtualDrifts == 0) && (countRealDrifts > 0))
		{			
			//------------------------------------------------
			// BUILD NEW QM
			//------------------------------------------------
			BuildNewQueryMesh(hashSTQMDataSamplesGlobalStorage);
			
			//compute new routes			
			currentRecommendations = STQM_TUNING_RECOMMENDATIONS.OLD_CLASSIFIER_NEW_ROUTES;
		}	
		
		Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Current recommendation: "+currentRecommendations+" --- ");
		return currentRecommendations;
	}        
	
    //-----------------------------------------------------------------------------
	
	private void BuildNewQueryMesh(Hashtable hashSTQMDataSamplesGlobalStorage)
	{	
		//-------------------------------------------------------------------
		// PREPARE SOME DATA FOR THE OPTIMIZER
		//-------------------------------------------------------------------
		DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
        java.util.Date date = new java.util.Date();
        String datetime = dateFormat.format(date);
        datetime = datetime.replaceAll(":", ".");
        datetime = datetime.replaceAll(" ", ".");
        datetime = datetime.replaceAll("/", ".");
        	//date.getYear()+"." +date.getMonth()+"." + date.getDate() + "." + date.getHours() + "." + date.getMinutes() + "." + date.getSeconds();
        String subDir = datetime;
        
		//create stats sample files necessary for the optimizer
		String statisticsSamplesSubDir = STQM.CreateNewOptimizerStatisticsSampleFiles(subDir, hashSTQMDataSamplesGlobalStorage);
		//create training tuples files necessary for the optimizer
		String trainingTuplesSamplesSubDir = STQM.CreateNewOptimizerTrainingTuplesFiles(subDir, hashSTQMDataSamplesGlobalStorage);			
		//update the input xml file with the latest settings		
		STQM.UpdateOptimizerInputXmlFileWithLatestSettings(statisticsSamplesSubDir, trainingTuplesSamplesSubDir);
	
		String inputfileName = STQM.OPTIMIZER_INPUT_XML_FILE;
		QM_SEARCH_STRATEGY searchHeuristic = QM_SEARCH_STRATEGY.ITERATIVE_IMPROVEMENT;
		int m = 0; //full enumeration subset size
		QM_SEARCH_START_SOLUTION startSolution = QM_SEARCH_START_SOLUTION.CONTENT_DRIVEN;
		
		//-------------------------------------------------------------------
		// CALL QM OPTIMIZER
		//-------------------------------------------------------------------			
		//compute a new QM, with the current QM as a start solution
		QueryMeshOptimizer qmOptimizer = new QueryMeshOptimizer(inputfileName);				
		//--------------------------------------------------------------------
		// PRODUCE GLOBAL QUERY MESH
		//--------------------------------------------------------------------
		Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Calling QM optimizer --- ");
		double timeOptimizerStarted = System.currentTimeMillis();	
		GlobalQueryMesh bestQM = qmOptimizer.produceGlobalQueryMesh(searchHeuristic, m, startSolution);			
		double timeOptimizerEnded = System.currentTimeMillis();	
		double timeOptimizerDuration = timeOptimizerEnded - timeOptimizerStarted;
		double timeOptimizerDurationDurationSec = timeOptimizerDuration / 1000;
		
		Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- QM optimizer is done ["+timeOptimizerDuration+"] msec or ["+timeOptimizerDurationDurationSec+"] sec --- ");
		//materialize just computed bestQM into xml file
		String outputXmlPlan = STQM.QM_SOLUTIONS_DIR + subDir + ".xml";
		qmOptimizer.MaterializeQMSolutionIntoXmlFile(bestQM, outputXmlPlan);
		STQM.NEW_QM_XML = outputXmlPlan;
		//--------------------------------------------------------------------
		// OPTIMIZER IS DONE
		//--------------------------------------------------------------------
	}
	
	//-----------------------------------------------------------------------------
	
}
