//--------------------------------------------------
// Bell.java
// Author: not known 
// Adopted By: Rimma V. Nehme rnehme@purdue.edu, 2007
//--------------------------------------------------
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.io.IOException;
import java.io.InputStream;
import java.io.DataInputStream;


public class Bell 
{

	private Bell() 
	{
	}

	/**
	 * Reads precomputed data into arrays.
	 */
	public static void initialize() throws IOException 
	{
		if ((Bell.indexesAfterRemove == null) || 
			(Bell.blocksBeforeRemove == null) || 
			(Bell.indexesAfterInsert == null)) 
		{
			//Open resource
			InputStream stream = Bell.class.getResourceAsStream("Bell.data");
			DataInputStream inStream = new DataInputStream(stream);

			// Read indexesAfterRemove short[][][] array
			Bell.indexesAfterRemove = new short[inStream.readInt()][][];
			for (int i = 0; i < Bell.indexesAfterRemove.length; i++) 
			{
				Bell.indexesAfterRemove[i] = new short[inStream.readInt()][];
				for (int j = 0; j < Bell.indexesAfterRemove[i].length; j++) 
				{
					Bell.indexesAfterRemove[i][j] = new short[inStream.readInt()];
					for (int k = 0; k < Bell.indexesAfterRemove[i][j].length; k++)
						Bell.indexesAfterRemove[i][j][k] = inStream.readShort();
				}
			}

			// Read blocksBeforeRemove byte[][][] array
			Bell.blocksBeforeRemove = new byte[inStream.readInt()][][];
			for (int i = 0; i < Bell.blocksBeforeRemove.length; i++) 
			{
				Bell.blocksBeforeRemove[i] = new byte[inStream.readInt()][];
				for (int j = 0; j < Bell.blocksBeforeRemove[i].length; j++) 
				{
					Bell.blocksBeforeRemove[i][j] = new byte[inStream.readInt()];
					inStream.read(Bell.blocksBeforeRemove[i][j]);
				}
			}

			// Read indexesAfterInsert short[][][][] array
			Bell.indexesAfterInsert = new short[inStream.readInt()][][][];
			for (int i = 0; i < Bell.indexesAfterInsert.length; i++) 
			{
				Bell.indexesAfterInsert[i] = new short[inStream.readInt()][][];
				for (int j = 0; j < Bell.indexesAfterInsert[i].length; j++) 
				{
					Bell.indexesAfterInsert[i][j] = new short[inStream.readInt()][];
					for (int k = 0; k < Bell.indexesAfterInsert[i][j].length; k++) 
					{
						Bell.indexesAfterInsert[i][j][k] = new short[inStream.readInt()];
						for (int l = 0; l < Bell.indexesAfterInsert[i][j][k].length; l++)
							Bell.indexesAfterInsert[i][j][k][l] = inStream.readShort();
					}
				}
			}

			// Close resource
			inStream.close();
			stream.close();
		}
	}

	// Bell Numbers ----------------------------------------------------------------------------------------------------

	/** First sixteen Bell numbers (n = 0, ..., 15) */
	private static int[] bellNumbers = {1,
									    1,
									    2,
									    5,
									    15,
									    52,
									    203,
									    877,
									    4140,
									    21147,
									    115975,
									    678570,
									    4213597,
									    27644437,
									    190899322,
									    1382958545};

	/**
	 * N-th Bell number. Method for practical use.
	 * 
	 * @param n
	 * @return n-th Bell number.
	 * 
	 * @throws IllegalArgumentException if n is outside of the desired range.
	 */
	public static int number(int n) 
	{
		if (n >= Bell.bellNumbers.length)
			throw new IllegalArgumentException("Value of the Bell number #" + n + " is not in type int range.");
		else if (n < 0)
			throw new IllegalArgumentException("n  less then zero.");
		else
			return Bell.bellNumbers[n];
	}

	/** The first Bell number (n = 0) */
	private static BigInteger[] bigBellNumbers = {BigInteger.ONE, BigInteger.ONE};

	/** The second Bell number (n = 1) */
	private static BigInteger[] bigBellTriangleRow = {BigInteger.ONE};

	/**
	 * Computes n-th Bell number using Bell triangle.
	 * 
	 * To compute n-th Bell number for the first time we use two arrays of size n +1 and n and perform (n^2 - n) / 2 additions. Once we have computed
	 * n-th Bell number, all predecessing Bell numbers are computed and stored in memory to avoid unnecessary recomputation.
	 * 
	 * @param n , works even if n = 1000, for n = 10000 I ran out of memory ;-)
	 * @return n-th Bell number.
	 * 
	 * @throws IllegalArgumentException if n is less than zero.
	 */
	public static BigInteger bigNumber(int n) 
	{
		if (n < 0)
			throw new IllegalArgumentException("n less then zero.");
		if (n >= Bell.bellNumbers.length) 
		{
			if (n >= Bell.bigBellNumbers.length) 
			{
				// Create bigger lookup table and copy old data
				int m = Bell.bigBellNumbers.length;
				BigInteger[] temp = new BigInteger[n + 1];
				System.arraycopy(bigBellNumbers, 0, temp, 0, m);
				Bell.bigBellNumbers = temp;
				// Fill in empty space
				for (int i = m; i < Bell.bigBellNumbers.length; i++) 
				{
					// Update Bell triangle row
					temp = new BigInteger[i];
					temp[0] = Bell.bigBellNumbers[i - 1];
					for (int j = 1; j < i; j++)
						temp[j] = temp[j - 1].add(Bell.bigBellTriangleRow[j - 1]);
					Bell.bigBellTriangleRow = temp;
					Bell.bigBellNumbers[i] = Bell.bigBellTriangleRow[i - 1];
				}
			}
			return Bell.bigBellNumbers[n];
		} 
		else 
		{
			return new BigInteger(Integer.toString(Bell.bellNumbers[n]));
		}
	}

	// Resticted growth strings ---------------------------------------------------------------------------------------

	/**
	 * Returns set partitions of the set of n elements encoded as restricted growth strings. Method for practical use.
	 * 
	 * Based on algorithm by George Hutchinson presented in CACM 6 (1963), 613-614 and described in Donald E. Knuth, TAOC Vol 7, pre-fascicle 3B,
	 * 25-27.
	 * 
	 * For complexity analysis see exercise 46 in Knuth's book.
	 * 
	 * @param n number of elements in the set.
	 * @return restricted growth stings representing set partitions.
	 * 
	 * @throws IllegalArgumentException if n is less than zero or greater than fiveteen.
	 */
	public static int[][] growthStrings(int n) 
	{
		if (n <= 0)
			throw new IllegalArgumentException("n less then zero.");
		else if (n >= Bell.bellNumbers.length)
			throw new IllegalArgumentException("Value of the Bell number #" + n
					+ " is not in type int range. There would be to many strings to represent.");
		else if (n == 1)
			return new int[1][1];
		else {
			// H1. [Initialize.]
			int[][] result = new int[Bell.number(n)][n];
			int[] a = new int[n];
			int[] b = new int[n];
			Arrays.fill(b, 1);
			int m = 1;
			int j;
			int index = 0;
			while (true) 
			{
				// H2. [Visit.]
				System.arraycopy(a, 0, result[index++], 0, n);
				if (a[n - 1] == m) 
				{
					// H4. [Find j.]
					j = n - 2;
					while (a[j] == b[j])
						j--;
					if (j == 0)
						break;
					else 
					{
						// H5. [Increase a_j.]
						a[j]++;
						// H6. [Zero out a_{j+1} ... a_n.]
						m = (a[j] == b[j]) ? b[j] + 1 : b[j];
						j++;
						while (j < n - 1) 
						{
							a[j] = 0;
							b[j] = m;
							j++;
						}
						a[n - 1] = 0;
					}
				} else
					// H3. [Increase a_n.]
					a[n - 1]++;
			}
			return result;
		}
	}

	/**
	 * Iterator over restricted growth strings.
	 * 
	 * @param n number of elements in the set.
	 * 
	 * @return an iterator over the partitions of the set represented as restricted growth strings. This code is based on the Iterator code in
	 * java.util.AbstractList.
	 */
	public static Iterator growthStringsIterator(int n) 
	{
		if (n <= 0)
			throw new IllegalArgumentException("n less then zero.");
		else if (n >= Bell.bellNumbers.length)
			throw new IllegalArgumentException("Value of the Bell number #" + n
					+ " is not in type int range. There would be to many strings to represent.");
		else
			return new Itr(n);
	}

	/**
	 * Simple Iterator class. Does not provide opportunity to remove elements from the container.
	 * <p>
	 * This code is based on the Iterator code in {@link java.util.AbstractList}.
	 * 
	 */
	private static class Itr implements Iterator 
	{

		/** Next element. */
		private int[] next;

		/** Indicates if there is a next element to be returned. */
		private boolean isNext;

		/** Helper variables. */
		int[] b;

		/**  */
		int m = 1;

		/**  */
		int j;

		/**	 */
		int index = 0;

		/**
		 * Constructor.
		 * 
		 * @param n arity of the set.
		 */
		public Itr(int n) 
		{
			this.next = new int[n];
			this.isNext = true;
			this.b = new int[n];
			Arrays.fill(this.b, 1);
		}

		/** Prepares next element for the iterator. */
		private void prepareNext() 
		{
			if (this.next.length == 1) 
			{
				this.isNext = false;
				return;
			} 
			else if (this.next[this.next.length - 1] == this.m) 
			{
				// [Find j.]
				this.j = this.next.length - 2;
				while (this.next[this.j] == this.b[this.j])
					this.j--;
				if (this.j == 0) 
				{
					this.isNext = false;
					return;
				} 
				else 
				{
					// [Increase this.next_j.]
					this.next[this.j]++;
					// [Zero out this.next_{j+1} ... this.next_n.]
					this.m = (this.next[this.j] == this.b[this.j]) ? this.b[this.j] + 1 : this.b[this.j];
					this.j++;
					while (this.j < this.next.length - 1) 
					{
						this.next[this.j] = 0;
						this.b[this.j] = this.m;
						this.j++;
					}
					this.next[this.next.length - 1] = 0;
				}
			} 
			else
				// [Increase this.next_n.]
				this.next[this.next.length - 1]++;
		}

		/**
		 * @see java.util.Iterator#hasNext()
		 */
		public boolean hasNext() 
		{
			return this.isNext;
		}

		/**
		 * @throws NoSuchElementException
		 * 
		 * @see java.util.Iterator#next()
		 */
		public Object next() 
		{
			if (this.isNext) 
			{
				int[] result = new int[this.next.length];
				System.arraycopy(this.next, 0, result, 0, this.next.length);
				prepareNext();
				return result;
			} 
			else
				throw new NoSuchElementException();
		}

		/**
		 * @throws UnsupportedOperationException
		 * 
		 * @see java.util.Iterator#remove()
		 */
		public void remove() 
		{
			throw new UnsupportedOperationException();
		}
	}

	// Index -> partition and partition -> index functions -----------------------------------------------------------

	/**
	 * Precomputed array of the sizes of subtrees for constructing restricted growth strings for sets of up to 15 elements.
	 */
	private static int[][] subTreeSize = {{1,2,5,15,52,203,877,4140,21147,115975,678570,4213597,27644437,190899322,1382958545},
			{1,3,10,37,151,674,3263,17007,94828,562595,3535027,23430840,163254885,1192059223},
			{1,4,17,77,372,1915,10481,60814,372939,2409837,16360786,116393205,865549453},
			{1,5,26,141,799,4736,29371,190497,1291020,9131275,67310847,516369838},{1,6,37,235,1540,10427,73013,529032,3967195,30785747,247126450},
			{1,7,50,365,2727,20878,163967,1322035,10949772,93197715},{1,8,65,537,4516,38699,338233,3017562,27499083},
			{1,9,82,757,7087,67340,649931,6376149},{1,10,101,1031,10644,111211,1176701},{1,11,122,1365,15415,175802},{1,12,145,1765,21652},
			{1,13,170,2237},{1,14,197},{1,15},{1}};

	/**
	 * Returns i-th set partition of the set of n elements encoded as restricted growth strings. Method for practical use.
	 * 
	 * @param i number of the requested partition (starting from 0)
	 * @param n number of elements in the set.
	 * @return restricted growth stings representing partition.
	 * 
	 * @throws IllegalArgumentException if n is less than zero or i is not a correct partition number.
	 */
	public static int[] growthString(int i, int n) 
	{
		if (n <= 0)
			throw new IllegalArgumentException("n less then zero.");
		else if ((i < 0) || (i >= Bell.number(n)))
			throw new IllegalArgumentException(i + " is not a correct partition number.");
		else 
		{
			int[] result = new int[n];
			int div;
			int rem = i;
			int max = 1;
			for (int j = n - 2; j >= 0; j--) 
			{
				div = rem / Bell.subTreeSize[max - 1][j];
				rem = rem % Bell.subTreeSize[max - 1][j];
				if (div > max) {
					// Adjust
					result[n - j - 1] = max;
					rem = rem + (Bell.subTreeSize[max - 1][j] * (div - max));
					max++;
				} 
				else 
				{
					// No need to adjust
					result[n - j - 1] = div;
					max = (max > div) ? max : ++max;
				}
			}
			return result;
		}
	}

	/**
	 * Returns index of the set partition (given as a restricted growth string). Method for practical use.
	 * 
	 * @param string representing partition
	 * @return index of the partition (strating from 0)
	 * 
	 * @throws IllegalArgumentException is string is too long or it is not a proper growth string.
	 */
	public static int indexOf(int[] string) 
	{
		if (string.length >= Bell.bellNumbers.length)
			throw new IllegalArgumentException("String too long.");
		else if (string[0] != 0)
			throw new IllegalArgumentException("Not a proper growth string.");
		else 
		{
			int result = 0;
			int max = 0;
			for (int i = 1; i < string.length; i++) 
			{
				result = result + (Bell.subTreeSize[max][string.length - i - 1] * string[i]);
				if (max == string[i] - 1)
					max++;
				else if (max < string[i])
					throw new IllegalArgumentException("Not a proper growth string.");
			}
			return result;
		}
	}

	/** Array of sizes of the subtrees for constructing restricted growth strings */
	private static ArrayList<ArrayList<BigInteger>> bigSubTreeSizes = new ArrayList<ArrayList<BigInteger>>();

	/**
	 * Returns the size of the subtree used for constructing restricted growth strings.
	 * 
	 * @param maxNode so far
	 * @param height of the subtree
	 * @return size of the subtree (number of leaves)
	 * 
	 * If we call method for the first time to compute size for given maxNode and height we do approximately ((maxNode + height)^2 + 1)/2 additions
	 * and multiplications. Later on, all so far computated values are already available in the memory, so computation is faster.
	 * 
	 * @throws IllegalArgumentException if maxNode or height is less than zero.
	 */
	private static BigInteger bigSubTreeSize(int maxNode, int height) 
	{
		// Check arguments
		if ((maxNode < 0) || (height < 0))
			throw new IllegalArgumentException("Wrong arguments.");
		// Check array
		if (maxNode + height >= Bell.bigSubTreeSizes.size()) {
			// Fill in array
			int size = maxNode + height;
			for (int i = Bell.bigSubTreeSizes.size(); i <= size; i++) 
			{
				// First column
				Bell.bigSubTreeSizes.add(new ArrayList<BigInteger>());
				Bell.bigSubTreeSizes.get(i).add(BigInteger.ONE);
				// Second column
				if (i > 0)
					Bell.bigSubTreeSizes.get(i - 1).add(new BigInteger(String.valueOf(i + 1)));
				// Next columns
				for (int j = i - 1; j > 0; j--) 
				{
					Bell.bigSubTreeSizes.get(j - 1).add(
							Bell.bigSubTreeSizes.get(j).get(i - j).add(
									Bell.bigSubTreeSizes.get(j - 1).get(i - j).multiply(new BigInteger(String.valueOf(j)))));
				}
			}
		}
		return Bell.bigSubTreeSizes.get(maxNode).get(height);
	}

	/**
	 * Returns i-th set partition of the set of n elements encoded as restricted growth strings.
	 * 
	 * @param i number of the requested partition (starting from 0)
	 * @param n number of elements in the set.
	 * @return restricted growth stings representing partition.
	 * 
	 * @throws IllegalArgumentException if n is less than zero or i is not a correct partition number.
	 */
	public static int[] bigGrowthString(BigInteger i, int n) 
	{
		if ((n < Bell.bellNumbers.length) && i.compareTo(new BigInteger(Integer.toString(Integer.MAX_VALUE))) <= 0)
			return Bell.growthString(i.intValue(), n);
		else if (n <= 0)
			throw new IllegalArgumentException("n less or equal zero.");
		else if ((i.compareTo(BigInteger.ZERO) < 0) || (i.compareTo(Bell.bigNumber(n)) >= 0))
			throw new IllegalArgumentException(i + " is not a correct partition number.");
		else 
		{
			int[] result = new int[n];
			BigInteger[] divAndRem = new BigInteger[2];
			divAndRem[1] = i;
			int max = 1;
			for (int j = n - 2; j >= 0; j--) 
			{
				divAndRem = divAndRem[1].divideAndRemainder(Bell.bigSubTreeSize(max - 1, j));
				int div = divAndRem[0].intValue();
				if (div > max) 
				{
					// Adjust
					result[n - j - 1] = max;
					divAndRem[1] = divAndRem[1].add(Bell.bigSubTreeSize(max - 1, j).multiply(new BigInteger(String.valueOf(div - max))));
					max++;
				} 
				else 
				{
					// No need to adjust
					result[n - j - 1] = divAndRem[0].intValue();
					max = (max > div) ? max : ++max;
				}
			}
			return result;
		}
	}

	/**
	 * Returns index of the set partition (given as a restricted growth string).
	 * 
	 * @param string representing partition
	 * @return index of the partition (starting from 0)
	 * 
	 * @throws IllegalArgumentException if string is not a proper growth string.
	 */
	public static BigInteger bigIndexOf(int[] string) 
	{
		if (string.length < Bell.bellNumbers.length)
			return new BigInteger(Integer.toString(Bell.indexOf(string)));
		else if (string[0] != 0)
			throw new IllegalArgumentException("Not a proper growth string.");
		else 
		{
			BigInteger result = BigInteger.ZERO;
			int max = 0;
			for (int i = 1; i < string.length; i++) 
			{
				result = result.add(Bell.bigSubTreeSize(max, string.length - i - 1).multiply(new BigInteger(String.valueOf(string[i]))));
				if (max == string[i] - 1)
					max++;
				else if (max < string[i])
					throw new IllegalArgumentException("Not a proper growth string.");
			}
			return result;
		}
	}

	// Removing elements ---------------------------------------------------------------------------------------------

	/** New index of the partition (growth string) after removing the particular element. */
	private static short[][][] indexesAfterRemove;

	/**
	 * Computes the new index of the growth string after removing the particular element. Method for practical use.
	 * 
	 * @param index of the growth string
	 * @param length before removal
	 * @param toRemove element (counting from 0)
	 * @return new index
	 * 
	 * @throws IllegalArgumentException if length is less than two or index of an element to remove is outside of a proper range or if growth string's
	 * index is not correct.
	 */
	public static int indexAfterRemove(int index, int length, int toRemove) 
	{
		if (length < 2)
			throw new IllegalArgumentException("Wrong size of the partition.");
		else if ((toRemove > length - 1) || (toRemove < 0))
			throw new IllegalArgumentException("Wrong index of element to be removed." + toRemove + " " + length);
		else if ((index < 0) || (index >= Bell.number(length)))
			throw new IllegalArgumentException("Wrong index of the partition.");
		else 
		{
			if (Bell.indexesAfterRemove == null)				
				try 
			{
					Bell.initialize();				
			} 
			catch (IOException e) 
			{
					throw new Error("Resource \"ca.ubc.cs.kisynski.Bell.data\" not available!");
			}
			if (length < Bell.indexesAfterRemove.length + 2)
				return Bell.indexesAfterRemove[length - 2][index][toRemove];
			else 
			{
				int result[] = new int[length - 1];
				int div;
				int rem = index;
				int max = 1;
				int current;
				int[] old2new = new int[length];
				if (toRemove > 0) 
				{
					// We are not removing the first element
					Arrays.fill(old2new, 1, length, -1);
					int difference = length - toRemove - 1;
					// Elements predecessing removed element
					for (int j = length - 2; j > difference; j--) 
					{
						div = rem / Bell.subTreeSize[max - 1][j];
						rem = rem % Bell.subTreeSize[max - 1][j];
						if (div > max) 
						{
							// Adjust
							result[length - j - 1] = max;
							old2new[max] = max;
							rem = rem + (Bell.subTreeSize[max - 1][j] * (div - max));
							max++;
						} 
						else 
						{
							// No need to adjust
							result[length - j - 1] = div;
							old2new[div] = div;
							max = (max > div) ? max : ++max;
						}
					}
					current = max;
					// Element to be removed
					div = rem / Bell.subTreeSize[max - 1][difference];
					rem = rem % Bell.subTreeSize[max - 1][difference];
					if (div > max) 
					{
						// Adjust
						rem = rem + (Bell.subTreeSize[max - 1][difference] * (div - max));
						max++;
					} 
					else
						// No need to adjust
						max = (max > div) ? max : ++max;
				} 
				else 
				{
					// We are removing the first element
					current = 0;
					Arrays.fill(old2new, -1);
				}
				// Rest of the elements (need to normalize)
				int difference = length - toRemove - 2;
				for (int j = difference; j >= 0; j--) 
				{
					div = rem / Bell.subTreeSize[max - 1][j];
					rem = rem % Bell.subTreeSize[max - 1][j];
					if (div > max) 
					{
						// Adjust
						if (old2new[max] == -1) 
						{
							old2new[max] = current;
							result[length - j - 2] = current++;
						} 
						else
							result[length - j - 2] = old2new[max];
						rem = rem + (Bell.subTreeSize[max - 1][j] * (div - max));
						max++;
					} 
					else 
					{
						// No need to adjust
						if (old2new[div] == -1) 
						{
							old2new[div] = current;
							result[length - j - 2] = current++;
						} 
						else
							result[length - j - 2] = old2new[div];
						max = (max > div) ? max : ++max;
					}
				}
				return Bell.indexOf(result);
			}
		}
	}

	/** Information about blocks before removing the particular element. */
	private static byte[][][] blocksBeforeRemove;

	/**
	 * Computes the new index of the growth string after removing the particular element together with information about partitions. Method for
	 * practical use.
	 * 
	 * @param index of the growth string
	 * @param length before removal
	 * @param toRemove element (counting from 0)
	 * @return new index and if element wasn't the only one in the block -1, else number of other blocks.
	 * 
	 * @throws IllegalArgumentException if length is less than two or index of an element to remove is outside of a proper range or if growth string's
	 * index is not correct.
	 */
	public static int[] indexAndDataAfterRemove(int index, int length, int toRemove) {
		if (length < 2)
			throw new IllegalArgumentException("Wrong size of the partition.");
		else if ((toRemove > length - 1) || (toRemove < 0))
			throw new IllegalArgumentException("Wrong index of element to be removed (" + toRemove + " from " + length + ").");
		else if ((index < 0) || (index >= Bell.number(length)))
			throw new IllegalArgumentException("Wrong index of the partition.");
		else 
		{
			if ((Bell.indexesAfterRemove == null) || (Bell.blocksBeforeRemove == null))
				try 
			{
					Bell.initialize();
				
			} 
			catch (IOException e) 
			{
				throw new Error("Resource \"ca.ubc.cs.kisynski.Bell.data\" not available!");
			}
			if (length < Bell.indexesAfterRemove.length + 2) 
			{
				int[] result = new int[2];
				result[0] = Bell.indexesAfterRemove[length - 2][index][toRemove];
				result[1] = Bell.blocksBeforeRemove[length - 2][index][toRemove];
				return result;
			} 
			else 
			{
				int oldString[] = new int[length];
				int newString[] = new int[length - 1];
				int div;
				int rem = index;
				int max = 1;
				int current;
				int[] old2new = new int[length];
				if (toRemove > 0) 
				{
					// We are not removing the first element
					Arrays.fill(old2new, 1, length, -1);
					// Elements predecessing removed element
					int difference = length - toRemove - 1;
					for (int j = length - 2; j > difference; j--) 
					{
						div = rem / Bell.subTreeSize[max - 1][j];
						rem = rem % Bell.subTreeSize[max - 1][j];
						if (div > max) 
						{
							// Adjust
							oldString[length - j - 1] = max;
							newString[length - j - 1] = max;
							old2new[max] = max;
							rem = rem + (Bell.subTreeSize[max - 1][j] * (div - max));
							max++;
						} 
						else 
						{
							// No need to adjust
							oldString[length - j - 1] = div;
							newString[length - j - 1] = div;
							old2new[div] = div;
							max = (max > div) ? max : ++max;
						}
					}
					current = max;
					// Element to be removed
					div = rem / Bell.subTreeSize[max - 1][difference];
					rem = rem % Bell.subTreeSize[max - 1][difference];
					if (div > max) 
					{
						// Adjust
						oldString[toRemove] = max;
						rem = rem + (Bell.subTreeSize[max - 1][difference] * (div - max));
						max++;
					} 
					else 
					{
						// No need to adjust
						oldString[toRemove] = div;
						max = (max > div) ? max : ++max;
					}
				} 
				else 
				{
					// We are removing the first element
					current = 0;
					Arrays.fill(old2new, -1);
				}
				// Rest of the elements (need to normalize)
				int difference = length - toRemove - 2;
				for (int j = difference; j >= 0; j--) 
				{
					div = rem / Bell.subTreeSize[max - 1][j];
					rem = rem % Bell.subTreeSize[max - 1][j];
					if (div > max) 
					{
						// Adjust
						if (old2new[max] == -1) 
						{
							old2new[max] = current;
							newString[length - j - 2] = current++;
						} 
						else
							newString[length - j - 2] = old2new[max];
						oldString[length - j - 1] = max;
						rem = rem + (Bell.subTreeSize[max - 1][j] * (div - max));
						max++;
					} 
					else 
					{
						// No need to adjust
						if (old2new[div] == -1) 
						{
							old2new[div] = current;
							newString[length - j - 2] = current++;
						} 
						else
							newString[length - j - 2] = old2new[div];
						oldString[length - j - 1] = div;
						max = (max > div) ? max : ++max;
					}
				}
				int[] result = new int[2];
				result[0] = Bell.indexOf(newString);
				// Collect data about blocks from the old string
				int nrPartitions = 0;
				for (int i = 0; i < oldString.length; i++)
					if ((oldString[i] == oldString[toRemove]) && (i != toRemove)) 
					{
						// Not a single element in the block
						result[1] = -1;
						break;
					} 
					else 
					{
						nrPartitions = (nrPartitions < oldString[i]) ? oldString[i] : nrPartitions;
					}
				if (result[1] != -1)
					// Single element in the block
					result[1] = nrPartitions;
				return result;
			}
		}
	}

	/**
	 * Normalizes restricted growth string after removal or insertion of some elements.
	 * 
	 * Method works in linear time with respect to the original string length. It uses additional array of the size max(string[]) + 1.
	 * 
	 * @param string to be normalized.
	 * 
	 * @throws IllegalArgumentException if string contains negative numbers.
	 */
	public static void normalizeString(int[] string) 
	{
		int max = 0;
		for (int i = 0; i < string.length; i++)
			if (string[i] < 0)
				throw new IllegalArgumentException("Not a proper growth string.");
			else
				max = (max > string[i]) ? max : string[i];
		int[] old2new = new int[max + 1];
		Arrays.fill(old2new, -1);
		int current = 0;
		for (int i = 0; i < string.length; i++)
			if (old2new[string[i]] == -1) 
			{
				old2new[string[i]] = current;
				string[i] = current++;
			} 
			else
				string[i] = old2new[string[i]];
	}

	/**
	 * Computes the new index of the growth string after removing the particular element.
	 * 
	 * @param index of the growth string
	 * @param length before removal
	 * @param toRemove element (counting from 0)
	 * @return new index
	 * 
	 * Method calls bigGrowthString(), normalizeString() and bigIndexOf(). They all could be combine into one, but I don't think that cost savings are
	 * worth doing this.
	 * 
	 * @throws IllegalArgumentException if length is less than two or index of an element to remove is outside of a proper range or if growth string's
	 * index is not correct..
	 */
	public static BigInteger bigIndexAfterRemove(BigInteger index, int length, int toRemove) {
		if ((length < Bell.bellNumbers.length) && index.compareTo(new BigInteger(Integer.toString(Integer.MAX_VALUE))) <= 0)
			return new BigInteger(Integer.toString(Bell.indexAfterRemove(index.intValue(), length, toRemove)));
		else if (length == 1)
			throw new IllegalArgumentException("Can not compute index of empty string.");
		else if ((toRemove > length - 1) || (toRemove < 0))
			throw new IllegalArgumentException("Wrong index of element to be removed.");
		else if ((index.compareTo(BigInteger.ZERO) < 0) || (index.compareTo(Bell.bigNumber(length)) >= 0))
			throw new IllegalArgumentException("Wrong index of the string.");
		else {
			int[] string = Bell.bigGrowthString(index, length);
			int[] removed = new int[string.length - 1];
			System.arraycopy(string, 0, removed, 0, toRemove);
			System.arraycopy(string, toRemove + 1, removed, toRemove, string.length - toRemove - 1);
			Bell.normalizeString(removed);
			return Bell.bigIndexOf(removed);
		}
	}

	/**
	 * Computes the new index of the growth string after removing the particular element together with information about blocks.
	 * 
	 * @param index of the growth string
	 * @param length before removal
	 * @param toRemove element (counting from 0)
	 * @return new index and if element wasn't the only one in the block -1, else number of other blocks
	 * 
	 * Method calls bigGrowthString(), normalizeString() and bigIndexOf(). They all could be combine into one, but I don't think that cost savings are
	 * worth doing this.
	 * 
	 * @throws IllegalArgumentException if length is less than two or index of an element to remove is outside of a proper range or if growth string's
	 * index is not correct.
	 */
	public static BigInteger[] bigIndexAndDataAfterRemove(BigInteger index, int length, int toRemove) {
		if ((length < Bell.bellNumbers.length) && index.compareTo(new BigInteger(Integer.toString(Integer.MAX_VALUE))) <= 0) {
			int[] data = Bell.indexAndDataAfterRemove(index.intValue(), length, toRemove);
			BigInteger[] result = new BigInteger[2];
			result[0] = new BigInteger(Integer.toString(data[0]));
			result[1] = new BigInteger(Integer.toString(data[1]));
			return result;
		} 
		else if (length == 1)
			throw new IllegalArgumentException("Can not compute index of empty string.");
		else if ((toRemove > length - 1) || (toRemove < 0))
			throw new IllegalArgumentException("Wrong index of element to be removed.");
		else if ((index.compareTo(BigInteger.ZERO) < 0) || (index.compareTo(Bell.bigNumber(length)) >= 0))
			throw new IllegalArgumentException("Wrong index of the string.");
		else 
		{
			BigInteger[] result = new BigInteger[2];
			int[] string = Bell.bigGrowthString(index, length);
			// Collect data about blocks
			int max = 0;
			for (int i = 0; i < string.length; i++)
				if ((string[i] == string[toRemove]) && (i != toRemove)) 
				{
					// Not a single element in the block
					result[1] = BigInteger.ONE.negate();
					break;
				} 
				else 
				{
					max = (max < string[i]) ? string[i] : max;
				}
			if (result[1] == null)
				// Single element in the block
				result[1] = new BigInteger(String.valueOf(max));
			int[] removed = new int[string.length - 1];
			System.arraycopy(string, 0, removed, 0, toRemove);
			System.arraycopy(string, toRemove + 1, removed, toRemove, string.length - toRemove - 1);
			Bell.normalizeString(removed);
			result[0] = Bell.bigIndexOf(removed);
			return result;
		}
	}

	// Inserting elements
	// ------------------------------------------------------------------------------------------------

	/** New indexes of the growth string after inserting a new element. */
	private static short[][][][] indexesAfterInsert;

	/**
	 * Computes an array of new indexes of the growth string after inserting a new element at position insertionIndex. If resulting array has length
	 * n, than first (n -1) growth strings have (n-2) blocks and the n-th one has (n - 1) blocks. Method for practical use.
	 * 
	 * @param index of the growth string
	 * @param length before insertion
	 * @param insertionIndex of the new element (counting from 0)
	 * 
	 * @return new indexes
	 * 
	 * @throws IllegalArgumentException if insertion index is not within the right scop or length is less than zero or index is not consistent with
	 * the length.
	 */
	public static int[] indexesAfterInsert(int index, int length, int insertionIndex) {
		if ((insertionIndex > length) || (insertionIndex < 0))
			throw new IllegalArgumentException("Wrong index of element to be inserted.");
		else if ((index < 0) || (index >= Bell.number(length)))
			throw new IllegalArgumentException("Wrong index of the string.");
		else 
		{
			if (Bell.indexesAfterInsert == null)
				try 
			{
					Bell.initialize();
			} 
			catch (IOException e) 
			{
					throw new Error("Resource \"ca.ubc.cs.kisynski.Bell.data\" not available!");
			}
			if (length < Bell.indexesAfterInsert.length) 
			{
				int[] result = new int[Bell.indexesAfterInsert[length][index][insertionIndex].length];
				for (int i = 0; i < Bell.indexesAfterInsert[length][index][insertionIndex].length; i++)
					result[i] = Bell.indexesAfterInsert[length][index][insertionIndex][i];
				return result;
			}
			// if (length == 0)
			// return new int[1];
			int[] string = Bell.growthString(index, length);
			// Compute sum of sizes of subtrees before insert
			int[] result;
			if (insertionIndex == 0) 
			{
				// We are inserting before the first element ---------------------------
				int max = 0;
				for (int i = 1; i < length; i++)
					if (max < string[i])
						max++;
				result = new int[max + 2];
				// Normalize and compute contribution from elements after insertion
				for (int i = 0; i < result.length; i++) 
				{
					int tempMax = 0;
					int[] old2new = new int[result.length];
					Arrays.fill(old2new, 0, result.length, -1);
					old2new[i] = 0;
					int current = 1;
					for (int j = insertionIndex; j < length; j++) 
					{
						int element = string[j];
						if (old2new[element] == -1) 
						{
							old2new[element] = current;
							element = current++;
						} 
						else
							element = old2new[element];
						result[i] = result[i] + (Bell.subTreeSize[tempMax][length - j - 1] * element);
						if (tempMax < element)
							tempMax++;
					}
				}
			} 
			else 
			{
				// We are inserting after some element ---------------------------------
				int maxBefore = 0;
				int sumBefore = 0;
				for (int i = 1; i < insertionIndex; i++) 
				{
					sumBefore = sumBefore + (Bell.subTreeSize[maxBefore][length - i] * string[i]);
					if (maxBefore < string[i])
						maxBefore++;
				}
				// Compute sum of sizes of subtrees after insert
				int max = maxBefore;
				int sumAfter = 0;
				for (int i = insertionIndex; i < length; i++) 
				{
					sumAfter = sumAfter + (Bell.subTreeSize[max][length - i - 1] * string[i]);
					if (max < string[i])
						max++;
				}
				result = new int[max + 2];
				// (1) Inserted element belongs to blocks represented by 0,...,maxBefore
				int subTreeSize = Bell.subTreeSize[maxBefore][length - insertionIndex];
				for (int i = 0; i <= maxBefore; i++)
					result[i] = sumBefore + (subTreeSize * i) + sumAfter;
				// (2) Inserted element belongs to blocks represented by (maxBefore + 1),...,(max + 1), we will need to normalize
				maxBefore++;
				Arrays.fill(result, maxBefore, result.length, sumBefore + (Bell.subTreeSize[maxBefore - 1][length - insertionIndex] * (maxBefore)));
				// Normalize and compute contribution from elements after insertion
				for (int i = maxBefore; i < result.length; i++) 
				{
					int tempMax = maxBefore;
					int[] old2new = new int[result.length];
					Arrays.fill(old2new, maxBefore, result.length, -1);
					old2new[i] = maxBefore;
					int current = maxBefore + 1;
					for (int j = insertionIndex; j < length; j++) 
					{
						int element = string[j];
						if (element >= maxBefore) {
							if (old2new[element] == -1) 
							{
								old2new[element] = current;
								element = current++;
							} 
							else
								element = old2new[element];
						}
						result[i] = result[i] + (Bell.subTreeSize[tempMax][length - j - 1] * element);
						if (tempMax < element)
							tempMax++;
					}
				}
			}
			return result;
		}
	}

	/**
	 * Computes an array of new indexes of the growth string after inserting a new element at position insertionIndex. If resulting array has length
	 * n, than first (n -1) growth strings have (n-2) blocks and the n-th one has (n - 1) blocks.
	 * 
	 * @param index of the growth string
	 * @param length before insertion
	 * @param insertionIndex of the new element (counting from 0)
	 * 
	 * @return new indexes
	 * 
	 * Method calls bigGrowthString(), normalizeString() and bigIndexOf(). They all could be combine into one, but I don't think that cost savings are
	 * worth doing this.
	 * 
	 * @throws IllegalArgumentException if insertion index is not within the right scop or length is less than zero or index is not consistent with
	 * the length.
	 */
	public static BigInteger[] bigIndexesAfterInsert(BigInteger index, int length, int insertionIndex) {
		if ((length < (Bell.bellNumbers.length - 1)) && index.compareTo(new BigInteger(Integer.toString(Integer.MAX_VALUE))) <= 0) {
			int[] data = Bell.indexesAfterInsert(index.intValue(), length, insertionIndex);
			BigInteger[] result = new BigInteger[data.length];
			for (int i = 0; i < result.length; i++)
				result[i] = new BigInteger(Integer.toString(data[i]));
			return result;
		} 
		else if ((insertionIndex > length) || (insertionIndex < 0))
			throw new IllegalArgumentException("Wrong index of element to be inserted.");
		else if ((index.compareTo(BigInteger.ZERO) < 0) || (index.compareTo(Bell.bigNumber(length)) >= 0))
			throw new IllegalArgumentException("Wrong index of the string.");
		else if (length == 0) 
		{
			BigInteger[] result = new BigInteger[1];
			result[0] = BigInteger.ZERO;
			return result;
		} 
		else 
		{
			int[] string = Bell.bigGrowthString(index, length);
			int max = 0;
			for (int i = 0; i < string.length; i++)
				max = (max < string[i]) ? string[i] : max;
			int[] inserted = new int[string.length + 1];
			System.arraycopy(string, 0, inserted, 0, insertionIndex);
			System.arraycopy(string, insertionIndex, inserted, insertionIndex + 1, string.length - insertionIndex);
			BigInteger[] result = new BigInteger[max + 2];
			for (int i = 0; i <= max + 1; i++) 
			{
				inserted[insertionIndex] = i;
				Bell.normalizeString(inserted);
				result[i] = Bell.bigIndexOf(inserted);
			}
			return result;
		}
	}
	
	/*
	public static void main(String args[])
	{
		BigInteger searchSpace = Bell.bigNumber(500);
		System.out.println(searchSpace);
	}
	*/
}