//--------------------------------------------------
// HybridSearch.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.QueryMeshOptimizer.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;


public class HybridSearch 
{
	//----------------------------------------------------------------------------------------------------
	
	Hashtable CACHED_ROUTES = new Hashtable();
	
	//----------------------------------------------------------------------------------------------------
	
    public enum QM_SEARCH_TYPE 
    {
    	QM_SEARCH_FULL_ENUMERATION,
    	QM_SEARCH_HYBRID,
    	QM_SEARCH_PURE_HEURISTIC
    }

    //----------------------------------------------------------------------------------------------------
    
    private QM_SEARCH_TYPE SEARCH_TYPE;    
    private QM_SEARCH_START_SOLUTION START_SOLUTION;
    
	private ArrayList minTrainingTuplePartitions;
	private LocalQueryMesh minQM;
	private double minQMCost = Double.MAX_VALUE;
	private LocalQueryMesh currentQM;
	private double currentQMCost;
		
	Hashtable hashClassifierRoutesMapping = null;
	DecisionTreeID3 currDT = null;
	ArrayList currRoutes = new ArrayList();
		
	//----------------------------------------------------------------------------------------------------
	
	public HybridSearch()
	{
		hashClassifierRoutesMapping = new Hashtable();
		minTrainingTuplePartitions = new ArrayList();
		
		//initialize cost as positive infinity
		minQMCost = Double.POSITIVE_INFINITY;
		currentQMCost = Double.POSITIVE_INFINITY;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	//------------------------------------------------------------------
	//  QUERY MESH HYBRID SEARCH REPRESENTS THE ACTUAL SEARCH FRAMEWORK
	//  FOR QUERY MESH
	//------------------------------------------------------------------
	//  SearchQm(m, n, Heuristic)
	

	SimpleQueryPlanSpecification defaultQueryPlan;
	SimpleQueryPlanSpecification queryPlan;
	DecisionTreeID3 dt;
	double totalQueryMeshCost = 0.0;
	int numTrainingTuples;
	S<String> s;
	//ArrayList allPartitions;
	
	//----------------------------------------------------------------------------------------------------
	//--------------------------------------------------------------
	// if m = 1 (full heuristic),
	// we pick a starting point in the middle of the solution space
	//
	// if m = n (full enumeration)
	// if m = k (first, enumerate, up to k subsets and then employ heuristc)
	//--------------------------------------------------------------
	
	public LocalQueryMesh runSearchQM(QM_SEARCH_STRATEGY heuristic,
									  QM_SEARCH_START_SOLUTION startSolution,
									  SimpleOperatorInputSpecification input, 
									  ArrayList trainingTupleSet, 
									  ArrayList partitionSpace, 
									  List xmlOperators, 
									  Hashtable xmlInputs,
									  int m) // n = cardinality of training tuple set
	{
		//initialize query mesh for the current stream
		LocalQueryMesh streamQM = new LocalQueryMesh(input.getId());
		
		//optimizer start solution
		START_SOLUTION = startSolution;
		
		if (m == 0)
		{
			SEARCH_TYPE = QM_SEARCH_TYPE.QM_SEARCH_PURE_HEURISTIC;
		}
		else if (m == -1) //-1 is a flag for full enum
		{
			SEARCH_TYPE = QM_SEARCH_TYPE.QM_SEARCH_FULL_ENUMERATION;
		}
		else
		{
			SEARCH_TYPE = QM_SEARCH_TYPE.QM_SEARCH_HYBRID;
		}
		
		//----------------------------------------------------------
		//                     FULL ENUMERATION
		//----------------------------------------------------------
	
		if (SEARCH_TYPE == QM_SEARCH_TYPE.QM_SEARCH_FULL_ENUMERATION)
		{
			ExhaustiveSearch es = new ExhaustiveSearch();
    		streamQM = es.runExhaustive(input, input.getTrainingTuples(), partitionSpace, xmlOperators, xmlInputs);        	
		}
		else 
		//----------------------------------------------------------
		//              PURE HEURISTIC ENUMERATION
		//----------------------------------------------------------		
		if (SEARCH_TYPE == QM_SEARCH_TYPE.QM_SEARCH_PURE_HEURISTIC)
		{
		    if (heuristic == QM_SEARCH_STRATEGY.ITERATIVE_IMPROVEMENT)
		    {
		    	// pure II algorithm
		    	IterativeImprovementSearch ii = new IterativeImprovementSearch();
		    	boolean isRandomStartSolution = false;
            	streamQM = ii.runII(START_SOLUTION,
            						input, 
            						input.getTrainingTuples(),
            						partitionSpace, 
            						xmlOperators, 
            						xmlInputs,
            						isRandomStartSolution);            	
		    }
		    else if (heuristic == QM_SEARCH_STRATEGY.SIMULATED_ANNEALING)
		    {
		    	// pure SA algorithm
		    	SimulatedAnnealingSearch sa = new SimulatedAnnealingSearch();
		    	boolean isRandomStartSolution = false;
		    	
		    	System.out.println("-------------------------------------------------");
		    	System.out.println("---             New Stream QM Search          ---");
		    	System.out.println("-------------------------------------------------");
            	streamQM = sa.runSA(START_SOLUTION,
            						input, 
            						input.getTrainingTuples(),
            						partitionSpace, 
            						xmlOperators, 
            						xmlInputs,
            						isRandomStartSolution);            	
		    }
		}
		else
		{
			//----------------------------------------------------------
			//                     HYBRID SEARCH
			//----------------------------------------------------------		
					
			//first do enumeation on subsets of size up to m, and then finish up
			//with a search heuristic 
			
			//first partial exhaustive
			ExhaustiveSearch es = new ExhaustiveSearch();
			streamQM = es.runPartialExhaustive(input, input.getTrainingTuples(),partitionSpace, xmlOperators, xmlInputs, m, CACHED_ROUTES);
		
		
			ArrayList partitions = new ArrayList();
			Hashtable hashTrainingTupleMapping = new Hashtable();
		
		try
		{		
			for(int q = 0; q < streamQM.getRoutes().size(); q++)
			{
				
				RouteSignature routeSignature = (RouteSignature)streamQM.getRoutes().get(q);
				QueryPlan currRoute = (QueryPlan)routeSignature.getQueryPlan();
				String rusterName = routeSignature.getRuster();
				rusterName = rusterName.replaceAll("ruster", "");
				rusterName = rusterName.substring(1, rusterName.length()-1); //remove [ and ]
				rusterName = rusterName.replaceAll(" ", "");
								
				String pid = "" + q;
				S currPartition = new S();
				
				String [] temp = null;
			    temp = rusterName.split(",");
			    for (int h=0; h < temp.length; h++)
			    {
			    	currPartition.add(temp[h]);
			    }
			      
				hashTrainingTupleMapping.put(pid,currPartition);
			}
			
			System.out.println("------------------------------------------------");
			
			Object [] currSubsets = new Object[hashTrainingTupleMapping.size()];
			System.out.print("Current partition: ");
			int count = 0;
			java.util.Enumeration en = hashTrainingTupleMapping.elements(); 
			while (en.hasMoreElements()) 
			{
				S currSubset = (S)en.nextElement();
				currSubsets[count++] = currSubset; 
				System.out.println(currSubset); 
			}
			System.out.println();
			System.out.println("------------------------------------------------");
	
			//-----------------------------------------------------------------
			//         2. DO DT AND ROUTES; PROCESSING
			//-----------------------------------------------------------------						

			/*
			currentQMCost = ComputeDTAndRoutes(currSubsets, 
											   input, 
											   xmlOperators, 
											   xmlInputs); //routes computed
			*/		
			//-----------------------------------------------------------------
			//         3. IF CURRENT QM HAS THE SMALLES COST, STORE IT
			//-----------------------------------------------------------------						
			/*
			if (currentQMCost < minQMCost)
			{
				streamQM.setDTClassifier(currDT);
				streamQM.setRoutes(currRoutes);
				streamQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);
				minQMCost = currentQMCost;
				minQM = streamQM;
			}
		    */
			//-----------------------------------------------------------------
			//               4. RUN UNTIL STOP CONDITION
			//-----------------------------------------------------------------
			int stopCondition = 10; //10 iterations is the STOPPING CONDITION
			int i = 0;
		
			while (i < stopCondition)
			{
				//-----------------------------------------------------------------
				//      5. PICK A RANDOM NEIGHBOR OF THE CURRENT SOLUTION
				//-----------------------------------------------------------------
				currSubsets = GetRandomNeighbor(hashTrainingTupleMapping);
							
				//-----------------------------------------------------------------
				//      6. ESTIMATE THE COST OF THE RANDOMLY PICKED NEIGHBOR
				//-----------------------------------------------------------------
				currDT = null;
				currRoutes = new ArrayList();
				currentQMCost = ComputeDTAndRoutes(currSubsets, 
												   input, 
												   xmlOperators, 
												   xmlInputs); //routes computed
								
				//-----------------------------------------------------------------
				//      7. UPDATE MIN SOLUTIONS BASED ON COSTS ESTIMATES
				//-----------------------------------------------------------------
				if (currentQMCost < minQMCost)
				{
					streamQM.setDTClassifier(currDT);
					streamQM.setRoutes(currRoutes);
					streamQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);					
					minQM = currentQM;
				}
				
				//increment num of iterations
				i++;
			}			
		}
		
		catch(Exception exp)
		{
			exp.printStackTrace();
		}
		}
		//--------------------------------------------------------------
		// LAST STEP: COMPUTE DEFAULT ROUTE FOR THE LOCAL QM STREAM
		//--------------------------------------------------------------
		ComputeDefaultRoute(input, 
				   			xmlOperators, 
				   			xmlInputs); //routes computed
		
		streamQM.setDefaultRoute(new RouteSignature(defaultQueryPlan.GetPlan(input.getId()), "default_route"));
		System.out.println("Default route: " + streamQM.getDefaultRoute());
		return streamQM;
	}
	
	//----------------------------------------------------------------------------------------------------

	public double ComputeDTAndRoutes(Object [] currSubsets, 
									 SimpleOperatorInputSpecification input,									 
									 List xmlOperators, 
									 Hashtable xmlInputs)
	{
		//reset mapping between DT leaves and the routes
		hashClassifierRoutesMapping = new Hashtable();
		ArrayList trainingTuples = input.getTrainingTuples();
		numTrainingTuples = trainingTuples.size() - 1;
		//---------------------------------------------------------------
	    // CREATE MAPPING BETWEEN TRAINING TUPLES AND SUBSETS IN THE CURRENT
		// PARTITION
		//---------------------------------------------------------------
		Hashtable trainingTuplesToSubsetsMapping = new Hashtable();
		for(int p=1; p <= numTrainingTuples; p++)
		{
			String trainingTupleId = "" + p;
			for(int i=0; i < currSubsets.length; i++)
			{
				if (currSubsets[i].toString().contains(trainingTupleId))
				{
					//create mapping between training tuple id and the subset 
					String partitionName = "ruster"+currSubsets[i].toString();
					trainingTuplesToSubsetsMapping.put(trainingTupleId, partitionName);
					break;
				}
			}
		}
		//-----------------------------------------------------------------
		//         2. ESTIMATE THE COST FOR THE START SOLUTION
		//-----------------------------------------------------------------
		
			//compute decision tree where target attribute is the partition of the tuple
			String inputFileName = "C:\\QueryMesh\\tmp\\qmInput"+input.getId()+".txt";
			String outputFileName = "C:\\QueryMesh\\tmp\\qmOutput"+input.getId()+".txt";

			for (int i=0; i<trainingTuples.size(); i++)
			{
				String[] stringTokens = (String[])trainingTuples.get(i);
				String stringToWrite = "";
				for(int j=0; j < stringTokens.length; j++)
				{   
					if (j == stringTokens.length - 1)
					{
						if (i > 0)
						{
	    					//retrieve the ruster (partition) name for the current training tuple
							String key = ""+i;
	    					String rusterName = (String)trainingTuplesToSubsetsMapping.get(key);
	    					//append the ruster as the last attribute (it will be a target attribute
	    					//in the decision tree construction
	    					stringToWrite += stringTokens[j] + "  "+ rusterName;
						}
						else
						{
							stringToWrite += stringTokens[j] + "  " + "RusterName";
						}
					}
					else
					{
						stringToWrite += stringTokens[j] + "  ";
					}
				}        			
				
				if(i == 0)
				{
					//delete sample dump (from previous runs)
					java.io.File file = new java.io.File (inputFileName);
			        file.delete ();
			        
					DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);
					DecisionTreeID3.WriteToDTDumpFile(inputFileName, "//**************************************");
				}
				else
				{
					DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);
				}
			}//end for
		
			//--------------------------------------------------------------------
			// COMPUTE DECISION TREE COST
			//--------------------------------------------------------------------
			dt = null;
			dt = DecisionTreeID3.optimizerCreateDT(input, inputFileName, outputFileName);
			double dtCost = dt.getDTCost(input.getStatisticsFileName());
	
			//set the current dt to the newly computed dt
			currDT = dt;
			
			//--------------------------------------------------------------------
			// COMPUTE ROUTES' COST
			//--------------------------------------------------------------------
			
			ArrayList inputs;
			double selectivity;
			Hashtable hashPreparedInputs = new Hashtable();
	
			Enumeration e = xmlInputs.elements();
			while (e.hasMoreElements()) 
			{  
				SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  // do whatever you need with x
				if (currInput != input)
				{
					//only 'prepare' input if it is NOT the current input operator
					
					//compute cardinality of the input
					double currCardinality = currInput.getCardinality();
					OptimizerOpInput preparedInput = new OptimizerOpInput(currInput.getId(), currCardinality);
					hashPreparedInputs.put(currInput.getId(), preparedInput);
				}
			}
			
			//------------------------------------------------------------------
			// ITERATE THROUGH POSSIBLE RUSTERS OF THE CURRENT INPUT STREAM
			// AND COMPUTE THE ROUTES' COSTS
			//------------------------------------------------------------------
			
			double totalRoutesCost = 0;
			for(int k = 0; k < currSubsets.length; k++)
			{
	    		//now (for every leaf of the decision tree) of the current input 
	    		//compute route cost
				String partitionName = "ruster"+currSubsets[k].toString();
								
				QueryPlan currPlan;
				//if route is NOT cached
				if (!CACHED_ROUTES.containsKey(partitionName))
				{				
					int currCardinality = input.computeCardinalityOfPartition(partitionName, outputFileName);
					OptimizerOpInput preparedInput = new OptimizerOpInput(input.getId(), currCardinality);
					hashPreparedInputs.put(input.getId(), preparedInput);
				
		    		//------------------------------------------------------------------
		    		// ALL INPUTS ARE PREPARED
		    		//------------------------------------------------------------------
		
					ArrayList partitionTuples = input.getPartitionTuples((S)currSubsets[k]);
					
		    		//------------------------------------------------------------------
		    		// NOW PREPARE OPERATORS
		    		//------------------------------------------------------------------
					ArrayList allOperators = new ArrayList();
		    		for(int b = 0; b < xmlOperators.size(); b++)
		    		{
		    			SimpleOperatorSpecification currOp = (SimpleOperatorSpecification)xmlOperators.get(b);
		    			ArrayList currOpInputs = new ArrayList();
		    			
		    			boolean containsMeshInput = false; //true if input to operator is the stream for which we are computing query mesh
		    			
		    			//assert 2 inputs
		    			assert currOpInputs.size() == 2;
		    			
		    			//-------------------------------------------------------------
		    			// 1st OP INPUT
		    			//-------------------------------------------------------------
		    			boolean isPartitionBasedEstimation1 = false;
		    			boolean isPartitionBasedEstimation2 = false;
		    			String inputId1 = (String)currOp.getInputs().get(0);
		    			ArrayList op1PartitionTuples = new ArrayList();
		    			if (inputId1.equalsIgnoreCase(input.getId()))
		    			{
		    				isPartitionBasedEstimation1 = true;
		    				op1PartitionTuples = partitionTuples;
		    			}	        				
		    			//retrieve op by id
		    			OptimizerOpInput inp1 = (OptimizerOpInput)hashPreparedInputs.get(inputId1);
		    			SimpleOperatorInputSpecification input1 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId1);						
		    			int joinAttributeIdx1 = currOp.getJoinAttributeIdxByInputId(inputId1);
		    			currOpInputs.add(inp1);
		    			//-------------------------------------------------------------
		    			// 2nd OP INPUT
		    			//-------------------------------------------------------------	        			
		    			String inputId2 = (String)currOp.getInputs().get(1);
		    			ArrayList op2PartitionTuples = new ArrayList();
		    			if (inputId2.equalsIgnoreCase(input.getId()))
		    			{
		    				isPartitionBasedEstimation2 = true;
		    				op2PartitionTuples = partitionTuples;
		    			}	        				
		    			//retrieve op by id
		    			OptimizerOpInput inp2 = (OptimizerOpInput)hashPreparedInputs.get(inputId2);
		    			SimpleOperatorInputSpecification input2 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId2);
		    			int joinAttributeIdx2 = currOp.getJoinAttributeIdxByInputId(inputId2);	        			
		    			currOpInputs.add(inp2);
		    			
		    			//now estimate selectivity of the operator
		    				        			
		    			double currOpSelectivity = -999;
		    			
		    			
		    			JoinSelectivityEstimator selectivityEstimator = new JoinSelectivityEstimator(
																				        			input1,
																				        			isPartitionBasedEstimation1, 									
																				        			joinAttributeIdx1,
																				        			op1PartitionTuples,
																									input2, 
																									isPartitionBasedEstimation2,
																									joinAttributeIdx2,
																									op2PartitionTuples,
																									null);
		    			
		    			currOpSelectivity = selectivityEstimator.getSelectivity();	        				
		    			OptimizerOp op = new OptimizerOp(currOp.toString(), "join", currOpInputs, currOpSelectivity);
		    			
		    			allOperators.add(op);
		    		}//end of operator iterations
		    		
		    		queryPlan = new SimpleQueryPlanSpecification(allOperators);
		    		
		    		//get the plan
		    		currPlan = queryPlan.GetPlan(input.getId());    			    			    		
		    		//---------------------------------------------------------
		    		//             CACHE THE ROUTE FOR THE FUTURE
		    		//---------------------------------------------------------
		    		CACHED_ROUTES.put(partitionName, currPlan);
		    		
				}
				else
				{
					//retrieve route from cache
					currPlan = (QueryPlan)CACHED_ROUTES.get(partitionName);
				}
				
				double routeCost =  currPlan.getCost();
	    		System.out.println("Cost :"+ routeCost);	
    		
	    		//add the plan to the curr routes list
	    		currRoutes.add(new RouteSignature(currPlan, partitionName));
	    		
	    		totalRoutesCost += routeCost;
			}//end of subsest iterations
			
			totalQueryMeshCost = dtCost + totalRoutesCost;
			System.out.println("Curent query mesh cost: " + totalQueryMeshCost);
			
			return totalQueryMeshCost;
	}

	//----------------------------------------------------------------------------------------------------
	
	public void ComputeDefaultRoute(SimpleOperatorInputSpecification input,									 
									List xmlOperators, 
									Hashtable xmlInputs)
	{
		
		//now (for every leaf of the decision tree) of the current input 
		//compute route cost
		int currCardinality = input.computeCardinalityFromStatsFile();
		OptimizerOpInput preparedInput = new OptimizerOpInput(input.getId(), currCardinality);
				
		//------------------------------------------------------------------
		// NOW PREPARE OPERATORS
		//------------------------------------------------------------------
		ArrayList allOperators = new ArrayList();
		for(int b = 0; b < xmlOperators.size(); b++)
		{
			SimpleOperatorSpecification currOp = (SimpleOperatorSpecification)xmlOperators.get(b);
			ArrayList currOpInputs = new ArrayList();
			
			boolean containsMeshInput = false; //true if input to operator is the stream for which we are computing query mesh
			
			//assert 2 inputs
			assert currOpInputs.size() == 2;
			
			//-------------------------------------------------------------
			// 1st OP INPUT
			//-------------------------------------------------------------
			boolean isPartitionBasedEstimation1 = false;
			boolean isPartitionBasedEstimation2 = false;
			String inputId1 = (String)currOp.getInputs().get(0);
			ArrayList op1PartitionTuples = new ArrayList();
							        				
			//retrieve op by id
			OptimizerOpInput inp1 = new OptimizerOpInput(inputId1,currCardinality);
			SimpleOperatorInputSpecification input1 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId1);						
			int joinAttributeIdx1 = currOp.getJoinAttributeIdxByInputId(inputId1);
			currOpInputs.add(inp1);
			//-------------------------------------------------------------
			// 2nd OP INPUT
			//-------------------------------------------------------------	        			
			String inputId2 = (String)currOp.getInputs().get(1);
			ArrayList op2PartitionTuples = new ArrayList();
				        				
			//retrieve op by id
			OptimizerOpInput inp2 = new OptimizerOpInput(inputId2,currCardinality);
			SimpleOperatorInputSpecification input2 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId2);
			int joinAttributeIdx2 = currOp.getJoinAttributeIdxByInputId(inputId2);	        			
			currOpInputs.add(inp2);
			
			//now estimate selectivity of the operator
				        			
			double currOpSelectivity = -999;
			
			
			JoinSelectivityEstimator selectivityEstimator = new JoinSelectivityEstimator(input1,
																	        			 isPartitionBasedEstimation1, 									
																	        			 joinAttributeIdx1,
																	        			 op1PartitionTuples,
																						 input2, 
																						 isPartitionBasedEstimation2,
																						 joinAttributeIdx2,
																						 op2PartitionTuples,
																						 null);
			
			currOpSelectivity = selectivityEstimator.getSelectivity();	        				
			OptimizerOp op = new OptimizerOp(currOp.toString(), "join", currOpInputs, currOpSelectivity);
			
			allOperators.add(op);
		}//end of operator iterations

		defaultQueryPlan = new SimpleQueryPlanSpecification(allOperators);
				
		//get the plan	    		
		QueryPlan currPlan = defaultQueryPlan.GetPlan(input.getId());    		
		double routeCost =  currPlan.getCost();
		System.out.println("Default route cost :"+ routeCost);	
	}	
	
	//----------------------------------------------------------------------------------------------------
	
	private Object [] GetRandomNeighbor(Hashtable hashTrainingTupleMapping)
	{
		Object [] currSubsets = null;
		
		currSubsets = HybridSearch.GetRandomNeighborMergeTwoSubsets(hashTrainingTupleMapping);
		
		if (currSubsets == null)
		{
			currSubsets = HybridSearch.GetRandomNeighborSplitSubset(hashTrainingTupleMapping);
		}
		
		return currSubsets;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public static Object [] GetRandomNeighborSplitSubset(Hashtable hashTrainingTupleMapping)
	{
		//here the heuristic is to split first subset (that contains more than one entry
		int size = hashTrainingTupleMapping.size() + 1;
		
		Object [] currSubsets = new Object[size];

		//Hashtable mergedHashTrainingTupleMapping = new Hashtable();
		int count = 0;
		int idx = 0;
		java.util.Enumeration en = hashTrainingTupleMapping.keys(); 
		String absorbingID = "";
		S splitSubset1 = new S();
		S splitSubset2 = new S();
		
		if (hashTrainingTupleMapping.size() == 1)
		{
			//one route for the entire stream situation --> split it in 1/2
			while (en.hasMoreElements()) 
			{
				String key = (String)en.nextElement();
				char[] subsets = key.toCharArray();
				for (int p = 0; p < subsets.length; p++)
				{
					if ((subsets[p] == ',') || (subsets[p] == ' ') || (subsets[p] == '[') || (subsets[p] == ']'))
						continue;
					
					if (p < (subsets.length/2))
					{						
						splitSubset1.add(subsets[p]);
					}
					else
					{
						splitSubset2.add(subsets[p]);
					}	
				}								
			}
			
			currSubsets[0] = splitSubset1;
			currSubsets[1] = splitSubset2;
			//System.out.print(currSubsets); 
			
			hashTrainingTupleMapping.clear();
			hashTrainingTupleMapping.put(splitSubset1.toString(), splitSubset1);
			hashTrainingTupleMapping.put(splitSubset2.toString(), splitSubset2);
			return currSubsets;
		}
		
		//if set to true, that means, we have already split the subsets, so
		//the rest of the subsets are just added (without splitting them)
		boolean skipTheRest = false;
		while (en.hasMoreElements()) 
		{
			String key = (String)en.nextElement();
			S currSubset = (S)((S)hashTrainingTupleMapping.get(key)).clone();

			if (currSubset.size() < 2)
			{
				currSubsets[count] = currSubset;
				count++;	
				continue;
			}
			
			if (!skipTheRest)
			{
				S splittingSubset = (S)currSubsets[count];
	
				Object [] splittingSubsetArray = splittingSubset.toArray();
				for(int i=0; i < splittingSubsetArray.length; i++)
				{
					//assign 1/2 to the first split subset
					if (i < splittingSubsetArray.length/2)
					{
						splitSubset1.add(splittingSubsetArray[i]);
					}
					else
					{
						splitSubset2.add(splittingSubsetArray[i]);
					}
				}
				skipTheRest = true;
			}
			currSubsets[count] = currSubset;

			count++;			
		}
		//System.out.print(currSubsets); 
		return currSubsets;
	}

	//----------------------------------------------------------------------------------------------------
	
	public static Object [] GetRandomNeighborMergeTwoSubsets(Hashtable hashTrainingTupleMapping)
	{
		//here the heuristic is to merge first two neighboring subsets
		int size = hashTrainingTupleMapping.size() - 1;
		if (size < 1)
		{
			return null;
		}
		
		Object [] currSubsets = new Object[size];
		S subsetOne = null;
		S subsetTwo = null;
		
		//Hashtable mergedHashTrainingTupleMapping = new Hashtable();
		int count = 0;
		int idx = 0;
		java.util.Enumeration en = hashTrainingTupleMapping.keys(); 
		String absorbingID = "";
		while (en.hasMoreElements()) 
		{
			String key = (String)en.nextElement();
			S currSubset = (S)hashTrainingTupleMapping.get(key);

			if(count == 0)
			{
				subsetOne = currSubset;
			}
			else if(count == 1)
			{				
				subsetTwo = currSubset;
				S mergedSubset = new S();
				
				for (int k = 0; k < subsetOne.size(); k++)
				{					
					mergedSubset.add(subsetOne.toArray()[k]); //add element of the first subset
				}
				for (int k = 0; k < subsetTwo.size(); k++)
				{
					mergedSubset.add(subsetTwo.toArray()[k]); //add element of the first subset
				}
				currSubsets[idx] = mergedSubset;	
				idx++;
			}
			else
			{		
				currSubsets[idx] = currSubset;
				idx++;
			}
			//System.out.print(currSubset); 
			count++;			
		}
		return currSubsets;
	}
	
	//----------------------------------------------------------------------------------------------------
	
}
	
