/*
 * Backtracking.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;


import java.util.*;

/**
 * The <code>Backtracking</code> class implements an abstraction of classical
 * backtracking.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class Backtracking implements java.io.Serializable {
    
    
    /**
     * Reference to the solver.
     */
    SolverClass solver;
    
    /**
     * The stack of alternatives, contains choice points.
     */
    protected Stack alternatives;
    
    
    /**
     * Creates a new instance of Backtracking.
     *
     * @param solverInstance a solver instance.
     */
    public Backtracking(SolverClass solverInstance) {
        this.alternatives = new Stack();
        this.solver       = solverInstance;
    }
    
    
    /*
      Il metodo gestisciFallimento() estrae una nuova alternativa dallo stack mediante,
      la chiamata al metodo getAlternative(), riporta il constraint Solver.store e le
      variabili allo stato in cui erano quando si  aperta l'alternativa
      a cui si riferisce il fallimento, poi inizia la risoluzione dall'elemento dello
      Solver.store in cui si trova la nuova alternativa.
      Se al momento dell'estrazione dell'alternativa lo stack  vuoto, viene sollevata
      un'eccezione (del linguaggio Java) EmptyStackException, in questo caso non ci sono pi
      alternative da provare, quindi questa eccezione viene gestita sollevando un'eccezione
      Failure, cio la computazione fallisce.
     
      Il metodo gestisciFallimentoSetof()  simile al precedente, si differenzia solo nel
      catch, infatti in questo caso, se non ci sono pi alternative nello stack, significa
      che sono gi state determinate tutte le possibili soluzione.
     */
    /**
     * Finds a new alternative and restores the constraints store to a correct
     * state, variables have the same values as they had at the moment when the
     * choice was open.
     *
     * @exception  Failure  if there aren't alternatives or the constraint
     *                      relusolution fails.
     */
    // * @see JSetL.Backtracking#gestisciFallimentoSetof
    protected void gestisciFallimento() throws Failure {
        try {
            this.manageGeneralBktrkFail();
        }
        catch(EmptyStackException e) {
            throw new Failure();
        }
    }
    
    
    /**
     * Finds a new alternative and restores the constraints store to a correct
     * state, variables have the same values as they had at the moment when the
     * alternative was open.
     *
     * @exception  Failure  if the constraint relusolution fails.
     * @see     JSetL.Backtracking#gestisciFallimento
     */
    protected void gestisciFallimentoSetof() throws Failure {
        try {
            this.manageGeneralBktrkFail();
        }
        catch(EmptyStackException e) {
            /* All the alternatives were explored, no failure raised */
            return;
        }
    }
    
    
    /**
     * Common operation for the backtracking. Finds a new alternative
     * and restores a correct enviroment for it.
     *
     * @exception  EmptyStackException  if no alternatives are available.
     * @exception  Failure  if the constraint relusolution fails.
     * @see     JSetL.Backtracking#gestisciFallimento
     * @see     JSetL.Backtracking#gestisciFallimentoSetof
     */
    private void manageGeneralBktrkFail() throws EmptyStackException, Failure {
        ChoicePoint c = this.getAlternative();
        this.restoreVarState(c.varsState);
        this.restoreStoreState(c.storeState.statoStore);
        ((Constraint) c.storeState.statoStore.get(c.storeState.posizione)).caseControl = c.cont;
        this.solver.storeInvariato = false;
        return;
    }
    
    
    /**
     * Adds a choice point to the alternatives stack.
     *
     * @param i   next resolution rule ("nondeterministic" choice).
     * @param v   a VarState, variables state at the moment of choice point creation.
     * @param s   a StoreState, store state at the moment of choice point creation.
     */
    protected void addChoicePoint(int i, VarState v, StoreState s) {
        ChoicePoint newCP = new ChoicePoint(i, v, s);
        this.alternatives.push(newCP);
        return;
    }
    
    
    /* Aggiunge un punto di scelta allo stack delle alternative ed 
       richiamato dalla classe NewConstraints cio dalla classe che permette
       all'utente di definire nuovi vincoli non deterministici. */
    /**
     * Adds a choice point to the alternatives stack, used by the Newconstraints
     * class.
     *
     * @param s a user-defined constraint.
     */
    public void addChoicePoint(Constraint s) {
        VarState    stateVarIn   = getVarState();
        StoreState  stateStoreIn = getStoreState(s);
        this.addChoicePoint(s.caseControl + 1, stateVarIn, stateStoreIn);
    }
    
    
    /**
     * Exctract a choice point from the alternatives stack.
     *
     * @return  The object at the top of the alternative stack.
     * @exception  EmptyStackException if the alternatives stack is empty.
     */
    protected ChoicePoint getAlternative() throws EmptyStackException {
        return (ChoicePoint)this.alternatives.pop();
    }
    
    
    /**
     * Gets the number of choice point in the alternatives stack.
     *
     * @return  The size of of the alternative stack.
     * @exception  EmptyStackException if the alternatives stack is empty.
     */
    protected int getNumOfAlternative() throws EmptyStackException {
        return this.alternatives.size();
    }
    
    
    /**
     * Clone the alternatives stack.
     *
     * @return  A clone of the alternatives stack.
     * @exception  EmptyStackException if the alternatives stack is empty.
     */
    protected Stack cloneAlternative() throws EmptyStackException {
        Stack tmpStack = new Stack();
        for(int i = 0; i < this.alternatives.size(); i++) {
            ChoicePoint cp = new ChoicePoint();
            cp.cont = ((ChoicePoint)this.alternatives.get(i)).cont;
            cp.varsState = (VarState)((VarState)((ChoicePoint)this.alternatives.get(i)).varsState).cloneVarState();
            cp.storeState = (StoreState)((StoreState)((ChoicePoint)this.alternatives.get(i)).storeState).cloneStoreState();
            tmpStack.add(cp);
        }
        return tmpStack;
    }
    
    
    /* Restituisce un elemento di tipo StatoVar che ha come attributi tre Vettori:
       un vettore di Lvar, uno di Set e uno di Lst. In ciascuno di questi vettori
       vengono memorizzate le variabili che non sono ancora inizializzate, in questo modo
       se in seguito, durante il backtracking, sar necessario tornare ad uno stato
       precedente, le variabili che prima del fallimento che ha innescato il backtracking
       erano non inizializzate torneranno tali. */
    /**
     * Get an object with references to (actually) all uninitialized variables.
     *
     * @return  An object VarState that reference the uninitialized variables.
     */
    protected VarState getVarState() {
        Vector statoLvar = new Vector();
        for(int i = 0; i < (Lvar.nonInizLvar).size(); i++) {
            if(!((Lvar)Lvar.nonInizLvar.get(i)).iniz)
                statoLvar.add((Lvar)Lvar.nonInizLvar.elementAt(i));
        }
        Vector statoSet = new Vector();
        for(int i = 0; i < (Set.nonInizSet).size(); i++) {
            if(!((Set)Set.nonInizSet.get(i)).iniz)
                statoSet.add((Set)Set.nonInizSet.elementAt(i));
        }
        Vector statoLst = new Vector();
        for(int i = 0; i < (Lst.nonInizLst).size(); i++) {
            if(!((Lst)Lst.nonInizLst.get(i)).iniz)
                statoLst.add((Lst)Lst.nonInizLst.elementAt(i));
        }
        Vector statoMultiset = new Vector();
        for(int i = 0; i < (MultiSet.nonInizMultiSet).size(); i++) {
            if(!((MultiSet)MultiSet.nonInizMultiSet.get(i)).iniz)
                statoMultiset.add((MultiSet)MultiSet.nonInizMultiSet.elementAt(i));
        }
        //	VarState statoVar = new VarState(statoLvar,statoSet,statoLst);
        VarState statoVar = new VarState(statoLvar,statoSet,statoLst,statoMultiset);
        
        return statoVar;
    }
    
    
    /**
     * Gets the state of the constraint store.
     * @param se a constraint.
     * @return An object StoreState, the constraint store state.
     * @see JSetL.StoreState
     */
    protected StoreState getStoreState(Constraint se) {
        if(se.bool == true) {
            se.bool = false;
            ConstraintStore newStore = new ConstraintStore(this.solver.store.cloneStore());
            StoreState    statoStore = new StoreState(newStore, this.solver.indexOf(se), this.solver);
            se.bool = true;
            return statoStore;
        }
        else {
            ConstraintStore newStore = new ConstraintStore(this.solver.store.cloneStore());
            StoreState statoStore = new StoreState(newStore, this.solver.indexOf(se), this.solver);
            return statoStore;
        }
    }
    
    
    /**
     * Gets the state of the constraint store.
     * @return An object StoreState, the constraint store state.
     * @see JSetL.StoreState
     */
    protected StoreState getStoreState() {
        ConstraintStore newStore   = new ConstraintStore(this.solver.store.cloneStore());
        StoreState      statoStore = new StoreState(newStore, 0, this.solver);
        return statoStore;
    }
    
    
    /**
     * Restore variables to uninitialized value.
     *
     * @param sv a variables state stored.
     */
    protected void restoreVarState(VarState sv) {
        for(int i = 0; i < sv.lvarNotInit.size(); i++) {
            ((Lvar) sv.lvarNotInit.get(i)).iniz = false;
            ((Lvar) sv.lvarNotInit.get(i)).val = null;
            ((Lvar) sv.lvarNotInit.get(i)).equ = null;
        }
        for(int i = 0; i < sv.setNotInit.size(); i++) {
            ((Set) sv.setNotInit.get(i)).iniz = false;
            ((Set) sv.setNotInit.get(i)).lista = null;
            ((Set) sv.setNotInit.get(i)).resto = null;
            ((Set) sv.setNotInit.get(i)).equ = null;
            ((Set) sv.setNotInit.get(i)).isInt = false;
            ((Set) sv.setNotInit.get(i)).inter = null;
        }
        for(int i = 0; i < sv.lstNotInit.size(); i++) {
            ((Lst) sv.lstNotInit.get(i)).iniz = false;
            ((Lst) sv.lstNotInit.get(i)).lista = null;
            ((Lst) sv.lstNotInit.get(i)).resto = null;
            ((Lst) sv.lstNotInit.get(i)).equ = null;
        }
        for(int i = 0; i < sv.multisetNotInit.size(); i++) {
            ((MultiSet) sv.multisetNotInit.get(i)).iniz = false;
            ((MultiSet) sv.multisetNotInit.get(i)).lista = null;
            ((MultiSet) sv.multisetNotInit.get(i)).resto = null;
            ((MultiSet) sv.multisetNotInit.get(i)).equ = null;
        }
        return;
    }
    
    /**
     * Restore the constraint store.
     *
     * @param cs a constraint store.
     */
    protected void restoreStoreState(ConstraintStore cs) {
        this.solver.store = cs;
        return;
    }
    
    
    /* I metodi updateStoreAlternative() e updateVarAlternative() aggiornano
       lo stack delle alternative, cio aggiungono in tutti i punti di scelta le
       nuove variabili non inizializzate introdotte e aggiungono i nuovi vincoli
       introdotti nel Solver. In questo modo se, dopo aver creato un' alternativa
       vengono create nuove variabili e aggiunti nuovi vincoli, questi non
       vengono persi se si ritorna a quel punto di scelta. */
    /**
     * Update all the choice point in the alternatives stack adding constraints
     * defined after an alternative creation.
     */
    protected void updateStoreAlternative() {
        for(int i = 0; i < this.alternatives.size(); i++) {
            for(int k = this.solver.storeSize; k < this.solver.size(); k++) {
                Constraint se = new Constraint();
                se.arg1 = ((Constraint) this.solver.get(k)).arg1;
                se.arg2 = ((Constraint) this.solver.get(k)).arg2;
                se.arg3 = ((Constraint) this.solver.get(k)).arg3;
                se.arg4 = ((Constraint) this.solver.get(k)).arg4;
                se.bool = ((Constraint) this.solver.get(k)).bool;
                se.cons = (int)((Constraint) this.solver.get(k)).cons;
                se.caseControl = (int)((Constraint) this.solver.get(k)).caseControl;
                se.ctx = (int)((Constraint) this.solver.get(k)).ctx;
                ((ChoicePoint)this.alternatives.get(i)).storeState.statoStore.add(se);
            }
        }
        return;
    }
    
    
    /**
     * Update all the choice point in the alternatives stack adding uninitialized
     * variables defined after an alternative creation.
     */
    protected void updateVarAlternative() {
        for(int i = 0; i < this.alternatives.size(); i++) {
            for(int k = 0; k < (Lvar.nonInizLvar).size(); k++) {
                if(!((Lvar)Lvar.nonInizLvar.get(k)).iniz &&
                !((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).lvarNotInit).contains((Lvar)Lvar.nonInizLvar.get(k)))
                    ((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).lvarNotInit).add((Lvar)Lvar.nonInizLvar.get(k));
            }
            for(int h = 0; h < (Set.nonInizSet).size(); h++) {
                if(!((Set)Set.nonInizSet.get(h)).iniz &&
                !((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).setNotInit).contains((Set)Set.nonInizSet.get(h)))
                    ((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).setNotInit).add((Set)Set.nonInizSet.get(h));
            }
            for(int j = 0; j < (Lst.nonInizLst).size(); j++) {
                if(!((Lst)Lst.nonInizLst.get(j)).iniz &&
                !((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).lstNotInit).contains((Lst)Lst.nonInizLst.get(j)))
                    ((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).lstNotInit).add((Lst)Lst.nonInizLst.get(j));
            }
            for(int h = 0; h < (MultiSet.nonInizMultiSet).size(); h++) {
                if(!((MultiSet)MultiSet.nonInizMultiSet.get(h)).iniz &&
                !((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).multisetNotInit).contains((MultiSet)MultiSet.nonInizMultiSet.get(h)))
                    ((Vector)((VarState) ((ChoicePoint)this.alternatives.get(i)).varsState).multisetNotInit).add((MultiSet)MultiSet.nonInizMultiSet.get(h));
            }
        }
        return;
    }
    
    
    /**
     * Delete the last <tt>anInteger</tt> items from the alternatives stack.
     *
     * @param   anInteger  the number of item to cut.
     */
    public void cut(int anInteger) {
        for(int i = 0; i < anInteger; i++) {
            getAlternative();
        }
        return;
    }
    
}


