/*
 * ConstraintStore.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;


import java.util.*;

/**
 * The <code>ConstraintStore</code> class implements a growable vector 
 * of constraints.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class ConstraintStore extends ConstraintsConjunction implements java.io.Serializable { 

    
    /**
      * Vector for storing the constraints.
     */ 
    public  ConstraintsConjunction store = new ConstraintsConjunction();   

    /**
      * Reference to the solver.
     */ 
     private SolverClass solver;
  
    /**
      * Vector of all variables used in not forall constraints.
     */     
    protected Vector notForallVars = new Vector(); 
    
    /**
      * Vector of all variables used in forall constraints.
     */         
    protected Vector forallVars = new Vector(); 
    
    
    
    /**
     * Creates a new instance of ConstraintStore.
     *
     * @param sc a solver instance.
     */    
    public ConstraintStore(SolverClass sc) {
	ConstraintsConjunction newS = new ConstraintsConjunction();
	this.store  = newS;
	this.solver = sc;
    }
  
    /**
     * Creates a new instance of ConstraintStore as reference to another constraints store.
     *
     * @param cs a constraints store instance.
     */      
    public ConstraintStore(ConstraintStore cs) {
	this.store  = cs.store;
	this.solver = cs.solver;
    }


    /* Member Methods */
    
    /**
     * Adds a constraint at the (end of) constraints store.
     *
     * @param se a constraint.
     */
    public void add(Constraint se) {
	this.store.add(se);
	return;
    }


    /**
     * Adds a constraint to the constraints store at position i.
     *
     * @param i  an integer index where insert the constraint.
     * @param se a constraint.
     */
    public void add(int i, Constraint se) {
	this.store.add(i, se);
	return;
    }


    /**
     * Gets the index on the constraints store for the given constraint.
     *
     * @param se a constraint.
     * @return the index of the first occurrence of the object argument in the 
     *         constraints store, returns -1 if the object is not found.
     */
    public int indexOf(Constraint se) {
	return this.store.indexOf(se);
    }

    
    /**
     * Returns the number of components in the vector used for storeing the constraints.
     *
     * @return the number of components in the store vector.
     */
    public int size() {
	return this.store.size();
    }


    /**
     * Returns the constraint with index i in the constraints store.
     * @param i an integer index.
     * @return the constraint at index i.
     */    
    public Object get(int i) {
	return this.store.get(i);
    }

    
    /**
     * Gets the constraints store.
     * @return the constraint at index i.
     */    
    public ConstraintsConjunction getStore() {
	return this.store;
    }    
    

    /**
     * Removes all components from the constraint store.
     */    
    public void removeAllElements() {
	this.store.removeAllElements();
	return;
    }


    /**
     * Clone the constraints store.
     * 
     * @return a clone of the current constraints store. 
     */    
    protected ConstraintStore cloneStore() {
	ConstraintStore newStore = new ConstraintStore(this.solver);
        for (int i = 0; i < this.solver.store.size(); i++) {    
	    Constraint se = (Constraint)((Constraint)this.get(i)).cloneConstraint();
	    newStore.add(se);
	}
        newStore.forallVars    = this.forallVars;
        newStore.notForallVars = this.notForallVars;        
        
	return newStore;
    }
    
    
    /**
     * Checks if there are forall constraints.
     * 
     * @return true if exist at least one forall constraint on the store, false 
     * otherwise. 
     */   
    protected boolean existForall() {
        int exist = 0;
        
        for (int i = 0; i < size(); i++) {                        
            if (((Constraint) get(i)).cons == Enviroment.forallCode) exist++;
        }    
        if (exist > 0) return true;
        else return false;
    }


    /**
     * Finds all the variables in the constraints store and keep tracks of 
     * variables used in forall constraints and variables not used in forall 
     * constraints (respectively vectors forallVars and notForallVars).
     */      
    protected void getConstraintStoreVars() {   
        // for (int i = 0; i < this.size(); i++) {
        for (int i = 0; i < this.store.size(); i++) {
            // Se il vincolo != forall memorizzo le vars in un vector
            if (((Constraint) get(i)).cons != Enviroment.forallCode) {
                ((Constraint)get(i)).findVars(((Constraint) get(i)).arg1, this.notForallVars);
                ((Constraint)get(i)).findVars(((Constraint) get(i)).arg2, this.notForallVars);
                ((Constraint)get(i)).findVars(((Constraint) get(i)).arg3, this.notForallVars);
                ((Constraint)get(i)).findVars(((Constraint) get(i)).arg4, this.notForallVars);  
            }
            else {
                // trattamento delle vars di un forall
                VarsForall actualForall = new VarsForall();
                
                // Vars quantificate universalmente
                if(((Constraint) get(i)).arg1 instanceof Lvar) {
                    actualForall.quantVars.add((Lvar)((Constraint) get(i)).arg1);                    
                }
                else {
                    // Se non  un Lvar  per forza una Lst
                    ((Constraint)get(i)).findVars((Lst)((Constraint) get(i)).arg1, actualForall.quantVars);
                    
                }
                
                /* arg3 pu essere: una lista di Lvar (quant. esistenziale), un vincolo o 
                   un vettore di vincoli. */
                // Vars quantificate esistenzialmente
                if(((Constraint) get(i)).arg3 instanceof Lst) {
                    ((Constraint)get(i)).findVars(((Constraint) get(i)).arg3, actualForall.quantVars);
                     
                    if(((Constraint) get(i)).arg4 instanceof Constraint) {                   
                        ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg4).arg1, actualForall.bodyVars);
                        ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg4).arg2, actualForall.bodyVars);
                        ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg4).arg3, actualForall.bodyVars);
                        ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg4).arg4, actualForall.bodyVars);  
                    }
                    else {
                        // Devo trattare un vettore di vincoli
                        for(int j = 0; j < ((ConstraintsConjunction)((Constraint) get(i)).arg4).size(); j++) {
                            ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg4).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg4).get(j)).arg1, actualForall.bodyVars);
                            ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg4).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg4).get(j)).arg2, actualForall.bodyVars);
                            ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg4).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg4).get(j)).arg3, actualForall.bodyVars);
                            ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg4).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg4).get(j)).arg4, actualForall.bodyVars);
                        }
                    }                    
                }
                else if(((Constraint) get(i)).arg3 instanceof Constraint) {                   
                    ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg3).arg1, actualForall.bodyVars);
                    ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg3).arg2, actualForall.bodyVars);
                    ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg3).arg3, actualForall.bodyVars);
                    ((Constraint)get(i)).findVars(((Constraint) ((Constraint) get(i)).arg3).arg4, actualForall.bodyVars); 
                }
                else {
                    // Devo trattare un vettore di vincoli
                    for(int j = 0; j < ((ConstraintsConjunction)((Constraint) get(i)).arg3).size(); j++) {
                        ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg3).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg3).get(j)).arg1, actualForall.bodyVars);
                        ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg3).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg3).get(j)).arg2, actualForall.bodyVars);
                        ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg3).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg3).get(j)).arg3, actualForall.bodyVars);
                        ((Constraint)((ConstraintsConjunction)((Constraint) get(i)).arg3).get(j)).findVars(((Constraint) ((ConstraintsConjunction) ((Constraint) get(i)).arg3).get(j)).arg4, actualForall.bodyVars);
                    }
                }
                this.forallVars.add(actualForall);
            }
        }
        return;
    }
   
    
    /**
     * Checks for objects intersection between to vectors.
     *
     * @param  v1 vector.
     * @param  v2 vector.
     * @return true the two vectors have at least one equals object, false 
     *         otherwise.  
     */      
    protected boolean ctrlVectorIntersection(Vector v1, Vector v2) {
        boolean found = false;
        int i = 0;
        while(!v1.isEmpty() && !v2.isEmpty() && !found && i < v2.size()) {          
            found = v1.contains((Lvar)v2.get(i));
            i++;
        }
        return found;
    }
    

    /* Verifica l'utilizzo corretto delle variabili presenti nei forall rispetto
       a tutti i vincoli presenti nel ConstraintStore.
       Tutte le vars quantificate da dei forall non possono venir utilizzate come 
       libere nella definizione di altri vincoli. */
    /**
     * Checks for a proper use of variables in constraints. For exaple universally 
     * quantified variables of a forall constraint cannot be used like free 
     * variables in other constraints on the same store.
     *
     * @return true if there is an uncorrect use of logical variables on the
     *         constraints.
     */
    protected boolean ctrlConstraintStoreVars() {
        boolean found = false;
        int i = 0;

        while (i < this.forallVars.size() && !found) {
            found = this.ctrlVectorIntersection(this.notForallVars, ((VarsForall)this.forallVars.get(i)).quantVars);
            i++;
        }

        if (!found && this.forallVars.size() > 1) {
            int intfound = 0;
            // Controllo delle vars tra i vari forall
            for (int j = 0; j < this.forallVars.size(); j++) {                
                for (int k = 0; k < ((VarsForall)this.forallVars.get(j)).quantVars.size(); k++) {                    
                    for (int h = 0; h < this.forallVars.size(); h++) {
                        if (h != j) {
                            if(!((VarsForall)this.forallVars.get(h)).quantVars.contains((Lvar)((VarsForall)this.forallVars.get(j)).quantVars.get(k))) {
                                if (((VarsForall)this.forallVars.get(h)).bodyVars.contains((Lvar)((VarsForall)this.forallVars.get(j)).quantVars.get(k))) intfound ++;
                            }
                        }
                    }
                }
            }
            if (intfound > 0) found = true;
        }
        return found;
    }
    
    
    protected boolean contains(Constraint c) {
        if (store.size() > 0) {
            for (int i = 0; i < store.size(); i++) {
                if (((Constraint)c).equals((Constraint)store.get(i))) return true;     
            }
        }
        return false;
    }

}
