/*
 * LibConstraints.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;


/**
 * The <code>LibConstraints</code> class allow the user to define new 
 * general constraints that extends the JSetL framework. 
 * These constraints are dealt with as the built-in 
 * constraints (they can be added to the constraint store and solved).
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class LibConstraints implements java.io.Serializable {
    
    /* Member Vars */
    /**
     * Reference to the solver.
     */    
    SolverClass Solver;    
    
    /**
     * Integer code for concat constraint.
     */       
    private static int concatCode = Enviroment.getMethodCode("libConstraintsCode");     

    /**
     * Integer code for size constraint.
     */           
    private static int sizeCode   = Enviroment.getMethodCode("libConstraintsCode");      
    
    
    /* Costructor */     
    /**
     * Creates a new instance of LibConstraints.
     *
     * @param solverInstance a solver instance.
     */    
    public LibConstraints(SolverClass solverInstance) {
	this.Solver = solverInstance;
    }    
    
    
    /* Member Methods */    
    
    /* Not deterministic concat */
    public static Constraint concat(Lst L1, Lst L2, Lst L3) {
	Constraint s = new Constraint(concatCode, L1, L2, L3); 
	return s;
    }


//    public static Constraint prefix(Lst L1, Lst L2) {
//	Lst L = new Lst();
//	return concat(L1, L, L2);
//    }
 

    public static Constraint size(Set S, Lvar N) {
	Constraint s = new Constraint(sizeCode, S, new Integer(0), N); 
	return s;
    }

   
    public static Constraint size(Set S, int N, Lvar M) {
	Constraint s = new Constraint(sizeCode, S, new Integer(N), M); 
	return s;
    }
    
    
    public static Constraint size(Set S, int N) {
	Constraint s = new Constraint(sizeCode, S, new Integer(0), new Integer(N)); 
	return s;
    }


    public static Constraint size(Set S, int N, int M) {
	Constraint s = new Constraint(sizeCode, S, new Integer(N), new Integer(M)); 
	return s;
    }


    protected void user_code(int c, Constraint s) throws Failure, NotDefinedConstraint {
	if (c == concatCode) concat(s); 
        else if (c == sizeCode) size(s); 
        else throw new NotDefinedConstraint();
    }

    
    /**
     * Rewriting rule for the resolution of a not deterministic concat constraint.
     *
     * @param st , a concat constraint.
     */       
    public void concat(Constraint st) {
	Lst L1 = (Lst)st.arg1;
	Lst L2 = (Lst)st.arg2;
	Lst L3 = (Lst)st.arg3;
	switch (st.caseControl) {
	case 0: 
	    Solver.B.addChoicePoint(st);
	    Solver.add(L1.eq(Lst.empty));      // L1.eq([])
	    Solver.add(L2.eq(L3));             // L2.eq(L3)
	    return;
	case 1:
	    Lvar X = new Lvar();
	    Lst R = new Lst();
	    Lst L3new = new Lst();
	    Solver.add(L1.eq(R.ins1(X)));     // L1.eq([X|R])
	    Solver.add(L3.eq(L3new.ins1(X))); // L3.eq([X|L3new])
	    Solver.add(concat(R,L2,L3new));   // concat(R,L2,L3new)
	    return;
	}
    }

    /**
     * Rewriting rule for the resolution of a size constraint.
     *
     * @param st , a size constraint.
     */   
    public void size(Constraint st) throws Failure { 
	Set S = (Set)st.arg1;
	int N = ((Integer)st.arg2).intValue(); 
	if (st.arg3 instanceof Integer) {
	    int M = ((Integer)st.arg3).intValue();
	    if(N == M){
		Solver.add(S.eq(Set.empty));   // S.eq({})
		return;
	    }
	    else {
		Lvar Y = new Lvar();
		Set SS = new Set();
		Solver.add(S.less(Y,SS));
		int NN = N+1;
		Solver.add(size(SS,NN,M));
		return;
	    }
	}
	else if (st.arg3 instanceof Lvar) {
	    Lvar M = (Lvar)st.arg3;
	    switch (st.caseControl) {
	    case 0: 
		Solver.B.addChoicePoint(st);
		Solver.add(M.eq(N));           // N.eq(M)
		Solver.add(S.eq(Set.empty));   // S.eq({})
		return;
	    case 1:
		Lvar Y = new Lvar();
		Set SS = new Set();
		Solver.add(M.neq(N));          // M.neq(N)
		Solver.add(S.less(Y,SS));
		int NN = N+1;
		Solver.add(size(SS,NN,M));
		return;
	    }
	}
    }
 
}
