/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;

/**
 * @author Bin Liu
 * This is a corresponding class of MemoryAndOutputRateCombined. The difference is to use 
 * simulated cost model to indicate the main memory usage of the system. Other factors
 * the same with the MemoryAndOutputRateCombined class
 */
public class SimMemoryAndOutputRate extends CostModel {

	// to record the previous total_number_of_tuple_results
	Hashtable previousResults = new Hashtable();

	public SimMemoryAndOutputRate(StatisticsTable st){
		this.st = st;
		this.requiredStats = new ArrayList();
		setStatistics();
		// initialize the previousResults hashtable
		for(Enumeration e = st.keys();e.hasMoreElements();){
			Machine m = (Machine) e.nextElement();
			previousResults.put(m,"0");
		}
	}

	public Iterator requiredStatistics() {
		return this.requiredStats.iterator();
	}

	protected void setStatistics() {
		super.setStatistics();
		
		//these two stats are already added in super class CostModel
		//this.requiredStats.add("USED_MEMORY");
		//this.requiredStats.add("MAX_MEMORY");
		
		this.requiredStats.add("NUMBER_OF_TUPLES_OUTPUTTED_TOTAL");
		this.requiredStats.add("JOIN_PARTITION_NUMBER");
		this.requiredStats.add("JOIN_PARTITION_SIZE");
		this.requiredStats.add("NUMBER_OF_TUPLES_IN_INPUT_QUEUES");
		this.requiredStats.add("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES");
	}
	
	public CostTable calculateCost(StatisticsTable st, DistributionTable dt) {
		this.dt = dt;
		this.st = st;
		CostTable ct = new CostTable(st.size());
		
		for(Enumeration e = st.keys();e.hasMoreElements();){
			Machine m = (Machine) e.nextElement();
			LightWeightStatistic lws = st.get(m);
			double mem = 0;
			double outputNum = 0;
			int partNum = 0;
			int partSize = 0;
			
			// here, we leave the real used main memory to show how much main memory been consumed
			// it does not been used in the cost model and the redistribution policy
			String memS = lws.get("USED_MEMORY");
			if (memS !=null) {
				mem = Double.parseDouble(memS);
			}
			
			// compute the simulated main memory usage and other cost statistics
			// we need to count all tuples in input_queue and output_queue of all operators
			int totalInQueue = 0;
			int totalOutQueue = 0;
			for(Enumeration en = lws.getKeys(); en.hasMoreElements();){
				LightWeightQueryObject xqo = (LightWeightQueryObject) en.nextElement();
				String inStr = lws.get(xqo,"NUMBER_OF_TUPLES_IN_INPUT_QUEUES");
				if (inStr != null){
					totalInQueue += Integer.parseInt(inStr);
				}
				
				//Consider the cost statistics related to the join operator only 
				if (xqo.getClassName().endsWith("SymmMwayHashJoinOperatorImp")) {
					String outStr = null;
					outStr = lws.get(xqo,"NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES");
					if (outStr != null){
						totalOutQueue += Integer.parseInt(outStr);
					}

					String partNumStr = lws.get(xqo,"JOIN_PARTITION_NUMBER");
					if (partNumStr !=null) {
						partNum = Integer.parseInt(partNumStr);
					}
					String partSizeStr = lws.get(xqo,"JOIN_PARTITION_SIZE");
					if (partSizeStr !=null) {
						partSize = Integer.parseInt(partSizeStr);
					}					
					String outNumS = lws.get(xqo,"NUMBER_OF_TUPLES_OUTPUTTED_TOTAL");
					if (outNumS !=null){
						outputNum = Double.parseDouble(outNumS);
					}										
				}				
			}

			if (totalOutQueue>15000){
				// might be transitant increase of queue size, here, we cut it to 15,000
				// the effect on the main memory usage is not linear
				totalOutQueue = 15000 + ((totalOutQueue-15000)/10);
			}
			int queSum = totalInQueue+totalOutQueue;
			double memRate = (queSum*1.0 + partSize)/1024.0;
			ct.putRawCost(m,memRate);			

			/*
			// (1) simply version of rate control
			ct.put(m,outputNum);
			System.out.println("Machine:" + m.getMachineName() + " UsedM:" + mem + " Outputed:" + outputNum + " PartNum:" + partNum);
			*/
			/*
			// (2)Compare the increase of output
			double rate = 0;
			double previousVal = Double.parseDouble((String)previousResults.get(m));
			rate = outputNum-previousVal;
			ct.put(m,rate);			
			System.out.println("Machine:" + m.getMachineName() + " UsedM:" + mem + " Outputed:" + outputNum + ":" + previousVal + " PartNum:" + partNum + " Rate:" + rate);
			previousResults.put(m,"" + outputNum);
			*/
			
			// (3)the real fair output rate should be the incremental part divide by the total number of partitions in the operator
			double rate = 0;
			double previousVal = Double.parseDouble((String)previousResults.get(m));
			if (partNum!=0){
				rate = ((outputNum)/partNum);//((outputNum-previousVal)/partNum); -- pure incremental
			}
			ct.put(m,rate);			
			//System.out.println("Machine: " + m.getMachineName() + " UsedM:SimM " + mem + " " + memRate + " Out:" + outputNum + ":" + " PNum:" + partNum + " Rate:" + rate);
			System.out.println("M: " + m.getMachineName() + " SimM: " +  memRate + " Out:" + outputNum + " PNum:" + partNum + " Rate:" + rate);
			previousResults.put(m,"" + outputNum);			
		}
		return ct;
	}

}
