/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.MigrationProtocol;

/**
 * @author Yali Zhu
 *
 * This class represents a subplan of a whole query plan tree.
 * 
 * After distribution manager optimizes the running query, a subplan that is
 * different from the old plan (delta change) is returned by optimizer, which is
 * then sent to each machine in the system. After receiving such a subplan at each
 * machine in the cluster, the migration manager on that machine would go through
 * several steps to migrate the query plan, including replacing the old sub-plan with
 * the new sub-plan.
 * 
 * Each sub-plan is uniquely marked by its set of input queues and set of output queues.
 * The queue ID of these queues will be used to identify the position of such a subplan
 * in the whole plan tree.
 * 
 * Assumptions: leaf operators can not be changed during query optimization. leaf operators can assumed to be select
 * or scan operators that do not need to be switched with other operators. 
 */
public class SubPlan implements Serializable{
	ArrayList inputQueueIDs;
	ArrayList outputQueueIDs;
	ArrayList operators;
	ArrayList rootOperators;
	ArrayList operatorIDs;
	ArrayList rootOperatorIDs;
	XATStreamOperator oldRootOperator;
	int oldRootOperatorID;
	// Record the old location a state is from. Hash key is the state ID.
	// This location is operator during optimization and is changed to machine
	// before the subplan is sent to each machine. 
	Hashtable statesOldLocations;
	
	public SubPlan() {
		inputQueueIDs = new ArrayList();
		outputQueueIDs = new ArrayList();
		operators = new ArrayList();
		rootOperators = new ArrayList();
		operatorIDs = new ArrayList();
		rootOperatorIDs = new ArrayList();
		statesOldLocations = new Hashtable();
	}
	
	public void addInputQID(String id) {
		if(!inputQueueIDs.contains(id)) {
			inputQueueIDs.add(id);
		}
		return;
	}
	
	public void addOutputQID(String id) {
		if(!outputQueueIDs.contains(id)) {
			outputQueueIDs.add(id);
		}
	}
	
	public void addOperator(XATStreamOperator op) {
		if(!operators.contains(op)) {
			operators.add(op);
			operatorIDs.add(Integer.toString(op.getObjectID()));
		}
	}
	
	public void addRootOperator(XATStreamOperator root) {
		if(!rootOperators.contains(root)) {
			rootOperators.add(root);
			rootOperatorIDs.add(Integer.toString(root.getObjectID()));
		}
	}

	public ArrayList getInputQueueIDs() {
		return inputQueueIDs;
	}
	
	public ArrayList getOutputQueueIDs() {
		return outputQueueIDs;
	}
	
	public ArrayList getOperators() {
		return operators;
	}
	
	public ArrayList getRootOperators() {
		return rootOperators;
	}
	
	public ArrayList getOperatorIDs() {
		return this.operatorIDs;
	}
	
	public ArrayList getRootOperatorIDs() {
		return this.rootOperatorIDs;
	}
	
	public XATStreamOperator getOldRootOperator() {
		return this.oldRootOperator;
	}
	
	public void setOldRootOperator(XATStreamOperator oldrootop) {
		this.oldRootOperator = oldrootop;
	}
	
	public int getOldRootOperatorID() {
		return this.oldRootOperatorID;
	}
	
	public void setOldRootOperatorID(int id) {
		this.oldRootOperatorID = id;
	}
	
	public void setStateLocation(String stateID, XATStreamOperator op){
		this.statesOldLocations.put(stateID, op);
	}
	
	public void setStateLocation(String stateID, Machine machine) {
		this.statesOldLocations.put(stateID, machine);
	}
	
	public Hashtable getStateLocationTable() {
		return this.statesOldLocations;
	}
	
	public String toString() {
		return ("\n\n== subplan == " 
				+ "\noperators: " + operators
				+ "\nrootOperators: " + rootOperators 
				+ "\noldRootOperator:" + oldRootOperator); 
				//+ "\nstateLocations: " + this.statesOldLocations);
	}
}
