/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution;

import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.SymmMwayHashJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.StatisticsTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
//import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;

/**
 * @author Bin Liu
 * Balance both memory usage and output rate by moving or pushing partitions across machines
 * or disks. Basically, we push low output_rate partitions into disks even the memory usage 
 * across machines are still balanced. In that case, it produces memory space for the high output_rate 
 * partitions to avoid these partitions being pushed to disks. 
 */
public class BalanceMemoryAndOutputRate extends XATRedistributionPolicy {

	//double threshold = 0.5;
	//Minimal used memory required to consider adaptation   
	double minRequiredMemorySize = 3.0; 
	double minRequiredPush = 3.0;
	// We do not want to over push partitions into disks. the idea is to push at most the Max_Needed_Mem - Mem_Space_Available
	int number_of_force_adaptation = 0; // usually for real memory cost model
	double amount_been_pushed = 0.0; // for simulated cost model
	double maximal_to_be_pushed = 6.0; // for simulated cost model
	
	// this is used to indicate the maximal main memory of the machine in the simulated cost model
	double maxMemorySize = 10.0;
	
	public BalanceMemoryAndOutputRate(){
		super();
	}
	/* (non-Javadoc)
	 * Non-partitioned adaptation, it is not being used by this class
	 */
	public DistributionTable redistribute(DistributionTable currTable,
			CostTable currCost, XATNode[] elegibleNodes,
			Machine[] elegibleMachines) {
		return null;
	}

	/* 
	 * 
	 */
	public ArrayList redistribute(DistributionTable currTable,
			CostTable currCost, StatisticsTable staTable,
			Machine[] elegibleMachines, Machine unionMachine) {
		
		ArrayList ret = new ArrayList(5);
		
		Machine maxM=null;
		Machine minM=null;
		
		double max = currCost.maxRawCost();
		double min = currCost.minRawCost();
		
		if (unionMachine!=null) {
			//need to separte the cost of the union machine cost if the union is the current minimal
			double unionCost = currCost.getRawCost(unionMachine);
			if (min==unionCost){
				min = currCost.minRawCostExcept(unionCost);
			}
			if (max == unionCost) {
				return null;
			}
		} 
		double rate = 0.0;
		int operID = 0;

		// do not redistribute if it is below threshold
		if (max < minRequiredMemorySize) {
			return null;
		}

		if ((min/max)>0.8) {
			if (max<minRequiredPush) {
				return null;
			}
			// too little difference between maximal and minimal, 
			// then check the output rate of all machines, if the difference among the machines 
			// reaches certain threshold, then push the partitions of the lowest output rate machine to disk
			double maxRate = currCost.maxCost();
			double minRate = currCost.minCost();
			if (unionMachine != null) {
				double unionC = currCost.getMachineCost(unionMachine);
				if (minRate == unionC) {
					minRate = currCost.minCostExcept(unionC);
				}
			}
			if ((maxRate/minRate)>2) {
				// we stop force_adaptations after a certain number -- for real memory cost model
				/*
				if (number_of_force_adaptation > 10){
					return null;
				}
				*/
				if (amount_been_pushed > maximal_to_be_pushed){ // for simulated memory cost model
					return null;
				}
				
				//ok, we push the partitions of minRate machine
				// (0) machine, (1) operatorID, (2) rate
				// find the minMachine
				Machine minRM = null;
				for (int i=0; i<elegibleMachines.length; i++){
					if (minRate==currCost.getMachineCost(elegibleMachines[i])) {
						minRM = elegibleMachines[i];
						break;
					}
				}
				if (minRM == null) {
					return null;
				}
				/*
				LightWeightStatistic Flws = staTable.get(minRM);
				String FstaVal =  Flws.get("MAX_MEMORY");
				double FmaxMinMachineMem = 0;
				if (FstaVal != null){
					FmaxMinMachineMem = Double.parseDouble(FstaVal);
				}
				if ((max/FmaxMinMachineMem)>0.8) {
					return null;
				} else {
					System.out.println("FF:min/minMachineMaxMem = " + (min/FmaxMinMachineMem) + "minMaxMem:" + FmaxMinMachineMem);
				}
				*/
				// find the first join operator
				XATNode[] nodes = currTable.getNodes();
				for (int i=0; i<nodes.length; i++){
					XATStreamOperator currOp = (XATStreamOperator)(nodes[i]).getOperator();
					if (currOp instanceof SymmMwayHashJoinOperatorImp){
						// yes, this is the operator we need,
						operID = currOp.getOperatorID();
						break;
					}
				}
				ret.add(minRM);
				ret.add(new String(""+operID));
				ret.add(new String("0.3"));
				
				double minRMRawCost = currCost.getRawCost(minRM);
				amount_been_pushed += minRMRawCost*0.3;
				
				number_of_force_adaptation ++; // for real memory cost model
				
				System.out.println("Current ratio:" +  (maxRate/minRate) + " Machine:" + minRM.getMachineName() + " NUM:" + number_of_force_adaptation + "SO FAR:" + amount_been_pushed);
				return ret;
			} else {
				return null;				
			}
		}
				
		// if the difference between max and min are large, then decide to adapt across machines
		// first, compute the rate
		rate = (max-min)/(2*max);
		
		// if it is too much data to be moved one time, i.e., >10M, reduce the rate
		//if ((rate*max)>10.0) {
		//	System.out.println("Too much data to be moved, Max:" + max + " rate:" + rate + "! reduce to 10M instead.");
		//	rate = 10.0/max;
		//}
				
		// find the sender and receiver machines
		for (int i=0; i<elegibleMachines.length; i++){
			if (max==currCost.getRawCost(elegibleMachines[i])) {
				maxM = elegibleMachines[i];
				continue;
			}
			if (min==currCost.getRawCost(elegibleMachines[i])) {
				minM = elegibleMachines[i];
				continue;
			}			
		}
		/*
		// we also do not want a distributed adaptation if the sender has a lower output rate comparing
		// with the receiver, to avoid unnecessary movement
		double senderRate = currCost.getMachineCost(maxM);
		double receiverRate = currCost.getMachineCost(minM);
		if (senderRate < receiverRate) {
			System.out.println("The senders output rate is lower than the receiver. Abort the distributed adaptation");
			return null;
		}
		*/
		/*
		// following for the real main memory cost model only.
		// if the receiver machine is too full, i.e., more than 80%, then it is not a good idea to adapt
		// or we can try to find another machine? now we simply stop adaptation
		LightWeightStatistic lws = staTable.get(minM);
		String staVal =  lws.get("MAX_MEMORY");
		double maxMinMachineMem = 0;
		if (staVal != null){
			maxMinMachineMem = Double.parseDouble(staVal);
		}
		if ((min/maxMinMachineMem)>0.8) {
			return null;
		} else {
			System.out.println("min/minMachineMaxMem = " + (min/maxMinMachineMem) + "minMaxMem:" + maxMinMachineMem);
		}
		*/
		// the following is for the simulated cost mode, kind of hard-coded for the maxMemorySize
		if (((min + (max-min)/2)/maxMemorySize) >0.8) {
			return null;
		} else {
			System.out.println("min/minMachineMaxMem = " + ((min + (max-min)/2)/maxMemorySize) + "minMaxMem:" + maxMemorySize);
		}
		
		// find the first join operator
		XATNode[] nodes = currTable.getNodes();
		for (int i=0; i<nodes.length; i++){
			XATStreamOperator currOp = (XATStreamOperator)(nodes[i]).getOperator();
			if (currOp instanceof SymmMwayHashJoinOperatorImp){
				// yes, this is the operator we need,
				operID = currOp.getOperatorID();
				break;
			}
		}

		// (0) sender Machine, (1) receiver Machine, (2) operator ID, (3) rate
		ret.add(maxM);
		ret.add(minM);
		ret.add(new String(""+ operID));
		ret.add(new String("" + rate));
		
		// increase thresthold for the next adaptation, debug code
		// threshold += 0.1;
		
		return ret;
	}
}
