/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui;

import java.awt.BorderLayout;
import javax.swing.JSplitPane;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JMenuBar;
import javax.swing.JComboBox;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.JLayeredPane;
import javax.swing.JToolBar;
import javax.swing.JButton;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.Timer;

import java.util.ArrayList;
import java.util.Hashtable;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileNotFoundException;

import diva.graph.basic.BasicGraphController;
import diva.graph.basic.BasicNodeRenderer;
import diva.graph.basic.BasicEdgeRenderer;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.manager.DistributionManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;

/** An abstract class for CAPE <code>GUI</code>s. An overview of <code>CAPE_GUI</code> features:
 * <ul>
 * <li><code>BasicGraphModelMaker</code> - makes the <code>diva.graph.basic.BasicGraphModel</code>s
 * for each of the <code>CAPE_GUI</code>'s loaded trees.
 * <li><code>FramesManager</code> - manages all of the open frames
 * and the making of new frames to display.
 * <li><code>GraphManager</code> - manages open graphs and the drawing of the
 * graphs to the tabbed pane.
 * <li><code>OptionsStorage</code> - stores all of the user's selected/deselected
 * options.
 * <li><code>PanelMaker</code> - makes all of the <code>CAPE_GUI</code>'s panels.
 * <li><code>StatisticStorage</code> - stores all statistics for the tree(s), queue(s), 
 * operator(s) and system.
 * <li><code>TableMaker</code> - handles making all of the tables.
 * <li><code>TreeStorage</code> - stores all of the operator(s) and queue(s) of all loaded trees.
 * </ul>
 * Each <code>CAPE_GUI</code> has a unique name. For example, if this <code>CAPE_GUI</code>
 * runs on the distribution manager, then the name would be "DISTRIBUTION_MANAGER".
 * The <code>CAPE_GUI</code> named "DISTRIBUTION_MANAGER" is flagged for the <code>BasicGraphModelMaker</code>
 * to add labels to the operators identifying which machine that specific operator is running on.
 * <code>CAPE_GUI</code>s running on machines other than the distribution manager
 * may be named "QUERY_ENGINE" and lack this machine identifier.
 * 
 * All derived classes must implement the init() method.
 * 
 * @author Maylene Waltz
 */
public abstract class CAPE_GUI extends JPanel
{
	/** How often to update all the queues and redraw the graphs 
	 * with the new queue values in milliseconds. */
	final private int UPDATE_QUEUES_MILLISECONDS = 1000;
	
	/** How often to update all the node colors to signify
	 * whether the node is inactive, active or running in milliseconds. */
	final private int UPDATE_NODE_COLOR_MILLISECONDS = 1000;
	
	/** How often to update all of the statistic table(s) in milliseconds. */
	final private int UPDATE_STATISTICS_MILLISECONDS = 1000;
	
	/** How often to check whether the trees need to be updated */
	final private int UPDATE_TREES_MILLISECONDS = 1000;
	
	/** The timer to update the statistic table(s). */
	private javax.swing.Timer _updateStatisticTableTimer;
	
	/** The timer to update queue sizes. */
	private javax.swing.Timer _updateQueuesTimer;
	
	/** The timer to update node status (color). */
	private javax.swing.Timer _updateNodesTimer;
	
	private Timer _updateTreesTimer;
	
	/** The <code>BasicGraphModelMaker</code> makes the <code>diva.graph.basic.BasicGraphModel</code>s
	 * for each of the <code>CAPE_GUI</code>'s loaded trees. */
	private BasicGraphModelMaker _bgmm = new BasicGraphModelMaker(this);
	
	/** The <code>FramesManager</code> manages all of the open frames
	 * and the making of new frames to display. */
	private FramesManager _fm = new FramesManager(this);
	
	/** The <code>GraphManager</code> manages open graphs and the drawing of the
	 * graphs to the tabbed pane. */
	private GraphManager _gm = new GraphManager(this);
	
	/** The <code>CAPE_GUI</code>'s menu bar. */
	private JMenuBarCAPE _menuBar = new JMenuBarCAPE(this);
	
	/** The <code>OptionsStorage</code> stores all of the user's selected/deselected options. */
	private OptionsStorage _os = new OptionsStorage();
	
	/** The <code>PanelMaker</code> makes all of the CAPE_GUI's panels. */
	protected PanelMaker _pm;
	
	/** The <code>CAPE_GUI</code>'s split pane.
	 * The top holds a tabbed pane of tree graphs and the bottom holds 
	 * a tabbed pane of statistic table(s). */
	private JSplitPane _splitPane;
	
	/** The <code>StatisticStorage</code> stores all of the statistics for the tree(s),
	 * queue(s), operator(s) and system. */
	private StatisticStorage _ss = new StatisticStorage(this);
	
	/** The <code>TableMaker</code> makes all of the tables for the <code>CAPE_GUI</code>.
	 * Table properties, such as column width and auto-resize mode, are set here. */
	private TableMaker _tm = new TableMaker(this);
	
	/** The <code>CAPE_GUI</code>'s tool bar. */
	private JToolBar _tb = new JToolBar();
	
	/** The <code>TreeStorage</code> stores all of the operator(s) and queue(s) of all loaded trees. */
	private TreeStorage _ts = new TreeStorage();
	
	/** The algebraic trees representing the query plans. */
	protected XATTree[] _XATTrees;
	
	/** The client's name. For example, "Machine 1". */
	protected String _clientName;
	
	/** The <code>SystemObject</code> is used for gathering system statistics. */
	protected SystemObject _systemObject;
	
	/** Extra information output to screen about the distribution. */
	protected String[] _distInfo;
	
	private DistributionManager dm;
	
	public static boolean fireStream = false;
	//------------------------------------------------------------------------------
	/** A table which maps stats files to their readers. */
	private Hashtable _fsReaders = null;
	
	/** A table which maps stats files to available tree options. */
	private Hashtable _fileToTreeOptionsMap = null;
	
	/** A table which maps stats files to available operator options. */
	private Hashtable _fileToOpOptionsMap = null;
	
	/** A table which maps stats files to available operators. */
	private Hashtable _fileToOperatorsMap = null;
	//------------------------------------------------------------------------------
	
	//------------------------------------------------------------------------------
	/** Constructor initializes the <code>CAPE_GUI</code>.
	 * Note that no trees have been loaded at this point. */
	public CAPE_GUI()
	{
		_XATTrees = createSampleTrees();
		_pm = new PanelMaker(this);
		initGUI();
	}
	//------------------------------------------------------------------------------	
	/** Constructor loads the given <code>trees</code>, sets the
	 * <code>CAPE_GUI</code>'s client name to the given <code>clientName</code>
	 * and initializes the <code>CAPE_GUI</code>.
	 * @param trees the trees to be loaded into the <code>CAPE_GUI</code>
	 * @param clientName the name of the client query engine machine. For example, "Machine 1". */
	public CAPE_GUI(XATTree[] trees, String clientName)
	{
		this._clientName = clientName;
		this._XATTrees = trees;
		_pm = new PanelMaker(this);
		initGUI();
	}
	
	//------------------------------------------------------------------------------	
	/** Constructor loads the given <code>trees</code>, sets the
	 * <code>CAPE_GUI</code>'s client name to the given <code>clientName</code>,
	 * sets the <code>CAPE_GUI</code>'s system object to the given <code>systemObject</code>
	 * and initializes the <code>CAPE_GUI</code>.
	 * @param trees the trees to be loaded into the <code>CAPE_GUI</code>
	 * @param clientName the name of the client query engine machine. For example, "Machine 1"
	 * @param systemObject the object used to gather system statistics */
	public CAPE_GUI(XATTree[] trees, String clientName, SystemObject systemObject)
	{
		this._XATTrees = trees;
		this._clientName = clientName;
		this._systemObject = systemObject;
		_pm = new PanelMaker(this);
		initGUI();
	}
	//------------------------------------------------------------------------------
	public CAPE_GUI(XATTree[] trees, String clientName, SystemObject systemObject, String[] distInfo)
	{
		this._XATTrees = trees;
		this._clientName = clientName;
		this._systemObject = systemObject;
		this._distInfo = (String[])distInfo.clone();
		_pm = new PanelMaker(this);
		initGUI();
	}
	//------------------------------------------------------------------------------
	public CAPE_GUI(XATTree[] trees, String clientName, SystemObject systemObject, DistributionManager dm)
	{
		this._XATTrees = trees;
		this._clientName = clientName;
		this._systemObject = systemObject;
		this.dm = dm;
		_pm = new PanelMaker(this);
		initGUI();
	}	
	//------------------------------------------------------------------------------
	/** Applies the user's customization selections to the graph 
	 * in the selected tab. The user can specify:
	 * <ul><li>Node Text Color
	 * <li>Node Text Size
	 * <li>Node Size and
	 * <li>Edge Text Size.
	 */
	public void applyCustomization()
	{
		try
		{
			GraphManager gm = getGraphManager();
			BasicGraphController controller = (BasicGraphController)gm.getGraph(gm.getGraphTabbedPane().getSelectedIndex()).getGraphPane().getGraphController();
			BasicNodeRenderer nodeRenderer = (BasicNodeRenderer)controller.getNodeController().getNodeRenderer();
			BasicEdgeRenderer edgeRenderer = (BasicEdgeRenderer)controller.getEdgeController().getEdgeRenderer();
			
			//apply customization to the visible graph
			nodeRenderer.setLabelTextColor(CAPEUtils.determineColor(((JComboBox)((ArrayList)_os.getOptions("CUSTOMIZE")).get(0)).getSelectedIndex()));
			nodeRenderer.setLabelTextSize(CAPEUtils.determineTextSize(((JComboBox)((ArrayList)_os.getOptions("CUSTOMIZE")).get(1)).getSelectedIndex()));
			edgeRenderer.setLabelTextSize(CAPEUtils.determineTextSize(((JComboBox)((ArrayList)_os.getOptions("CUSTOMIZE")).get(3)).getSelectedIndex()));
			nodeRenderer.setNodeShape(CAPEUtils.determineNodeSize(((JComboBox)((ArrayList)_os.getOptions("CUSTOMIZE")).get(2)).getSelectedIndex()));
			
			//rerender and redraw the customize graph(s)
			controller.getEdgeController().setEdgeRenderer(edgeRenderer);
			controller.getNodeController().setNodeRenderer(nodeRenderer);
			controller.rerender();
			gm.redrawGraph(gm.getGraph(gm.getGraphTabbedPane().getSelectedIndex()));
		}
		catch(Exception e){System.err.println("CAPE_GUI: Error customizing graph.");}
	}
	//------------------------------------------------------------------------------
	/** Ensures the update frequency is in milliseconds.
	 * The update frequency is the user-specified amount of time
	 * to plot a new point in the statistic plot(s).
	 * @return true if the frequency is in milliseconds; false, otherwise */
	public boolean checkFreqFormat(JTextField textField)
	{
		String text = textField.getText();
		boolean formatOk;
		if(!java.util.regex.Pattern.matches("\\d*\\p{Punct}++\\d*",
			text.subSequence(0,text.length()))){
			formatOk = true;
		}
		else formatOk = false;
		return formatOk;
	}
	//------------------------------------------------------------------------------
	/**
	 * Creates sample XATDag trees.
	 * @return an array of sample XATDag trees
	 */
	private XATDag[] createSampleTrees()
	{
		CreateSampleTrees createSampleTrees = new CreateSampleTrees();
		System.setProperty("DATA_MODEL", "XATMemoryQueueImp");
		return createSampleTrees.createSampleTrees();
	}
	//------------------------------------------------------------------------------
	/**
	 * Returns the BufferedReader used to read statistics from file.
	 * 
	 * Creation Date: (Jul 19, 2004 10:25:20 AM)
	 */
	public BufferedReader getFileStatsReader(File inFile) 
	{
		if (_fsReaders == null || !_fsReaders.contains(inFile)) 
		{
			return null;
		} 
		else 
		{
			return (BufferedReader)_fsReaders.get(inFile);
		}
	}
	
	//------------------------------------------------------------------------------
	/**
	 * Sets the file stats reader to read from specified file.
	 * 
	 * Creation Date: (Jul 19, 2004 10:35:28 AM)
	 * @param inFile File
	 * @throws FileNotFoundException
	 */
	public void setFileStatsReader(File inFile) throws FileNotFoundException 
	{
		if (_fsReaders == null)
			_fsReaders = new Hashtable();
		try 
		{
			if (!_fsReaders.contains(inFile)) 
			{
				_fsReaders.put(inFile, new BufferedReader(new FileReader(inFile)));
			}
		} 
		catch (IOException ioe) 
		{
			// file was not found
			throw new FileNotFoundException();
		}
	}
	//------------------------------------------------------------------------------
	/**
	 * Reads available statistics for all selected files and stores them in 
	 * appropriate maps.
	 * 
	 * Creation Date: (Jul 20, 2004 10:51:50 AM)
	 * @author Natasha Bogdanova
	 */
	public ArrayList getFileStats() 
	{
		String[] files = (String[])_os.getOptions("PLOT_FILE_STATS_SELECTED_FILES");
		ArrayList result = new ArrayList();
		for (int i = 0; i < files.length; i++) 
		{
			try 
			{
				File currFile = new File(files[i]);
				// create a BufferedReader and store in hash table
				setFileStatsReader(currFile);
				BufferedReader in = getFileStatsReader(currFile);
				// read off statistics
				String line = new String();
				boolean readOpOptions = false;
				boolean readLastLine = false;
				ArrayList operators = new ArrayList();
				ArrayList options = new ArrayList();
				
				for (int l = 0; !readLastLine; l++) 
				{
					// skip first two lines
					if (l < 2) { /* do nothing */ }
					else if (l == 2) 
					{
						// get options from header
						String[] optionsStr = line.split(",");
						for (int k = 1; k < optionsStr.length; k++)
							options.add(optionsStr[k]);
						result.add(options);
						readLastLine = true;
					}
					else 
					{
						/*
						String[] vals = line.split(",");
						if (vals[0] == "Tree") {
							// get tree options
							ArrayList treeOptions = new ArrayList();
							for (int opt = 1; opt < vals.length; opt++) {
								if (vals[opt] != "------")
									treeOptions.add(options.get(opt));
							}
							// store tree options
							_fileToTreeOptionsMap.put(currFile, treeOptions);
							readLastLine = true;
						}
						else if (!readOpOptions) {
							// get first operator
							operators.add(vals[0].split(" "));
							// get operator options
							ArrayList opOptions = new ArrayList();
							for (int opt = 1; opt < vals.length; opt++) {
								if (vals[opt] != "------")
									opOptions.add(options.get(opt));
							}
							// store operator options
							_fileToOpOptionsMap.put(currFile, opOptions);
							readOpOptions = true;
						}
						else { 
							// get operator
							operators.add(vals[0].split(" "));
						}
						*/
					}
				}
				// store operators
				if (_fileToOperatorsMap == null)
					_fileToOperatorsMap = new Hashtable();
				if (!_fileToOperatorsMap.containsKey(currFile))
					_fileToOperatorsMap.put(currFile, operators);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}	
		return result;	
	}
	//------------------------------------------------------------------------------
	/**
	 * Returns all the available tree options for the given file, or null if
	 * no options exist.
	 * 
	 * Creation Date: (Jul 20, 2004 10:54:55 AM)
	 * @author Natasha Bogdanova
	 * @param inFile File
	 * @return ArrayList
	 */
	public ArrayList getTreeOptions(File inFile) 
	{
		if (_fileToTreeOptionsMap == null)
			return null;
		else if (_fileToTreeOptionsMap.containsKey(inFile))
			return (ArrayList)_fileToTreeOptionsMap.get(inFile);
		else
			return null;
	}
	//------------------------------------------------------------------------------
	/**
	 * Returns all the available operator options for the given file, or null 
	 * if no options exist.
	 * 
	 * Creation Date: (Jul 20, 2004 10:55:45 AM)
	 * @author Natasha Bogdanova
	 * @param inFile File
	 * @return ArrayList
	 */
	public ArrayList getOpOptions(File inFile) 
	{
		if (_fileToOpOptionsMap == null)
			return null;
		else if (_fileToOpOptionsMap.containsKey(inFile))
			return (ArrayList)_fileToOpOptionsMap.get(inFile);
		else
			return null;
	}
	//------------------------------------------------------------------------------
	/**
	 * Returns all the available operators for the given file, or null if no 
	 * options exist.
	 * 
	 * Creation Date: (Jul 20, 2004 11:05:13 AM)
	 * @author Natasha Bogdanova
	 * @param inFile File
	 * @return ArrayList
	 */
	public ArrayList getOperators(File inFile) 
	{
		if (_fileToOperatorsMap == null)
			return null;
		else if (_fileToOperatorsMap.containsKey(inFile))
			return (ArrayList)_fileToOperatorsMap.get(inFile);
		else
			return null;
	}
	//------------------------------------------------------------------------------
	
	/** Returns the <code>FramesManager</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>FramesManager</code> manages all of the open frames
	 * and the making of new frames to display.
	 * @return the <code>FramesManager</code> that manages all open frames
	 * and the making of new frames to display */	
	public FramesManager getFramesManager()
	{
		return _fm;
	}
	//------------------------------------------------------------------------------
	/** Returns the <code>GraphManager</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>GraphManager</code> manages open graphs and the drawing of the
	 * graphs to the tabbed pane.
	 * @return the <code>GraphManager</code> that manages open graphs and the drawing of the
	 * graphs to the tabbed pane */	
	public GraphManager getGraphManager()
	{
		return _gm;
	}
	//------------------------------------------------------------------------------
	
	/** Returns the <code>CAPE_GUI</code>'s distribution information.
	 * @return the <code>CAPE_GUI</code>'s distribution information */	
	public String[] getDistInfo()
	{
		return _distInfo;
	}
	//------------------------------------------------------------------------------	
	
	/** Returns the <code>CAPE_GUI</code>'s menu bar.
	 * @return the <code>CAPE_GUI</code>'s menu bar */	
	public JMenuBar getJMenuBar()
	{
		return _menuBar.createJMenuBarCAPE();
	}	
	
	//------------------------------------------------------------------------------

	/** Returns the <code>BasicGraphModelMaker</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>BasicGraphModelMaker</code> makes the <code>diva.graph.basic.BasicGraphModel</code>s
	 * for each of the <code>CAPE_GUI</code>'s loaded trees.
	 * @return the <code>BasicGraphModelMaker</code> that makes the graph model
	 * for each of the <code>CAPE_GUI</code>'s loaded trees */
	public BasicGraphModelMaker getModelMaker()
	{
		return _bgmm;
	}

	//------------------------------------------------------------------------------
	
	/** Returns the name of the <code>CAPE_GUI</code>. For example, if this <code>CAPE_GUI</code>
	 * runs on the distribution manager, then the name would be "<code>DISTRIBUTION_MANAGER</code>".
	 * The <code>CAPE_GUI</code> named "<code>DISTRIBUTION_MANAGER</code>" is flagged for the <code>BasicGraphModelMaker</code>
	 * to add labels to the operators identifying which machine that specific operator is running on.
	 * <code>CAPE_GUI</code>s running on quey machines are named
	 * "<code>QUERY_ENGINE</code>" and lack this machine identifier.
	 * @return the name of the <code>CAPE_GUI</code> */	
	public abstract String getName();
	
	//------------------------------------------------------------------------------
	
	/** Returns the <code>OptionsStorage</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>OptionsStorage</code> stores all of the user's selected/deselected
	 * options. Examples of options are:
	 * <ul><li>Which tree, queue, operator and system statistic to display in the
	 * statistic table(s),
	 * <li>Which tree(s), queue(s) or operator(s) to plot selected statistics for, or
	 * <li>Which graph customizations to apply to the graph in the selected tab.
	 * </ul>
	 * @return the <code>OptionsStorage</code> where all user selected/deselected
	 * options are stored */	
	public OptionsStorage getOptionsStorage()
	{
		return _os;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the <code>PanelMaker</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>PanelMaker</code> makes all of the <code>CAPE_GUI</code>'s panels.
	 * @return the <code>PanelMaker</code> that makes all of the <code>CAPE_GUI</code>'s panels */	
	public PanelMaker getPanelMaker()
	{
		return _pm;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the <code>CAPE_GUI</code>'s split pane.
	 * The top holds a tabbed pane of tree graphs and the bottom holds 
	 * a tabbed pane of statistic table(s).
	 * @return the <code>CAPE_GUI</code>'s split pane. The top holds a tabbed pane
	 * of tree graphs and the bottom holds a tabbed pane of statistic table(s). */	
	public JSplitPane getSplitPane()
	{
		return _splitPane;
	}	
	
	//------------------------------------------------------------------------------
	
	/** Returns the <code>StatisticStorage</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>StatisticStorage</code> stores all statistics for the tree(s), queue(s), 
	 * operator(s) and system.
	 * @return the <code>StatisticStorage</code> that stores all statistics 
	 * for the tree(s), queue(s), operator(s) and system */	
	public StatisticStorage getStatisticStorage()
	{
		return _ss;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the <code>SystemObject</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>SystemObject</code> is used to gather system statistics.
	 * @return the system object of this <code>CAPE_GUI</code> */
	public SystemObject getSystemObject()
	{
		return _systemObject;
	}

	//------------------------------------------------------------------------------
	
	/** Returns the <code>TableMaker</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>TableMaker</code> handles making all of the tables.
	 * Table properties such as column width and auto-resize mode are set here.
	 * The table is placed in a scroll pane and these properties are set here as well.
	 * @return the <code>TableMaker</code> that handles making all tables */	
	public TableMaker getTableMaker()
	{
		return _tm;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the <code>CAPE_GUI</code>'s tool bar. 
	 * @return this <code>CAPE_GUI</code>'s tool bar */
	public JToolBar getToolBar()
	{
		return _tb;
	}

	//------------------------------------------------------------------------------
	
	/** Returns the <code>TreeStorage</code> associated with this <code>CAPE_GUI</code>.
	 * The <code>TreeStorage</code> stores all of the operator(s) and queue(s) of all loaded trees.
	 * @return the <code>TreeStorage</code> that stores all loaded trees */	
	public TreeStorage getTreeStorage()
	{
		return _ts;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the timer that updates node status every <code>UPDATE_NODE_COLOR_MILLISECONDS</code>.
	 * Inactive nodes are gray, active nodes are red and running nodes are green.
	 * @return the timer that updates node status */
	public javax.swing.Timer getUpdateNodesTimer()
	{
		return _updateNodesTimer;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the timer that updates queue sizes every <code>UPDATE_QUEUES_MILLISECONDS</code>. 
	 * @return the timer that updates queue sizes */
	public javax.swing.Timer getUpdateQueuesTimer()
	{
		return _updateQueuesTimer;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the timer that updates the statistic table(s) every <code>UPDATE_STATISTICS_MILLISECONDS</code>. 
	 * @return the timer that updates the statistic table(s) */
	public javax.swing.Timer getUpdateStatisticTableTimer()
	{
		return _updateStatisticTableTimer;
	}

	//------------------------------------------------------------------------------
	
	/** Returns the loaded tree at the given <code>index</code>. 
	 * @return the loaded tree at the given <code>index</code> */	
	public XATTree getXATTree(int index)
	{
		return _XATTrees[index];	
	}

	//------------------------------------------------------------------------------
	
	/** Returns the number of loaded trees in the <code>CAPE_GUI</code>.
	 * @return the number of loaded trees */	
	public int getXATTreeCount()
	{
		return _XATTrees.length;
	}
	
	//------------------------------------------------------------------------------
	
	/** Makes the timer to update the queue sizes every <code>UPDATE_QUEUES_MILLISECONDS</code>.
	 * @return the timer to update the queue sizes */
	protected javax.swing.Timer makeUpdateQueuesTimer()
	{
		javax.swing.Timer timer = new javax.swing.Timer(UPDATE_QUEUES_MILLISECONDS, new UpdateQueuesListener(this));
		return timer;
	}
	
	//------------------------------------------------------------------------------
	
	/** Makes the timer to update node status every <code>UPDATE_NODE_COLOR_MILLISECONDS</code>.
	 * <ul><li>gray - inactive node
	 * <li>red - active node
	 * <li>green - running node
	 * </ul>
	 * @return the timer to update node status */
	protected javax.swing.Timer makeUpdateNodesTimer()
	{
		javax.swing.Timer timer = new javax.swing.Timer(UPDATE_NODE_COLOR_MILLISECONDS, new UpdateNodeColorListener(this));
		return timer;
	}
	
	//------------------------------------------------------------------------------
	
	protected Timer makeUpdateTreesTimer() 
	{
		this._updateTreesTimer = new Timer(UPDATE_TREES_MILLISECONDS, new UpdateTreesListener(this));
		return this._updateTreesTimer;
	}
	
	//------------------------------------------------------------------------------
	
	/** Makes the timer to update the statistic table(s) every <code>UPDATE_STATISTICS_MILLISECONDS</code>.
	 * @return the timer to update the statistic table(s) */
	protected javax.swing.Timer makeUpdateStatisticTableTimer()
	{
		_updateStatisticTableTimer = new javax.swing.Timer(UPDATE_STATISTICS_MILLISECONDS, new UpdateStatisticTableListener(this));
		return _updateStatisticTableTimer;
	}
	
	//------------------------------------------------------------------------------
	
	/** Sets this <code>CAPE_GUI</code>'s split pane to the given <code>splitPane</code>.
	 * @param the split pane to set to this <code>CAPE_GUI</code>'s split pane */
	protected void setSplitPane(JSplitPane splitPane)
	{
		_splitPane = splitPane;
	}
	
	//------------------------------------------------------------------------------
	
	/** Creates this <code>CAPE_GUI</code>'s toolbar.
	 * @return the toolbar specific to this <code>CAPE_GUI</code> */
	protected java.awt.Component makeToolbar()
	{
		JToolBar tb = new JToolBar();
		tb.setLayout(new BoxLayout(tb, BoxLayout.X_AXIS));
		tb.setRollover(true);
		tb.setFloatable(false);
		tb.setBorderPainted(true);
		tb.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createEmptyBorder(3,3,0,3),
			BorderFactory.createEtchedBorder()));
		
		java.net.URL plotURL = CAPE_GUI.class.getResource("images/plot.png");
		ImageIcon plot = new ImageIcon(plotURL, "Plot Statistics");
		JButton pB = new JButton("Plot ", plot);
		pB.setIconTextGap(5);
		pB.setBorder(BorderFactory.createEmptyBorder(2,2,2,2));
		pB.setToolTipText("Plot Statistics");
		pB.addActionListener(new CAPEListener(this));
		pB.setActionCommand("WHAT_TO_PLOT");
		
		java.net.URL treeURL = CAPE_GUI.class.getResource("images/tree.png");
		ImageIcon tree = new ImageIcon(treeURL, "Tree Statistics");
		JButton tSB = new JButton("Tree Statistics ", tree);
		tSB.setIconTextGap(3);
		tSB.setBorder(BorderFactory.createEmptyBorder(2,2,2,2));
		tSB.setToolTipText("Tree Statistics");
		tSB.addActionListener(new CAPEListener(this));
		tSB.setActionCommand("DISPLAY_TREE_STATS");
		
		java.net.URL sysURL = CAPE_GUI.class.getResource("images/system.png");
		ImageIcon sys = new ImageIcon(sysURL, "System Statistics");
		JButton sSB = new JButton("System Statistics", sys);
		sSB.setIconTextGap(3);
		sSB.setBorder(BorderFactory.createEmptyBorder(2,2,2,2));
		sSB.setToolTipText("System Statistics");
		sSB.addActionListener(new CAPEListener(this));
		sSB.setActionCommand("DISPLAY_SYSTEM_STATS");
		
		tb.add(pB);
		tb.add(tSB);
		tb.add(sSB);
		
		return tb;
	}
	
	//------------------------------------------------------------------------------
	
	/** Initializes the <code>CAPE_GUI</code>. */
	protected void initGUI()
	{
		setLayout(new BorderLayout());
		JLayeredPane layeredPane = new JLayeredPane();
		layeredPane.setLayout(new BorderLayout(3,3));
		layeredPane.setOpaque(true);
		
		//JSplitPane topSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, 
		//										 true, 
		//										 _pm.makePanel("GRAPH"), 
		//										 _pm.makePanel("USERINPUT"));
		//topSplitPane.setDividerLocation(700);
		
		//JSplitPane splitPane = 
		//	new JSplitPane(JSplitPane.VERTICAL_SPLIT, 
		//				   true,
		//				   _pm.makePanel("GRAPH"),
		//	 			   _pm.makePanel("STATS"));
		//
		// nb
		JSplitPane splitPane = 
					new JSplitPane(JSplitPane.VERTICAL_SPLIT, 
								   true,
								   _pm.makePanel("GRAPH"),
								   _pm.makePanel("CURRENT_CONFIGURATION"));
		setSplitPane(splitPane);
	  	
		layeredPane.add(makeToolbar(), BorderLayout.PAGE_START);
		layeredPane.add(splitPane, BorderLayout.CENTER);
		
		add(layeredPane);
		
		JPanel panel = (JPanel)((JTabbedPane)_pm.getPanel("CURRENT_CONFIGURATION").getComponent(0)).getComponentAt(0);
		// make the operators update timer
/*		Timer opTimer = new Timer(
				2000, 
				new UpdateCurrentInfoListener(this, (JPanel)panel.getComponent(0)));
		opTimer.start();*/
		
/*		// make the scheduler adaptation update timer
		Timer schedTimer = new Timer(
				500, 
				new UpdateCurrentInfoListener(this, (JPanel)panel.getComponent(1)));
		schedTimer.start();*/
		
		// make the plan optimization update timer
		Timer optTimer = new Timer(
				500, 
				new UpdateCurrentInfoListener(this, (JPanel)panel.getComponent(0)));
		optTimer.start();
		// make the distribution update timer
		Timer distTimer = new Timer(
				500, 
				new UpdateCurrentInfoListener(this, (JPanel)panel.getComponent(1)));
		distTimer.start();
		
		/*
		//make the result update timer
		Timer resultDisplayTimer = new Timer(
				2000, 
				new UpdateCurrentInfoListener(this, (JPanel)panel.getComponent(4)));
		resultDisplayTimer.start();
		*/
	}
	
	//------------------------------------------------------------------------------
	
	public abstract void init();
	
	//------------------------------------------------------------------------------
	
	public boolean makeStatisticsFromFile(String fname) 
	{
		File inFile = new File(fname);
		if (inFile.exists()) 
		{
			try {
				// create a BufferedReader to read the contents of the file
				BufferedReader in = new BufferedReader(new FileReader(inFile));
				String temp = new String();

				String distributionPattern = in.readLine();
				String timeStamp = in.readLine().substring(11);
				temp = in.readLine();
				/*
				
				// read in available statistics headers
				String currStatHeader = new String();
				ArrayList statsHeaders = new ArrayList();
				boolean first = true;
				boolean done = false;
				while(i < contents.length && !done) {
					if (contents[i] == ',' && !first) {
						statsHeaders.add(currStatHeader);
						currStatHeader = new String();
					}
					else if (contents[i] == '\n' && !first) {
						statsHeaders.add(currStatHeader);
						currStatHeader = new String();
						done = true;
					}
					else {
						currStatHeader += contents[i];;
					}
					i++;
				}
				lineCount++;
				done = false;
				
				// read in operators
				ArrayList currOperator = new ArrayList();
				String currString = new String();
				while (i < contents.length && !done) {
					if (contents[i] == ' ') {
						currOperator.add(currString);
						currString = new String();
					}
					else if (contents[i] == ',') {
						currOperator.add(currString);
						currString = new String();
						done = true;
					}
					else {
						currString += contents[i];
					} 
				}
				
				
				_gui.getOptionsStorage().setOptions("PLOT_FILE_STATS_OPTIONS", options);
				fm.makeNewFrame("PLOT_FILE_STATS_OPTIONS", "Select Statistics to Plot",
					new Point(new Double(tk.getScreenSize().getWidth()/2-350).intValue(), 
							  new Double(tk.getScreenSize().getHeight()/2-180).intValue()));
				fm.removeOpenFrame("PLOT_FILE_STATS");
				*/
				return true;
			} 
			catch (IOException ioe) 
			{
				return false;
			}
		} 
		else 
		{
			return false;
		}
	}
	
	//------------------------------------------------------------------------------
	
	
}