/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
//import java.util.ArrayList;
//import java.util.Iterator;

import javax.swing.JFrame;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;

/** This <code>CAPE_GUI</code> is for the distribution manager.
 * The <code>CAPE_GUI</code>'s name is "<code>DISTRIBUTION_MANAGER</code>" which
 * flags it for the <code>BasicGraphModelMaker</code>
 * to add labels to the operators identifying which machine that specific operator is running on.
 * An overview of <code>CAPE_GUI</code> features:
 * <ul>
 * <li><code>BasicGraphModelMaker</code> - makes the <code>diva.graph.basic.BasicGraphModel</code>s
 * for each of the <code>CAPE_GUI</code>'s loaded trees.
 * <li><code>FramesManager</code> - manages all of the open frames
 * and the making of new frames to display.
 * <li><code>GraphManager</code> - manages open graphs and the drawing of the
 * graphs to the tabbed pane.
 * <li><code>OptionsStorage</code> - stores all of the user's selected/deselected
 * options.
 * <li><code>PanelMaker</code> - makes all of the <code>CAPE_GUI</code>'s panels.
 * <li><code>StatisticStorage</code> - stores all statistics for the tree(s), queue(s), 
 * operator(s) and system.
 * <li><code>TableMaker</code> - handles making all of the tables.
 * <li><code>TreeStorage</code> - stores all of the operator(s) and queue(s) of all loaded trees.
 * </ul>
 * 
 * @author Maylene Waltz
 */
public class CAPE_GUIDistributionManager extends CAPE_GUI implements Runnable
{
		
	//------------------------------------------------------------------------------
	
	/** The <code>CAPE_GUI</code>'s name is "DISTRIBUTION_MANAGER". 
	 * This flags this <code>CAPE_GUI</code> to include machine names on the operators
	 * to identify which machine that specific operator is running on. */
	private static String _GUIName = new String("DISTRIBUTION_MANAGER");
	
	//------------------------------------------------------------------------------
	
	/** Constructor initializes the <code>CAPE_GUI</code>.
	 * Note that no trees have been loaded at this point. */
	public CAPE_GUIDistributionManager()
	{
		initGUI();
	}
	
	//------------------------------------------------------------------------------
	
	/** Constructor loads the given <code>trees</code>, sets the
	 * <code>CAPE_GUI</code>'s client name to the given <code>clientName</code>
	 * and initializes the <code>CAPE_GUI</code>.
	 * @param trees the trees to be loaded into the <code>CAPE_GUI</code>
	 * @param clientName the name of the client query engine machine. For example, "Machine 1". */
	public CAPE_GUIDistributionManager(XATTree[] trees, String clientName)
	{
		this._XATTrees = trees;
		this._clientName = clientName;
		initGUI();
	}
	
	//------------------------------------------------------------------------------
	
	/** Constructor loads the given <code>trees</code>, sets the
	 * <code>CAPE_GUI</code>'s client name to the given <code>clientName</code>,
	 * sets the <code>CAPE_GUI</code>'s system object to the given <code>systemObject</code>
	 * and initializes the <code>CAPE_GUI</code>.
	 * @param trees the trees to be loaded into the <code>CAPE_GUI</code>
	 * @param clientName the name of the client query engine machine. For example, "Machine 1"
	 * @param systemObject the object used to gather system statistics */
	public CAPE_GUIDistributionManager(XATTree[] trees, String clientName, SystemObject systemObject){
		this._XATTrees = trees;
		this._clientName = clientName;
		this._systemObject = systemObject;
		initGUI();
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the name of the <code>CAPE_GUI</code>. For example, if this <code>CAPE_GUI</code>
	* runs on the distribution manager, then the name would be "<code>DISTRIBUTION_MANAGER</code>".
	* The <code>CAPE_GUI</code> named "<code>DISTRIBUTION_MANAGER</code>" is flagged for the <code>BasicGraphModelMaker</code>
	* to add labels to the operators identifying which machine that specific operator is running on.
	* <code>CAPE_GUI</code>s running on quey machines are named
	* "<code>QUERY_ENGINE</code>" and lack this machine identifier.
	* @return the name of the <code>CAPE_GUI</code> */	
	public String getName()
	{
		return _GUIName;
	}
	
	//------------------------------------------------------------------------------
	
	/** Initializes the <code>CAPE_GUI</code>. */
	public void init(){
		//Set the look and feel
		JFrame.setDefaultLookAndFeelDecorated(false);
		
		//Create and setup window
		JFrame frame = new JFrame("CAPE Distribution Manager");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		//Create and setup content pane
		CAPE_GUIDistributionManager gui = this;
		gui.setOpaque(true);
		frame.setJMenuBar(gui.getJMenuBar());
		frame.setContentPane(gui);
		
		//Display window
		frame.pack();
		Toolkit tk = Toolkit.getDefaultToolkit();
		frame.setSize(new Dimension(new Double(tk.getScreenSize().getWidth()*0.7).intValue(), 
			new Double(tk.getScreenSize().getHeight()*0.8).intValue()));
		frame.setLocation(new Point(new Double(tk.getScreenSize().getWidth()*0.15).intValue(),
			new Double(tk.getScreenSize().getHeight()*0.1).intValue()));
		frame.setVisible(true);
		frame.toFront();
		
		//Draw the trees
		getSplitPane().setDividerLocation(0.60);
		
		try
		{
			for(int i=0; i<getXATTreeCount(); i++){
				System.err.println(i);
				getGraphManager().drawModel(getModelMaker().makeModel(getXATTree(i)),getXATTree(i));
		
		}
		
		}
		catch(Exception e)
		{
			e.printStackTrace();
			System.err.println("No trees have been loaded.");}
		
		
		getStatisticStorage().initialize();
		makeUpdateQueuesTimer().start();
		makeUpdateNodesTimer().start();
		makeUpdateStatisticTableTimer();
	}
	
	//------------------------------------------------------------------------------
	
	/** Main method. */
	public static void main(String[] args)
	{
	
		javax.swing.SwingUtilities.invokeLater(new Runnable() 
		{
			 CAPE_GUIDistributionManager gui = new CAPE_GUIDistributionManager();
			
			 public void run() 
			 {
				 gui.init();
			 }
		 });	
	}
	
	//------------------------------------------------------------------------------
	
	/** Run method. */
	public void run()
	{	
		javax.swing.SwingUtilities.invokeLater(new Runnable() 
		{
			 public void run() 
			 {
				 init();
			 }
		 });	
	}
	
	//------------------------------------------------------------------------------
	
}