/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage;

import java.util.HashMap;
import java.util.ArrayList;
import java.util.Set;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;

/** The <code>TreeStorage</code> class stores all of the trees' operators and queues.
 * The <code>HashMap</code>'s 'key' is the <code>XATTree</code> and the 'value'
 * is a <code>HashMap</code>. This second <code>HashMap</code>'s 'key' is a <code>String</code>
 * ('<code>OPERATORS</code>' or '<code>QUEUES</code>') and 'value' is an <code>ArrayList</code> of either operators
 * or queues. The list of operators and queues is unordered.
 * 
 * @author Maylene Waltz
 */
public class TreeStorage{
	
	/** The key is an <code>XATTree</code> and value is an <code>ArrayList</code>
	 * of <code>XATQueryObjects</code> (operators or queues). */
	private HashMap _trees;
	
	/** Initializes the <code>TreeStorage HashMap</code>. */
	public TreeStorage(){
		_trees = new HashMap();
	}
	
	/** Returns the <code>XATStreamOperator</code> with the given <code>objectID</code>. */
	public Object getOperator(String objectID){
		try{
			//for every operator see if its OBJECT_ID matches the given objectID
			for(Iterator ops = getAllOperators(); ops.hasNext();){
				XATStreamOperator op = (XATStreamOperator)ops.next();
				if(op.getProperty("OBJECT_ID").equals(objectID)){
					return op;
				}
			} 
		}catch(Exception e){System.err.println("The XATStreamOperator with the given objectID does not exist.");}
		//never gets here
		return new Object();
	}
	
	/** Returns the <code>XATQueue</code> with the given <code>objectID</code>.
	 * This unique id (i.e. T1_O3_Q0) is constructed by the string:
	 * "T" + <code>TREE_ID</code> + "_O" + OBJECT_ID + "_Q" + queueNumber. 
	 * @return the queue with the given <code>objectID</code> */
	public Object getQueue(String objectID){
		try{
			//ids[0] = 'T' + TREE_ID, ids[1] = 'O' + OBJECT_ID and ids[2] = 'Q' queueNumber
			String[] ids = objectID.split("\\p{Punct}");
			
			//get the tree with this queue
			XATTree tree = (XATTree)getTree(ids[0].substring(1));
			
			//get this tree's operator(s)
			ArrayList ops = (ArrayList)getOperators(tree);
			
			//for each operator see if its OBJECT_ID matches ids[1]
			for(Iterator opsIt = ops.iterator(); opsIt.hasNext();){
				XATStreamOperator op = (XATStreamOperator)opsIt.next();
				//if the OBJECT_ID matches ids[1]
				if(op.getProperty("OBJECT_ID").equals(ids[1].substring(1))){
					//get the operators output queues
					for(Iterator oqs = op.getOutputQueues(); oqs.hasNext();){
						XATQueue q = (XATQueue)oqs.next();
						//if this queue's number matches ids[2] we found our queue!
						if(q.getQueueNumber() == Integer.parseInt(ids[2].substring(1))){
							return q;
			}}}}
		}catch(Exception e){e.printStackTrace();System.err.println("The XATQueue with the given objectID does not exist.");}
		//never gets here
		return new Object();
	}
	
	/** Returns the <code>XATTree</code> with the given <code>objectID</code>. */
	public Object getTree(String objectID){
		try{
			for(Iterator trees = getAllTrees(); trees.hasNext();){
				XATTree tree = (XATTree)trees.next();
				if(tree.getProperty("TREE_ID").equals(objectID)){
					return tree;
				}
			}
		}catch(Exception e){System.err.println("The XATTree with the given objectID does not exist.");}
		//never gets here
		return new Object();
	}
	
	/** Returns an <code>ArrayList</code> of <code>XATQueryObjects</code> (operators) that belong
	 * to the given <code>tree</code>.
	 * @param key - edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
	 * @return Object - java.util.ArrayList */
	public Object getOperators(XATTree tree){
		return ((HashMap)_trees.get(tree)).get("OPERATORS");
	}
	
	/** Returns an <code>ArrayList</code> of <code>XATQueryObjects</code> (queues) that belong
	 * to the given <code>tree</code>.
	 * @param key - edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
	 * @return Object - java.util.ArrayList */
	public Object getQueues(XATTree tree){
		return ((HashMap)_trees.get(tree)).get("QUEUES");
	}
	
	/** Returns an <code>Iterator</code> of all the operators. 
	 * @return java.util.Iterator */
	public Iterator getAllOperators(){
		ArrayList operators = new ArrayList();
		//for each tree's HashMap
		for(Iterator it = _trees.values().iterator(); it.hasNext();){
			ArrayList opsList = (ArrayList)((HashMap)it.next()).get("OPERATORS");
			//for each operator in an ArrayList
			for(Iterator it2 = opsList.iterator(); it2.hasNext();){
				//add the operator to the operators ArrayList
				operators.add(it2.next());
			}
		}return operators.iterator();
	}
	
	/** Returns an <code>Iterator</code> of all the queues.
	 * @return java.util.Iterator */
	public Iterator getAllQueues(){
		ArrayList queues = new ArrayList();
		//for each tree's HashMap
		for(Iterator it = _trees.values().iterator(); it.hasNext();){
			ArrayList qList = (ArrayList)((HashMap)it.next()).get("QUEUES");
			//for each queue in an ArrayList
			for(Iterator it2 = qList.iterator(); it2.hasNext();){
				//add the queue to the queues ArrayList
				queues.add(it2.next());
			}
		}return queues.iterator();
	}
	
	/** Returns an <code>Iterator</code> of all the trees.
	 * @return java.util.Iterator */
	public Iterator getAllTrees(){
		return getKeys().iterator();
	}
	
	/** Returns a <code>Set</code> of keys (or <code>XATTree</code>s) of this <code>HashMap</code>. 
	 * @return java.util.Set */
	public Set getKeys(){
		return _trees.keySet();
	}
	
	/** Puts the given tree's operators into the storage <code>HashMap</code>.
	 * @param key - edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
	 * @param ops - java.util.ArrayList */
	public void setOperators(XATTree key, ArrayList ops){
		//if this tree is not yet in the HashMap, then add it
		if(_trees.get(key) == null){
			HashMap treeMap = new HashMap();
			treeMap.put("OPERATORS", ops);
			_trees.put(key, treeMap);
		}else{
			((HashMap)_trees.get(key)).put("OPERATORS", ops);
		}
	}
	
	/** Puts the given tree's queues into the storage <code>HashMap</code>.
	 * @param key - edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
	 * @param qs - java.util.ArrayList */
	public void setQueues(XATTree key, ArrayList qs){
		//if this tree is not yet in the HashMap, then add it
		if(_trees.get(key) == null){
			HashMap treeMap = new HashMap();
			treeMap.put("QUEUES", qs);
			_trees.put(key, treeMap);
		}else{
			((HashMap)_trees.get(key)).put("QUEUES", qs);
		}
	}
}