/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.util.HashMap;
import javax.swing.JFrame;

import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.Point;

/** The <code>FramesManager</code> manages open frames and 
 * the making of new frames. The <code>HashMap</code>'s key is a unique
 * frame name and value is the actual <code>JFrame</code>.
 * 
 * @author Maylene Waltz
 */
public class FramesManager{

	/** The CAPE_GUI */
	private CAPE_GUI _gui;
	
	/** A <code>HashMap</code> that holds the open frames (or windows). */
	private HashMap _openFrames;
	
	/** The number of open plot windows. */
	private int _openPlotsCount = 0;
	
	/** Constructor that initializes the <code>HashMap</code> to hold open frames
	 * and associates the given <code>gui</code> with this <code>FramesManager</code>. 
	 * @param gui the <code>CAPE_GUI</code> to associate with this <code>FramesManager</code> */
	public FramesManager(CAPE_GUI gui){
		_gui = gui;
		_openFrames = new HashMap();
	}
	
	/** Adds the given <code>frame</code> to the <code>HashMap</code> of open frames.
	 * The <code>frame</code>'s name is the key and the <code>frame</code> is the value.
	 * @param frame the frame to be added to the <code>HashMap</code> of open frames */
	private void addOpenFrame(JFrame frame){
		_openFrames.put(frame.getName(), frame);
	}
	
	/** Decrements the count variable for the number of plot windows open. */
	public void decrementOpenPlotsCount(){
		_openPlotsCount--;
	}
	
	/** Returns the open frame with the given <code>key</code>.
	 * @return the open frame associated with the given <code>key</code> */
	public JFrame getFrame(String key){
		return (JFrame)_openFrames.get(key);	
	}
	
	/** Returns the number of open plot windows. */
	public int getOpenPlotsCount(){
		return _openPlotsCount;
	}
	
	/** Increments the count variable for the number of plot windows open. */
	public void incrementOpenPlotsCount(){
		_openPlotsCount++;
	}
	
	/** Returns whether or not the frame with the given <code>key</code>
	 * is open or not.
	 * @param key the <code>key</code> to match to see if the frame is open
	 * @return <code>true</code>, if the frame with the given <code>key</code> is open; 
	 * <code>false</code>, otherwise */
	public boolean isOpenFrame(String key){
		if(_openFrames.containsKey(key)) return true;
		else return false;
	}
	
	/** Makes a new frame and adds it to the <code>HashMap</code> of open frames
	 * using the given <code>key</code>. Sets the frame's title to the given <code>frameTitle</code>.
	 * @param key the key to use to add this frame to the <code>HashMap</code> of open frames
	 * @param frameTitle the title to use for this frame
	 * @return the newly created frame */
	public JFrame makeNewFrame(String key, String frameTitle){
		if(_gui.getPanelMaker().makePanel(key) == null){
			return null;
		}else{
			JFrame frame = new JFrame(frameTitle);
			frame.setName(key);
			frame.addWindowListener(new CloseWindowListener());
			frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			frame.getContentPane().add(_gui.getPanelMaker().makePanel(key));
			frame.pack();
			frame.setVisible(true);
			addOpenFrame(frame);
			return frame;
		}	
	}
	
	/** Makes a new frame and adds it to the <code>HashMap</code> of open frames
	 * using the given <code>key</code>. Sets the frame's title to the given <code>frameTitle</code>.
	 * Places the frame at the given <code>location</code> on the screen.
	 * @param key the key to use to add this frame to the <code>HashMap</code> of open frames
	 * @param frameTitle the title to use for this frame
	 * @param location the point at which to place the frame on the screen
	 * @return the newly created frame */
	public JFrame makeNewFrame(String key, String frameTitle, Point location)
	{
		if(_gui.getPanelMaker().makePanel(key) == null)
		{
			//System.out.println("**** DOES NOT ATTEMPT TO MAKE NEW FRAME ****");
			return null;
		}
		else
		{
			//System.out.println("**** ATTEMPTS TO MAKE NEW FRAME ****");
			JFrame frame = new JFrame(frameTitle);
			frame.setName(key);
			frame.setLocation(location);
			frame.addWindowListener(new CloseWindowListener());
			frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			
			//frame.getContentPane().removeAll();
			frame.getContentPane().add(_gui.getPanelMaker().makePanel(key));
			frame.pack();
			frame.setVisible(true);
			addOpenFrame(frame);
			return frame;
		}
	}
	
	/** Removes the given <code>frame</code> from the <code>HashMap</code> of open frames. 
	 * @param frame the frame to remove from the <code>HashMap</code> of open frames */
	public void removeOpenFrame(JFrame frame){
		_openFrames.remove(frame.getName());
		frame.dispose();
	}
	
	/** Removes the frame with the given <code>key</code> from the <code>HashMap</code> of open frames.
	 * @param key the key of the frame to remove from the <code>HashMap</code> of open frames. */
	public void removeOpenFrame(String key){
		((JFrame)_openFrames.get(key)).dispose();
		_openFrames.remove(key);
	}
	
	/** Resets the counter for the number of open plot windows. */
	public void resetOpenPlotsCount(){
		_openPlotsCount = 0;
	}
	
	/** Focuses the frame with the given <code>key</code> to the front (or top).
	 * @param key the key of the frame to focus to the front */
	public void toFront(String key){
		((JFrame)_openFrames.get(key)).toFront();
	}
	
	/** The listener to handle window/frame closing events.
	 * Removes the window being closed from the <code>FramesManager</code>'s open frames. */
	public class CloseWindowListener extends WindowAdapter{
		public void windowClosing(WindowEvent e){
			removeOpenFrame((JFrame)e.getComponent());
		}
	}
}