/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import javax.swing.JComponent;
import javax.swing.JTabbedPane;

import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;

import java.awt.Rectangle;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;

/** <code>JTabbedPaneCAPE</code> manages the creation of tabbed panes, 
 * adding tabs to tabbed panes, and handles mouse interaction on the tabs.
 * 
 * @author Maylene Waltz
 */
public class JTabbedPaneCAPE extends JTabbedPane implements MouseListener{
	
	/** The <code>CAPE_GUI</code>. */
	private CAPE_GUI _gui;
	
	/** Constructor makes an empty tabbed pane with the default tab
	 * placement of <code>JTabbedPane.TOP</code>. */
	JTabbedPaneCAPE(){
		super();
		addMouseListener(this);
	}
	
	/** JTabbedPaneCAPE constructor makes a tabbed pane with the specified <code>tabPlacement</code> 
	 * and <code>tabLayoutPolicy</code>.
	 * @param tabPlacement the location of the tabs (i.e. JTabbedPane.TOP)
	 * @param tabLayoutPolicy the layout policy for the tabs (i.e. JTabbedPane.WRAP_TAB_LAYOUT) */
	JTabbedPaneCAPE(int tabPlacement, int tabLayoutPolicy){
		super(tabPlacement, tabLayoutPolicy);
		addMouseListener(this);
	}
	
	/** JTabbedPaneCAPE constructor makes a tabbed pane with the specified <code>tabPlacement</code> 
	 * and <code>tabLayoutPolicy</code> and associates the given Rainstorm
	 * <code>gui</code> with this tabbed pane.
	 * @param tabPlacement the location of the tabs (i.e. JTabbedPane.TOP)
	 * @param tabLayoutPolicy the layout policy for the tabs (i.e. JTabbedPane.WRAP_TAB_LAYOUT)
	 * @param gui the <code>CAPE_GUI</code> this tabbed pane is associated with */
	JTabbedPaneCAPE(int tabPlacement, int tabLayoutPolicy, CAPE_GUI gui){
		super(tabPlacement, tabLayoutPolicy);
		addMouseListener(this);
		_gui = gui;
	}
	
	/** Adds a tab with the given <code>title</code> and <code>component</code> to the tabbed pane.
	 * @param title the title for the tab
	 * @param component the component to place in the tab's panel */
	public void addTab(String title, JComponent component){
		super.addTab(title, new CAPETab(), component);	
	}
	
	/** Adds a tab with the given <code>title</code> and <code>component</code> to the tabbed pane.
	 * The <code>paneType</code> specifies the type of tabbed pane. For example, "QUERY" for
	 * the tabbed pane of query graphs and "STATS" for the statistics tabbed pane.
	 * @param title the title for the tab
	 * @param component the component to place in the tab's panel
	 * @param paneType the type of pane specifies the type of tab to use */
	public void addTab(String title, JComponent component, String paneType){
		super.addTab(title, new CAPETab(paneType), component);
	}
	
	/** Adds a tab at the specified <code>index</code> in the tabbed pane with the given <code>component</code>.
	 * @param index the position in the tabbed pane to add the tab
	 * @param component the component to place in the tab's panel */
	public void addTab(int index, JComponent component){
		this.addTab(new Integer(index).toString(), component);
	}
	
	/** Adds a tab with the given <code>component</code> and sets the tab title to:
	 * <ul>
	 * <li>For trees: 'T' + TREE_ID
	 * </ul>
	 * If the given <code>queryObject</code> is an <code>XATTree</code>, then the title is set to 'T' + TREE_ID.
	 * The <code>paneType</code> specifies the type of tabbed pane. For example, "QUERY" for
	 * the tabbed pane of query graphs and "STATS" for the statistics tabbed pane.
	 * @param title the tab title
	 * @param queryObject the query object to associate with this tab
	 * @param component the component to place in the tab's panel
	 * @param paneType the type of pane specifies the type of tab to use
	 */
	public void addTab(String title, Object queryObject, JComponent component, String paneType){
		super.addTab(title, new CAPETab(queryObject, paneType), component);
	}
	
	/** Handles the <code>mouseClicked</code> event on a tab.
	 * If the <code>paneType</code> is "STATS" then hitting 'X' closes the statistics table.
	 * If the <code>paneType</code> is "QUERY" then hitting 'X' closes the graph and removes the graph
	 * from CAPE's open graphs. */
	public void mouseClicked(MouseEvent e){
		int tabNum = getUI().tabForCoordinate(this, e.getX(), e.getY());
		if(tabNum < 0) return;
		
		//----------- paneType is "QUERY" ----------//
		if(((CAPETab)getIconAt(tabNum)).getPaneType() == "QUERY"){
			Rectangle rect = ((CAPETab)getIconAt(tabNum)).getBounds();
			Rectangle closeRect = new Rectangle((int)rect.getX(), (int)rect.getY(), rect.width, rect.height);
			//close tab
			if(closeRect.contains(e.getX(), e.getY())){
				//if this is the last tab open stop timers
				if(_gui.getGraphManager().getGraphTabbedPane().getTabCount() == 1){
					_gui.getUpdateNodesTimer().stop();
					_gui.getUpdateQueuesTimer().stop();
					_gui.getUpdateStatisticTableTimer().stop();
				}
				_gui.getGraphManager().removeGraph(tabNum);
				this.removeTabAt(tabNum);
			} else return;
			
		//---------- paneType is "STATS" ----------//
		} else if(((CAPETab)getIconAt(tabNum)).getPaneType() == "STATS"){
			Rectangle rect = ((CAPETab)getIconAt(tabNum)).getBounds();
			Rectangle closeRect = new Rectangle((int)rect.getX(), (int)rect.getY(), rect.width, rect.height);
			//close tab
			if(closeRect.contains(e.getX(), e.getY())){
				//if this is the last tab open stop timer
				if(_gui.getGraphManager().getStatsTabbedPane().getTabCount() == 1){
					_gui.getUpdateStatisticTableTimer().stop();
				}
				this.removeTabAt(tabNum);
			}
		}
	}
	/** Handles the <code>mouseEntered</code> event on a tab. Does nothing. */
	public void mouseEntered(MouseEvent e){}
	/** Handles the <code>mouseExited</code> event on a tab. Does nothing. */
	public void mouseExited(MouseEvent e){}
	/** Handles the <code>mousePressed</code> event on a tab. Does nothing. */
	public void mousePressed(MouseEvent e){}
	/** Handles the <code>mouseReleased</code> event on a tab. Does nothing. */
	public void mouseReleased(MouseEvent e){}
}

