/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics;

//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
//import edu.wpi.cs.cs4432.project3.rewritingrules.SwitchOperatorRules;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller.XATBox;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.rewritingrules.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;

/**
 * @author Yali Zhu
 * April 1, 2006
 * 
 * This heuristics go through the tree and find a pair of WTransitiveHashJoinOperators, namely lowerjoin and highjoin. 
 * If lowerjoin is the left child of higherjoin, the two joins are switched if:
 *   rightstate of higherjoin < rightstate of lowerjoin
 * 
 * If lowerjoin is the right child of higherjoin, the two joins are switched if:
 *   rightstate of higherjoin < rightstate of lowerjoin
 * 
 * The logic is that the two joins are switched if the larger join can generate smaller intermediate state than 
 * the lower join does. It assumes that all join operators have about the same selectivities. 
 * 
 * Right now this class is somewhat hard-coded to cope with the case that the tree has two WTransitiveHashJoinOperators,
 * three input streams, three split operators and one union operator as the root. 
 * During the experiments, the arrival rate of input stream C is being changed over time. So it may be necessary
 * to switch the two joins several times. The input rates of stream A and B are kept the same. 
 * 
 */
public class SwitchJoinByStateSizeHeuristics extends Heuristic{
	
	public SwitchJoinByStateSizeHeuristics() {
		super();
	}
	/**
	 * Go through each node of the tree and apply the switch join heuristics
	 * 
	 * @param tree to be traversed
	 * @return boolean whether heuristics was applied
	 */
	public boolean ApplyHeuristic(XATBox newbox, XATTree oldtree) {
		
	    //System.out.println("Apply Hueristics:" + this);
		this.box = newbox;
		this.tree = oldtree;

		XATStreamOperator[] operators;
		
		//If efficient migration is used
		if(box == null) {
			XATNode[] nodes = oldtree.convertToArray();
			if(nodes.length == 0)
				return false;
			
			operators = new XATStreamOperator[nodes.length];
			for(int i=0; i<nodes.length; i++) {
				operators[i] = nodes[i].getStreamOperator();
			}	
		} 
		//If old migration strategy is used (need to make a copy of current plan), get the operator from the box
		else {
			operators = box.getAllOperators();
		}
		
		// check if there's anything in the box		
		if(operators.length == 0)
			return false;
		
		return switchJoins(operators);
	}
	
	/**
	 * This method compare a node and if it is a join operator that has join operators
	 * as its children, switch that node with the child that has the highest selectivity
	 * 
	 * @param node parent
	 * @return whether any switching of joins is done
	 */
	public boolean switchJoins(XATStreamOperator[] operators)
	{	
		try{
			for(int i=0; i<operators.length; i++) {
			
				if(operators[i] instanceof WTransitiveHashJoinOperator) {
					//System.out.println ("Found a join operator");
					XATNode inode = operators[i].getNode();
				
					// if already at leaf level, continue to the next operator in the box
					if(inode.getOpChildrenSize() == 0) {
						continue;
					}
								
					XATNode[] cnodes = inode.getOpChildren();
						
					for(int j=0; j<cnodes.length; j++) {
					    if (cnodes[j].getStreamOperator() instanceof WTransitiveHashJoinOperator) {
					        System.out.println("switch parent join with child at position: " + j);
				            XATStreamOperator upperjoin = inode.getStreamOperator();
				            XATStreamOperator lowerjoin = cnodes[j].getStreamOperator();
				            upperjoin.cleanAllAccumulatedTuples();
					        if (this.box == null) {
/** CHUAN LEI - Couldn't find DistributedExecutionController, so pass 'null' as the last parameter into the following method calls*/
					            SwitchOperatorRules.Switch(upperjoin, lowerjoin, 0, null, null);
					        } else {
/** CHUAN LEI - Couldn't find DistributedExecutionController, so pass 'null' as the last parameter into the following method calls*/
					            SwitchOperatorRules.Switch(upperjoin, lowerjoin, this.box.getStateCounter(), this.box, null);
					        }
					        //System.out.println("switch " + inode + " " + cnodes[0]);
					        return true;
					    }
					}
				}
			}
		} catch(Exception e) {
			e.printStackTrace();	
		}	
		return false;
	}
	
	/**
	 * check if the given box have any two consecutive join operators that can be switched.
	 * 
	 * the method returns TRUE right away if it finds the first chance of switch two join operators.
	 * it returns FALSE is none of such chance can be found.
	 */
	public boolean CheckHeuristic(XATBox box, XATTree tree) {	
		return true;
	}	
}
