/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

/**
 * @author Yali Zhu
 * 
 * This class is the partial join entry generated by the dynamic programming method
 * to exhaustively search for the qualified mjoin query plan.
 */
public class DPJoinEntry {
	
	private static final double COST_TUPLE_INSERT = 0.0002;
	private static final double COST_TUPLE_DELETE = 0.0002;
	private static final double COST_TUPLE_JOIN = 0.0022;
	
	int streamsize;         // indicates how many streams involved in this partial join
	double memoryCost;  // memory cost of this partial join according to cost model
	double CPUCost;     // cpu cost of this partial join according to cost model.
	double treememorycost;
	double treecpucost;
	double outputRate;  // output rate of this partial join according to cost model.
	int inputSize;
	double[] inputRates;
	double[][] selectivities;
	long timeWindow;
	int[] childrenstreamsizes;
	// String with input names and parenthese, representing partial join order.
	// For example: ((AB)C) means that A joins B first, the results then join with C.
	// Each left parenthesis must be ended with a matching right parenthesis.
	// This can later be converted to a join tree using a stack.
	String pstring;
	
	private int[][] probeOrders;
	private double[][] probeSelectivities;
	
	public DPJoinEntry(int streamsize, String pstring) {
		
		// allocate just enough space for representing stream name 
		//(assumed to be single character} as well as parenthesis.
		//pstring = new char[length + (length-1)*2];
		this.streamsize = streamsize;
		this.pstring = pstring;
	}
	
	public void setPString(String s) {
		this.pstring = s;
	}

	/**
	 * If no exising character at the position, just insert the character.
	 * Otherwise, shift the current character and all characters after it to the left.
	 * @param position
	 */
	public void insertOrderCharacter(int position) {
		
	}
	
	public void setStreamSize(int s) {
		this.streamsize = s;
	}
	
	public int getStreamSize() {
		return this.streamsize;
	}
	
	public double getOutputRate() {
		return this.outputRate;
	}
	
	public void setInputSize(int i) {
		this.inputSize = i;
	}
	
	public void setInputRates(double[] rates) {
		this.inputRates = rates;
	}
	
	public void setInputStreamSizes(int[] streamsizes) {
		this.childrenstreamsizes = streamsizes;
	}
	
    /**
     * calculate memory and cpu costs by combining the two input partial join orders.
     * 
     * entryL and entryR are treated as two inputs.
     * 
     * entryL: (AB)  entryR: ((CD)E)
	 * combined entry:  ((AB)((CD)E))
     * @param entryL
     * @param entryR
     */
	public void calculateNewCombinedBinaryCosts(DPJoinEntry entryL, DPJoinEntry entryR) {
		
	}
	
	/**
	 * calculate memory and cpu costs of the new combined multi-way join.
	 * 
	 * entry 1 is treated as one input, while entry 2 is treates as multiple input, which means
	 * that the outside most parenthesis of entry 2 is removed before entry 2 is combined with entry 1.
	 * 
	 * entry1: (AB)  entry2: ((CD)E)
	 * combined entry:  ((AB)(CD)E)
	 * 
	 * The modified IK join ordering algorithm for mjoin needs to be first invoked to compute the join ordering 
	 * inside the new root mjoin. The cost of memory and cpu can then be computed according to cost models.
	 * 
	 * @param entry1
	 * @param entry2
	 */
	public void calculateNewCombinedMultiwaycosts(DPJoinEntry entry1, DPJoinEntry entry2) {
		
	}
	
	
	
	public void calculateOutputRate() {
		outputRate = 1;
		for(int i=0; i<inputRates.length; i++)
			outputRate *= inputRates[i] * timeWindow/1000;
		
		for(int i=0; i<this.selectivities.length; i++) {
			for(int j=i+1; j<this.selectivities.length; j++) {
				outputRate *= selectivities[i][j];
			}
		}
		
		outputRate = outputRate *1000 / timeWindow;
	}
	
	public void setMemoryCost(double m) {
		this.memoryCost = m;
	}
	
	public void setCPUCost(double c) {
		this.CPUCost = c;
	}
	
	public void setOutputRate(double r) {
		this.outputRate = r;
	}

	public void calculateCPUCost() {
		this.CPUCost = 0;
		
		for(int i=0; i<this.inputSize; i++) {
			// first the cost of inserting and deleting tuples to/from states
			this.CPUCost += this.inputRates[i]*this.childrenstreamsizes[i] * (COST_TUPLE_INSERT + COST_TUPLE_DELETE);
			
			double inter_results = this.inputRates[i] * this.childrenstreamsizes[i];
			double temp;
			// then the cost of probing each of the remaining states
			for(int k=0; k<(this.inputSize-1); k++) {
				int probedinput = this.probeOrders[i][k];
				double statesize = this.inputRates[probedinput] * this.timeWindow/1000;
				temp = inter_results * statesize * this.probeSelectivities[i][k];
				this.CPUCost += temp * (COST_TUPLE_JOIN);
				inter_results = temp;
			}
		}
	}

	public void calculateMemoryCost() {
		this.memoryCost = 0;
		
		for(int i=0; i<inputRates.length; i++)
			memoryCost += inputRates[i] * this.timeWindow/1000;

	}

	public void calculateProbeOrders() {
		this.probeOrders = new int[this.inputSize][this.inputSize - 1];
		//this.joinGraph = null;
		JoinGraph joinGraph;
		joinGraph = new JoinGraph(this.selectivities, this.inputRates, true);
		joinGraph.setTimeWindow(this.timeWindow);
		
		//this.joinGraph.setArrivalRates(this.arrivalRates);
		//this.joinGraph.setSelectivities(this.selectivityMatrix);
		
		for (int i = 0; i < this.inputSize; i++) {
			this.probeOrders[i] = joinGraph.computeIthJoinOrder(i);
		}
		
		//printProbeOrders();
		calculateProbeSelectivities();
		//printProbeSelectivities();
	}

	/**
	 * Compute the correct selectivities given the computed probe sequences.
	 * Also set up the random generator for each selectivity;
	 * This selectivity matrix is used at runtime for looking up certain selectivity between two consecutive 
	 * stream pairs in a join sequence.
	 */
	private void calculateProbeSelectivities() {
		this.probeSelectivities = new double[this.inputSize][this.inputSize - 1];
		
		for(int i=0; i<this.inputSize; i++) {
			int index0 = this.probeOrders[i][0];
			probeSelectivities[i][0] = this.selectivities[i][index0];
			for(int j=1; j<this.inputSize-1; j++) {
				int indexj = this.probeOrders[i][j];
				probeSelectivities[i][j] = this.selectivities[i][indexj];
				for(int k=0; k<j; k++) {
					int indexk = this.probeOrders[i][k];
					probeSelectivities[i][j] *= this.selectivities[indexj][indexk];
				}
			}
		}
	}

	private void printProbeOrders() {
		for(int i=0; i<this.inputSize; i++) {
			System.out.print("join order for input "+i + ": ");
			for(int k=0; k<this.inputSize - 1; k++) {
				System.out.print(" -> " + this.probeOrders[i][k]);
			}
			System.out.println();
		}
	}

	private void printProbeSelectivities() {
		for(int i=0; i<this.inputSize; i++) {
			System.out.print("join selectivity for input "+i + ": ");
			for(int k=0; k<this.inputSize - 1; k++) {
				System.out.print(" -> " + this.probeSelectivities[i][k]);
			}
			System.out.println();
		}
	}
	
	public void setTreeMemoryCost(double m) {
		this.treememorycost = m;
	}
	
	public void setTreeCPUCost(double c) {
		this.treecpucost =c;
	}
}
