/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATSimpleQueueImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ComplexOperator.JoinGraph;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ComplexOperator.ProbabilisticWMJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ComplexOperator.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator.WMJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;


/**
 * @author Yali Zhu
 *
 * Initial optimization: apply two-layer optimization to the multiple joins, depends on the pre-defined running mode given
 * in the query plan configuration file. An example of such query plan configuration file is given in
 * resources\MJoinTree\pmjoin-3.xml
 * 
 */
public abstract class MultiJoinOptimizer {
	
	private static final double COST_TUPLE_INSERT = 0.0002;
	private static final double COST_TUPLE_DELETE = 0.0002;
	private static final double COST_TUPLE_JOIN = 0.0022;
	
	protected XATDag queryPlan;
	protected String queueModel;
	protected long timewindow; // in milliseconds
	protected String optimizationMode;
	protected JoinGraph joinGraph;
	
	protected double memorythreshold = 800;
	protected double CPUthreshold = 1;
	
	protected boolean treeIsChanged = false;
	Hashtable streamqueues = new Hashtable();
	Hashtable allselectivities = new Hashtable();
	
	public MultiJoinOptimizer(XATDag queryplan){
		this.queryPlan = queryplan;
		
		if(queryplan == null) {
			return;
		}
		
		//figure out stream input queues
		XATNode[] nodes = queryplan.convertToArray();
		
		for(int i=0; i<nodes.length; i++) {
			Object[] children = nodes[i].getChildren();
			
			for(int k=0; k<children.length; k++) {
				if(children[k] instanceof String) {
					XATStreamOperator operator = nodes[i].getStreamOperator();
					XATQueue queue = operator.getInputQueue(k);
					streamqueues.put(children[k], queue);
				}
			}
		}
	}
	
	public abstract XATDag optimize(XATDag tree);
	
	/**
	 * Assumes the starting query plan only has a single mjoin operator
	 * 
	 * This method finds a qualifed query plan using different optimization method, which is specified in
	 * query plan config file.
	 *
	 */
	public void initialOptimization() {
	
		XATNode[] nodes = queryPlan.convertToArray();
		
		if(nodes.length == 1 && nodes[0].getStreamOperator() instanceof WMJoinOperator) {
			WMJoinOperator op = (WMJoinOperator)nodes[0].getStreamOperator();
			int numOfInputs = op.getNumOfInputs();
			int[][] probeSequence = new int[numOfInputs][numOfInputs - 1];
			
		    //generate join graph representation
			//compute join orderings for starting mjoin operator
			probeSequence = new int[numOfInputs][numOfInputs - 1];
			this.joinGraph = new JoinGraph(op.getSelectivityMatrix(), op.getArrivalRates(), op.getCartesionProductStatus());
			this.joinGraph.setTimeWindow(op.getTimeWindow());

			for (int i = 0; i < numOfInputs; i++) {
				probeSequence[i] = this.joinGraph.computeIthJoinOrder(i);
			}		
			op.setProbeSequences(probeSequence, null);
		}
	}
	
	/**
	 * initial optimization of the query plan after the query is being parsed from the query plan configuration file.
	 * 
	 * the initial optimization steps implemented so far:
	 * 1. Go through the query plan dag and optimize the ProbabilisticWMJoinOperator
	 *    according to the running mode given in the query plan config file.
	 *    An mjoin can be run in mjoin mode, binary tree mode or mjoin mode.
	 */
	
//	public void initOptimization() {
//	
//
//	}

	
	

	public void setQueueModel(String queuemodel) {
		this.queueModel = queuemodel;
		System.setProperty("DATA_MODEL", queuemodel);
	}

	/**
	 * This method computes the BTree with the minimal state sizes. 
	 * It is similar to the method generateMinStateBTree. 
	 * The difference is that instead of returning a real query tree, it returns a list of join entries.
	 * This is for the purpose of saving the unnecessary costs of creating a real query tree if the generated 
	 * BTree is in fact not qualified, and therefore needs to be further optimized to a qualifed mjoin tree.
	 * 
	 * The tree is generated based on the current join graph.
	 * If the BTree is indeed qualifed, the list of entries can be converted to a real query tree by calling
	 * the method  convertEntryListToRealQueryTree().
	 * @return a list of mjoinentries, each have their associated memory and cpu costs
	 */
	protected ArrayList generateMinStateBTreeEntries() {
		
		this.allselectivities.clear();
		
		//first initilized selectivity hashtable
		for(int i=0; i<this.joinGraph.numNodes; i++) {
			for(int j=0; j<this.joinGraph.numNodes; j++) {
				this.allselectivities.put(""+i+" "+j, ""+this.joinGraph.selectivities[i][j]);
			}
		}

		// first generate edge list from selectivity matrix
		ArrayList edgelist = new ArrayList();
		ArrayList joinentrylist = new ArrayList();
		double[][] selectivities = this.joinGraph.getSelectivities();
		double[] rates = this.joinGraph.getArrivalRates();
		long timewindow = this.joinGraph.getTimeWindow();
		
		// if use cartesian product
		if (this.joinGraph.getUseCartesianProductStatus()) {
			for(int i=0; i<selectivities.length; i++) {
				for (int j=i+1; j<selectivities.length; j++) {
						Edge edge = new Edge(i, j, rates[i], rates[j], selectivities[i][j]);
						edgelist.add(edge);
				}
			}
		} else {
			for (int i = 0; i < selectivities.length; i++) {
				for (int j=i+1; j<selectivities.length; j++) {
					if (selectivities[i][j] < 1) {
						Edge edge = new Edge(i, j, rates[i], rates[j], selectivities[i][j]);
						edgelist.add(edge);
					}
				}
			}
		}

		// the ID of the join entry starts from next number larger than the
		// largest input id.
		int minJoinID = rates.length;
		int joinID = minJoinID;

		// go through the edgelist, pick the smallest weighted edge and make a
		// join entry out of it.
		// remove the edge from the edgelist and update other edges connecting
		// to this edge in the edgelist.
		while (edgelist.size() > 0) {

			// find the edge with the smallest weight
			int size = edgelist.size();
			int minindex = 0;
			double minweight = ((Edge) edgelist.get(0)).weight;
			for (int i = 1; i < size; i++) {
				double temp = ((Edge) edgelist.get(i)).weight;
				if (temp < minweight) {
					minindex = i;
					minweight = temp;
				}
			}

			// make a join entry from the smallest weighted edge
			Edge edge = (Edge) edgelist.remove(minindex);
			////DagNode newnode = new DagNodeImp();
			////WMJoinOperator newoperator = new WMJoinOperator(newnode);
			////rootnode = newnode;

			double[][] selectivitymatrix = new double[2][2];
			double[] arrivalrates = new double[2];
			selectivitymatrix[0][1] = edge.selectivity;
			selectivitymatrix[1][0] = edge.selectivity;
			selectivitymatrix[0][0] = 1;
			selectivitymatrix[1][1] = 1;
			arrivalrates[0] = edge.rate1;
			arrivalrates[1] = edge.rate2;
			////newoperator.setRunningMode("mjoin");
			////newoperator.setNumOfInputs(2);
			////newoperator.setSelectivityMatrix(selectivitymatrix);
			////newoperator.setArrivalRates(arrivalrates);
			////newoperator.setTimeWindow(this.timewindow);
			////newoperator.setupStates();
			////newoperator.computeProbeSequence();

			
			int[] inputindex = new int[2];
			inputindex[0] = edge.v1;
			inputindex[1] = edge.v2;
			int[] inputstreamsize = new int[2];
			inputstreamsize[0] = edge.streamsize1;
			inputstreamsize[1] = edge.streamsize2;
			MJoinEntry newjoinentry = new MJoinEntry(joinID, inputindex, inputstreamsize, timewindow, null);
			joinID++;
			newjoinentry.setInputRates(arrivalrates);
			newjoinentry.setSelectivities(selectivitymatrix);
			newjoinentry.setTimeWindow(this.joinGraph.timeWindow);
			newjoinentry.calculateProbeOrders();
			newjoinentry.calculateCPUCost();
			newjoinentry.calculateMemoryCost();
			newjoinentry.calculateOutputRate();
			
			//newjoinentry.lchildID = edge.v1;
			//newjoinentry.rchildID = edge.v2;
			joinentrylist.add(newjoinentry);

			// if the input to this new join is the intermediate output from
			// another join, mark the parent/child relationship in each join entry
			if ((edge.v1 > minJoinID) || (edge.v2 > minJoinID)) {
				for (int i = 0; i < joinentrylist.size(); i++) {
					MJoinEntry entry = (MJoinEntry) joinentrylist.get(i);
					if (entry.entryID == edge.v1 || entry.entryID == edge.v2) {
						entry.parentID = newjoinentry.entryID;
					}
				}
			}

			//now update the edgelist
			for (int i = 0; i < edgelist.size(); i++) {
				Edge nextedge = (Edge) edgelist.get(i);
				if(nextedge == edge) {
					edgelist.remove(nextedge);
					i--;
					continue;
				}
				if (nextedge.v1 == edge.v1 || nextedge.v1 == edge.v2) {
					nextedge.v1 = newjoinentry.entryID;
					nextedge.streamsize1 = edge.streamsize1 + edge.streamsize2;
					nextedge.rate1 = edge.weight * timewindow / 1000;
					nextedge.updateWeight();
					continue;
				}
				if (nextedge.v2 == edge.v1 || nextedge.v2 == edge.v2) {
					nextedge.v2 = newjoinentry.entryID;
					nextedge.streamsize2 = edge.streamsize1 + edge.streamsize2;
					nextedge.rate2 = edge.weight * timewindow / 1000;
					nextedge.updateWeight();
				}
			}

			//now combine edges with the same vertices
			for (int i = 0; i < edgelist.size(); i++) {
				Edge edge1 = (Edge) edgelist.get(i);
				for (int j = i + 1; j < edgelist.size(); j++) {
					Edge edge2 = (Edge) edgelist.get(j);

					if ((edge1.v1 == edge2.v1 && edge1.v2 == edge2.v2)
							|| (edge1.v1 == edge2.v2 && edge1.v2 == edge2.v1)) {
						edge1.selectivity = edge1.selectivity * edge2.selectivity;
						this.allselectivities.put(""+edge1.v1+" "+edge1.v2, ""+edge1.selectivity);
						edge1.updateWeight();
						edgelist.remove(j);
						break;
					}
				}
			}
		}
		return joinentrylist;
	}
	
	/**
	 * Generate a binary join tree using the min-state algorithm. Upon return, a
	 * binary join tree is generated and has replaced the current
	 * ProbabilisticWMJoinOperator and its node in the current query plan tree.
	 * 
	 * @param operator
	 * @param node
	 * @return the root node of the min-state binary join tree.
	 * @deprecated
	 */
	protected XATNode generateMinStateBTree(WMJoinOperator operator, XATNode node) {

		double[][] selectivities = operator.getSelectivityMatrix();
		double[] rates = operator.getArrivalRates();
		XATNode rootnode = null;

		// first generate edge list from selectivity matrix
		ArrayList edgelist = new ArrayList();
		ArrayList joinentrylist = new ArrayList();

		//TODO change this so the choice of cartesianproduct makes a difference
		// in the edgelist.
		//so far the cartesianproduct is included in the edge list.
		for (int i = 0; i < selectivities.length; i++) {
			for (int j = i + 1; j < selectivities.length; j++) {
				if (selectivities[i][j] > 0) {
					Edge newedge = new Edge(i, j, rates[i], rates[j], selectivities[i][j]);
					edgelist.add(newedge);
				}
			}
		}

		// the ID of the join entry starts from next number larger than the
		// largest input id.
		int minJoinID = rates.length;
		int joinID = minJoinID;

		// go through the edgelist, pick the smallest weighted edge and make a
		// join entry out of it.
		// remove the edge from the edgelist and update other edges connecting
		// to this edge in the edgelist.
		// connect queues and set parent/child relationships between nodes.
		// when finish, the edgelist is empty and a binary join tree is
		// generated and has replaced
		// the current ProbabilisticWMJoinOperator and its node in the current
		// query plan tree.
		while (edgelist.size() > 0) {

			// find the edge with the smallest weight
			int size = edgelist.size();
			int minindex = 0;
			double minweight = ((Edge) edgelist.get(0)).weight;
			for (int i = 1; i < size; i++) {
				double temp = ((Edge) edgelist.get(i)).weight;
				if (temp < minweight) {
					minindex = i;
					minweight = temp;
				}
			}

			// make a join entry from the smallest weighted edge
			Edge edge = (Edge) edgelist.remove(minindex);
			DagNode newnode = new DagNodeImp();
			WMJoinOperator newoperator = new WMJoinOperator(newnode);
			rootnode = newnode;

			double[][] selectivitymatrix = new double[2][2];
			double[] arrivalrates = new double[2];
			selectivitymatrix[0][1] = edge.selectivity;
			selectivitymatrix[1][0] = edge.selectivity;
			selectivitymatrix[0][0] = 1;
			selectivitymatrix[1][1] = 1;
			arrivalrates[0] = edge.rate1;
			arrivalrates[1] = edge.rate2;
			newoperator.setRunningMode("mjoin");
			newoperator.setNumOfInputs(2);
			newoperator.setSelectivityMatrix(selectivitymatrix);
			newoperator.setArrivalRates(arrivalrates);
			newoperator.setTimeWindow(this.timewindow);
			newoperator.setupStates();
			newoperator.computeProbeSequence();

			joinID++;
			BinaryJoinEntry newjoinentry = new BinaryJoinEntry(joinID, edge.v1,
					edge.v2, newoperator);
			newjoinentry.lchildID = edge.v1;
			newjoinentry.rchildID = edge.v2;
			joinentrylist.add(newjoinentry);

			// if the input to this new join is the intermediate output from
			// another join
			// mark the parent/child relationship in each join entry
			if ((edge.v1 > minJoinID) || (edge.v2 > minJoinID)) {
				for (int i = 0; i < joinentrylist.size(); i++) {
					BinaryJoinEntry entry = (BinaryJoinEntry) joinentrylist
							.get(i);
					if (entry.joinID == edge.v1 || entry.joinID == edge.v2) {
						entry.parentID = newjoinentry.joinID;
					}
				}
			}

			//			if((edge.v2 > minJoinID)) {
			//				for(int i=0; i < joinentrylist.size(); i++) {
			//					BinaryJoinEntry entry = (BinaryJoinEntry)joinentrylist.get(i);
			//					if (entry.joinID == edge.v2) {
			//						entry.parentID = newjoinentry.joinID;
			//						break;
			//					}
			//				}
			//			}

			//now update the edgelist
			for (int i = 0; i < edgelist.size(); i++) {
				Edge nextedge = (Edge) edgelist.get(i);
				if (nextedge.v1 == edge.v1 || nextedge.v1 == edge.v2) {
					nextedge.v1 = joinID;
					nextedge.rate1 = 2 * edge.weight * timewindow / 1000;
					nextedge.updateWeight();
					continue;
				}
				if (nextedge.v2 == edge.v1 || nextedge.v2 == edge.v2) {
					nextedge.v2 = joinID;
					nextedge.rate2 = 2 * edge.weight * timewindow / 1000;
					nextedge.updateWeight();
				}
			}

			//now combine edges with the same vertices
			for (int i = 0; i < edgelist.size(); i++) {
				Edge edge1 = (Edge) edgelist.get(i);
				for (int j = i + 1; j < edgelist.size(); j++) {
					Edge edge2 = (Edge) edgelist.get(j);

					if ((edge1.v1 == edge2.v1 && edge1.v2 == edge2.v2)
							|| (edge1.v1 == edge2.v2 && edge1.v2 == edge2.v1)) {
						edge1.selectivity = edge1.selectivity
								* edge2.selectivity;
						edge1.updateWeight();
						edgelist.remove(j);
						break;
					}
				}
			}
		}

		// finally connect the parents and children
		for (int i = 0; i < joinentrylist.size(); i++) {
			BinaryJoinEntry joinentry = (BinaryJoinEntry) joinentrylist.get(i);

			// connect to parent
			if (joinentry.parentID > 0) {
				int parentid = joinentry.parentID;

				for (int k = 0; k < joinentrylist.size(); k++) {
					BinaryJoinEntry parententry = (BinaryJoinEntry) joinentrylist
							.get(k);

					if (parententry.joinID == parentid) {
						try {
							Constructor[] c1 = Class.forName(this.queueModel)
									.getConstructors();
							Class[] cl;
							Constructor construct;
							int count = 0;
							do {
								construct = c1[count];
								cl = construct.getParameterTypes();
								count++;
							} while (cl.length != 1);

							XATQueue newqueue = (XATQueue) construct
									.newInstance(new Object[] { joinentry.operator });
							joinentry.operator.setOutputQueue(newqueue);
							joinentry.operator.getNode().setParent(
									parententry.operator.getNode());
							if (joinentry.joinID == parententry.lchildID) {
								parententry.operator
										.setLeftInputQueue(newqueue);
								parententry.operator.getNode().insertChild(
										joinentry.operator.getNode(), 0);
							} else if (joinentry.joinID == parententry.rchildID) {
								parententry.operator
										.setRightInputQueue(newqueue);
								parententry.operator.getNode().appendChild(
										joinentry.operator.getNode());
							}
						} catch (Exception e) {
							e.printStackTrace();
							System.exit(-1);
						}
					}
				}
			}

			// connect to the corresponding children of the original
			// probablisticwmjoinoperator (the first parameter of this method)
			if (joinentry.lchildID < minJoinID) {
				XATQueue queue = operator.getInputQueue(joinentry.lchildID);
				queue.unregister(operator);
				List owners = queue.getQueueOwner();
				for (int j = 0; j < owners.size(); j++) {
					if (owners.get(j) instanceof XATStreamOperator) {
						XATStreamOperator op = (XATStreamOperator) owners
								.get(j);
						if (op != operator) {
							op.getNode()
									.setParent(joinentry.operator.getNode());
							joinentry.operator.getNode().insertChild(
									op.getNode(), 0);
						} else {
							owners.set(j, joinentry.operator);
						}
					}
				}
				joinentry.operator.setLeftInputQueue(queue);

				// Add stream children of the new operator
				Object stream = operator.getNode().getChild(joinentry.lchildID);
				joinentry.operator.getNode().insertChild(stream, 0);
			}

			if (joinentry.rchildID < minJoinID) {
				XATQueue queue = operator.getInputQueue(joinentry.rchildID);
				queue.unregister(operator);
				List owners = queue.getQueueOwner();
				for (int j = 0; j < owners.size(); j++) {
					if (owners.get(j) instanceof XATStreamOperator) {
						XATStreamOperator op = (XATStreamOperator) owners
								.get(j);
						if (op != operator) {
							op.getNode()
									.setParent(joinentry.operator.getNode());
							joinentry.operator.getNode().appendChild(
									op.getNode());
						} else
							owners.set(j, joinentry.operator);
					}
				}
				joinentry.operator.setRightInputQueue(queue);

				// Add stream children of the new operator
				Object stream = operator.getNode().getChild(joinentry.rchildID);
				joinentry.operator.getNode().insertChild(stream, 1);
			}
		}
		return rootnode;
	}

	protected String getBTreeString(XATNode rootnode, String indent) {
		if (rootnode == null)
			return indent + "\n";
		String output = indent + rootnode.toString() + "\n";
		String newIndent = indent + "  ";
		Object[] children = rootnode.getChildren();
		for (int i = 0; i < children.length; i++) {
			if(children[i] instanceof XATNode)
				output = output + getBTreeString((XATNode)children[i], newIndent);
		}
		return output;
	}

	protected XATDag convertEntryListToRealQueryTree(ArrayList joinentrylist) {
		XATDag tree = new XATDagImp();
		double[] rates = this.joinGraph.getArrivalRates();
		int minJoinID = rates.length;
		
		// generate operator for each joinentry.
		for(int i=0; i<joinentrylist.size(); i++) {
			MJoinEntry entry = (MJoinEntry)joinentrylist.get(i);
			entry.createEntryOperator();
		}
		
		// connect queues and set parent/child relationships between nodes.
		// when finish, the edgelist is empty and a binary join tree is
		// generated and has replaced
		// the current WMJoinOperator and its node in the current
		// query plan tree.
		
		// connect the parents and children
		for (int i = 0; i < joinentrylist.size(); i++) {
			MJoinEntry joinentry = (MJoinEntry) joinentrylist.get(i);
			//WMJoinOperator wmoperator = joinentry.operator;
			
			//if this is the root operator
			if(joinentry.parentID < 0) {
				tree.setRoot(joinentry.operator.getNode());
			}
			//otherwise connect to parent
			else if (joinentry.parentID > 0) {
				int parentid = joinentry.parentID;

				for (int k = 0; k < joinentrylist.size(); k++) {
					MJoinEntry parententry = (MJoinEntry) joinentrylist.get(k);

					if (parententry.entryID == parentid) {
						try {
							Constructor[] c1 = Class.forName(this.queueModel).getConstructors();
							Class[] cl;
							Constructor construct;
							int count = 0;
							do {
								construct = c1[count];
								cl = construct.getParameterTypes();
								count++;
							} while (cl.length != 1);

							XATQueue newqueue = (XATQueue)construct.newInstance(new Object[]{joinentry.operator});
							joinentry.operator.setOutputQueue(newqueue);
							joinentry.operator.getNode().setParent(parententry.operator.getNode());
							int cindex;
							//if this is a parent entry of the current entry
							if ((cindex = parententry.IndexOfChildID(joinentry.entryID)) != -1) {
								parententry.operator.setNthInputQueue(newqueue, cindex);
								parententry.operator.getNode().insertChild(joinentry.operator.getNode(), cindex);
							}
						} catch (Exception e) {
							e.printStackTrace();
							System.exit(-1);
						}
					}
				}
			}

			// connect to the stream input queues.
			for(int m=0; m<joinentry.inputSize; m++) {
				int cid = joinentry.childrenIDs[m];
				if( cid < minJoinID) {
					String cstring = new String(""+(cid + 'A'));
					//Assuming simple queue is used
					XATQueue streamqueue = (XATQueue)streamqueues.get(cstring);
					joinentry.operator.setNthInputQueue(streamqueue, m);
					joinentry.operator.getNode().insertChild(cstring, m);
				}
			}
		}
	
		return tree;
	}
	
	/**
	 * @param joinOrders
	 * @param joinSelectivities
	 * @return A dag tree containing one mjoin operator with the given jonOrders and probeselectivities.
	 *         The operator is connected to the same stream queues as the old query plan.
	 */
	protected XATDag convertJoinOrderToQueryTree(int[][] joinOrders, double[][] joinSelectivities) {
	
		XATDag dag = new XATDagImp();
		XATNode node = new XATNodeImp();
		WMJoinOperator wmoperator = new WMJoinOperator(node);
		
		wmoperator.setArrivalRates(this.joinGraph.arrivalRates);
		wmoperator.setAsRootOperator();
		wmoperator.setCartesianProductStatus(this.joinGraph.useCartesianProductStatus);
		wmoperator.setNumOfInputs(this.joinGraph.numNodes);
		wmoperator.setProbeSequences(joinOrders, joinSelectivities);
		wmoperator.setTimeWindow(this.joinGraph.timeWindow);
		wmoperator.setSelectivityMatrix(this.joinGraph.selectivities);
		wmoperator.setupStates();
		
		// connect to the stream input queues.
		for(int m=0; m<this.joinGraph.numNodes; m++) {
			String cstring = new String(""+(m + 'A'));
			
			//Assuming simple queue is used
			XATQueue streamqueue = (XATQueue)streamqueues.get(cstring);
			wmoperator.setNthInputQueue(streamqueue, m);
			wmoperator.getNode().insertChild(cstring, m);
		}
		dag.setRoot(node);
		return dag;
	}
	
	/**
	 * @param list of join entries
	 * @return the total memory cost of the list of join entries.
	 */
	protected double computeMemoryCost(ArrayList entries){
		double memorycost = 0;
		
		for(int i=0; i<entries.size(); i++) {
			memorycost += ((MJoinEntry)entries.get(i)).getMemoryCost();
		}
		return memorycost;
	}
	
	/**
	 * @return Memory cost of the query plan with one single mjoin.
	 */
	protected double computeMJoinMemoryCost() {
		double memorycost = 0;
		
		for(int i=0; i<this.joinGraph.arrivalRates.length; i++)
			memorycost += this.joinGraph.arrivalRates[i] * this.joinGraph.timeWindow/1000;

		return memorycost;
	}

	/**
	 * @param list of join entries
	 * @return the total cpu cost of the list of join entries.
	 */
	protected double computeCPUCost(ArrayList entries) {
		double cpucost = 0;
		
		for(int i=0; i<entries.size(); i++) {
			cpucost += ((MJoinEntry)entries.get(i)).getCPUCost();
		}
		return cpucost;
	}
	
	/**
	 * 
	 * @param probeOrders
	 * @param probeSelectivities
	 * @return CPU cost of the query plan with one single mjoin.
	 */
	protected double computeMJoinCPUCost(int[][] probeOrders, double[][] probeSelectivities) {
		double cpucost = 0;
		
		int size = this.joinGraph.numNodes;
		
		for(int i=0; i<size; i++) {
			// first the cost of inserting and deleting tuples to/from states
			cpucost += this.joinGraph.arrivalRates[i]*(COST_TUPLE_INSERT + COST_TUPLE_DELETE);
			
			double inter_results = this.joinGraph.arrivalRates[i];
			double temp;
			// then the cost of probing each of the remaining states
			for(int k=0; k<(size-1); k++) {
				int probedinput = probeOrders[i][k];
				double statesize = this.joinGraph.arrivalRates[probedinput] * this.joinGraph.timeWindow/1000;
				temp = inter_results * statesize * probeSelectivities[i][k];
				cpucost += temp * (COST_TUPLE_JOIN);
				inter_results = temp;
			}
		}		
		return cpucost;
	}
	
	public void setJoinGraph(JoinGraph g) {
		this.joinGraph = g;
	}

	public void printProbeOrders(int[][] probeOrders) {
		int inputSize = probeOrders.length;
		for(int i=0; i<inputSize; i++) {
			System.out.print("join order for input "+i + ": ");
			for(int k=0; k<inputSize - 1; k++) {
				System.out.print(" -> " + probeOrders[i][k]);
			}
			System.out.println();
		}
	}

	public void printProbeSelectivities(double[][] probeSelectivities) {
		int inputSize = probeSelectivities.length;
		for(int i=0; i<inputSize; i++) {
			System.out.print("join selectivity for input "+i + ": ");
			for(int k=0; k<inputSize - 1; k++) {
				System.out.print(" -> " + probeSelectivities[i][k]);
			}
			System.out.println();
		}
	}
}
