/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.Enumeration;
import java.util.Vector;

import com.sun.org.apache.xalan.internal.xsltc.runtime.Hashtable;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.CostEstimator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;

/**
 * MJoinDynamicAlgorithm is an adaptation of the Selinger Algorithm to MJoin 
 * Given an MJoin with n streams S1, S2, .... Sn. 
 * Now, let Si be any one streams that belongs to S1, S2, .... Sn.
 * This algorithm is used to generate a join order that processes the tuples from 
 * the stream Si.
 * NOTE: For an illustration of how this algorithm works please refer to my qualifier report on DAVIS.
 * user2.dsrg.DISSERTATION.venky.Research-Qualifier\Report.
 * 
 * @author Venkatesh Raghavan 
 * @email venky@cs.wpi.edu
 * @version 1.0, Summer 2007
 * 
 */
public class MJoinDynamicAlgorithm 
{	
	// The hash table of plans that are generated in each iteration
	// NOTE: I manipulate this to have 
	public static Hashtable plansTable;
	protected static CostEstimator costEst = new CostEstimator();
	
	/**
	 * Generate pairs of size = current_size + 1
	 * Example: Let there be a join graph JG1 with n=5 streams S1, S2, S3, S4 and S5. 
	 * Let the edges in JG1 be S1-S2, S1-S3, S2-S4, S3-S5
	 * Now, given plan is (S1 JOIN S3) 
	 * So the size of the plan is 2 (# of streams in the join) 
	 * We want to generate all plans having size three that have this plan as its left-most branch.
	 * That is, say, ((S1 JOIN S3) JOIN S2). 
	 * @param _input
	 * @param _graph
	 * @param _plans 
	 * @return A list of plans whose size is current-size + 1
 	 */
	public static Vector<QueryPlan> generatePairs(Plan _input, JoinGraph _graph)
	{
		Vector<QueryPlan> plans = new Vector<QueryPlan>();
		/*
		 * Here we want to do the following:
		 * 1. We would want to generate all the streams that are not involved in the given plan.
		 *    For example, in our example above, (S1 JOIN S3) does not include the streams S2, S4 and S5.
		 * 2. Since the common heuristic is to not consider cartesian product.
		 *    In our example since there is no edge in the join graph S1-S4 or S3-S4,
		 *    therefore we only consider S2 and S5.
		 * Here the result of not visited node is a set {S2,S5}    
		 */
		Vector<Source> remainingSources = Utils.notVisitedSources(_graph, _input);
		BinaryJoinNodeImpl joinNode;
		QueryPlan nQueryPlan;
		 
		   
		/*
		 * Now, having the set {S2, S5} we can now generate plans:
		*/
		if(remainingSources !=null && remainingSources.size() >0)
		{
			 /*
			  ** --- ((S1 JOIN S3) JOIN S2)
			  * --- ((S1 JOIN S3) JOIN S5)
			  * To do this:
			  */
			
			for(int i=0; i<remainingSources.size();i++)
			{				
				// 1. create an new binary node
				joinNode = new BinaryJoinNodeImpl();
				// 2. we clone the original plan
				// 3. Add the clone as its left child
				joinNode.addLeftChild(_input.clone());
				// 4. Add the next candidate in the set as the right child
				joinNode.addRightChild(remainingSources.elementAt(i));
				nQueryPlan = new QueryPlan(joinNode);
				// 5. Add the plan to the list
				plans.add(nQueryPlan);
			}			
		}
		// Return the list of plans
		return plans;
	}
	

	/**
	 * For example, if in the list of plans there exists two plans:
	 * 1. ((S1 JOIN S3) JOIN S2) 
	 * 2. ((S1 JOIN S2) JOIN S1)
	 * Both these plans have the same streams {S1, S2, S3}
	 * Therefore, we can just keep the cheapest among them for the join pair {S1, S2, S3}. 
	 * @param _plans
	 * @param _graph
	 * @return A vector of plans
	 */
	public static Vector<QueryPlan> prune(Vector<QueryPlan> _plans, JoinGraph _graph)
	{
		plansTable = new Hashtable();
			
		if(_plans != null && _plans.size() >0)
		{
			// For each plan in the list do:
			QueryPlan nextPlan, otherPlan;
			String name;
			for(int i=0; i<_plans.size();i++)
			{
				nextPlan = _plans.get(i);
				// 1. Estimate cost of the plan
				nextPlan.setCost(costEst.generateCost(nextPlan,_graph));
				// 2. Generate a name for the plan
				// For example, ((S1 JOIN S3) JOIN S2) has the name S1:S2:S3
				// Note that the stream name has all the participant streams (in sorted order) separated by the delimiter ":"
				// WHY? Because 
				//  PlanA: ((S1 JOIN S3) JOIN S2)
				//  PlanB: ((S1 JOIN S2) JOIN S3)
				// Both will have the name S1:S2:S3 and therefore we can easily determine who is cheaper.
				name = Utils.sortedName(nextPlan);
				// So: 
				if(plansTable.get(name) == null)
				{
					//if the plan name does not exist add it to the hash table
					plansTable.put(name, nextPlan);
				}else
				{
					//if the plan name does exist then keep the cheaper alternative
					otherPlan = (QueryPlan)plansTable.get(name);
					
					if(otherPlan.getCost() > nextPlan.getCost())
					{
						plansTable.put(name, nextPlan);
					}
				}
			}
		}
		Vector<QueryPlan> result = new Vector<QueryPlan>();
		// Generate the vector of plans from the hash table
		Enumeration enums = plansTable.keys();
		while(enums.hasMoreElements())
		{
			result.add((QueryPlan)plansTable.get(enums.nextElement()));
		}
		return result;
	}
	
	/**
	 * This algorithms starts bottom-up. Starts with plans with size =1.
	 * In MJoin the start stream is predetermined based on the pipeline one in processing.
	 * @param _plans
	 * @param _graph
	 * @return
	 */
	public static QueryPlan iterate(Vector<QueryPlan> _plans, JoinGraph _graph)
	{
		Vector<QueryPlan> newPlans = new Vector<QueryPlan>();
		int count =1;
		int maxCount = _graph.getVerticesCount()-1;
		while(count <= maxCount)
		{
			if(_plans!=null && _plans.size() >0)
			{
				newPlans = new Vector<QueryPlan>();
				// Increase the size by 1.
				// That is, if you currently have plans of size 2, then now look at plans of size 3.
				// The iteration is truncated by the number of streams in teh Join Graph.
				for(int i=0; i<_plans.size(); i++)
				{
					newPlans.addAll(MJoinDynamicAlgorithm.generatePairs(_plans.elementAt(i), _graph));
				}
				// Prune plans that have identical leafs
				// For example, 
				//  PlanA: ((S1 JOIN S3) JOIN S2)
				//  PlanB: ((S1 JOIN S2) JOIN S3)
				// Both will have the name sources {S1, S2, S3} 
				_plans = MJoinDynamicAlgorithm.prune(newPlans,_graph);
			}
			count++;
		}
		return _plans.get(0);
	}
	
	/**
	 * For a given join graph JG with streams S1, S2, ... Sn
	 * And, start stream Si.
	 * Generate the optimal join order of for processing tuples from Si.
	 * @param _graph
	 * @param _start
	 * @return
	 */
	public static QueryPlan order(JoinGraph _graph, Stream _start)
	{
		int numberOfSource = _graph.getVerticesCount();
		QueryPlan startPlan = new QueryPlan(_start);
		Vector<QueryPlan> ps = new Vector<QueryPlan>();
		// Start the whole process by creating the first plan with just the start stream.
		ps.add(new QueryPlan(_start));
		return MJoinDynamicAlgorithm.iterate(ps,_graph);
	}
}
