/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.Enumeration;
import java.util.Vector;


import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.JoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.WeightedEdgeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.WeightedJoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.WeightedEdge;


/**
 * Minimum Spanning Tree Algorithms
 * @author Venkatesh Raghavan (venky@cs.wpi.edu)
 * @version 1.0, July 2007
 */
public class MinimumSpanningTree {

	/**
	 * Generates a weighted join graph from a given join graph.
	 * @param _graph a join graph that has no weight
	 * @return 	  	 a weighted join graph
	 */
	public static WeightedJoinGraphImpl computeWeightedJoinGraph(JoinGraph _graph)
	{
		WeightedJoinGraphImpl result = new WeightedJoinGraphImpl();
		
		Enumeration enums = _graph.getEdges();
		
		Edge nextEdge;
		while(enums.hasMoreElements())
		{
			nextEdge = (Edge) enums.nextElement();
			try{
			result.addEdge(nextEdge.getVertexA(), nextEdge.getVertexB(), nextEdge.getValue());
	
			}catch(Exception e)
			{
				e.printStackTrace();
			}
		}
		
		enums = _graph.getVertices();
		while(enums.hasMoreElements())
		{
			result.addVertex((Vertex) enums.nextElement());
		}
		return result;
	}
	
	/**
	 * Returns an acyclic join graph that is the minimum spanning tree for a given cyclic join graph.
	 * This is an implementation of 
	 * <a href="http://en.wikipedia.org/wiki/Kruskal's_algorithm"> Kruskal's_algorithm</a> 
	 * @param _weightedGraph a weighted join graph
	 * @return               an acyclic join graph  
	 */
	public static JoinGraph kruskalAlgorithm(WeightedJoinGraphImpl _weightedGraph)
	{
		JoinGraphImpl result = new JoinGraphImpl();
		
		// In each iteration of the Kruskal's_algorithm the next cheapest edge is considered.
		WeightedEdgeImpl cheapestEdge;
		// The cheapest weight in each iteration.
		double minWeight;
		// Index of this cheapestEdge
		int indexOfMinWeightEdge =0; 
		
		// The resultant set of edges that makes up the acyclic join graph (solution).
		Vector<WeightedEdgeImpl> minSpanningEdges = new Vector<WeightedEdgeImpl>();
		
		// Many implementations of the Kruskal's_algorithm use coloring the vertices as the
		// proceed with the algorithm. Instead of color an integer index can also be used
		// to as a flag mark.
		
		int flagIndex = 0;
		
		Enumeration enums = _weightedGraph.getEdges();
		// Get all the edges in th ejoin graph
		Vector<WeightedEdgeImpl> edges = new Vector<WeightedEdgeImpl>();
		while(enums.hasMoreElements())
		{
			edges.add((WeightedEdgeImpl)enums.nextElement());
		}
		
		enums = _weightedGraph.getVertices();
		// Get all the vertices in the join graph
		Vector<Vertex> vertices = new Vector<Vertex>();
		while(enums.hasMoreElements())
		{
			vertices.add((Vertex)enums.nextElement());
		}
		
		if(edges!=null && edges.size()>0 && vertices!=null && vertices.size()>0)
		{
			// Data structure to hold the coloring or flaging value each vertex.
			int[] flags = new int[vertices.size()];
			for(int i=0; i<vertices.size(); i++) {
				flags[i] = flagIndex;
			}
			
			int totalNumberOfEdge = edges.size();
			for(int k=0; k<totalNumberOfEdge;k++)
			{
				minWeight = Double.MAX_VALUE;
				/* Pick the edge with the minimum edge weight*/
				
				for(int i=0; i<edges.size();i++)
				{
					if(edges.get(i).getWeight()<minWeight)
					{
							minWeight = edges.get(i).getWeight();
							indexOfMinWeightEdge = i;
					}
				}
				// Remove the edge from the list of all edge.
				cheapestEdge = edges.remove(indexOfMinWeightEdge);
				
				// Let,
				// minSpanningEdges(S for short) be all the edges in the solution.
				// cheapestEdge(C for short) be the cheapestEdge in the current iteration i.e. C(u,v) 
				// if FORALL edge E(x,y) such x != u and x != v and y !=x and y!=v
				// MEANING: there is no edge in set S with either u or v as one of its node.
				// Then it is safe to add in S
				// increase the flagIndex (i.e. change color)
				// and mark the vertices u and v with the same flag (or color)
				
				if(flags[vertices.indexOf(cheapestEdge.getVertexA())] ==0 && flags[vertices.indexOf(cheapestEdge.getVertexB())] ==0)
				{
					minSpanningEdges.add(cheapestEdge);
					flagIndex++;
					flags[vertices.indexOf(cheapestEdge.getVertexA())] = flagIndex;
					flags[vertices.indexOf(cheapestEdge.getVertexB())] = flagIndex;
					continue;
				}
				
				// Let,
				// minSpanningEdges(S for short) be all the edges in the solution.
				// cheapestEdge(C for short) be the cheapestEdge in the current iteration i.e. C(u,v) 
				// if FORALL edge E(x,y) such x == u || y == u
				// MEANING there exist an edge in S with u as on of its vertex.
				// Safe to add as v is not in S and therefore no chance of forming a cycle.
				// Mark v with the same flag (or color) as u
				
				if(flags[vertices.indexOf(cheapestEdge.getVertexA())] !=0 && flags[vertices.indexOf(cheapestEdge.getVertexB())] ==0)
				{
					minSpanningEdges.add(cheapestEdge);
					flags[vertices.indexOf(cheapestEdge.getVertexB())] =flags[vertices.indexOf(cheapestEdge.getVertexA())];
					continue;
				}
				

				// Let,
				// minSpanningEdges(S for short) be all the edges in the solution.
				// cheapestEdge(C for short) be the cheapestEdge in the current iteration i.e. C(u,v) 
				// if FORALL edge E(x,y) such v == u || y == v
				// MEANING there exist an edge in S with v as on of its vertex.
				// Safe to add as u is not in S and therefore no chance of forming a cycle.
				// Mark v with the same flag (or color) as v
				
				if(flags[vertices.indexOf(cheapestEdge.getVertexA())] ==0 && flags[vertices.indexOf(cheapestEdge.getVertexB())] !=0)
				{
					minSpanningEdges.add(cheapestEdge);
					flags[vertices.indexOf(cheapestEdge.getVertexA())] =flags[vertices.indexOf(cheapestEdge.getVertexB())];
					continue;
				}
				

				// Let,
				// minSpanningEdges(S for short) be all the edges in the solution.
				// cheapestEdge(C for short) be the cheapestEdge in the current iteration i.e. C(u,v) 
				// Example 1: A--B C--D--E
				// 			  Flag of B,A = 1
				// 			  Flag of C,D,E = 2
				// Example 2: B--A--C
				//			  Flag B,A,C = 1
				// Now let the next cheapest edge is B--C
				// In Example 1, this will not form a cycle, while in Example 2, it will form a cycle B--A--C--B
				// So we only consider when the flags(or color) are different.
				// After adding the edge we give a new flag(or color) to the new chain.
				// In Example 1, A--B--C--D--E
				// 			     Flag of A,B,C,D,E = 3
				
				if(flags[vertices.indexOf(cheapestEdge.getVertexA())] != flags[vertices.indexOf(cheapestEdge.getVertexB())])
				{
					minSpanningEdges.add(cheapestEdge);
					flagIndex++;
					int flagA = flags[vertices.indexOf(cheapestEdge.getVertexA())];
					int flagB = flags[vertices.indexOf(cheapestEdge.getVertexB())];
					for(int j=0; j<flags.length;j++)
					{
						if(flags[j] == flagA)
							flags[j] = flagIndex;
						else if (flags[j] == flagB)
							flags[j] = flagIndex;
					}
					continue;
				}
			}
			
			for(int i=0; i<vertices.size();i++)
			{
				result.addVertex(vertices.get(i));
			}
			
			for(int i=0; i< minSpanningEdges.size();i++)
			{
				try{
					result.addEdge(minSpanningEdges.get(i).getVertexA(), minSpanningEdges.get(i).getVertexB(), minSpanningEdges.get(i).getValue());
				}catch(Exception e)
				{
					e.printStackTrace();
				}
			}
			return result;
		}else
		{
			return null;
		
		}
		
	}
	/**
	 * Given a join graph, the algorithm first calculates the weighted join graph and then evaluates the
	 * minimum spanning tree for that weighted join graph.
	 * @param _graph
	 * @return
	 */
	public static JoinGraph kruskalAlgorithm(JoinGraph _graph)
	{
		WeightedJoinGraphImpl weightedGraph = MinimumSpanningTree.computeWeightedJoinGraph(_graph);
		/*
		System.out.println("---- Weighted Graph-----");
		Enumeration enums = weightedGraph.getEdges();
		while(enums.hasMoreElements())
		{
			((WeightedEdgeImpl) enums.nextElement()).print();
		}
		System.out.println("---------------");
		*/
		return MinimumSpanningTree.kruskalAlgorithm(weightedGraph);
	}
}
