/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.Hashtable;


import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;

/**
 * All functions that helpful to other algorithms.
 * @author Venkatesh Raghavan 
 * @email venky@cs.wpi.edu
 * @version 1.0, Summer 2007
 */
public class Utils {


	/**
	 * Generate the name of a plan based on its leaves (source)
	 * For example, ((S1 JOIN S3) JOIN S2) has the name S1:S2:S3
	 * Note that the stream name has all the participant streams (in sorted order) separated by the delimiter ":"
	 * @param _plan
	 * @return
	 */
	public static String sortedName(QueryPlan _plan)
	{
		Vector sources = _plan.getSources();
		String value = new String();
		if(sources != null && sources.size()>0)
		{
			int count = sources.size();
			String[] sortedName = new String[count];
			int j;
			for(int i=0;i<count;i++)
			{
				value = ((Source) sources.get(i)).getLabel();
				j =i-1;
				while(j>=0 && sortedName[j].compareTo(value) >0)
				{
					sortedName[j+1] = sortedName[j];
					j--;
				}
				sortedName[j+1]=value;
			}
			value = sortedName[0];

			for(int i=1;i<count;i++)
			{
				value += ":"+sortedName[i];
			}
		}
		return value;
	}

	/**
	 * Is the _current element in a a given set _set
	 * @param _set
	 * @param _current
	 * @return true is yes and false if no
	 */
	public static boolean inVector(Vector _set, Object _current)
	{
		if(_set != null && _set.size() >0 && _current !=null)
		{
			for(int i=0;i<_set.size();i++)
			{
				if(_set.get(i).toString().compareTo(_current.toString()) ==0)
				{
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Given a join graph with n vertices and e edges.
	 * Also given a set _set of edges (this is a subset of the set of all vertices)
	 * Also given a vertex _newVertex
	 * Now, find if the _newVertex has an edge in the given join graph from the 
	 * vertices in the set _set.
	 * @param _graph
	 * @param _set
	 * @param _newVertex
	 * @return true is edge exist from the vertices in set _set to _newVertex, false otherwise.
	 */
	public static boolean hasEdge(JoinGraph _graph, Vector _set, Vertex _newVertex)
	{
		boolean found = false;
		Vector<Vertex> neighbors;
		if(_set !=null && _newVertex != null)
		{
			neighbors = _graph.getNeighbors(_newVertex);
			for(int i=0; i<_set.size();i++)
			{
				found = Utils.inVector(neighbors, _set.elementAt(i));
				if(found)
				{
					return found;
				}
			}
		}
		return found;
	}

	/**
	 * Let there be a join graph JG1 with n=5 streams S1, S2, S3, S4 and S5. 
	 * Let the edges in JG1 be S1-S2, S1-S3, S2-S4, S3-S5
	 * Now, given plan is (S1 JOIN S3) 
	 * Here we want to do the following:
	 * 1. We would want to generate all the streams that are not involved in the given plan.
	 *    For example, in our example above, (S1 JOIN S3) does not include the streams S2, S4 and S5.
	 * 2. Since the common heuristic is to not consider cartesian product.
	 *    In our example since there is no edge in the join graph S1-S4 or S3-S4,
	 *    therefore we only consider S2 and S5.
	 * Here the result of not visited node is a set {S2,S5}    
	 * @param _graph
	 * @param _plan
	 * @return
	 */
	public static Vector<Source> notVisitedSources(JoinGraph _graph, Plan _plan)
	{
		Vector<Source> result = new Vector<Source>();

		Vector visited = _plan.getSources();

		Enumeration enums = _graph.getVertices();
		Object nextSource;
		while(enums.hasMoreElements())
		{
			nextSource = enums.nextElement();
			if(nextSource instanceof Source)
			{
				if(!Utils.inVector(visited,nextSource))
				{
					if( Utils.hasEdge(_graph, visited, (Source) nextSource))
					{
						result.add((Source)nextSource);
					}
				}
			}
		}
		return result;
	}

	/**
	 * For a given plan print it pretty
	 * @param _plans
	 */
	public static void print(Vector<QueryPlan> _plans)
	{
		System.out.println("-------------------------");
		if(_plans !=null && _plans.size()>0)
		{
			System.out.println("******************");
			for(int i=0; i<_plans.size();i++)
			{
				_plans.elementAt(i).print();
			}
			System.out.println("******************");
		}
		System.out.println("-------------------------");
	}
	
	
	/**
	 * Given an integer n, it generates all the partitions of that number
	 * eg for 3 , the partitions would be ((1,1,1), (1,2))
	 * @param n
	 * @return A Hashtable of all partitions
	 */
	public static Hashtable<String,ArrayList<Integer>> generateBellPartitions(int n)
	{
		Hashtable<String,ArrayList<Integer>> PT = new Hashtable();
		ArrayList<Integer> initial;
		initial = new ArrayList<Integer>();
		for(int i=0;i<n;i++)
		{
			initial.add(0);
		}
		
		PT = iterateBellPartitions(n, PT, 0, initial);
		return PT;
	}
	
	/**
	 *  
	 * @param n
	 * @param PT
	 * @param index
	 * @param current
	 * @return
	 */
	public static Hashtable<String,ArrayList<Integer>> iterateBellPartitions(int n, Hashtable<String,ArrayList<Integer>> PT,int index, ArrayList<Integer> current)
	{
		if(current != null)
		{
		int sum;
		String key;
		ArrayList<Integer> sequence = (ArrayList<Integer>) current.clone();	
		for(int i =1; i<= n-1;i++)
		{
			sequence.set(index, i);
			sum = Utils.calculateSum(sequence);
			if(sum == n)
			{
				key = Utils.sortArrayList((ArrayList<Integer>)sequence.clone());
				if (PT.get(key)== null)
				{
					PT.put(key, (ArrayList<Integer>)sequence.clone());
				}
				return PT;
			}else if(sum > n)
			{
				return PT;
			}else
			{
				iterateBellPartitions(n,PT,index+1,sequence);
			}
			}
		}	
		return PT;
	}
	
	/**
	 * Calculates the sum of all the elements in an array
	 * @param array
	 * @return sum
	 */
	private static int calculateSum (ArrayList array) {
		int sum = 0;
		for (Iterator i = array.iterator (); i.hasNext (); ) {
			Integer val = (Integer) i.next ();
			sum += val.intValue ();
		}
		return sum;
	}
	
	/**
	 * Sorts the elements of an array and returns the sorted array
	 * @param array
	 * @return sorted array
	 */
	
	public static String sortArrayList (ArrayList array)
	{
		Collections.sort(array);
		String stringArray = array.toString();
		return stringArray;
	}

	public static void printHashtable(Hashtable h)
	{

		System.out.println("######## HASH TABLE #######");

		Enumeration Enum = h.elements();

		while(Enum.hasMoreElements())
		{
			System.out.println(Enum.nextElement());
		}
		System.out.println("############################");
	}
	public static Hashtable h[];

}
