/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.Utils;
public class CostEstimator {
	private double cost = 0;
	
    public CostEstimator()
    {
    	
    }
    /**
     * This functions takes in an plan and the join graph to return the 
     * cummulative cost of executing the plan. For example, if the current plan is
     * ((A Join B) Join (C Join D)), then the cummulative cost is the 
     * cost(A Join B) + cost(C Join D) + cost (resultOf(A Join B) Join resultOf(C Join D))   
     * @param _currentPlan the plan whose cost needs to be determined.
     * @param _graph is the join graph that has the selectivity between the different sources
     * @return the cummulative cost of executing the plan
     */
	public double generateCost(Plan _currentPlan, JoinGraph _graph)
	{
		this.cost = 0.0;
		//print join graph
		//_graph.print();
		// Calculate iteratively in a depth-first fashion the cost of the plan.
		_currentPlan.setCost(this.evaluateCost(_currentPlan, _graph));
		return _currentPlan.getCost();
	}
	
	/**
	 * This is a recursive program that calculates the cost of plan in a dept-first fashion.
	 * @param _currentPlan the plan whose cost needs to be determined.
	 * @param _graph is the join graph that has the selectivitty between the different sources
	 * @return cardinality of the join
	 */
	private double evaluateCost(Plan _currentPlan, JoinGraph _graph)
	{
		//System.out.println("---");
		//_currentPlan.print();
		//System.out.println("---");
		double subPlanCost;
		double joinCost = 0;
		double cardinality =0;
		double leftCost, rightCost, selectivity;
		Node rootNode = _currentPlan.getRoot(); 
		
		if(rootNode instanceof SourceImpl)
		{
			// Sources are the leaves of the query plan. 
			// Their cost is the cardinality of the source. For example, if the source is a
			// stream then the cost is the input rate, while the cost of a relation is size of table.
			//System.out.println(rootNode.getLabel() + ": " + ((SourceImpl)rootNode).getCardinality());
			return ((SourceImpl) _currentPlan.getRoot()).getCardinality();
		}
		else if(rootNode instanceof Plan)
		{
			// If the node is a plan, then the cost of the subplan must first be determined.
			subPlanCost = this.evaluateCost((Plan)_currentPlan.getRoot(), _graph);
			((Plan)_currentPlan.getRoot()).setCost(subPlanCost);
			//System.out.println(rootNode.getLabel() + ": " + ((Plan)rootNode).getCardinality());
			return subPlanCost; 
		}
		else if(rootNode instanceof BinaryJoinNodeImpl)
		{			
			Plan lPlan, rPlan;
			
			if( (((BinaryJoinNodeImpl) rootNode).getLeftChild() != null) && (((BinaryJoinNodeImpl) rootNode).getRightChild() != null))
			{
				if(((BinaryJoinNodeImpl) rootNode).getLeftChild() instanceof Plan)
				{
					lPlan = (Plan) ((BinaryJoinNodeImpl) rootNode).getLeftChild();
					leftCost = this.evaluateCost(lPlan,_graph);
					lPlan.setCost(leftCost);
				}
				else
				{
					lPlan = new QueryPlan((Node) ((BinaryJoinNodeImpl) rootNode).getLeftChild());
					leftCost = this.evaluateCost(lPlan,_graph);
				}
				
			
				if(((BinaryJoinNodeImpl) rootNode).getRightChild() instanceof Plan)
				{
					rPlan = (Plan) ((BinaryJoinNodeImpl) rootNode).getRightChild();
					rightCost = this.evaluateCost(rPlan,_graph);
					rPlan.setCost(rightCost);
				}else
				{
					rPlan = new QueryPlan((Node) ((BinaryJoinNodeImpl) rootNode).getRightChild());
					rightCost = this.evaluateCost(rPlan,_graph);
				}
				
				//rPlan = new QueryPlan((Node) ((BinaryJoinNodeImpl) rootNode).getRightChild());
				
				
				selectivity = Utils.binaryJoinSelectivity(lPlan.getSources(), rPlan.getSources(), _graph);
				
							 
				//joinCost = leftCost * rightCost;
				
				//cardinality = joinCost  * selectivity;
				//cost = cost +joinCost;
				
				
				joinCost = leftCost * rightCost * selectivity;
				this.cost = joinCost;
				System.out.println("Left Cost "+ leftCost +" Right Cost "+ rightCost+ " Selectivity "+selectivity+ " JOIN: "+ joinCost);
			}
			// return cardinality;
			return joinCost;
		}
		return -1;
	}
	
/**
	 * This function returns the cummulative product of the join selectivities.
	 * For example, A Join B, the selectivity is just \sigma(AB), while
	 * the cummulative selectivity of plan ((A Join B) Join C) is \sigma(AC) * \sigma(BC).
	 * Note that the \sigma(AB) is already incorporated in the cost of A Join B. Also, as the cost
	 * is generated bottom-up, we don't consider \sigma(AB). 
	 * @param leftSources sources in the left hand side of the join operator.
	 * @param rightSources sources in the right hand side of the join operator.
	 * @param _graph is the join graph that specifies the individual selectivities between
	 * the various sources.
	 * @return
	 */
	private double binaryJoinSelectivity(Vector leftSources, Vector rightSources, JoinGraph _graph)
	{
		//System.out.println("Selectivity");
		double selectivity = 1;
		if(leftSources != null && rightSources!= null && leftSources.size() >0 && rightSources.size() >0)
		{
			SourceImpl left, right;
			Edge tempEdge;
			for(int i=0; i< leftSources.size(); i++)
			{
				left = (SourceImpl) leftSources.get(i);
				for(int j=0; j<rightSources.size();j++)
				{
					right = (SourceImpl) rightSources.get(j);
					try
					{
						if(_graph.hasEdge(left,right))
						{
							tempEdge = _graph.getEdge(left, right);
							if(tempEdge != null)
							{
								//System.out.print(tempEdge.getSelectivity() +" * ");
								selectivity *= tempEdge.getValue();
							}
						}
					}catch(Exception e)
					{
						e.printStackTrace();
					}
				}
			}
			
			//System.out.println("\n"+selectivity);
			return selectivity;
		}else
		{
			return 0.0;
		}
	}
}
