/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Source;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SourceImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.ProbabilisticSingleStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;
/**
 * <pre>
 * Abstract super class for all QueryPlanGenerators
 * Provides some basic functionality for each type of QueryPlan
 *
 * Each class that extends this QueryPlanGenerator must do the following tasks so
 * that it will properly fit in with any other code that uses the QueryPlanGenerators.
 *
 * 1. Implement the setDefaultProperties method.  Note that each property's value
 * must be a string, even in cases where an integer or double makes sense.  This
 * is used to encapsulate all property values in one method.  Java provides
 * very good conversion methods from Strings to any other data type so this isnt a major problem.
 * 
 * In this method, only the properties that are UNIQUE to this query generator should be set.  Any new 
 * QueryPlanGenerator may use any property that it wishes.  Please try to avoid duplicate
 * properties.  Properties common to all query generators are set by the superclass and should be set
 * by calling super.setDefaultProperties() within the subclasses setDefaultProperties()  
 * 				******************************************************
 * Failure to call super.setDefaultProperties()  will result in the QueryPlanGenerator not having
 * the correct propertie set and will cause problems later on 
 * 				******************************************************
 *
 * 2. Implement the setProperty(String, String) method.  This method is
 * used to validate the property before it is set.  Within the individual
 * generator, you only need to validate the generator's unique properties.  Any
 * common properties are validated by the super class using super.setProperty().
 * Please see the documentation within setProperty(String, String) for more details on
 * how the method should be implemented.
 *
 * 3.  Implement makeQueryPlan(). This method contains all of the logic for
 * making the new query plan. The implementation should make use of the properties that
 * were set. While there are no measures in place to ensure the properties are used, doing so 
 * allows the the generator to be properly configured from the outside. 
 *
 * </pre>
 */
public abstract class QueryPlanGeneratorImp extends XATQueryObjectImp implements QueryPlanGenerator, java.io.Serializable{
	private edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag dag;
	protected String datamodel;
/**
 * RandomGraphGen constructor comment.
 */
protected QueryPlanGeneratorImp() {		
	super();
	dag = new XATDagImp();	
}
/**
 * Iterates through the operators in an XATTree and
 * sets the REAL_SELECTIVITY and RESULT_DELAY parameters
 * Basically just calls the customizeOperator method which will recursively
 * traverse the tree <p>
 *
 * This method will also connect a Source operator to each leaf of the query plan and
 * set the name of each source to be A, B, C, etc.  This will correspond to a Stream in
 * the Stream Generator <p>
 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 * @throws Exception:  Throw any exception we get from customizeOperator
 */
public void connectTreeToStreams() throws Exception {
    /** the STREAMS property contains all of the names of the streams*/
    ArrayList streamList = new ArrayList();
    StringTokenizer st = new StringTokenizer(this.getProperty("STREAMS"), ",");
    while (st.hasMoreTokens()) {
        streamList.add(st.nextToken());
    }
    String[] streams = (String[]) streamList.toArray(new String[] {
    });
    int streamID = 0;
    /** if there are more leaves than streams, then just go back to the first stream*/
    XATNode[] nodes = dag.getLeaves();
    for (int i = 0; i < nodes.length; i++) {
    	/** if leaf operator is already source operator, just set the stream name */
		if (nodes[i].getOperator() instanceof Source) {
//			((Source) nodes[i].getOperator()).setName(streams[streamID]);
//			streamID++;
//			if (streamID >= streams.length) {
//				streamID = 0;
//			}
		}
    	else if (nodes[i].getOperator() instanceof ProbabilisticSingleStreamOperator) {
			((ProbabilisticSingleStreamOperator)nodes[i].getOperator()).setLeaf(true);
			XATNode newNode = new XATNodeImp();
			dag.addNode(newNode, nodes[i]);
			Source s = new SourceImp(newNode);
			newNode.setOperator(s);
			s.setName(streams[streamID]);
			((ProbabilisticSingleStreamOperator)nodes[i].getOperator()).setStreamName(streams[streamID]);
			streamID++;
			if (streamID >= streams.length) {
				streamID = 0;
			}
    	} 
    	/** 
    	 * NEED TO MODIFY:
    	 * The following case should first check if the leaf operator is an instance of join operator.
    	 */
    	else {
			XATNode leftNode = new XATNodeImp();
			Source leftSrc = new SourceImp(leftNode);
			leftNode.setOperator(leftSrc);
			leftSrc.setName(streams[streamID]);
			dag.addNode(leftNode, nodes[i]);

			streamID++;
			if (streamID >= streams.length) {
				streamID = 0;
			}

			XATNode rightNode = new XATNodeImp();
			Source rightSrc = new SourceImp(rightNode);
			rightNode.setOperator(rightSrc);
			rightSrc.setName(streams[streamID]);
			dag.addNode(rightNode, nodes[i]);

			streamID++;
			if (streamID >= streams.length) {
				streamID = 0;
			}
    	}
    }
}
/**
 * Insert the method's description here.
 * 
 * @return boolean
 * @param o1 java.lang.Object
 */
public boolean equals(Object o1) {
	String thisClass = this.getClass().getName();
	String otherClass = o1.getClass().getName();
	if(thisClass.equals(otherClass)){
		return true;
	}
	return false;
}
/**
 * Insert the method's description here.
 * Creation date: (5/1/2003 2:00:48 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag
 */
public edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag getDag() {
	return dag;
}
/**
* <pre>
* creates N query plans according to the properties supported by the QueryPlanGenerator
* <b>pre</b>:
* the properties of the query plan have been initialized accordingly
* <b>post</b>:
* The Enumeration returned will consist of randomly generated IQueryPlans
* </pre>
* @return Enumeration of IQueryPlans
* @param int howMany how many IQueryPlans to generate
* @exception QueryPlanGeneratorException if either a property of the generator was not initialized or
* if there is a conflict in values ex: max < min
*/
public Enumeration makeNQueryPlans(int howMany)
    throws QueryPlanGeneratorException {
    Vector queryPlans = new Vector();
    try {
        for (int i = 0; i < howMany; i++) {
                queryPlans.addElement(makeQueryPlan());
        }            
        
    } catch (QueryPlanGeneratorException e) {
        throw e;
    }
    return queryPlans.elements();

}
/**
 * Makes a single query plan.  Each QueryPlanGenerator implements
 * this method in order to generate a query plan in their own manner
 * 
 * @return XATTree
 * @exception QueryPlanGeneratorException 
 */
public abstract XATTree makeQueryPlan() throws QueryPlanGeneratorException;
/**
 * <pre>
 * Return the correct factory as specified by a System parameter called
 * "edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator"
 * The property name can be accessed through QueryPlanGenerator.factoryID 
 * if no property is set then an exception is thrown and tree generation cannot continue
 * <pre>
 * @return IQueryPlanGenerator a handle to an IQueryPlanGenerator that can be used to create query plans with
 * certain parameters.
 */
public static QueryPlanGenerator newInstance() throws QueryPlanGeneratorException {

    try {
        String systemProp = System.getProperty(QueryPlanGeneratorImp.factoryID);
		System.out.println(systemProp);
		System.out.flush();
        if (systemProp != null) {
            Class factoryClass = Class.forName(systemProp);
            return (QueryPlanGenerator) factoryClass.newInstance();
        } else {
            throw new QueryPlanGeneratorException("ERROR: The edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator system property was not set");
        }
    } catch (ClassNotFoundException e) {
        throw new QueryPlanGeneratorException("The specified Factory cannot be found");

    } catch (Exception e1) {
        throw new QueryPlanGeneratorException(e1.getMessage());
    }
}
/**
 * Sets the properties and default values that are available for ALL QueryPlanGenerators
 */
protected void setDefaultProperties(){
	super.setDefaultProperties();
    addProperty("STREAMS","ABC");   
    addProperty("QUERY_ID","1");   
}
/**
* sets the given property to the new value.  This method is responsible
* for validating the given property.  An exception is thrown if
* the value is not correct.
* Validation should be done in the following manner: 
* First the method should check if the object supports the given property by calling
* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
* be thrown. <p>
*
* Once it is known that the object supports the property, the value must be validated.  Each implementor of
* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
* will reduce duplicate code. <p>
*
* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
* it is up to the subclass that created the property to validate it (as described above).
* @param name the name of the property to set
* @param value the new value of the property
* @throws Exception if the given property cannot be found or
* if the value is not valid for the given property.
*/
public void setProperty(java.lang.String name, java.lang.String value)
    throws java.lang.Exception {
    //first see if the QueryGenerator supports the property name	
    if (isValidPropertyName(name)) {
        int min = 0, max = 1;
        /*
        * if they are trying to set the MAX_NUMBER_OF_OPERATORS or MIN_NUMBER_OF_OPERATORS
        * verify it is a valid int and that the max >= min
        */
        if (name.equals("MAX_NUMBER_OF_OPERATORS")) {
            try {
                max = Integer.parseInt(value);
                min = Integer.parseInt(getProperty("MIN_NUMBER_OF_OPERATORS"));
                if (max >= min) {
                    setNewPropertyValue(name, value);
                } else {
                    throw new NumberFormatException();

                }
            } catch (NumberFormatException e) {
                throw new NumberFormatException(
                    "The property named "
                        + name
                        + " requires an int value that is >= the min value ("
                        + min
                        + ")");
            }
        } else
            if (name.equals("MIN_NUMBER_OF_OPERATORS")) {
                try {
                    min = Integer.parseInt(value);
                    max = Integer.parseInt(getProperty("MAX_NUMBER_OF_OPERATORS"));
                    if ((min > 0) && (max >= min)) {
                        setNewPropertyValue(name, value);
                    } else {
                        throw new NumberFormatException();

                    }
                } catch (NumberFormatException e) {
                    throw new NumberFormatException(
                        "The property named "
                            + name
                            + " requires an int value that is <= the max value ("
                            + max
                            + ")");
                }
            } else {
                //the property name is valid, but its value is validated in the super class
                super.setProperty(name, value);
            }
    } else
        throw new Exception(
            "The QueryPlanGenerator does not support the property named " + name);

}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("QueryPlanGenerator",this.getClass().getName(),true);
	}
	
	public void setDataModel(String model){
	
		this.datamodel = model;
	}
	
	public String getDataModel() {
		return this.datamodel;
	}

}
