/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queuemanager;

import java.io.FileWriter;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPersistentQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.StorageManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidStorageManagerInstanceException;


/**
 * <h2>Overview</h2>
 * <p>The purpose of the queue manager is to serve as an umbrella for all 
 * queues that are used while the CAPE system is running. It handles all the 
 * storage and tuple processing so that the queues and operators do not need to
 * know anything about how the information is actually stored. Tuples that come
 * into the queue are deferred to the queue manager. They are put into main 
 * memory, until a set memory threshold is reached, at which point the tuples 
 * are sent to disk.</p>
 * <h2>Storage Management</h2>
 * <p>All storage on disk is managed by the <code>StorageManager</code> 
 * component. The storage manager writes tuples to a random access file and 
 * provides several tools for accessing data quickly and efficiently. 
 * Specifically, the queue manager uses the methods <code>retrieveFromDisk</code>
 * and <code>saveToDisk</code>.</p>
 * <p>Each queue maintains one file where all its tuples are stored. In 
 * general, a queue that stores some tuples in main memory and some on disk 
 * would have the following layout:</p>
 * <img src="http://davis.wpi.edu/~dsrg/CAPE/images/doc/QM_queue_layout_with_cursors.png">
 * <h2>Decision Making</h2>
 * <p>The queue manager contains a <code>MEMORYTHRESHOLD</code> variable, which
 * represents how much memory should be free while the system is running. Once 
 * the amount of free memory goes below this threshold, incoming tuples are put 
 * to disk. There is a method, <code>updateMemoryThreshold</code>, which is 
 * responsible for keeping track of the most recent memory information for the 
 * threshhold. This method can be edited to specify different conditions for 
 * the memory threshold. It is called every time a decision needs to be made by 
 * the queue manager, so the threshold value always remains up-to-date. In 
 * general, the methods responsible for decision making are 
 * <code>checkMemory</code>, <code>enqueue</code>, <code>getMemoryThreshold</code>,
 * <code>setMemoryThreshold</code>, and <code>updateMemoryThreshold</code>.</p>
 * <h2>Enqueue/Dequeue</h2>
 * <p>Tuples are always enqueued to the end of the queue. If the memory 
 * threshold is reached, then incoming tuples are output to a file as a group. 
 * In other words, if an operator wants to enqueue 100 tuples to its output 
 * queue but there is not enough free memory, then all 100 tuples will go to 
 * disk, regardless of whether some might have fit in main memory.</p>
 * <p>All the customers on a queue maintain cursors into the queue's contents.
 * This is used to maintain concurrent access to a single queue by multiple 
 * operators. The cursor represents the index of the tuple which the operator 
 * should dequeue next. It is important to note that the cursor points to a 
 * <i>tuple's index</i>, not the index in the queue's <code>contents</code> 
 * array, which stores both the tuples in main memory and the place holder to 
 * the queue's file. Every time tuples are dequeued from a queue's file by a 
 * customer, it is put into main memory so that future customers that need to 
 * access the same tuples do not need to read from disk.</p>
 * <p>The important methods for enqueueing and dequeueing tuples in the storage
 * manager mimic the queue access methods of the <code>XATPersistentQueue</code>, namely,
 * <code>dequeue</code>, <code>dequeueN</code>, <code>enqueue</code>, 
 * <code>peek</code>, <code>peekLast</code>, and <code>peekN</code>. The methods 
 * <code>skip</code> and <code>skipN</code> are not implemented in the queue
 * manager because they only change a customer's cursor and do not actually 
 * retreive any information from the queue.
 * 
 * All methods are synchronized for multi-thread accessing.
 * 
 * Creation date: (May 24, 2004 9:50:27 AM)
 * @author: Natasha Bogdanova
 */
public final class XATQueueManager implements Serializable {

	/** This is the maximum memory that should be used before sending 
	 *  tuples to disk. */
	private static long MEMORYTHRESHOLD;
	
	/** The one and only XATQueueManager object */
	private static XATQueueManager queueManager = null;
	
	/** A list of all the queues managed by this QM */
	private List queues;
	
	/** Mapping from queues to partition numbers */
	private Hashtable queueToPartitionMapping = null;
	
	/** 
	 * Mapping from schema to Storage Manager instance.
	 * This is necessary because each instance of the Storage Manager can only
	 * store one type of tuples, i.e. the fields cannot change. 
	 */
	private Hashtable queueToSMMapping = null;
	
	/** The storage manager for this QM */
	//private static StorageManager storageManager = null;
	
	
	/**
	 * XATQueueManager constructor.
	 * 
	 * Creation Date: (Jun 7, 2004 3:52:21 PM)
	 */
	public XATQueueManager() {
		//storageManager = new StorageManager();
		queueToSMMapping = new Hashtable();
		this.updateMemoryThreshold();
		queues = new ArrayList();
		queueToPartitionMapping = new Hashtable();
	}
	
	/**
	 * Adds a queue for the queue manager to watch.
	 *
	 * Creation Date: (May 24, 2004 11:52:48 AM)
	 * @param queue XATPersistentQueue
	 */
	public synchronized void addQueue(XATPersistentQueueImp queue) {
		queues.add(queue);
	}
	
	/**
	 * Returns true or false, depending on whether there is enough free
	 * memory available based on the MEMORYTHRESHOLD.
	 * 
	 * Creation Date: (Jun 10, 2004 10:39:08 AM)
	 * @return boolean
	 */
	private synchronized boolean checkMemory() {
		this.updateMemoryThreshold();
		if (Runtime.getRuntime().freeMemory() > MEMORYTHRESHOLD)
			return true;
		else
			return false;
	}
	
	/**
	 * Dequeues a number of tuples from specified queue in an array.
	 * 
	 * Creation Date: (May 25, 2004 2:42:44 PM)
	 * @return XATTuple[]
	 * @param queue XATPersistentQueue
	 * @param howMany int
	 * @param customer Object
	 */
	public synchronized XATTuple[] dequeue(XATPersistentQueueImp queue, int number, Object customer) {
		
		String str = "DEQUEUE: ";
		boolean readFromFile = false;

		// check if the queue is being managed by the queue manager
		// if not, then add it to be managed
		if (!this.isRegistered(queue))
			this.addQueue(queue);
			
		//System.out.println("Trying to dequeue: " + number);

		// get cursor
		int cursor = queue.getCursor(customer);
		int newCursor = cursor;
		//System.out.println("Cursor: " + cursor);
		// put the actual results in this array
		XATTuple[] result = new XATTuple[number];
		int rInd = 0;
		
		// the PH index
		int phIndex = queue.getPHIndex();
		
		//System.out.println("PH Index: " + phIndex);
		//System.out.println("Queue Size (customer): " + queue.getQueueSize(customer));
		//System.out.println("Queue Size: " + queue.getQueueSize());
		//if (phIndex > 0)
		//	System.out.println("Class: " + queue.getContents().get(phIndex).getClass());
		// check validity
		if(cursor + number > queue.getQueueSize())
			throw new XATQueueException("Not enough tuples to dequeue!");
		// if asked to read from middle of file, adjust so that tuples at
		// the beginning are read into memory as well (but these will not
		// appear in the results)
		if(phIndex >= 0 && cursor > phIndex && 
			cursor <= phIndex + ((PlaceHolder)(queue.getContents().get(phIndex))).getNumTuples()) {
			
			newCursor = phIndex;
			number += cursor - newCursor;
			
			System.out.println("Adjusted cursor and number to dequeue. (BAD)");
		}
		
		// keeps track of how many tuples are left to be read
		int leftToRead = number;
		
		// put the full results in this array (only used if reading from disk)
		XATTuple[] fullResult = new XATTuple[number];
		int frInd = 0;
		
		// get tuples		
		if (queue.getPHIndex() < 0) {
			// just read off the result from main memory
			for (int i = cursor; i < queue.getContents().size() && leftToRead > 0; i++) {
				result[rInd] = (XATTuple)queue.getContents().get(i);
				rInd++;
				leftToRead--;
			}
			str += result.length + " tuples read from main memory.";

		}
		else {
			// read in any tuples before the place holder
			for(int i = newCursor; i < phIndex && leftToRead > 0; i++){
				if (i >= cursor) {
					result[rInd] = (XATTuple)queue.getContents().get(i);
					rInd++;
				}
				fullResult[frInd] = (XATTuple)queue.getContents().get(i);
				frInd++;
				leftToRead--;
			}
			
			try {
				if (leftToRead > 0) {
					readFromFile = true;
						
					//System.out.println("Purge: " + purge);
					// read in tuples from file
					XATTuple[] tempResult = this.getFromFile(queue, leftToRead, true);
				
					//System.out.println("Read from File: " + tempResult.length);
				
					// add tuples from tempResult to result
					for (int j = 0; j < tempResult.length; j++) {
						if (phIndex + j >= cursor) {
							result[rInd] = tempResult[j];
							rInd++;
						}
						fullResult[frInd] = tempResult[j];
						frInd++;
						leftToRead--;
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
				leftToRead -= ((PlaceHolder)queue.getContents().get(phIndex)).getNumTuples();
			}
			
			if (leftToRead > 0) {
				// read anything after the place holder, if necessary
				for (int i = 1; leftToRead > 0 && (phIndex + i) < queue.getContents().size(); i++, leftToRead--) {
					result[rInd] = (XATTuple)queue.getContents().get(phIndex + i);
					fullResult[frInd] = (XATTuple)queue.getContents().get(phIndex + i);
					rInd++;
					frInd++;
				}
			}
			str += result.length + " tuples read from disk.";

		}
		
		// if tuples were read from file, restructure the queue
		// to put read tuples into main memory
		//if (phIndex >= 0 && newCursor + fullResult.length >= phIndex) {
		if (readFromFile) {
			List newTuples = new ArrayList();
			// add everything up to the placeHolder
			for (int i = 0; i < newCursor; i++) {
				newTuples.add(queue.getContents().get(i));
			}
			for (int i = 0; i < fullResult.length; i++) {
				newTuples.add(fullResult[i]);
			}
			
			// add the place holder if necessary and everything after it
			PlaceHolder ph = (PlaceHolder)queue.getContents().get(phIndex);
			if (newCursor + fullResult.length < phIndex + ph.getNumTuples()) {
			//if (!purge) {
				// update PH info
				ph.setNumTuples(ph.getNumTuples() - (newCursor + fullResult.length - phIndex));
				newTuples.add(ph);
				queue.setPHIndex(newTuples.size()-1);
				// add everything after place holder
				for (int i = phIndex + 1; i < queue.getContents().size(); i++) {
					newTuples.add(queue.getContents().get(i));
				}
			}
			else {
				// don't add place holder
				for (int i = newCursor + fullResult.length - ph.getNumTuples() + 1; i < queue.getContents().size(); i++) {
					newTuples.add(queue.getContents().get(i));
				}
				queue.setPHIndex(-1);
			}

			queue.setTuples(newTuples);
		}
		
		try {
			FileWriter out = new FileWriter("queue_history.txt", true);
			out.write(str + "  -----> " + queue.getQueueLayout() + "\n");
			out.close();
		} catch (IOException e) {
			System.out.println(e);
			e.printStackTrace();
		}
		queue.cleanTuples();
		return result;
	}
	
	/**
	 * Dequeues one tuple from specified queue.
	 * 
	 * Creation Date: (May 25, 2004 2:49:55 PM)
	 * @return XATTuple
	 * @param queue XATPersistentQueue
	 * @param customer Object
	 */
	public synchronized XATTuple dequeue(XATPersistentQueueImp queue, Object customer) throws XATQueueException {
		
		// get tuple
		XATTuple result = this.dequeue(queue, 1, customer)[0];

		return result;
	}

	/**
	 * Dequeues a number of tuples from specified queue in a list.
	 *
	 * Creation Date: (May 25, 2004 2:49:44 PM)
	 * @return List
	 * @param queue XATPersistentQueue
	 * @param number int
	 */
	public synchronized List dequeueN(XATPersistentQueueImp queue, int number, Object customer) {

		List result = new ArrayList();
		XATTuple[] tempResult = this.dequeue(queue, number, customer);
		
		// convert the array into a list
		for (int i = 0; i < tempResult.length; i++) {
			result.add(tempResult[i]);
		}
		
		return result;

	}

	/**
	 * Enqueues a list of tuples into specified queue.
	 * 
	 * Creation Date: (May 25, 2004 11:33:09 AM)
	 * @param queue XATPersistentQueue
	 * @param tuples List
	 */
	public synchronized void enqueue(XATPersistentQueueImp queue, List tuples) {
		//dummyTuple = (XATTuple)tuples.get(0);
		//String str = "ENQUEUE: "; 
		
		// make sure the queue is registered with the queue manager,
		// if it's not, add it to be managed
		if (!this.isRegistered(queue))
			this.addQueue(queue);
		
		// if there is enough main memory left, simply add the tuples to
		// the queue's tuple list
		if (this.checkMemory()) {
			queue.addTuples(tuples);
			//str += tuples.size() + " tuples written to main memory.";
		}
		else { // otherwise, write the tuples to disk
			this.sendToFile(queue, tuples);
			if (queue.getPHIndex() > -1) {
				PlaceHolder ph = (PlaceHolder)(queue.getContents().get(queue.getPHIndex()));
				ph.setNumTuples(ph.getNumTuples() + tuples.size());
			}
			else {
				queue.addTuple(new PlaceHolder(tuples.size()));
			}
			//str += tuples.size() + " tuples written to disk.";
		}
		
		/*
		try {
			FileWriter out = new FileWriter("queue_history.txt", true);
			out.write(str + "  -----> " + queue.getQueueLayout() + "\n");
			out.close();
		} catch (IOException e) {
			System.out.println(e);
			e.printStackTrace();
		}
		*/
	}
	
	/**
	 * Enqueues a tuple into specified queue.
	 * 
	 * Creation Date: (May 25, 2004 1:19:47 PM)
	 * @param queue XATPersistentQueue
	 * @param tuple XATTuple
	 */
	public synchronized void enqueue(XATPersistentQueueImp queue, XATTuple tuple) {
		//dummyTuple = tuple;
		//String str = "ENQUEUE: ";
		// make sure the queue is registered with the queue manager,
		// if it's not, add it to be managed
		if (!this.isRegistered(queue))
			this.addQueue(queue);
		
		// if there is enough main memory left, simply add the tuples to
		// the queue's tuple list
		if (this.checkMemory()) {
			queue.addTuple(tuple);
			//str += "1 tuple written to main memory.*";
		}
		else { // otherwise, write the tuples to memory
			this.sendToFile(queue, (XATTupleImp)tuple);
			if (queue.getPHIndex() > -1) {
				PlaceHolder ph = (PlaceHolder)(queue.getContents().get(queue.getPHIndex()));
				ph.setNumTuples(ph.getNumTuples() + 1);
			}
			else {
				queue.addTuple(new PlaceHolder(1));
			}
			//str += "1 tuple written to disk.*";
		}
		/*
		try {
			FileWriter out = new FileWriter("queue_history.txt", true);
			out.write(str + "  -----> " + queue.getQueueLayout() + "\n");
			out.close();
		} catch (IOException e) {
			System.out.println(e);
			e.printStackTrace();
		}
		*/
	}
	
	/**
	 * Enqueues an array of tuples into specified queue.
	 * 
	 * Creation Date: (May 25, 2004 1:23:10 PM)
	 * @param queue XATPersistentQueue
	 * @param tuples XATTuple[]
	 */
	public synchronized void enqueue(XATPersistentQueueImp queue, XATTuple[] tuples) {
		//dummyTuple = tuples[0];
		//String str = "ENQUEUE: ";
		// make sure the queue is registered with the queue manager,
		// if it's not, add it to be managed
		if (!this.isRegistered(queue))
			this.addQueue(queue);
		
		// if there is enough main memory left, simply add the tuples to
		// the queue's tuple list
		if (this.checkMemory()) {
			for(int i = 0; i < tuples.length; i++)
				queue.addTuple(tuples[i]);
			//str += tuples.length + " tuples written to main memory.";
		}
		else { // otherwise, write the tuples to memory
			// convert from XATTuple[] to XATTupleImp[]
			XATTupleImp[] tempTuples = new XATTupleImp[tuples.length];
			for (int i = 0; i < tuples.length; i++) {
				tempTuples[i] = (XATTupleImp) tuples[i]; 
			}
			this.sendToFile(queue, tempTuples);
			if (queue.getPHIndex() > -1) {
				PlaceHolder ph = (PlaceHolder)(queue.getContents().get(queue.getPHIndex()));
				ph.setNumTuples(ph.getNumTuples() + tuples.length);
			}
			else {
				queue.addTuple(new PlaceHolder(tuples.length));
			}
			//str += tuples.length + " tuples written to disk.";
		}
		/*
		try {
			FileWriter out = new FileWriter("queue_history.txt", true);
			out.write(str + "  -----> " + queue.getQueueLayout() + "\n");
			out.close();
		} catch (IOException e) {
			System.out.println(e);
			e.printStackTrace();
		}
		*/
	}

	/**
	 * Gets the given number of tuples from the queue's file. This will delete
	 * the tuples from disk once they are read.
	 * Note: This DOES NOT dequeue from the beginning of the queue, just from
	 * the external file of the queue.
	 * 
	 * Creation Date: (Jun 1, 2004 3:33:56 PM)
	 * 
	 * @param queue XATPersistentQueue
	 * @param number int
	 * @return java.util.List
	 * 
	 */
	public synchronized XATTuple[] getFromFile(XATPersistentQueueImp queue, int number, boolean purge) 
		throws InvalidStorageManagerInstanceException {
		
		// first, make sure there is a storage manager for the queue's schema
		StorageManager sm = null;
		
		if (queueToSMMapping == null || !queueToSMMapping.containsKey(queue)
			|| queueToSMMapping.get(queue) == null) {
			throw new InvalidStorageManagerInstanceException("No instance of Storage Manager for this queue's schema.");
		}
		else {
			sm = (StorageManager)queueToSMMapping.get(queue);
		}
		
		int partitionNumber;
		
		// look up the partition number
		if (!queueToPartitionMapping.containsKey(queue))
			return new XATTuple[0];
		else
			partitionNumber = ((Integer)queueToPartitionMapping.get(queue)).intValue();
		//System.out.println("Attempting to retrieve partition number: " + partitionNumber);
		
		// make sure there are enough tuples to get from disk
		// if not, set to max number that can be retrieved
		if (number > sm.size(partitionNumber))
			number = (int)sm.size(partitionNumber);
		//if (number > storageManager.size(partitionNumber))
		//	number = (int)storageManager.size(partitionNumber);

		
		XATTuple[] result = new XATTuple[number];
		
		// get the tuples from file
		try {
			//result = storageManager.retrieveFromDisk(partitionNumber, number, purge);
			result = sm.retrieveFromDisk(partitionNumber, number, purge);
			//System.out.println("Attempted to read from file: " + partitionNumber);
		} catch (Exception e) {
			System.out.println(e.toString());
			e.printStackTrace();
		}

		return result;		
	}
	
	/**
	 * Gets the memory threshold.
	 * 
	 * Creation Date: (Jun 29, 2004 10:49:51 AM)
	 * @return long
	 */
	public synchronized long getMemoryThreshold() {
		return MEMORYTHRESHOLD;
	}

	/**
	 * Gets the number of queues managed by this QM.
	 * 
	 * Creation Date: (Jun 23, 2004 11:31:10 AM)
	 * @return int
	 */
	public synchronized int getNumQueues() {
		return this.queues.size();
	}
	
	/**
	 * Returns the queue at the given index;
	 * 
	 * Creation Date: (Jun 23, 2004 1:54:43 PM)
	 * @param index int
	 * @return XATPersistentQueue
	 */
	public synchronized XATPersistentQueueImp getQueue(int index) {
		return (XATPersistentQueueImp)this.queues.get(index);
	}
	
	/**
	 * Gets a handle for the queue manager.
	 * 
	 * Creation Date: (May 24, 2004 11:05:55 AM)
	 * @return XATQueueManager
	 */
	public static XATQueueManager getQueueManager() {		
		if(queueManager == null) {
			queueManager = new XATQueueManager();
		}
		
		return queueManager;
	}
	
	/**
	 * Returns the total number of tuples stored on disk from all queues.
	 * 
	 * Creation Date: (Jun 23, 2004 11:34:34 AM)
	 * @return int
	 */
	public synchronized int getTotalOnDisk() {
		XATPersistentQueueImp currQ;
		PlaceHolder ph;
		int total = 0;
		for (int i = 0; i < this.queues.size(); i++) {
			currQ = (XATPersistentQueueImp)this.queues.get(i);
			if (currQ.getPHIndex() >= 0) {
				ph = (PlaceHolder)currQ.getContents().get(currQ.getPHIndex());
				total += ph.getNumTuples();
			}
		}
		return total;
	}
	
	/**
	 * Returns the total number of tuples that currently exist in all
	 * queues.
	 * 
	 * Creation Date: (Jun 23, 2004 11:36:12 AM)
	 * @return int
	 */
	public synchronized int getTotalTuples() {
		XATPersistentQueueImp currQ;
		int total = 0;
		for (int i = 0; i < this.queues.size(); i++) {
			currQ = (XATPersistentQueueImp)this.queues.get(i);
			total += currQ.getQueueSize();
		}
		return total;
	}
	
	/**
	 * Checks whether the given queue is managed by the 
	 * queue manager.
	 * 
	 * Creation Date: (May 24, 2004 11:55:32 AM)
	 * @param queue XATPersistentQueue
	 * @return boolean
	 */
	public synchronized boolean isRegistered(XATPersistentQueueImp queue) {
		if (queues.contains(queue))
			return true;
		else
			return false;
	}
	
	/**
	 * Peek at the given number of tuples in the queue. This is analagous
	 * to dequeue, but this will not increment or cleanup cursors. If 
	 * tuples have to be read from disk, then the queue will be 
	 * restructured to put those tuples in main memory.
	 * 
	 * Creation Date: (Jun 15, 2004 11:48:23 AM)
	 * @param queue XATPersistentQueue
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	public synchronized XATTuple[] peek (XATPersistentQueueImp queue, int number, Object customer) {
		return this.dequeue(queue, number, customer);
	}
	
	/**
	 * Retrieves a tuple but also keeps the tuple in the queue. If the
	 * tuple is read from disk, the queue is restructured to put it in
	 * main memory.
	 *
	 * Creation Date: (Jun 15, 2004 1:17:50 PM)
	 * @param queue XATPersistentQueue
	 * @param customer Object
	 * @return XATTuple
	 */
	public synchronized XATTuple peek(XATPersistentQueueImp queue, Object customer) {
		return this.dequeue(queue, customer);
	}

	/**
	 * Peek at the last tuple in the queue. This is for XATSubQueueImp.
	 * XATSubQueueImp needs to merge a "subtuple" into the last tuple 
	 * (which must also be a "subtuple") to form a "full tuple".
	 * This is analagous to dequeue, but this will not increment or 
	 * cleanup cursors. If the tuple is read from disk, the queue is 
	 * restructured to put it in main memory.
	 * 
	 * Creation Date: (Jun 15, 2004 1:20:56 PM)
	 * @param queue XATPersistentQueue
	 * @return XATTuple
	 */
	public synchronized XATTuple peekLast(XATPersistentQueueImp queue) {
		
		int number = 1;
		
		// check if the queue is being managed by the queue manager
		// if not, then add it to be managed
		if (!this.isRegistered(queue))
			this.addQueue(queue);

		// get cursor
		int cursor = queue.getQueueSize() - 1;
		int newCursor = cursor;
		
		// put the actual results in this array
		XATTuple[] result = new XATTuple[number];
		int rInd = 0;
		
		// the PH index
		int phIndex = queue.getPHIndex();
	
		// check validity
		if(cursor + number > queue.getQueueSize())
			throw new XATQueueException("Not enough tuples to dequeue!");
		// if asked to read from middle of file, adjust so that tuples at
		// the beginning are read into memory as well (but these will not
		// appear in the results)
		if(phIndex >= 0 && cursor > phIndex && 
			cursor <= phIndex + ((PlaceHolder)(queue.getContents().get(phIndex))).getNumTuples()) {
			
			newCursor = phIndex;
			number += cursor - newCursor;
		}
		
		// keeps track of how many tuples are left to be read
		int leftToRead = number;
		
		// put the full results in this array (only used if reading from disk)
		XATTuple[] fullResult = new XATTuple[number];
		int frInd = 0;
		
		// get tuples		
		if (queue.getPHIndex() < 0) {
			// just read off the result from main memory
			for (int i = cursor; i < queue.getContents().size() && leftToRead > 0; i++) {
				result[rInd] = (XATTuple)queue.getContents().get(i);
				rInd++;
				leftToRead--;
			}
		}
		else {
			// read in any tuples before the place holder
			for(int i = newCursor; i < phIndex && leftToRead > 0; i++){
				if (i >= cursor) {
					result[rInd] = (XATTuple)queue.getContents().get(i);
					rInd++;
				}
				fullResult[frInd] = (XATTuple)queue.getContents().get(i);
				frInd++;
				leftToRead--;
			}
			
			try {
				if (leftToRead > 0) {
					// read in tuples from file
					XATTuple[] tempResult = this.getFromFile(queue, leftToRead, true);
					
					System.out.println("Read from File: " + tempResult.length);
					
					// add tuples from tempResult to result
					for (int j = 0; j < tempResult.length; j++) {
						if (phIndex + j >= cursor) {
							result[rInd] = tempResult[j];
							rInd++;
						}
						fullResult[frInd] = tempResult[j];
						frInd++;
						leftToRead--;
					}
				}
			} catch (Exception e) {
				System.out.println(e.toString());
				leftToRead -= ((PlaceHolder)queue.getContents().get(phIndex)).getNumTuples();
			}
			
			if (leftToRead > 0) {
				// read anything after the place holder, if necessary
				for (int i = 1; leftToRead > 0; i++, leftToRead--) {
					result[rInd] = (XATTuple)queue.getContents().get(phIndex + i);
					fullResult[frInd] = (XATTuple)queue.getContents().get(phIndex + i);
					rInd++;
					frInd++;
				}
			}
		}
		
		// if tuples were read from file, restructure the queue
		// to put read tuples into main memory
		if (phIndex >= 0 && newCursor + fullResult.length >= phIndex) {
			List newTuples = new ArrayList();
			// add everything up to the placeHolder
			for (int i = 0; i < newCursor; i++) {
				newTuples.add(queue.getContents().get(i));
			}
			for (int i = 0; i < fullResult.length; i++) {
				newTuples.add(fullResult[i]);
			}
			
			// add the place holder if necessary and everything after it
			PlaceHolder ph = (PlaceHolder)queue.getContents().get(phIndex);
			if (newCursor + fullResult.length < phIndex + ph.getNumTuples()) {
				// update PH info
				ph.setNumTuples(ph.getNumTuples() - (newCursor + fullResult.length - phIndex));
				newTuples.add(ph);
				// add everything after place holder
				for (int i = phIndex + 1; i < queue.getContents().size(); i++) {
					newTuples.add(queue.getContents().get(i));
				}
			}
			else {
				// don't add place holder
				for (int i = newCursor + fullResult.length - ph.getNumTuples() + 1; i < queue.getContents().size(); i++) {
					newTuples.add(queue.getContents().get(i));
				}
			}

			queue.setTuples(newTuples);
		}
	
		return result[0];

	}
	/**
	 * Retrieve a number of tuples but also keeps the tuples in the queue.
	 * 
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @return List
	 * @param number int
	 * @param customer java.lang.Object
	 */
	public synchronized List peekN(XATPersistentQueueImp queue, int number, Object customer) {
		return this.dequeueN(queue, number, customer);
	}

	
	
	/**
	 * Prints out all the queues managed by this queue manager.
	 * 
	 * Creation Date: (May 25, 2004 10:24:05 AM)
	 */
	public synchronized void printQueues() {
		for (int i = 0; i < queues.size(); i++)
			System.out.println(queues.get(i).toString());
	}
	
	/**
	 * Removes a queue from the queue manager.
	 * 
	 * Creation Date: (May 24, 2004 11:53:20 AM)
	 * @param queue XATPersistentQueue
	 */
	public synchronized void removeQueue(XATPersistentQueueImp queue) {
		queues.remove(queue);
	}
	
	/**
	 * Writes the given tuples to the queue's file.
	 * 
	 * Creation Date: (Jun 1, 2004 3:36:05 PM) 
	 * @param queue XATPersistentQueue
	 * @param tuples XATTupleImp[]
	 */
	public synchronized void sendToFile(XATPersistentQueueImp queue, XATTupleImp[] tuples) {
		
		int partitionNumber = 0;
		
		// get the queue's partition number
		if (queueToPartitionMapping.containsKey(queue)) {
			partitionNumber = ((Integer)queueToPartitionMapping.get(queue)).intValue();
		}
		else {
			// get a new, random partition number
			Random rand = new Random();
			do {
				partitionNumber = Math.abs(rand.nextInt());
			} while (queueToPartitionMapping.containsValue(new Integer(partitionNumber)));
			
			// add the new partition number to the mapping
			queueToPartitionMapping.put(queue, new Integer(partitionNumber));
			//System.out.println("Assigned partition number: " + partitionNumber);
		}
		
		
		//if (storageManager == null) {
		//	storageManager = new StorageManager(tuples[0]);
		//}
		
		
		// get the appropriate instance of the storage manager and use it
		// to save tuples to disk
		if (queueToSMMapping == null) {
			queueToSMMapping = new Hashtable();
		}
		StorageManager sm = null;
		if (queueToSMMapping.containsKey(queue)) {
			sm = (StorageManager)queueToSMMapping.get(queue);
		}
		else {
			sm = new StorageManager(tuples[0]);
			queueToSMMapping.put(queue, sm);
		}
		
		
		try {
			//storageManager.saveToDisk(partitionNumber, tuples);
			sm.saveToDisk(partitionNumber, tuples);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Writes the given tuples to the queue's file.
	 * 
	 * Creation Date: (Jun 10, 2004 10:58:06 AM) 
	 * @param queue XATPersistentQueue
	 * @param tuples java.util.List
	 */
	public synchronized void sendToFile(XATPersistentQueueImp queue, List tuples) {
		// turn the List into an array
		XATTupleImp[] tupleArray = new XATTupleImp[tuples.size()];
		for (int i = 0; i < tuples.size(); i++)
			tupleArray[i] = (XATTupleImp)tuples.get(i);
		
		this.sendToFile(queue, tupleArray);
	}

	/**
	 * Writes the given tuple to the queue's file.
	 * 
	 * Creation Date: (Jun 10, 2004 10:55:31 AM)
	 * @param queue XATPersistentQueue
	 * @param tuple XATTuple
	 */
	public synchronized void sendToFile(XATPersistentQueueImp queue, XATTupleImp tuple) {
		XATTupleImp[] tuples = {tuple};
		sendToFile(queue, tuples);
	}
	
	/**
	 * Sets the memory threshold.
	 * 
	 * Creation Date: (Jul 1, 2004 11:55:27 AM)
	 * @param newThreshold long
	 */
	public synchronized void setMemoryThreshold(long newThreshold) {
		MEMORYTHRESHOLD = newThreshold; 
	}
	
	/**
	 * Sets the queues for this queue manager to handle.
	 * 
	 * Creation Date: (May 24, 2004 11:25:30 AM)
	 * @param newQueues java.util.List
	 */
	public synchronized void setQueues(List newQueues) {
		queues = newQueues;
	}
	
	/**
	 * This is where the memory threshold is set and updated, if it needs
	 * to be. Any changes to the threshold should be put here.
	 * 
	 * Creation Date: (Jun 29, 2004 10:49:24 AM)
	 */
	public synchronized void updateMemoryThreshold() {
		// set the memory threshold to half of the total available memory
		//MEMORYTHRESHOLD = Runtime.getRuntime().totalMemory()/2;
		
		// set the memory threshold to 0, i.e. all tuples should be sent
		// to main memory
		//MEMORYTHRESHOLD = 0;
		
		// set the memory threshold to the maximum long value, i.e. all 
		// tuples should be sent to disk
		MEMORYTHRESHOLD = Long.MAX_VALUE;
		
		// randomize when tuples are sent to disk
		//Random rand = new Random();
		//boolean bool = rand.nextBoolean();
		//if (bool) {
		//	MEMORYTHRESHOLD = 0;
		//}
		//else {
		//	MEMORYTHRESHOLD = Long.MAX_VALUE;
		//}
	}
}
