/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;

/**
 * The chain scheduler is a scheduler designed to minimize input queues by
 * grouping operators into chains to process the most tuples in the least amount
 * of time. This is done by creating progress charts for every possible path
 * in the <tt>XATTree</tt> and calculating the lower envelope of the chart created
 * using the Average Processing Time and Selectivity of each operator. If there are
 * multiple progress charts, each chain from every progress chart are ordered by slope
 * and then processed using a greedy algorithm. This scheduler does not focus on throughput
 * so there is a high possibility that tuples are not processed through the root node
 * until there is a stall in the streaming data, or some other external event.
 *
 * Creation date: (4/15/2003 4:07:00 PM)
 * @author: Tim Sutherland
 * @version: 1.1.1
 */
public class ChainScheduler extends XATScheduler {
/** A vector of all of the partial tree objects */
private java.util.Vector partialTrees;
/** A vector of chain objects */	
private java.util.Vector chains;
/** A vector of all of the ProgressChart objects */
private java.util.Vector progressCharts;
/** The current chain that we are working on */
private int currentChain;
/** The current operator that we are working on */
private int currentOperator;
/** The current workload that the chain is processing */
private int workload;
/** The number of milliseconds to wait before re-evaluating the XATTree */
private long evaluationThreshold;
/** The last time in the system that the makeDecision() method was called */
private long lastSystemTime;
/** Whether or not to print output to console */
private boolean verbose;
/**
 * Default constructor
 */
public ChainScheduler() {
	super();
	lastSystemTime = System.currentTimeMillis();
	evaluationThreshold = 500;

	verbose = false;
	//System.out.println("*********** CHAIN SCHEDULER *******************");
		
}
/**
 * Creates a vector of all the partial trees from the given XATTree
 * We want to create a partial tree for all of the possible paths in the XATTree
 * So it can be analyzed for our Chain Algorithm.
 * This only works for non DAG trees at the moment.
 * 
 * Creation date: (4/19/2003 9:26:08 AM)
 * @return java.util.Vector- The partial trees from the given XATTree
 */
public Vector createPartialTrees(XATTree tree) {

	if(verbose){
	System.out.println("Creating Partial Trees from:");
	System.out.println(tree.toString());
	}
	
	Vector ret = new Vector();
	XATNode[] leaves = tree.getLeaves();
	XATNode parent;
	XATNode newParent, newChild;
	XATTree[] partialTree = new XATTree[leaves.length];

	// Go through all of the leaves of the tree and make a partial
	// tree.

	for(int i=0; i<leaves.length ; i++){

		partialTree[i] = new XATTreeImp();
		newParent = new XATNodeImp();
		newParent.setOperator(leaves[i].getOperator());
		partialTree[i].setRoot(newParent);
		parent = leaves[i].getParent();

		// Recurse through the tree and attach the parent nodes
		// to the new tree.

		while(parent != null){
			newChild = new XATNodeImp();
			newChild.setOperator(parent.getOperator());
			try{

			newParent.setOpChild(0,newChild);
			newParent = newParent.getOpChild(0);
			}
			catch (edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException ndnee){
				System.out.println("Error: " + ndnee.getMessage());

			}
			
			newChild.setParent(newParent.getParent());
			parent = parent.getParent();
			
		}

		if(verbose)
		System.out.println("Partial Tree " + i + "\n" + partialTree[i].toString());
		ret.add(i, partialTree[i]);
	}
	
	// Returns a vector of XATTrees
	return ret;
}
/**
 * Creates a new set of chains. This method is to be called when we want
 * new chain values to schedule over.
 * Creation date: (4/19/2003 9:41:08 AM)
 */
private void evaluateTree() {

	if(verbose)
		System.out.println("Evaluating Tree Information");

	Vector currentChain = new Vector();
	Vector sortedChain = new Vector();
	progressCharts = new Vector();
	chains = new Vector();

	this.workload = 0;
	this.currentOperator = 0;
	this.currentChain = -1;

	// Create all of the partial trees for the XATTree given
	partialTrees = createPartialTrees(getTree());

	// For each partial tree, make a progress chart, calculate its chain and add it to our chain
	if(verbose)
		System.out.println("Creating Progress Charts");
		
	for(int i=0; i < partialTrees.size() ; i++){

		//For each partial tree, make the progress chart
		progressCharts.add(i, new ProgressChart((XATTree) partialTrees.elementAt(i)));
		if(verbose)
			System.out.println("Finished Making Progress Chart " + i);
		// Get the chain from the progress chart
		currentChain = ((ProgressChart) progressCharts.elementAt(i)).getChain();
		if(verbose)
			System.out.println("Retrieving Progress Chart " + i);
		//Add these chains to our overall chain collection from all partial trees
		for(int j =0; j< currentChain.size(); j++){
			chains.addElement((Chain) currentChain.elementAt(j));
		}

	}
	if(verbose){
	System.out.println("Here are the chains unsorted:");
	System.out.println(chains.toString());
	}

	// Sort the chain in ascending order based on slope for the greedy algorithm

	double curr, max;
	int maxChain;
	int overallSize = chains.size();

	for(int k =0; k < overallSize ; k++){
		curr = 0;
		max = 0;
		maxChain = 0;
		for(int l = 0; l < chains.size() ; l++){
			curr = ((Chain) chains.elementAt(l)).getSlope();
			max = Math.max(curr, max);
			if(curr == max)
				maxChain = l;
		}

		sortedChain.addElement(chains.elementAt(maxChain));
		chains.removeElementAt(maxChain);
	}

	chains = sortedChain;

	if(verbose){
	System.out.println("Sorting the Chain");
	System.out.println("Here are the chains sorted:");
	System.out.println(chains.toString());
	System.out.println("Completed Tree Evaluation");
	}
		
}
/**
 * Gets the number of milliseconds before we re-evaluate the XATTree
 * Creation date: (4/23/2003 11:10:41 AM)
 * @return long- Number of milliseconds to wait.
 */
public long getEvaluationThreshold() {
	return evaluationThreshold;
}
/**
 * This is called by the Execution Controller.  This method decides
 * which operator to run next and how many tuples that operator should dequeue.
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision
 */
public XATSchedulerDecision makeNextDecision() throws XATSchedulerException {

	XATSchedulerDecision decision = new XATSchedulerDecision();
	Chain currChain = null;
	long newSystemTime = System.currentTimeMillis();

	if((newSystemTime - lastSystemTime) > evaluationThreshold)
		evaluateTree();

	lastSystemTime = newSystemTime;

	while(true){

		int i = 0;

		// See if we are currently working on a chain
		// If we are on a chain, schedule the next operator
		// and return it.
		// If we are on the last operator in the chain, reset to find the next chain
		// to operate on, and continue.

		if(currentChain >= 0){
			
			if(currentOperator > ((Chain) chains.elementAt(currentChain)).getNumberOfOperators() - 1){
				currentChain = -1;
				currentOperator = 0;
				workload = 0;
			}
			else{
				currChain = ((Chain) chains.elementAt(currentChain));
	    		decision.setNextOperator(currChain.getOperator(currentOperator));
	    		workload = calculateWorkLoad(currChain.getOperator(currentOperator));
      			decision.setNumberOfTuples(workload);
				currentOperator++;
				return decision;
			}
		}
		//Otherwise find the current chain that has work to do and schedule the
		//chains first operator
		
		while(workload <= 0 && i < chains.size()){
			currChain = ((Chain) chains.elementAt(i));
			currentChain = i;
			currentOperator = 0;
			// See if the chain has a workload at the top of the chain
			workload = calculateWorkLoad(currChain.getOperator(currentOperator));
			i++;

		}

		//If no chain has a workload, we are done.
		if(workload <= 0)
			throw new XATSchedulerException("No more work to do"); 

	    decision.setNextOperator(currChain.getOperator(currentOperator));		
        decision.setNumberOfTuples(workload);
		currentOperator++;
		return decision;

	}
}
/**
 * Sets the number of milliseconds to wait before re-evaluating the XATTree.
 * Creation date: (4/23/2003 11:10:41 AM)
 * @param newEvaluationThreshold long
 */
public void setEvaluationThreshold(long newEvaluationThreshold) {
	evaluationThreshold = newEvaluationThreshold;
}
/**
 * Sets the tree for the XATScheduler Class
 * Creation date: (2/22/03 1:46:59 PM)
 * @param newTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public void setTree(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree newTree) {
	super.setTree(newTree);
	evaluateTree();
}
}
