/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsReference;

/**
 * The progress chart class represents the chart made from a partial algebra tree
 * by getting the Selectivity and ATP of each operator. We can use this information to
 * calculate the lower envelope between points and create our chains.
 * Creation date: (4/15/2003 4:20:00 PM)
 * @author: Tim Sutherland
 * @version: 1.1.1
 */
public class ProgressChart {

/** An array of progress chart points that are in the chart. */
private ProgressChartPoint[] pcp;
/** The partial tree from which to create the progress chart */
private XATTree tree;
/** A node array to represent all of the nodes in the partial tree */
private XATNode[] node;
/** Statistics gatherer reference */
private StatisticsGatherer stats;
/** A vector of chain objects */
private Vector chain;
/** Default Constructor
 * @param XATTree- the partial tree for the progress chart
 */	
public ProgressChart(XATTree tree){

		this.tree = tree;
		this.node = tree.convertToArray();
		this.stats = StatisticsGatherer.getStatisticsGather();
		pcp = new ProgressChartPoint[this.tree.size()];
		createChart();
		calculateLowerEnvelope();

	}
/**
 * Calculates the lower envelope of between points on the chart
 * It then creates the chains that we need to give to the <tt>ChainScheduler</tt>
 */ 
public void calculateLowerEnvelope(){

		//System.out.println("Calculating Lower Envelope");

		Chain currChain = null;
		chain = new Vector();
		double slope, slopeMax =Integer.MIN_VALUE;
		int indexMax = 0;
		int i = 0;

		// Create all of the chains, and put them in the chain variable

		for(i = 0; i < (pcp.length-1) ;){
			for(int j = i+1; j < pcp.length ; j++){

				// Get the slope between the starting point and the current point

				if(pcp[j].getRelativeAverageProcessTime() - pcp[i].getRelativeAverageProcessTime() == 0)
					slope = 0;
				else
					slope = -1 * ((pcp[j].getRelativeSelectivity() - pcp[i].getRelativeSelectivity()) /
						 (pcp[j].getRelativeAverageProcessTime() - pcp[i].getRelativeAverageProcessTime()));

				if(slope >= slopeMax){
					slopeMax = slope;
					indexMax = j;
					
				}

			} // end inner loop


			// Make the new chain based on the max slope found above.
			
			currChain = new Chain();
			currChain.setSlope(slopeMax);
			
			// Add all of the operators in the slope of the chain, except the last operator, which
			// will be used in the next chain..
			
			for(int k = i; k< indexMax; k++){
				currChain.addOperator(pcp[k].getOperator());
			}

			// Unless the operator is alone, in which case add it anyways so we have an operator in the chain!
			//if(i == indexMax)
			//	currChain.addOperator(pcp[i].getOperator());

			chain.addElement(currChain);

			// Get ready to find the next chain
			i = indexMax;
			slopeMax = Integer.MIN_VALUE;

		} // end outer loop

		// Add the last operator to the chain, set the slope to "super huge" cause its the last operator

		currChain = new Chain();
		currChain.setSlope(Integer.MAX_VALUE);
		currChain.addOperator(pcp[pcp.length -1].getOperator());

		chain.addElement(currChain);

		// Lower envelope is done!
	}
/**
 * Creates the chart by analyzing the partial tree and its statistics.
 *
 */
public void createChart() {
    //System.out.println("Creating Chart");

    double selectivity, avgProcessTime;
    double currentSelectivity = 1.0;
    double currentTime = 0;

    // For every operator, find out its stats and put this
    // information into the ProgressChartPoints

//    System.out.println("Progress Chart:");

    for (int i = 0; i < pcp.length; i++) {

        selectivity =
            Double.parseDouble(
                stats.getValue(
                    "SELECTIVITY",
                    (XATStreamOperator) node[i].getOperator(),
                    StatisticsReference.AVERAGE));
        avgProcessTime =
            Double.parseDouble(
                stats.getValue(
                    "AVERAGE_TUPLE_PROCESSING_TIME",
                    (XATStreamOperator) node[i].getOperator(),
                    StatisticsReference.AVERAGE));

        if(selectivity > 0)
      	  currentSelectivity *= selectivity;
      	  
        currentTime += avgProcessTime;

        pcp[i] = new ProgressChartPoint();
        pcp[i].setSelectivity(selectivity);
        pcp[i].setAvgProcessTime(avgProcessTime);
        pcp[i].setOperator((XATStreamOperator) node[i].getOperator());
        pcp[i].setRelativeAverageProcessTime(currentTime);
        pcp[i].setRelativeSelectivity(currentSelectivity);
    //    System.out.println("Point " + i + ": " + pcp[i].toString());	
    }
}
/**
 * Returns a vector of chains for this Chart
 * Creation date: (4/17/2003 11:00:34 AM)
 * @return Vector
 */
public Vector getChain() {
	return chain;
}
}
