/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.NameValuePair;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsReference;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
/**
 * This class is responsible for determining which scheduling algorithm
 * to run next.  The various subclasses will each implement a certain heuristic 
 * @author: Brad Pielech
 */
public abstract class XATExecutionAdapter extends XATQueryObjectImp {
	/** 
	 * these are the actual scheduling algorithms
	 * the key is the scheduler's category name and the
	 * value is the actual XATScheduler object
	*/
	protected Hashtable schedulers;
	private ConfigFileParser config;
	private Machine machineConfig;
	
	/** this is the index of the current scheduling algorithm*/
	protected Object currentCategory;
	/**
	 * This is Hashtable where the key is the
	 * category name and the value is the algorithm's score (as a Double). 
	 */
	protected Hashtable scores;
	/* this is an array list of XATPreferences, the order does not really matter */
	private ArrayList preferences;
	protected edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree[] tree;
	private final double QUANTIFIER_PENALTY = 0.25;
	private final double QUANTIFIER_OK = 1;
	private java.util.ArrayList leftToRun;
	protected boolean debug;
	
public XATExecutionAdapter(ConfigFileParser parser, XATTree inTree) {
	super();
	config = parser;
	try {	
		ArrayList trees = new ArrayList();
    	trees.add(inTree);
		tree = (XATTree[])trees.toArray(new XATTree[] {});
		
		//tree[0] = inTree;	
		init();
	} catch (Exception e) {
		e.printStackTrace();
	}
}
/**
 * XATExecutionAdapter constructor comment.
 */
public XATExecutionAdapter(Machine config, XATTree[] tree)
{
	super();
	machineConfig = config;
	try
	{
		this.tree = tree;
		initDistribution();
	}
	catch(Exception e)
	{
		System.err.println(e.toString());
	}
}

/**
 * returns the category of the current scheduler
 * @return java.lang.Object
 */
public Object getCategory() {
	return currentCategory;
}

/**
 * Periodically the execution controller will request the Adapter to
 * change the algorithm b/c the previous one was not working well (ex: starvation).
 * Therefore, this method is similar to getNextScheduler, but getNext could result
 * in the same scheduler being used as the previous.  This method will guarentee a different
 * algorithm <p>
 * This can create a problem if there is only 1 algorithm available.  Therefore, it throws
 * an IllegalStateException if a switch is not possible. 
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler
 * @exception IllegalStateException if there is only 1 scheduler and a change is not possible
 */
public abstract XATScheduler forceChange() throws IllegalStateException;
/**
 * Returns the current Scheduler
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler
 */
public XATScheduler getCurrentScheduler() {
	return (XATScheduler)schedulers.get(currentCategory);
}
/**
 * Retrieves the score for the currentCategory
 * @return double
 */
protected double getCurrentScore() {
	return ((Double)scores.get(currentCategory)).doubleValue();
}
/**
 * Returns the index of the Scheduling algorithm that should be used next.
 * Each subclass will override this method b/c this is where the actual logic of switching 
 * takes place.  This method updates the currentCategory class variable and then
 * return the next scheduler using getCurrentScheduler().  
 * This is how schedulers are referenced internally. <p>
 *
 * This method will return the next algorithm that has not run yet.  Once an algorithm
 * has run, it will be removed from this list.  The exception is the first algorithm, which
 * will run twice.  Once the leftToRun list is empty, this algorithm will return null.  Therefore,
 * the caller should first call <tt>isLeftToRun</tt> to see if any algorithms still have to run.
 * If there are none, then the subclass should use its own logic.  
 * @return XATScheduler the next algorithm to use.  
 */
public XATScheduler getNextScheduler(){
	currentCategory = leftToRun.remove(0);
	return getCurrentScheduler();
	
}
/**
 * Return an Enumeration of the scheduling algorithms supported in the system
 * The Enumeration contains the category names, not the actual algorithms
 * @return java.util.Hashtable
 */
public Enumeration getSchedulers() {
	return schedulers.keys();
}

public Enumeration getSchedulerObjects() {
	return schedulers.elements();
}

/**
 * convienence method to retrieve a statistic for a tree
 * @return double
 * @param quantifier int
 */
private double getTreeStatistic(String name, int quantifier) {
	return Double.parseDouble(stats.getValue(name, tree[0], quantifier));
}

private double getTreeStatistic(String name, int quantifier, int treeNumber) {
	return Double.parseDouble(stats.getValue(name, tree[treeNumber], quantifier));
}
/** 
 * initialize the scheduling algorithms and the corresponding properties. First
 * parser XML and figure out the various algorithms that will be used.  Then
 * instantiate those schedulers, configure them, and store them in the
 * schedulers array. <p>
 * @throws Exception any number of exceptions can come from setting properties or parsing XML.  They
 * are just passed to somebody else to handle
 */
protected void init() throws Exception{
		schedulers = new Hashtable();
		scores = new Hashtable();
		leftToRun = new ArrayList();
		boolean isfirst = true;
        for(Iterator i = config.getSchedulingAlgorithms(); i.hasNext();){
			NameValuePair pair = (NameValuePair)i.next();
	        //make a new category in the stats gatherer
	        stats.addNewCategory(pair.getName());
     		XATScheduler s = (XATScheduler) Class.forName(pair.getValue()).newInstance();
     		s.setTree(tree[0]);     		
     		schedulers.put(pair.getName(), s);     		
     		//the first scheduler listed will be the first algorithm used
     		if(isfirst){
				isfirst = false;
				currentCategory = pair.getName();
     		} else{
	     		//dont want to put the first algorithm into the left to run list
				leftToRun.add(pair.getName());
     		}
     		scores.put(pair.getName(), new Double(0));
        }
        
        //now configure each scheduler
        for(Iterator it = config.getSchedulingProperties();it.hasNext();){
			NameValuePair pair = (NameValuePair)it.next();
			for (Iterator i = schedulers.values().iterator(); i.hasNext();){
				XATScheduler next = (XATScheduler)i.next();
				next.setProperty(pair.getName(), pair.getValue());
			}
			
        }
        preferences = new ArrayList();
        for(Iterator i = config.getPreferences(); i.hasNext();){
			preferences.add(i.next());
        }
        for(Iterator it = config.getExecutionAdapterProperties();it.hasNext();){
			NameValuePair pair = (NameValuePair)it.next();
			setProperty(pair.getName(), pair.getValue());
        }
	        
        //add the current (first) category to the left to run so that it may run twice
        //leftToRun.add(currentCategory);
        debug = (new Boolean(this.getProperty("DEBUG"))).booleanValue();
}

/** 
 * initialize the scheduling algorithms and the corresponding properties. First
 * parser XML and figure out the various algorithms that will be used.  Then
 * instantiate those schedulers, configure them, and store them in the
 * schedulers array. <p>
 * @throws Exception any number of exceptions can come from setting properties or parsing XML.  They
 * are just passed to somebody else to handle
 */
protected void initDistribution() throws Exception{
		schedulers = new Hashtable();
		scores = new Hashtable();
		leftToRun = new ArrayList();
		boolean isfirst = true;
        for(Iterator i = machineConfig.getSchedulingAlgorithms(); i.hasNext();){
			XATSchedulingAlgorithm sa = (XATSchedulingAlgorithm) i.next();
			NameValuePair pair = new NameValuePair(sa.getAlgorithmName(),sa.getAlgorithmClassName());
			
	        //make a new category in the stats gatherer
	        stats.addNewCategory(pair.getName());
     		XATScheduler s = (XATScheduler) Class.forName(pair.getValue()).newInstance();
     		s.setTrees(tree);     
     		
     		for(int x=0; x<tree.length; x++) {
     			tree[x].addSchedulerCategory(pair.getName());		
     		}
     		schedulers.put(pair.getName(), s);     		
     		//the first scheduler listed will be the first algorithm used
     		if(isfirst){
				isfirst = false;
				currentCategory = pair.getName();
     		} else{
	     		//dont want to put the first algorithm into the left to run list
				leftToRun.add(pair.getName());
     		}
     		scores.put(pair.getName(), new Double(0));
        }
        
        //now configure each scheduler
        for(Enumeration e = machineConfig.getProperties();e.hasMoreElements();){
			String name = (String) e.nextElement();

			for (Iterator i = schedulers.values().iterator(); i.hasNext();){
				XATScheduler next = (XATScheduler)i.next();
				if(next.isValidPropertyName(name)){
					String value = machineConfig.getProperty(name);
					next.setProperty(name,value);
				}
			}
        }
        
        /*
        for(Iterator it = config.getSchedulingProperties();it.hasNext();){
			NameValuePair pair = (NameValuePair)it.next();
			for (Iterator i = schedulers.values().iterator(); i.hasNext();){
				XATScheduler next = (XATScheduler)i.next();
				next.setProperty(pair.getName(), pair.getValue());
			}
			
        }
        */
        preferences = new ArrayList();
        for(Iterator i = machineConfig.getSchedulingPreferences(); i.hasNext();){
			preferences.add(i.next());
        }
        
        //for(Iterator it = config.getExecutionAdapterProperties();it.hasNext();){
		//	NameValuePair pair = (NameValuePair)it.next();
		//	setProperty(pair.getName(), pair.getValue());
        //}
	        
        //add the current (first) category to the left to run so that it may run twice
        //leftToRun.add(currentCategory);
        debug = (new Boolean(this.getProperty("DEBUG"))).booleanValue();
}
/**
 * Returns true if there are still algorithms that have not run yet.
 * Returns false if all algorithms have had a chance to run.
 * @return boolean
 */
protected boolean isLeftToRun() {
	//return the opposite of isEmpty
	return !leftToRun.isEmpty();
}
	/**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    */   
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		//the weight of the previous values in calculating average
		addProperty("DEBUG", "false");	
	}
/**
 * Updates the score for the scheduler that just ran.  The formula
 * that is used is zscore = (mean of scheduler - mean of overall) / (max - min of overall)
 * then multiply each zscore by the quantifier and weight and store that
 * as the score
 */
protected void updateScores() {

    Object oldCategory = stats.getCurrentCategory();
    /**
     * next calculate the zscore for each statistic of the preferences
     * for each category except the currentCategory
    */
    double score = 0;
    for (Enumeration e = schedulers.keys(); e.hasMoreElements();) {
        Object nextCategory = e.nextElement();        
      	//only update score for the schedulers that are not the current scheduler  
        if (!nextCategory.equals(currentCategory)) {
//	        System.out.println("Next category to update: " + nextCategory);
//	        System.out.println("Current Category is: " + currentCategory);
            for (int i = 0; i < preferences.size(); i++) {
                XATPreference p = (XATPreference) preferences.get(i);
                stats.setCurrentCategory(StatisticsGatherer.OVERALL);
                //get the mean and min/max for the overall category
                double overallMean =
                    getTreeStatistic(p.getStatistic(), StatisticsReference.AVERAGE);
                double min = getTreeStatistic(p.getStatistic(), StatisticsReference.LOWEST);
                double max = getTreeStatistic(p.getStatistic(), StatisticsReference.HIGHEST);
                stats.setCurrentCategory(nextCategory);
                //get the mean for this category
                double curMean =
                    getTreeStatistic(p.getStatistic(), StatisticsReference.AVERAGE);
                double zscore = 0;
                if ((max - min) != 0) {
                    zscore = (curMean - overallMean) / (max - min);
                }
                //the quantifier will be QUANTIFIER_OK if the zscore was less than 0 and we wanted to
                //minimize the statistic or the zscore was > 0 and we wanted to maximize it
                double quantifier =
                    (zscore * p.getQuantifier() > 0) ? QUANTIFIER_OK : QUANTIFIER_PENALTY;
                score += Math.abs(zscore) * quantifier * p.getWeight();
            }
            if(debug){
	            System.out.println("The score for " + nextCategory + " is " + score);
            }
            scores.put(nextCategory, new Double(score));
            score = 0;
        }
    }
    if(debug){
	    System.out.println("The score for the currentScheduler, " + currentCategory + ", is " + scores.get(currentCategory) + "\n");
    }
    //try to keep the statistics current by resetting the category
    stats.setCurrentCategory(oldCategory);
}
}
