/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidArgumentException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidIndexException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.NotEnoughBytesException;
/**
 * This class represents a file on the disk. It opens a file channel on
 * the file and writes records into the file. The records are written in
 * the form of byte arrays. This class is used by the StorageManager to 
 * store and retreive records from the file. The methods it implements 
 * will be similar to IStore.
 * 
 * @author Nishant Mehta and Abhiskek Kumar; edited by Natasha Bogdanova (6/9/04)
 */
public class DiskFile implements IRecordStore {
	
	/* Used to store the index of the records within the file.
	 * Basically stores the byte offset in the file.
	 * Is used to access randomly in the file. */	
	protected Hashtable indexTable = null;
	/* Stores the name of the file. */
	protected String fileName = null;
	/* Stores the instance of the Factory. */
	protected Factory factory = null;
	/* Stores the current size of the DiskFile in no of bytes */
	protected long partitionSize = 0;
	/* Stores the reference to the RandomAccessFile. */
	protected RandomAccessFile file = null;
	/* Stores the fileChannel. */
	protected FileChannel fc = null;
	/* Stores the file object for the file. */
	protected File objectFile = null;	
	/* Stores the current size of the diskFile in number of tuples 
	 * excluding deleted tuples. */
	protected long currentSize = 0;
	/* Stores the true size of the diskfile it includes the deleted tuples*/
	protected long trueSize = 0;
	/**
	 * Constructs a new DiskFile given a file name.
	 * 
	 * @param fileName String
	 */
	public DiskFile(String fileName,Factory factory)
	{				
		this.fileName = fileName; 	
		this.factory = factory;
		indexTable = new Hashtable();
		try {
			this.objectFile = new File(fileName);	
			this.objectFile.deleteOnExit();
			this.file = new RandomAccessFile(objectFile,"rw");
		} catch (FileNotFoundException e) {
			System.out.println("Exception "+e);
			e.printStackTrace();	
		}		
		fc = file.getChannel();			
		this.partitionSize = 0;
		this.currentSize = 0;	
		this.trueSize = 0;
	}	
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#cleanup()
	 */
	public void cleanup()
	{
		try {
			fc.close();
			file.close();		
			//System.out.println("Deleting temporary files");
			//System.out.println(this.objectFile.delete());
		} catch (IOException e) {
			System.out.println("Exception "+e);
			e.printStackTrace();
		}
	}		
	/**
	 * @see java.lang.Object#finalize()
	 */
	public void finalize() {
	try {
		fc.close();
		file.close();
		//System.out.println("File Deleting"+objectFile.getName());	
		} catch (java.io.IOException e) {
			System.out.println("Exception " + e);
			e.printStackTrace();
		}
	}
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#purgeBucketFromDisk()
	 */
	public void purgeDiskFile()
	{
		//Here we delete the file and remove all the records from the file
		//make a new file with the same name and set the partition size to zero
		//again		
		try {
			fc.close();
			file.close();
			fc = null;
			file = null;
			objectFile.delete();			
			//Making a new File with the same name
			try {
				this.file = new RandomAccessFile(this.fileName,"rw");
				this.objectFile = new File(fileName);	
				this.objectFile.createNewFile();
			} catch (FileNotFoundException e) {
				System.out.println("Exception "+e);
				e.printStackTrace();	
			}		
			fc = file.getChannel();			
			this.partitionSize = 0;
			this.currentSize = 0;
			this.indexTable = new Hashtable();
			this.trueSize = 0;
			
		} catch (java.io.IOException e) {
			System.out.println("Purge Problem");
			System.out.println("Exception " + e);
			e.printStackTrace();
		}					 	
	}
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#retrieveFromDisk(boolean)
	 */
	public Record[] retrieveFromDisk(boolean purgeTuples) {		
		System.out.println("!!!");
		int noOfRecords = indexTable.size();
		Record[] arrayRecord = new Record[(int)this.currentSize];
		try {		
			//Reset the File Channel position
			fc.position(0);			
			if (fc.size() > Integer.MAX_VALUE)
			{
				System.out.println("Will not work the file size is too big");
				System.out.println("Cannot fit into an integer");
			}
			// Note this is a hack I should  not covert the file size to int.
			// Files can get very large with streaming systems			
			ByteBuffer bb =
				ByteBuffer.allocateDirect((int) fc.size());				
			int nbytes = fc.read(bb);			
			//System.out.println(" The number of bytes read " +nbytes);									
			byte[] buffer = new byte[nbytes];	
			bb.rewind();
			bb.get(buffer);
			//Setting this to null so that it can be collected by the
			//Grabage collector
			bb = null;
			//System.out.println("The length of the buffer "+buffer.length);			
			for(int i=0,j=0;i<noOfRecords;i++)
			{				
				IndexEntry ie = (IndexEntry) indexTable.get(new Integer(i));

				if(!ie.isDeleted())
				{
					//System.out.println("StartOffset "+ie.startOffsetInFile+ " endOffset "+ie.endOffsetInFile);
					arrayRecord[j] = new Record(buffer, ie.startOffsetInFile, 
												ie.endOffsetInFile, ie.getPosition());	
					j = j + 1;
				}							
			}						
			//System.out.println("--------------------------");
			//System.out.println(arrayRecord[0].size());
			//System.out.println(arrayRecord[1].size());
			//System.out.println("--------------------------");			
			//ByteArrayInputStream bais = new ByteArrayInputStream(arrayRecord[1].getByteArray());							
			//ObjectInputStream oi = new ObjectInputStream(bais);		
			//System.out.println("---------Work man common---------");			
			//System.out.println(oi.readInt());
											
		} catch (IOException e) {			
			System.out.println("Exception"+e);
			e.printStackTrace();
		}		
		//if (purge) {
		//	// delete the contents of the file
		//	this.purgeDiskFile();
		//}
		return(arrayRecord);	
	}	
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#retrieveFromDisk(int, boolean)
	 * Retreive the records from the front of the file and pruges the retreived tuples
	 */
	public Record[] retrieveFromDisk(int noOfRecords, boolean purgeTuples) {			
		/*
		IndexEntry enCurr;
		for (Enumeration en = indexTable.elements(); en.hasMoreElements();) {
			enCurr = (IndexEntry)en.nextElement();
			System.out.println(enCurr.toString() + "  " + enCurr.size());
		}
		*/
		long noOfBytes = 0;
		// calculate how many bytes are needed to be read
		// first, make sure noOfRecords is valid
		if (noOfRecords > indexTable.entrySet().size())
			noOfRecords = indexTable.entrySet().size();
		for (int i = 0; i < noOfRecords; i++) {
			IndexEntry ie = (IndexEntry)indexTable.get(new Integer(i));
			if (ie != null) {
				if (ie.isDeleted())
					noOfRecords++;
				else
					noOfBytes += ie.size();
			}
			else {
				break;
			}
		}
		
		//System.out.println("Number of bytes to retrieve: " + noOfBytes);
		//System.out.println("Current file size: " + partitionSize);		
		try {
			// pass on the work to the method which deals with number of bytes
			Record[] arrayRecord = retrieveFromDisk(noOfBytes, purgeTuples);
			return arrayRecord;
		} catch (NotEnoughBytesException e) {
			System.out.println("Exception: " + e.toString());
			e.printStackTrace();
		}
		return(new Record[0]);	
	}	
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#retrieveFromDisk(long, boolean)
	 *The no of bytes does not include deleted index entries
	 *
	 */
	public Record[] retrieveFromDisk(long noOfBytes, boolean purgeTuples)
		throws NotEnoughBytesException {		
		int countRecords = 0;
		Record[] arrayRecord = null;
		int deletedRecords = 0;
		long noOfBytesToRetreive = noOfBytes;
		//At the end of the routine the numberOfBytesToRetreive will contain
		//the actual number of bytes to retreive from the file		
		//Calculating the number of records that need to be retreived
		for(int i=0;i< indexTable.size();i++)
		{
			IndexEntry ie = (IndexEntry) indexTable.get(new Integer(i));
			if(!ie.isDeleted())
			{
				noOfBytes  = noOfBytes - ie.size();	
				//System.out.println(noOfBytes);
				if(noOfBytes >= 0)
				{
					countRecords = countRecords + 1;		
					noOfBytesToRetreive = noOfBytesToRetreive + ie.size();
				}
				if(noOfBytes < 0)
				{
					break;			
				}
			}
			//if deleted we still read the bytes from the file but we do not consider them as valid
			else
			{		
					noOfBytesToRetreive = noOfBytesToRetreive + ie.size();
					deletedRecords = deletedRecords + 1;
			}
		}	
		//System.out.println("Record count: " + countRecords);
		//System.out.println("Number deleted: " + deletedRecords);
		//System.out.println("Number of bytes to retrieve: " + noOfBytesToRetreive);
		//This is the array that we return
		arrayRecord = new Record[countRecords];
		int[] indexesToPurge = new int[countRecords];
		//Retreiveing bytes from file setting position for file channel			
		try{
			fc.position(0);			
			if(noOfBytesToRetreive > Integer.MAX_VALUE)
			{
				System.out.println("Note the no of bytes are too many are to many");
				System.out.println("Integer cannot handle it");
			}	
			// Note this is a hack I should  not covert the file size to int
			//Files can get very large with streaming systems			
			ByteBuffer bb =
				ByteBuffer.allocateDirect((int) noOfBytesToRetreive);				
			int nbytes = fc.read(bb);			
			//System.out.println(" The number of bytes read " +nbytes);									
			byte[] buffer = new byte[nbytes];	
			bb.rewind();					
			bb.get(buffer);
			//Setting this to null so that it can be collected by the
			//Grabage collector
			bb = null;
			for(int i = 0,j = 0; i < countRecords + deletedRecords; i++)
			{				
				IndexEntry ie = (IndexEntry) indexTable.get(new Integer(i));
				if(!ie.isDeleted())
				{
					// set the index entry to be deleted if needed
					if (purgeTuples) 
						indexesToPurge[j] = i;						
					//System.out.println("Start Index "+ie.startOffsetInFile+"EndIndex "+ie.endOffsetInFile);
					arrayRecord[j] = new Record(buffer, ie.startOffsetInFile, ie.endOffsetInFile, ie.getPosition());					
					j = j + 1;
				}			
			}
			//System.out.print("Purging indexes: ");
			//for (int x = 0; x < indexesToPurge.length; x++) 
			//	System.out.print(x + " ");
			//System.out.print("\n");
		} catch (IOException e) {			
			System.out.println("Exception"+e);
			e.printStackTrace();
		}								
		//System.out.println("The number of bytes to retrive "+noOfBytesToRetreive);	
		//System.out.println("The number of records "+countRecords);	
		
		// purge the records if needed
		if (purgeTuples) {
			try {
				this.purgeFromDiskFile(indexesToPurge);
			} catch (InvalidIndexException e) {
				System.out.println("Exception: " + e.toString());
				e.printStackTrace();	
			}
		}
		//Removes the file from the disk
		if (currentSize <= 0 && purgeTuples) {
			//System.out.println("Purging disk file " + fileName);
			this.purgeDiskFile();
		} else {
			//System.out.println("NOT purging disk file " + fileName);
		}
		return(arrayRecord);
	}
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#retrieveFromDisk(int, long, boolean)
	 */
	public Record[] retrieveFromDisk(int startIndex, long noOfBytes, boolean purgeTuples)
			throws  NotEnoughBytesException, InvalidIndexException {				
		int countRecords = 0;
		Record[] arrayRecord = null;
		int deletedRecords = 0;
		long noOfBytesToRetreive = 0;		
		//Calculating the number of records that need to be retreived
		if(startIndex > indexTable.size())
		{
			throw new InvalidIndexException();
		}
		IndexEntry ie = (IndexEntry) indexTable.get(new Integer(startIndex));
		if(ie.isDeleted())
			throw (new InvalidIndexException());
		for(int i=startIndex;i< indexTable.size();i++)
		{
			ie = (IndexEntry) indexTable.get(new Integer(i));
			if(!ie.isDeleted())
			{
				noOfBytes  = noOfBytes - ie.size();					
				if(noOfBytes >= 0)
				{
					countRecords = countRecords + 1;			
					noOfBytesToRetreive = noOfBytesToRetreive + ie.size();
				}
				if(noOfBytes < 0)
				{
					break;			
				}
			}
			//if delete we still read the bytes from the file but we do not consider them as valid
			else
			{		
					noOfBytesToRetreive = noOfBytesToRetreive + ie.size();		
					deletedRecords = deletedRecords + 1;
			}
		}		
		//This is the array that we return
		arrayRecord = new Record[countRecords];
		int[] indexesToPurge = new int[countRecords];		
		//Retreiveing bytes from file setting poistion for file channel to the start position
		//Another hack here converting the 
		try{			
			IndexEntry indexEntry = (IndexEntry) indexTable.get(new Integer(startIndex));	
			int startOffset = (int) indexEntry.getStartOffsetInFile();
			fc.position(startOffset);			
			if(noOfBytesToRetreive > Integer.MAX_VALUE)
			{
				System.out.println("Note the no Of bytes are too many are to many");
				System.out.println("Integer cannot handle it");
			}				
			// Note this is a hack I should  not covert the file size to int
			// Files can get very large with streaming systems			
			ByteBuffer bb =
				ByteBuffer.allocateDirect((int) noOfBytesToRetreive);				
			int nbytes = fc.read(bb);			
			//System.out.println(" The number of bytes read " +nbytes);
			byte[] buffer = new byte[nbytes];	
			bb.rewind();									
			bb.get(buffer);
			//Setting this to null so that it can be collected by the
			//Grabage collector
			bb = null;
			int offsetInByteArray = 0;						
			for(int i=startIndex,j=0;i<startIndex+countRecords+deletedRecords;i++)
			{				
				indexEntry = (IndexEntry) indexTable.get(new Integer(i));
				if(!indexEntry.isDeleted())
				{					
					// set the index entry to be deleted if needed
					if (purgeTuples) 
						indexesToPurge[j] = i;

					//System.out.println("Start Index "+indexEntry.startOffsetInFile+"EndIndex "+indexEntry.endOffsetInFile);
					//You cannot use startOffset any more because of the length of the byte array
					arrayRecord[j] = new Record(buffer,offsetInByteArray,indexEntry.size()-1,indexEntry.getPosition());										
					j = j + 1;
					offsetInByteArray = offsetInByteArray + indexEntry.size();					
				}
				//Still increment the offset because the byte array 
				//can also contain deleted records
				else
					offsetInByteArray = offsetInByteArray + indexEntry.size();																	
			}											
		} catch (IOException e) {			
			System.out.println("Exception"+e);
			e.printStackTrace();
		}																						
		//System.out.println("The number of bytes to retrive "+noOfBytesToRetreive);		
		//System.out.println("The number of records "+countRecords);		
		// purge the records if needed
		if (purgeTuples) {
			try {
				this.purgeFromDiskFile(indexesToPurge);
			} catch (InvalidIndexException e) {
				System.out.println("Exception: " + e.toString());
				e.printStackTrace();	
			}
		}
		return(arrayRecord);			
	}
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#size()
	 */
	public long size() {
		return this.currentSize;
	}
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#saveToDisk(Record[])
	 */
	public void saveToDisk(Record[] arrayRecord) 
		throws java.io.IOException,InvalidArgumentException{
		if(arrayRecord == null)
			throw (new InvalidArgumentException());
		//Build the index;
		long tempSize = partitionSize;				
		for(int index=(int) this.trueSize,j=0;index<arrayRecord.length+trueSize;index++,j++)
		{
			//Note the second argument stores the byte offset in the file
			IndexEntry ie =	new IndexEntry(
					index,
					tempSize,
					tempSize + arrayRecord[j].length()-1);									
			this.indexTable.put(new Integer(index),	ie);	
			//System.out.println("StartOffset "+ie.startOffsetInFile+ " endOffset "+ie.endOffsetInFile);
			tempSize = tempSize + arrayRecord[j].length();
		}

		//Write the complete array together for efficiency reasons
		byte[] byteArray = factory.recordsToBytes(arrayRecord);
		//System.out.println("The length of byteArray "+byteArray.length);		
		ByteBuffer bb = ByteBuffer.allocateDirect(byteArray.length);
		bb.put(byteArray);	
		// reset the pointer in bytebuffer
		bb.rewind();						
		//Write the ByteBuffer in filechannel
		//Setting the file position to the end of the file
		//before writing data into the file
		long endPosition = fc.size();
		fc.position(endPosition);
		//Writing the byte buffer to the end of file
		fc.write(bb);		
		//Forces the data to be written to the file
		//fc.force(true);						
		//Printing the index
		//Enumeration enumeration = this.indexTable.elements();
		//while (enumeration.hasMoreElements()) {
		//	IndexEntry element = (IndexEntry) enumeration.nextElement();
		//	System.out.println("Index "+element);			
		//}

		//Updating the currentSize of the disk file						
		this.currentSize = this.currentSize + arrayRecord.length;
		this.trueSize = this.trueSize + arrayRecord.length;
		//Updating the Partition Size
		partitionSize = tempSize; 
	}
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#purgeFromDiskFile(int[])
	 */
	public void purgeFromDiskFile(int[] recordIndexes)
		throws NotEnoughBytesException, InvalidIndexException
	  {
		if(recordIndexes.length > indexTable.size())
		{
			System.out.println("Not enough recrods to delete");
			throw new NotEnoughBytesException();
		}					
		for(int i=0;i<recordIndexes.length;i++)
		{
			IndexEntry indexEntry = 
				((IndexEntry)this.indexTable.get(new Integer(recordIndexes[i])));
			if(indexEntry == null)
			 	throw(new InvalidIndexException());
			 	
			if(indexEntry.isDeleted())
			{
			 	System.out.println("Tuple already deleted ");
			 	throw (new InvalidIndexException());
			}			
			else
			{
			 	//Decrementing size
			 	indexEntry.setDeleted(true);
			 	this.currentSize = this.currentSize - 1;
			 	//Note we do not decrement trueSize here
			}
		}	
	}
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IRecordStore#sizeInBytes()
	 */
	public long sizeInBytes() {		
		return(objectFile.length());	
	}
}