/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;

import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidIndexException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidArgumentException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.NotEnoughBytesException;

/**
 * Defines the interface between the StorageManger and the DiskFile
 * @author Nishant Mehta and Abhishek Kumar; edited by Natasha Bogdanova (6/9/04)
 */
public interface IRecordStore {
	
	/**
 	 * Retrieves all the records from the RecordStore. If purge = true,
	 * then the records are also deleted.
	 * 
	 * @return Record[]
	 * @param purge boolean
	 */
 	public Record[] retrieveFromDisk(boolean purge);	

 	/**
	  * Retrieves a given number of tuples from the RecordStore. If 
	  * purge = true, then the records are also deleted. 
	  * 
	  * Creation Date: (Jun 8, 2004 10:17:41 AM)
	  * @author Natasha Bogdanova
	  * @param noOfRecords int
	  * @param purge boolean
	  * @return Record[]
	  */
	public Record[] retrieveFromDisk(int noOfRecords, boolean purge);

	/**
	 * Retrieves the records equivalent to the number of bytes. If
	 * purge = true, then the records are also deleted.
	 * 
	 * @return Record[]
	 * @param noOfBytes long
	 * @param purge boolean
	 * @throws NotEnoughBytesException
	 */
	public Record[] retrieveFromDisk(long noOfBytes, boolean purge) throws NotEnoughBytesException;
 
	/**
	 * Retreives the records equivalent to the given number of bytes, 
	 * starting at the requested index entry. If purge = true, then 
	 * the records are also deleted.
	 * 
	 * @param startIndex int
	 * @param noOfBytes long
	 * @param purge boolean
	 * @return Record[]
	 * @throws NotEnoughBytesException
	 * @throws InvalidIndexException
	 */ 
	public Record[] retrieveFromDisk(int startIndex, long noOfBytes, boolean purge)
		throws NotEnoughBytesException,InvalidIndexException;	
  
	/**
	 * Retrieves the size of the RecordStore in number of records stored.
	 * 
	 * @return long
	 */
	public long size();
 
	/**
	 * Retrieves the size of the RecordStore as number of bytes.
	 * 
	 * @return long
	 */
	public long sizeInBytes();
  
	/**
	 * Writes the records to disk.
	 * 
	 * @param arrayRecord Record[]
	 * @throws InvalidArgumentException
	 * @throws IOException
	 */
	public void saveToDisk(Record[] arrayRecord)
		throws InvalidArgumentException, java.io.IOException;

	/**
	 * Deletes all the records from the DiskFile and removes the DiskFile.
	 */
	public void purgeDiskFile();
 
	/**
	 * Deletes the records that correspond to the given indexes from the
	 * DiskFile.
	 * 
	 * @param recordIndexes int[]
	 * @throws NotEnoughBytesException
	 * @throws InvalidIndexException
	 */  
	public void purgeFromDiskFile(int[] recordIndexes)
		throws NotEnoughBytesException, InvalidIndexException;

	/**
	 * Used to clean up the records.
	 */
	public void cleanup();

}
