/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidIndexException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidPartitionException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.NotEnoughBytesException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.StorageManagerNotInitializedException;

/**
 * Defines the interface for the Storage Manager.
 * 
 * @author Nishant Mehta and Abhishek Kumar; edited by Natasha Bogdanova (6/9/04)
 */
public interface IStore {
	
	/**
	 * Initializes the storage manager. Must be called before using the
	 * storage manager.
	 */
	public void initialize(); 
	 
	/**
	 * Deletes all the temporary files and closes all file handles.
	 */
	public void cleanup(); 
			
	/**
	 * Retrieves the bucket corressponding to the partionNumber from the disk.
	 * If purge = true, the tuples are also deleted from the file.
	 * 
	 * @param partitionNumber int
	 * @param purge boolean
	 * @return XATTupleImp[]
	 * @throws InvalidPartitionException
	 */
	public XATTupleImp[] retrieveFromDisk(int partitionNumber, boolean purge) throws InvalidPartitionException,StorageManagerNotInitializedException;
	 
	/**
	 * Retrieves a given number of tuples from the desired partition.
	 * If purge = true, the tuples are also deleted from the file.
	 * 
	 * Creation Date: (Jun 8, 2004 10:36:58 AM)
	 * @author Natasha Bogdanova
	 * @param partitionNumber int
	 * @param noOfRecords int
	 * @param purge boolean
	 * @return XATTupleImp[]
	 * @throws InvalidPartitionException
	 * @throws NotEnoughBytesException
	 */
	 public XATTupleImp[] retrieveFromDisk(int partitionNumber, int noOfRecords, boolean purge) throws InvalidPartitionException, NotEnoughBytesException;
	 
	/**
	 * Retrieves the tuples from the bucket corressponding to the 
	 * partionNumber from the disk. The number of tuples returned will
	 * be equal to or less than the noOfBytes/(tuple size). The tuples
	 * will be returned from the begining of the file. If purge = true,
	 * the tuples are also deleted from the file.
	 * 
	 * @param partitionNumber int
	 * @param noOfBytes long
	 * @param purge boolean
	 * @return XATTupleImp[]
	 * @throws InvalidPartitionException
	 * @throws NotEnoughBytesException
	 */
	public XATTupleImp[] retrieveFromDisk(int partitionNumber, long noOfBytes, boolean purge) throws InvalidPartitionException ,NotEnoughBytesException;
	 
	/**
	 * Retrieves the tuples from the bucket corressponding to the 
	 * partionNumber from the disk. The number of tuples returned will
	 * be equal to or less than the noOfBytes/(tuple size). The tuples
	 * will be returned starting from the index, given as startIndex.
	 * If purge = true, the tuples are also deleted from the file.
	 * 
	 * @param partitionNumber int
	 * @param startIndex int
	 * @param noOfBytes long
	 * @param purge boolean
	 * @return XATTupleImp[]
	 * @throws InvalidPartitionException
	 * @throws NotEnoughBytesException
	 * @throws InvalidIndexException
	 */
	public XATTupleImp[] retrieveFromDisk(int partitionNumber,int startIndex, long noOfBytes, boolean purge) throws InvalidPartitionException, NotEnoughBytesException,InvalidIndexException;	
	  
	/**
	 * Retrieves the size of the bucket corresponding to the partionNumber
	 * from the disk. The size is the number of valid tuples on the disk 
	 * that not deleted. If the partition does not exist on the disk it 
	 * returns zero.
	 * 
	 * @param partitionNumber int
	 * @return long
	 */
	public long size(int partitionNumber);
	 
	/**
	 * Saves the tuples to the bucket indicated by the partitionNumber.
	 * If the bucket already exists, this will append the tuples to the bucket.
	 * Else this will create a new bucket.
	 * 
	 * @param partitionNumber int
	 * @param arrayTuple XATTupleImp[]
	 * @throws IllegalArgumentException
	 */
	public void saveToDisk(int partitionNumber, XATTupleImp[] arrayTuple) throws IllegalArgumentException,StorageManagerNotInitializedException;
	
	/**
	 * Deletes all the tuples from the partition and deletes the file too.
	 * 
	 * @param partitionNumber int
	 * @throws InvalidPartitionException;
	 */	
	public void purgeBucketFromDisk(int partitionNumber) throws InvalidPartitionException;
	 
	/**
	 * Deletes the tuples which correspond to the given indexes from the
	 * given partition.
	 * 
	 * @param partitionNumber int
	 * @param tupleIndexes int[]
	 * @throws NotEnoughBytesException
	 * @throws InvalidPartitionException
	 * @throws InvalidIndexException
	 */  
	public void purgeBucketFromDisk(int partitionNumber, int[] tupleIndexes) throws NotEnoughBytesException,InvalidPartitionException,InvalidIndexException	;
	 
	/**
	 * Returns the total number of tuples stored on the disk currently.
	 * 
	 * @return long
	 */
	public long totalSize();
  
}
