/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;


/**
 * Used to store the metainformation about the tuple class
 * This class is used to store the type information to model the tuple class
 * Note the tuple is a container of objects so it also has to be modeled as a tree
 * Thus to store the type information we model this also using a tree
 * of symbol entries. This class is a helper that helps us to build this tree
 * of symbol entries This class will be used by the factory to create record objects
 * This class stores the mapping to convert from tuples to records and vice versa
 * 
 * @author Nishant Mehta and Abhishek Kumar
 */

// Note we also need a mechanism to type cast the object to primitive
// data types once read from the file


public class MetaInformation {
	/*Stores the root the symbol entry tree*/
	protected SymbolEntry root;
	/*Stores the reference to the Metainformation*/
	//protected static MetaInformation mi = null;
	/**
	 * USed to create a new MetaInformation class
	 * Note this class is a singleton too
	 * @param t
	 * @param varName
	 */
	protected MetaInformation(Object t, String varName) {
		root = this.makeSymbolTree(t, varName);			
	}
	
	/*
	public static MetaInformation getMetaInformation(Object t,String varName)
	{
		if(mi == null)	{
				mi = new MetaInformation(t,varName);
			}
		return(mi);			
	}
	*/
	/**
	 * Constructs the Symbol Tree for the object passed in as a parameter
	 * @param t
	 * @param varName
	 * @return
	 */
	public SymbolEntry makeSymbolTree(Object t, String varName) {
		Class theClass;
		SymbolEntry se = null;
		
		theClass = t.getClass();
		if (this.isSimpleType(t)) {
			se = this.makeNewSymbolEntry(t, theClass.getName(), varName);
		}	else if (this.isComplexType(t)) {
			ComplexSymbolEntry cse = null; 
			if (theClass.isArray()) {
				cse = (ComplexSymbolEntry) makeNewArraySymbolEntry(t, theClass, varName);
				Object[] array = (Object[]) t;					
				for (int i = 0; i < array.length; i++) {					
					cse.addChild(this.makeSymbolTree(array[i], varName + "[" + i + "]"));
				}
				se = cse;
			}
			
			if (t instanceof Vector) {
				Vector v = (Vector) t;
				cse =(ComplexSymbolEntry) new VectorSymbolEntry(t, theClass.getName(), varName);
				for (int i = 0; i < v.size(); i++) {					
					cse.addChild( this.makeSymbolTree(v.get(i), varName + "[" + i + "]"));
				}
				se = cse;
			}
			if (t instanceof ArrayList) {
				ArrayList al = (ArrayList) t;
				cse =(ComplexSymbolEntry) new ArrayListSymbolEntry(t, theClass.getName(), varName);
				for (int i = 0; i < al.size(); i++) {
					cse.addChild(this.makeSymbolTree(al.get(i), varName + "[" + i + "]"));
				}
				se = cse;
			}
		}	else { // t is an instance of a class that is a collection //hetrogenous types	
			ComplexSymbolEntry cse = (ComplexSymbolEntry) new ClassSymbolEntry(t, theClass.getName(), varName);
			//class entry iterate`through all the fields
			Field[] fields = theClass.getDeclaredFields();
			for (int i = 0; i < fields.length; i++) {
				fields[i].setAccessible(true);
				Object obj = null;
				try {
					obj = fields[i].get(t);
					//This is to take care of null references
					if(obj == null)	{
						//System.out.println("The member field " + varName +"  "+fields[i].getName()+" does not have any value");
					}
					else {
						//add the symbol entry for the element to the 
						//list of child symbol entries
						cse.addChild(this.makeSymbolTree(obj, fields[i].getName()));
						se = cse;
					}
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				}				
			}			
		}
		return (se);
	}

	/** 
	 * @see java.lang.Object#finalize()
	 */
	public void finalize() {

	}

	/**
	 * Method isSimpleType.
	 * @param typeName
	 * @return boolean
	 */

	public boolean isSimpleType(Object t) {
		if (t instanceof String)
			return true;
		if (t instanceof Integer)
			return true;
		if (t instanceof Float)
			return true;
		if (t instanceof Long)
			return true;
		if (t instanceof Boolean)
			return true;		
		else
			return false;
	}

	/**
	 * This contains all the supported types
	 * @param typeName
	 * @return boolean
	 */
	public boolean isComplexType(Object t) {
		//		System.out.println("Checking for type name "+ typeName);		
		if (t.getClass().isArray())
			return true;
		if (t instanceof Vector)
			return true;
		if (t instanceof ArrayList)
			return true;
			
		else
			return false;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return(root.returnString());
	}

	/**
	* Makes new Symbol Entry used to construct entries for sile types 
	* @param object
	* @param string
	* @param string2
	* @return SymbolEntry
	*/
	private SymbolEntry makeNewSymbolEntry(Object t, String typeName, String varName) {
		//		System.out.println("Creating for type " + typeName);
		if (t instanceof String) {
			return (new StringSymbolEntry(t, typeName, varName));
		}

		if (t instanceof Integer) {
			return (new IntegerSymbolEntry(t, typeName, varName));
		}

		if (t instanceof Boolean) {
			return (new BooleanSymbolEntry(t, typeName, varName));
		}
		if (t instanceof Float) {
			return (new FloatSymbolEntry(t,	typeName, varName));
		}
		if (t instanceof Long) {
			return (new LongSymbolEntry(t, typeName, varName));
		}

		return (null);

	}
	
	private SymbolEntry makeNewArraySymbolEntry(Object t,Class theClass, String varName) {	
		SymbolEntry se = null;
		if(theClass.isArray()) {
			if(t instanceof XATDataValue[])	{
				se = new XATDataValueArraySymbolEntry(t,theClass.getName(),varName);		
			}
			else if(t instanceof Object[]) {
				se = new XATDataValueArraySymbolEntry(t,theClass.getName(),varName);				
			}
			else if (t instanceof long[]) {
				System.out.println("Not Supported array type as yet please add a class");
			}
			else if (t instanceof int[]) {
				System.out.println("Not Supported array type as yet please add a class");
			} else {
				System.out.println("Not Supported array type as yet please add a class");
			}

		}
		return(se);
	}
	

	/**
	 * Returs the root of the symbol tree
	 * @return
	 */
	public SymbolEntry getSymbolEntryTree() {
		return(this.root);		
		
	}

	/**
	 * 
	 * @return MetaInformation
	 */		
	/*
	public static MetaInformation getMetaInformation() {
		if(mi !=null)
		return(mi);
		else return(null);
	}
	*/

}
